/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>

#include <glib.h>
#include <gmodule.h>

#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>

#include "Device.hpp"
#include "Omni.hpp"

int fDebugOutput;
int fDoOutput;

typedef struct _HWMAR
{
  float fLeftClip;
  float fBottomClip;
  float fRightClip;
  float fTopClip;
  float fxWidth;
  float fyHeight;
} HWMARGINS;

typedef struct _HWRES
{
  float xRes;
  float yRes;
  float fScanDots;  // number of dots in scan line
} HWRESOLUTION;

typedef struct _PRTMODE
{
  int iBitCount;
  int iPlanes;
} PRINTMODE;

/*-----------------------------------------------------------------------------*/
/*                                                                             */
/* Function: FindBandSize                                                      */
/*  This function figures out the appropriate band size based on the amount    */
/*  of memory that can be occupied by the band and bitmap information.         */
/*  bitmap, and the specifics of the hardware                                  */
/*                                                                             */
/*-----------------------------------------------------------------------------*/



ULONG FindBandSize(ULONG ulYHeight, ULONG ulXWidth, ULONG ulBitsPerPel, ULONG ulNumPlanes,
                            ULONG ulModulus, ULONG ulMemoryLimit)
{
    ULONG ulSizeScanLine;
    ULONG ulMemoryNeeded;
    ULONG ulNumLinesFit;

    if (0 == ulModulus)
    {
       ulModulus = 1;
    }
    // figure out how much memory is needed fore each line
    ulSizeScanLine = ((ulBitsPerPel*ulXWidth+31)/32)*ulNumPlanes*4;

    // Figure out how much memory is needed for the page
    ulMemoryNeeded = ulYHeight * ulSizeScanLine;

    // How many lines can fit in the size given?
    ulNumLinesFit = ulMemoryLimit / ulSizeScanLine;

    if (0 == ulNumLinesFit)
       // Minimum of 1 scan line
       ulNumLinesFit = 1;

    if (ulNumLinesFit <= ulModulus)
       // Not enough lines... Promote it to a modulus.
       ulNumLinesFit = ulModulus;
    else
       // Bump down the number of lines so that it is a modulus.
       ulNumLinesFit -= ulNumLinesFit % ulModulus;

    if ((ulYHeight % ulNumLinesFit) * 100 / ulYHeight <= 15)
    {
       USHORT  usBumpUp;

       usBumpUp  = ulYHeight % ulNumLinesFit;
       usBumpUp += ulModulus - 1;
       usBumpUp /= ulModulus;
       usBumpUp *= ulModulus;
       ulNumLinesFit += usBumpUp;
    }

    return ulNumLinesFit ;  // return the number of lines we want for the band
}


int
main (int argc, char *argv[])
{
#define MAX_SCAN_LINES 1024
   const char         *pszPPDFilename;        /* PPD environment variable */
   ppd_file_t         *ppd;                   /* PPD file */
   int                 fdInput;               /* File descriptor */
   cups_raster_t      *pRasterStream     = 0; /* Raster stream to read from */
   cups_page_header_t  pageHeader;            /* Page header from file */
   cups_page_header_t  pageHeaderCurrent;     /* Page header from file */
   GModule            *hmodDevice        = 0;
   Device             *pDevice           = 0;
   PFNNEWDEVICEWARGS   pfnNewDeviceWArgs = 0;
   int                 iPageNumber       = 0;
   unsigned int        uiScanline        = 0;
   unsigned int        iNumScanLines     = MAX_SCAN_LINES;
   PBYTE               pbBits            = 0;
   PBYTE               pbBitsToSend      = 0;
   PBITMAPINFO2        pbmi2Bitmap       = 0;
   int                 iNumColors        = 0;
   int                 icbBitmapScanLine = 0;
   int                 iBytesToAllocate  = 0;
   RECTL               rectlPageLocation;
   HWRESOLUTION        hwRes;
   DeviceResolution   *pDevRes           = 0;                // Device Resolution stuff
   HardCopyCap        *pHCC              = 0;                   // Hard copy information
   unsigned long       ulBytesRead       = 0;
   bool                bFoundForm        = false;
   char                cParamString[1024];
   int                 i,
                       j,
                       k;
   char               *ptemp1            = 0;
   char                temp2;
   unsigned int        xRes,
                       yRes;
   char               *pszCommandLine    = 0;
   bool                fTopOfBand        = false;
   FILE               *outputFile        = 0;
   int                 rc                = 0;

   memset(cParamString, 0, 1024);


   if (  argc < 6
      || argc > 7
      )
   {
      /*
       * We don't have the correct number of arguments; write an error message
       * and return.
       */

      cerr << "ERROR: CUPSToOmni job-id user title copies options [file]" << endl;
      return 1;
   }

   Omni::initialize ();

   pszCommandLine = strstr((const char *)argv[5],(const char *)"debug");

   if(pszCommandLine)  // enable debugging output
       fDebugOutput = 1;
   else
       fDebugOutput = 0;

   pszCommandLine = strstr((const char *)argv[5],(const char *)"no-output");

   if(pszCommandLine)  // enable debugging output
       fDoOutput = 0;
   else
       fDoOutput = 1;


   if(fDebugOutput) outputFile = fopen("/tmp/omni_debug.out", "w+b");

   if(fDebugOutput)
   {
      for (i = 0 ; i < argc ; i++)
      {
          fprintf(outputFile, "Arg %d = %s \n", i, argv[i]);
      }
   }

   /*
    * Get the PPD file and figure out which driver to use...
    */

   if ((pszPPDFilename = getenv ("PPD")) == NULL)
   {
      cerr << "ERROR: Fatal error: PPD environment variable not set!" << endl;

      rc = 1;
      goto done;
   }

   if ((ppd = ppdOpenFile (pszPPDFilename)) == NULL)
   {
      cerr << "ERROR: Fatal error: Unable to load PPD file " << pszPPDFilename << endl;

      rc = 1;
      goto done;
   }

   if(fDebugOutput) fprintf(outputFile," successful ppd open \n");

   if (ppd->nickname == NULL)
   {
      cerr << "ERROR: Fatal error: No NickName attribute in PPD file " << pszPPDFilename << endl;
      ppdClose (ppd);

      rc = 1;
      goto done;
   }

   if(fDebugOutput) fprintf(outputFile," The nickname is %s \n", ppd->nickname);

   hmodDevice = g_module_open (ppd->nickname, (GModuleFlags)0);

   if(fDebugOutput)
     if(hmodDevice)
     {
        fprintf(outputFile," successful driver load \n");
     }
     else
     {
        fprintf(outputFile," unsuccessful driver load \n");
     }

   if (!hmodDevice)
   {
      cerr << "Error: unable to load " << ppd->nickname << endl;
      cerr << "g_module_error returns " << g_module_error () << endl;

      rc = 1;
      goto done;
   }

   g_module_symbol (hmodDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

   if(fDebugOutput)
     if(pfnNewDeviceWArgs)
     {
        fprintf(outputFile,"set up driver successful \n");
     }
     else
     {
        fprintf(outputFile,"set up driver was unsuccessful \n");
     }


   if (!pfnNewDeviceWArgs)
   {
      cerr << "Error: " << ppd->nickname << " is not a valid library" << endl;

      rc = 1;
      goto done;
   }

   /*
    * Open the page stream...
    */
   if (argc == 7)
   {
      if ((fdInput = open (argv[6], O_RDONLY)) == -1)  //open the raster file
      {
         perror ("ERROR: Unable to open raster file - ");
         sleep (1);

         rc = 1;
         goto done;
      }
   }
   else
   {
      fdInput = 0;
   }

   pRasterStream = cupsRasterOpen (fdInput, CUPS_RASTER_READ);

   if(!pRasterStream)
   {
      if(fDebugOutput) fprintf(outputFile,"cupsRasterOpen Failed \n");
   }

   iPageNumber = 0;


   /*
    * Process pages as needed...
    */

   while (cupsRasterReadHeader (pRasterStream, &pageHeader))
   {
 //      if( (int) pageHeader.PageSize[0] < 1) // flag that output has been completed
//           break;

/***********************************************************************************************/
/*                                                                                             */
/*   Current Monochrome selection is:                                                          */
/*      -o ColorModel=Gray                                                                     */
/*                                                                                             */
/*                                                                                             */
/*   We are currently using the FORM_ strings and such to invoke Omni                          */
/*                                                                                             */
/*    Formulate a string to pass into the driver based on the following parameter list         */
/*                                                                                             */
/*  - form:                                                                                    */
/*                                                                                             */
/*     form=FORM_LETTER or form=FORM_A4                                                        */
/*                                                                                             */
/*  - resolution:                                                                              */
/*                                                                                             */
/*     resolution=RESOLUTION_360_X_360 or resolution=RESOLUTION_180_X_180                      */
/*                                                                                             */
/*  - media:                                                                                   */
/*                                                                                             */
/*     media=MEDIA_PLAIN or media=MEDIA_GLOSSY                                                 */
/*                                                                                             */
/*  - dithering:                                                                               */
/*                                                                                             */
/*     dither=DITHER_STUCKI_DIFFUSION or dither=DITHER_STEINBERG_DIFFUSION                     */
/*                                                                                             */
/*  - tray:                                                                                    */
/*                                                                                             */
/*     tray=TRAY_MANUAL_FEEDER                                                                 */
/*                                                                                             */
/*  - printmode:                                                                               */
/*     printmode=PRINT_MODE_1_ANY (**) or printmode=PRINT_MODE_24_CMYK                         */
/***********************************************************************************************/

   if(iPageNumber == 0)  // we need to do preliminary driver setup here
   {
       // get the form name

       for(i = 0 ; i < ppd->num_sizes && !bFoundForm ; i++)
       {
           if(( ppd->sizes[i].width == pageHeader.PageSize[0] )
               && (ppd->sizes[i].length == pageHeader.PageSize[1] ))
           {
               strcat(cParamString, "form=");
               strcat(cParamString, ppd->sizes[i].name);
               strcat(cParamString, " ");
               bFoundForm = true;
           }
       }
      // get the resolution
       for(i = 0 ; i < ppd->num_groups ; i++)
       {
           for (j = 0 ; j < ppd->groups[i].num_options ; j++)
           {
               if(!strcmp("Resolution", ppd->groups[i].options[j].text))
               {
                   for (k = 0 ; k < ppd->groups[i].options[j].num_choices ; k++)
                   {
                       ptemp1 = strchr(ppd->groups[i].options[j].choices[k].code, '[');  // get the string for the
                                                                            // [resx resy]setpagedevice
                       if(ptemp1)
                       {
                           ptemp1++;
                           while(!isdigit(*ptemp1))
                           {
                               ptemp1++;
                           }
                           xRes = atoi(ptemp1);
                           while(isdigit(*ptemp1))
                           {
                               ptemp1++;
                           }
                           while(!isdigit(*ptemp1))
                           {
                               ptemp1++;
                           }
                           yRes = atoi(ptemp1);

                           if((xRes == pageHeader.HWResolution[0]) &&  (yRes == pageHeader.HWResolution[1]))
                           {
                             strcat(cParamString, "resolution=");
                             strcat(cParamString, ppd->groups[i].options[j].choices[k].choice);
                             strcat(cParamString, " ");
                             break;
                           }
                       }
                   }
               }
           }
       }

       // get the media type
       for(i = 0 ; i < ppd->num_groups ; i++)
       {
           for (j = 0 ; j < ppd->groups[i].num_options ; j++)
           {
//             if(fDebugOutput) fprintf(outputFile,"media key= %s \n",ppd->groups[i].options[j].text );
               if(!strcmp("Media Type", ppd->groups[i].options[j].text))
               {
                   for (k = 0 ; k < ppd->groups[i].options[j].num_choices ; k++)
                   {
                      if(fDebugOutput) fprintf(outputFile,"MediaType = %s \n", pageHeader.MediaType );
                      if(fDebugOutput) fprintf(outputFile,"MediaType Val = %s \n", ppd->groups[i].options[j].choices[k].text );
                      if(!strcmp(pageHeader.MediaType, ppd->groups[i].options[j].choices[k].text))
                      {
                          strcat(cParamString, "media=");
                          strcat(cParamString, ppd->groups[i].options[j].choices[k].choice);
                          if(fDebugOutput) fprintf(outputFile,"media concat = %s \n",ppd->groups[i].options[j].choices[k].choice );
                          strcat(cParamString, " ");
                          break;

                      }
                   }
               }
           }
       }
     // get the dither type
     for(i = 0 ; i < ppd->num_groups ; i++)
     {
         for (j = 0 ; j < ppd->groups[i].num_options ; j++)
         {
//             if(fDebugOutput) fprintf(outputFile,"dither key= %s \n",ppd->groups[i].options[j].text );
             if(!strcmp("Dither", ppd->groups[i].options[j].text))
             {
                 for (k = 0 ; k < ppd->groups[i].options[j].num_choices ; k++)
                 {
                    if(fDebugOutput) fprintf(outputFile,"dither = %s \n", pageHeader.OutputType );
                    if(fDebugOutput) fprintf(outputFile,"Dither Val = %s \n", ppd->groups[i].options[j].choices[k].text );
                    if(!strcmp(pageHeader.OutputType, ppd->groups[i].options[j].choices[k].text))
                    {
                        strcat(cParamString, "dither=");
                        strcat(cParamString, ppd->groups[i].options[j].choices[k].choice);
                        strcat(cParamString, " ");
                        break;

                    }
                 }
             }
         }
     }

       // get the tray selection
       for(i = 0 ; i < ppd->num_groups ; i++)
       {
           for (j = 0 ; j < ppd->groups[i].num_options ; j++)
           {
//               if(fDebugOutput) fprintf(outputFile,"Media Source = %s \n",ppd->groups[i].options[j].text );
               if(!strcmp("Media Source", ppd->groups[i].options[j].text))
               {
                   for (k = 0 ; k < ppd->groups[i].options[j].num_choices ; k++)
                   {
                      if(fDebugOutput) fprintf(outputFile,"MediaClass = %s \n", pageHeader.MediaClass );
                      if(fDebugOutput) fprintf(outputFile,"Media Source Val = %s \n", ppd->groups[i].options[j].choices[k].text );
                      if(!strcmp(pageHeader.MediaClass, ppd->groups[i].options[j].choices[k].text))
                      {
                          strcat(cParamString, "tray=");
                          strcat(cParamString, ppd->groups[i].options[j].choices[k].choice);
                          strcat(cParamString, " ");
                          break;

                      }
                   }
               }
           }
       }
       if(pageHeader.cupsBitsPerPixel == 1)
       {
           strcat(cParamString, "printmode=PRINT_MODE_1_ANY");
       }
       else
       {
         pszCommandLine = 0;              // dither failsafe
         pszCommandLine = strstr((const char *)argv[5],(const char *)"dither");

         if(pszCommandLine && !strstr((const char *)cParamString,(const char *)"dither"))
         {
             pszCommandLine =0;
             pszCommandLine = strstr((const char *)argv[5],(const char *)"dither");
             if(pszCommandLine)
             {
                 ptemp1 = 0;
                 strcat(cParamString, " ");
                 ptemp1 = strchr((const char *) pszCommandLine, ' ');
                 if(ptemp1)            // additional parameters on the line
                 {
                     temp2 = *ptemp1;
                     *ptemp1 = '\0';
                     strcat(cParamString, pszCommandLine);
                     *ptemp1 = temp2;
                 }
                 else
                 {
                     strcat(cParamString, pszCommandLine);     // no additional parameters on the line
                 }

             }
         }
       }
       if(fDebugOutput) fprintf(outputFile, "input = %s \n", argv[5]);

       if(pageHeader.cupsBitsPerPixel != 1)
       {
           pszCommandLine =0;
           pszCommandLine = strstr((const char *)argv[5],(const char *)"printmode");
           if(pszCommandLine)
           {
               ptemp1 = 0;
               strcat(cParamString, " ");
               ptemp1 = strchr((const char *) pszCommandLine, ' ');
               if(ptemp1)        // additional parameters on the line
               {
                   temp2 = *ptemp1;
                   *ptemp1 = '\0';
                   if(strlen(pszCommandLine) > strlen("printmode = "))
                   {
                       strcat(cParamString, pszCommandLine);
                       *ptemp1 = temp2;
                   }
               }
               else
               {
                   strcat(cParamString, pszCommandLine); // no additional parameters on the line
               }
           }
       }

       pszCommandLine = 0;
       pszCommandLine = strstr((const char *)argv[5],(const char *)"InputSlot");
       if(pszCommandLine)
       {
           ptemp1 = 0;
           strcat(cParamString, " ");
           ptemp1 = strchr((const char *) pszCommandLine, ' ');
           if(ptemp1)              // additional parameters on the line
           {
               temp2 = *ptemp1;
               *ptemp1 = '\0';
               strcat(cParamString, pszCommandLine);     // no additional parameters on the line
               *ptemp1 = temp2;
           }
           else
           {
               strcat(cParamString, pszCommandLine);
           }
       }
/*********************  Failsafe for media type selection **************************************/

       pszCommandLine = 0;
       pszCommandLine = strstr((const char *)argv[5],(const char *)"media");

       if(pszCommandLine && !strstr((const char *)cParamString,(const char *)"media"))
       {
           pszCommandLine = 0;          // check for MEDIA_ for now if we have to override media settings
           pszCommandLine = strstr((const char *)argv[5],(const char *)"MEDIA_");
           if(pszCommandLine)
           {
               for(i = 0 ; i < ppd->num_groups ; i++)
               {
                   for (j = 0 ; j < ppd->groups[i].num_options ; j++)
                   {
                       if(!strcmp("Media Type", ppd->groups[i].options[j].text))
                       {
                           for (k = 0 ; k < ppd->groups[i].options[j].num_choices ; k++)
                           {

                              if(fDebugOutput)  fprintf(outputFile, "compare - %s \n",ppd->groups[i].options[j].choices[k].choice );
                              if(!strncmp(pszCommandLine, ppd->groups[i].options[j].choices[k].choice,
                                  strlen(ppd->groups[i].options[j].choices[k].choice) ))
                              {
                                  strcat(cParamString, " media=");
                                  strcat(cParamString, ppd->groups[i].options[j].choices[k].choice);
                                  strcat(cParamString, " ");
                                  break;

                              }
                           }
                       }
                   }
               }

           }

       }

       pDevice = pfnNewDeviceWArgs (cParamString, true);    // true for non-pdc implementation

       if(fDebugOutput)
         if(pDevice)
         {
            fprintf(outputFile,"set up of new driver properties successful \n");
         }
         else
         {
            fprintf(outputFile,"set up of new driver properties was unsuccessful \n");
         }

       if (!pDevice)
       {
          cerr << "Error: Could not instantiate a new device." << endl;

          rc = 1;
          goto done;
       }


       pageHeaderCurrent.cupsColorSpace        = pageHeader.cupsColorSpace;
       pageHeaderCurrent.cupsWidth             = pageHeader.cupsWidth;
       pageHeaderCurrent.cupsHeight            = pageHeader.cupsHeight;
       pageHeaderCurrent.cupsBitsPerColor      = pageHeader.cupsBitsPerColor;
       pageHeaderCurrent.cupsBytesPerLine      = pageHeader.cupsBytesPerLine;
       pageHeaderCurrent.cupsBitsPerPixel      = pageHeader.cupsBitsPerPixel;
       strcpy(pageHeaderCurrent.MediaClass, pageHeader.MediaClass);
       strcpy(pageHeaderCurrent.MediaType, pageHeader.MediaType);
       strcpy(pageHeaderCurrent.OutputType, pageHeader.OutputType);
       pageHeaderCurrent.HWResolution[0]       = pageHeader.HWResolution[0];
       pageHeaderCurrent.HWResolution[1]       = pageHeader.HWResolution[1];
       pageHeaderCurrent.PageSize[0]           = pageHeader.PageSize[0];
       pageHeaderCurrent.PageSize[1]           = pageHeader.PageSize[1];
       pageHeaderCurrent.ImagingBoundingBox[0] = pageHeader.ImagingBoundingBox[0];
       pageHeaderCurrent.ImagingBoundingBox[1] = pageHeader.ImagingBoundingBox[1];
       pageHeaderCurrent.ImagingBoundingBox[2] = pageHeader.ImagingBoundingBox[2];
       pageHeaderCurrent.ImagingBoundingBox[3] = pageHeader.ImagingBoundingBox[3];
   }

    if(fDebugOutput) fprintf(outputFile,"parameter string = %s\n", cParamString);

    if(fDebugOutput) fprintf(outputFile,"HW Res x = %u  - y = %u \n", pageHeader.HWResolution[0], pageHeader.HWResolution[1]);
    if(fDebugOutput) fprintf(outputFile,"page size = %d x %d \n",(int) pageHeader.PageSize[0], (int) pageHeader.PageSize[1]);
    if(fDebugOutput) fprintf(outputFile,"Bounding Box = %u %u %u %u \n", pageHeader.ImagingBoundingBox[0], pageHeader.ImagingBoundingBox[1],
                                         pageHeader.ImagingBoundingBox[2], pageHeader.ImagingBoundingBox[3]);

    if(fDebugOutput) fprintf(outputFile,"cupsColorSpace       = %d \n", pageHeader.cupsColorSpace);
    if(fDebugOutput) fprintf(outputFile,"CupsWidth              =%d \n", pageHeader.cupsWidth);
    if(fDebugOutput) fprintf(outputFile,"CupsHeight             =%d \n", pageHeader.cupsHeight);
    if(fDebugOutput) fprintf(outputFile,"cupsBitsPerColor       =%d \n", pageHeader.cupsBitsPerColor);
    if(fDebugOutput) fprintf(outputFile,"CupsBytesPerLine       =%d \n", pageHeader.cupsBytesPerLine);
    if(fDebugOutput) fprintf(outputFile,"cupsBitsPerPixel       =%d \n", pageHeader.cupsBitsPerPixel);
    if(fDebugOutput) fprintf(outputFile,"MediaClass             =%s \n", pageHeader.MediaClass);
    if(fDebugOutput) fprintf(outputFile,"MediaType              =%s \n", pageHeader.MediaType);
    if(fDebugOutput) fprintf(outputFile,"OutputType             =%s \n", pageHeader.OutputType);

      // @TBD
///// "form = " ????????????????????????
///// "tray = " pageHeader.MediaClass
///// "media = " pageHeader.MediaType
///// "resolution = " pageHeader.cupsCompression
///// "dither = " pageHeader.OutputType

//    pDevice = pfnNewDeviceWArgs (pszJobProperties, false);

//    if (!pDevice)
//    {
//       cerr << "Error: Could not instantiate a new device." << endl;
//
//       rc = 1;
//       goto done;
//    }

//      iNumScanLines = min (MAX_SCAN_LINES, (int)(pageHeader.cupsHeight - uiScanline));
      iNumColors    = 1 << pageHeaderCurrent.cupsBitsPerPixel;

      if(fDebugOutput) fprintf(outputFile,"Number of colors = %d \n", iNumColors);

      if (!pbmi2Bitmap)
      {
         // Allocate a full bitmap info 2 and bitmap header message
         iBytesToAllocate = sizeof (BITMAPINFO2);
         if (256 >= iNumColors)
            iBytesToAllocate += (iNumColors - 1) * sizeof (RGB2);

         pbmi2Bitmap = (PBITMAPINFO2)calloc (1, iBytesToAllocate);
         if (!pbmi2Bitmap)
            goto done;

         // Initialize it
         pbmi2Bitmap->cbFix         = sizeof (BITMAPINFO2)
                                    - sizeof (((PBITMAPINFO2)0)->argbColor);

         pbmi2Bitmap->cx            = pageHeaderCurrent.cupsWidth;
         pbmi2Bitmap->cPlanes       = 1;
         pbmi2Bitmap->cBitCount     = pageHeaderCurrent.cupsBitsPerPixel;
         pbmi2Bitmap->cclrUsed      = 1 << pbmi2Bitmap->cBitCount;
         pbmi2Bitmap->cclrImportant = 0;                            // All are important

         // Read the color table, if there is one
         // @TBD
      }

      icbBitmapScanLine = (((pbmi2Bitmap->cx) * pbmi2Bitmap->cBitCount + 31) >> 5) << 2;


      iPageNumber++;
      uiScanline = 0;

      if (1 == iPageNumber)
      {
         pDevice->beginJob ();
      }

      // get the hardware resolution

      pDevRes = pDevice->getCurrentResolution();

      hwRes.xRes      = pDevRes->getXRes();
      hwRes.yRes      = pDevRes->getYRes();
      hwRes.fScanDots = pDevRes->getScanlineMultiple();   // need head pass information



      // get the hardware page margins information
      // **** might not currently need this info
      pHCC = pDevice->getCurrentForm ()->getHardCopyCap ();

      if(iPageNumber == 1)
      {
          // need to calculate the correct bandsize needed for the device
          iNumScanLines = FindBandSize(pageHeaderCurrent.cupsHeight,
                                      pageHeaderCurrent.cupsWidth,
                                      24,       // always use 24 bits for source bitmap (for now)
                                      1,      // number of planes
                                      (unsigned)hwRes.fScanDots,
                                      4000*1024);      //eight meg buffer
      }

      if(fDebugOutput) fprintf(outputFile, "bandsize = %d \n", iNumScanLines);

      if (!pbBits)
      {
         iBytesToAllocate = icbBitmapScanLine * iNumScanLines;

         pbBits = (PBYTE)calloc (1, iBytesToAllocate);
         if (!pbBits)
         {
             if(fDebugOutput) fprintf(outputFile, "failed to allocate bitmap memory\n");
             goto done;
         }
      }

      if(fDebugOutput) fprintf(outputFile, "bytes per c - %d - us - %d \n", pageHeaderCurrent.cupsBytesPerLine, icbBitmapScanLine);

      uiScanline = pageHeaderCurrent.cupsHeight-1 ;

      rectlPageLocation.xLeft   = 0;
      rectlPageLocation.xRight  = pbmi2Bitmap->cx;

      while (uiScanline >= 0 )
      {

         if(fDebugOutput) fprintf(outputFile, "uiScanline = %u\n", uiScanline);
         if(fDebugOutput) fprintf(outputFile, "pbmi2Bitmap->cx = %d\n",pbmi2Bitmap->cx );

         pbmi2Bitmap->cy = iNumScanLines < uiScanline  ? iNumScanLines : uiScanline ;

         rectlPageLocation.yTop    = uiScanline;

         rectlPageLocation.yBottom = uiScanline - pbmi2Bitmap->cy;

         if(fDebugOutput) fprintf(outputFile, "pbmi2Bitmap->cy = %d\n",pbmi2Bitmap->cy );

         ulBytesRead = 0;

         ulBytesRead = cupsRasterReadPixels (pRasterStream,
                                             pbBits,
                                             (pageHeaderCurrent.cupsBytesPerLine)*pbmi2Bitmap->cy);

         if(fDebugOutput) fprintf(outputFile, "bytes to read = %ld \n", (long) (pageHeaderCurrent.cupsBytesPerLine*pbmi2Bitmap->cy));
         if(fDebugOutput) fprintf(outputFile, "bytes read    = %ld \n", ulBytesRead);

         if( ulBytesRead != (ULONG) (pageHeaderCurrent.cupsBytesPerLine*pbmi2Bitmap->cy))
         {
             cerr << "Error reading bitmap data" << endl;
             cerr << "read " << ulBytesRead <<  " bytes of raster data" << endl;
             cerr << "needed " << (pageHeaderCurrent.cupsBytesPerLine*pbmi2Bitmap->cy) << " bytes of raster data \n" << endl;
             break;
         }
                  // Test to see if CUPS had anymore data for us even though the page routine is called.
         if(fDebugOutput) fprintf(outputFile, "************** Page Number - %d \n", iPageNumber);
         if((iPageNumber != 1) && ulBytesRead && fTopOfBand)
         {
             fTopOfBand = false;
             pDevice->newFrame ();
             if(fDebugOutput) fprintf(outputFile, "doing new frame \n");

         }


         // Flip the band
//         for( i = 0 ; i < pbmi2Bitmap->cy ; i++)
//         {
//           memcpy((void *) &pbBitsToSend[i*icbBitmapScanLine],
//                  (void *) &pbBits[pageHeader.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)],
//                  pageHeader.cupsBytesPerLine);
//         }


         if(pageHeaderCurrent.cupsBitsPerPixel != 1)
         {
             // allocate memory to flip the band in
             pbBitsToSend = (PBYTE) malloc(icbBitmapScanLine*pbmi2Bitmap->cy );

             if(!pbBitsToSend)
             {
                 if(fDebugOutput) fprintf(outputFile, "failed to allocate flip bitmap memory\n");
                 goto done;
             }
             // flip and rotate BGR to RGB in the output band
             for( i = 0 ; i < pbmi2Bitmap->cy ; i++)
             {
               for(j = 0 ; j < (int) pageHeaderCurrent.cupsBytesPerLine-2; j+=3)
               {
                   int p = j+2;

                   pbBitsToSend[i*icbBitmapScanLine+j] =   // b --> r
                                   pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+p];
                   pbBitsToSend[i*icbBitmapScanLine+p] =   // r --> b
                                   pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+j];
                   pbBitsToSend[i*icbBitmapScanLine+j+1] =  // g --> g
                                   pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+j+1];

               }
             }

             if(fDebugOutput) fprintf(outputFile, "middle cy = %u\n", pbmi2Bitmap->cy);
             // Send it to the device

             uiScanline -= pbmi2Bitmap->cy;  // add in number of lines read
             if(fDoOutput)

                pDevice->rasterize (pbBitsToSend,
                                    pbmi2Bitmap,
                                    &rectlPageLocation,
                                    BITBLT_BITMAP);

             if(fDebugOutput) fprintf(outputFile, "middle2 cy = %u\n", pbmi2Bitmap->cy);

             if(pbBitsToSend)
             {
                 free(pbBitsToSend);
                 pbBitsToSend = 0;
             }
         }
         else
         {
             pbBitsToSend = (PBYTE) malloc(icbBitmapScanLine*pbmi2Bitmap->cy );

             if(!pbBitsToSend)
             {
                 if(fDebugOutput) fprintf(outputFile, "failed to allocate flip bitmap memory\n");
                 goto done;
             }
         for( i = 0 ; i < pbmi2Bitmap->cy ; i++)
         {
           memcpy((void *) &pbBitsToSend[i*icbBitmapScanLine],
                  (void *) &pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)],
                  pageHeaderCurrent.cupsBytesPerLine);
         }

             uiScanline -= pbmi2Bitmap->cy;  // add in number of lines read
             if(fDoOutput)

                pDevice->rasterize (pbBitsToSend,
                                    pbmi2Bitmap,
                                    &rectlPageLocation,
                                    BITBLT_BITMAP);
             if(pbBitsToSend)
             {
                 free(pbBitsToSend);
                 pbBitsToSend = 0;
             }
         }

         if(fDebugOutput) fprintf(outputFile, "End pbmi2Bitmap->cy = %d\n",pbmi2Bitmap->cy );
         if(fDebugOutput) fprintf(outputFile, "End uiScanline = %u\n", uiScanline);
         //uiScanline += pbmi2Bitmap->cy;  // add in number of lines read
         if(!uiScanline)
             break;

      }  // end while to read raster data

      fTopOfBand = true;
      if(fDebugOutput) fprintf(outputFile, "Through first loop\n");
   }  // end read of ppd file


   if(fDebugOutput) fprintf(outputFile, "dropping out\n");
   pDevice->endJob ();

done:
   if (ppd)
   {
      ppdClose (ppd);
      ppd = 0;
   }

   if (pRasterStream)
   {
      /*
       * Close the raster stream...
       */
      cupsRasterClose (pRasterStream);
      pRasterStream = 0;
   }

   if (fdInput != 0)
   {
      close (fdInput);
      fdInput = 0;
   }

   if (pbmi2Bitmap)
   {
      free (pbmi2Bitmap);
      pbmi2Bitmap = 0;
   }

   delete pDevice;

   if (hmodDevice)
   {
      g_module_close (hmodDevice);
      hmodDevice = 0;
   }

   if (0 == iPageNumber)
   {
      cerr << "ERROR: No pages found!" << endl;
   }
   else
   {
      cerr << "INFO: Ready to print." << endl;
   }
   if(fDebugOutput) fprintf(outputFile, "Closing output stream\n");
   if(fDebugOutput && outputFile) fclose(outputFile);

   Omni::terminate ();

   return 0;
}
