/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DeviceCollation.hpp"
#include "StringResource.hpp"
#include <stdio.h>
#include <string.h>

/* Function prototypes...
*/
static int    omniNameToID          (char             *pszId);
static int    standardNameToID      (char             *pszId);

#define STRING_OF(x) (apszCollationNames[x - DeviceCollation::COLLATION_UNLISTED])

static char *apszCollationNames[] = {
   "COLLATION_UNLISTED",
   "COLLATION_UNCOLLATED",
   "COLLATION_TRAY_COLLATION"
};

#ifdef USE_STANDARD_NAMES

static char *apszStandardCollationNames[] = {
   "unlisted",
   "uncollated",
   "tray-collation"
};

#endif

DeviceCollation::
DeviceCollation (Device     *pDevice,
                 int         id,
                 BinaryData *pbdData,
                 bool        fSimulationRequired)
{
   pDevice_d             = pDevice;
   id_d                  = id;
   pbdData_d             = pbdData;
   fSimulationRequired_d = fSimulationRequired;
}

DeviceCollation::
~DeviceCollation ()
{
   pDevice_d             = 0;
   id_d                  = 0;
   pbdData_d             = 0;
   fSimulationRequired_d = false;
}

char * DeviceCollation::
getName (Device *pDevice,
         int     id)
{
   return StringResource::getString (pDevice->getLanguageResource (),
                                     StringResource::STRINGGROUP_COLLATIONS,
                                     id);
}

char * DeviceCollation::
getName ()
{
   return StringResource::getString (pDevice_d->getLanguageResource (),
                                     StringResource::STRINGGROUP_COLLATIONS,
                                     id_d);
}

bool DeviceCollation::
isID (int id)
{
   return id == id_d;
}

int DeviceCollation::
getID ()
{
   return id_d;
}

char * DeviceCollation::
getDeviceID ()
{
   return 0;
}

BinaryData * DeviceCollation::
getData ()
{
   return pbdData_d;
}

bool DeviceCollation::
needsSimulation ()
{
   return fSimulationRequired_d;
}

DeviceCollation * DeviceCollation::
create (char *pszId)
{
   return this->createV (nameToID (pszId));
}

int DeviceCollation::
nameToID (char *pszId)
{
   if (  !pszId
      || !*pszId
      )
   {
      return COLLATION_UNLISTED;
   }

   int iCollationID = omniNameToID (pszId);

   if (COLLATION_UNLISTED != iCollationID)
      return iCollationID;

   iCollationID = standardNameToID (pszId);

   return iCollationID;
}

int
omniNameToID (char *pszId)
{
   static short int asiEntries[] = {
      DeviceCollation::COLLATION_TRAY_COLLATION,
      DeviceCollation::COLLATION_UNCOLLATED
   };

   int iLow  = 0;
   int iMid  = (int)dimof (asiEntries) / 2;
   int iHigh = (int)dimof (asiEntries) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, STRING_OF (asiEntries[iMid]));

      if (0 == iResult)
      {
         return asiEntries[iMid];
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceCollation::COLLATION_UNLISTED;
}

typedef struct _CollationMap {
   int   iOmniId;
   char *pszStandardName;
} COLLATIONMAP, *PCOLLATIONMAP;

static COLLATIONMAP vaStandardCollationMapping[] = {
   { DeviceCollation::COLLATION_TRAY_COLLATION, "tray-collation" },
   { DeviceCollation::COLLATION_UNCOLLATED,     "uncollated"     },
   { DeviceCollation::COLLATION_UNLISTED,       "unlisted"       }
};

int
standardNameToID (char *pszId)
{
   int   iLow    = 0;
   int   iMid    = (int)dimof (vaStandardCollationMapping) / 2;
   int   iHigh   = (int)dimof (vaStandardCollationMapping) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, vaStandardCollationMapping[iMid].pszStandardName);

      if (0 == iResult)
      {
         return vaStandardCollationMapping[iMid].iOmniId;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceCollation::COLLATION_UNLISTED;
}

char * DeviceCollation::
IDToName (int id)
{
   id -= COLLATION_UNLISTED;

   if (  0 <= id
      && id < (int)dimof (apszCollationNames)
      )
   {
#ifdef USE_STANDARD_NAMES
      return apszStandardCollationNames[id];
#else
      return apszCollationNames[id];
#endif
   }
   else
   {
      static char achUnknown[21];

      sprintf (achUnknown, "Unknown (%d)", id + COLLATION_UNCOLLATED);

      return achUnknown;
   }
}

string DeviceCollation::
toString (ostringstream& oss)
{
   oss << "{"
       << STRING_OF (id_d)
       << ", fSimulationRequired_d = " << fSimulationRequired_d
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const DeviceCollation& const_self)
{
   DeviceCollation& self = const_cast<DeviceCollation&>(const_self);
   ostringstream    oss;

   os << self.toString (oss);

   return os;
}
