/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DeviceDuplex.hpp"
#include "StringResource.hpp"
#include <stdio.h>
#include <string.h>

/* Function prototypes...
*/
static int    omniNameToID          (char             *pszId);
static int    standardNameToID      (char             *pszId);

#define STRING_OF(x) (apszDuplexNames[x - DeviceDuplex::DUPLEX_UNLISTED])

static char *apszDuplexNames[] = {
   "DUPLEX_UNLISTED",
   "DUPLEX_SIMPLEX",
   "DUPLEX_LONG_EDGE",
   "DUPLEX_SHORT_EDGE"
};

#ifdef USE_STANDARD_NAMES

static char *apszStandardDuplexNames[] = {
   "unlisted",
   "duplex-simplex",
   "duplex-long-edge",
   "duplex-short-edge"
};

#endif

DeviceDuplex::
DeviceDuplex (Device     *pDevice,
              int         id,
              BinaryData *pbdData,
              bool        fSimulationRequired)
{
   pDevice_d             = pDevice;
   id_d                  = id;
   pbdData_d             = pbdData;
   fSimulationRequired_d = fSimulationRequired;
}

DeviceDuplex::
~DeviceDuplex ()
{
   delete pbdData_d;

   pDevice_d             = 0;
   id_d                  = 0;
   pbdData_d             = 0;
   fSimulationRequired_d = false;
}

char * DeviceDuplex::
getName (Device *pDevice,
         int     id)
{
   return StringResource::getString (pDevice->getLanguageResource (),
                                     StringResource::STRINGGROUP_DUPLEXES,
                                     id);
}

char * DeviceDuplex::
getName ()
{
   return StringResource::getString (pDevice_d->getLanguageResource (),
                                     StringResource::STRINGGROUP_DUPLEXES,
                                     id_d);
}

bool DeviceDuplex::
isID (int id)
{
   return id == id_d;
}

int DeviceDuplex::
getID ()
{
   return id_d;
}

char * DeviceDuplex::
getDeviceID ()
{
   return 0;
}

BinaryData * DeviceDuplex::
getData ()
{
   return pbdData_d;
}

bool DeviceDuplex::
needsSimulation ()
{
   return fSimulationRequired_d;
}

DeviceDuplex * DeviceDuplex::
create (char *pszId)
{
   return this->createV (nameToID (pszId));
}

int DeviceDuplex::
nameToID (char *pszId)
{
   if (  !pszId
      || !*pszId
      )
   {
      return DUPLEX_UNLISTED;
   }

   int iDuplexID = omniNameToID (pszId);

   if (DUPLEX_UNLISTED != iDuplexID)
      return iDuplexID;

   iDuplexID = standardNameToID (pszId);

   return iDuplexID;
}

int
omniNameToID (char *pszId)
{
   static short int asiEntries[] = {
      DeviceDuplex::DUPLEX_LONG_EDGE,
      DeviceDuplex::DUPLEX_SHORT_EDGE,
      DeviceDuplex::DUPLEX_SIMPLEX
   };

   int iLow  = 0;
   int iMid  = (int)dimof (asiEntries) / 2;
   int iHigh = (int)dimof (asiEntries) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, STRING_OF (asiEntries[iMid]));

      if (0 == iResult)
      {
         return asiEntries[iMid];
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceDuplex::DUPLEX_UNLISTED;
}

typedef struct _DuplexMap {
   int   iOmniId;
   char *pszStandardName;
} DUPLEXMAP, *PDUPLEXMAP;

static DUPLEXMAP vaStandardDuplexMapping[] = {
   { DeviceDuplex::DUPLEX_LONG_EDGE,  "duplex-long-edge"       },
   { DeviceDuplex::DUPLEX_SHORT_EDGE, "duplex-short-edge"      },
   { DeviceDuplex::DUPLEX_SIMPLEX,    "duplex-simplex"         }
};

int
standardNameToID (char *pszId)
{
   int   iLow    = 0;
   int   iMid    = (int)dimof (vaStandardDuplexMapping) / 2;
   int   iHigh   = (int)dimof (vaStandardDuplexMapping) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, vaStandardDuplexMapping[iMid].pszStandardName);

      if (0 == iResult)
      {
         return vaStandardDuplexMapping[iMid].iOmniId;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceDuplex::DUPLEX_UNLISTED;
}

char * DeviceDuplex::
IDToName (int id)
{
   id -= DUPLEX_UNLISTED;

   if (  0 <= id
      && id < (int)dimof (apszDuplexNames)
      )
   {
#ifdef USE_STANDARD_NAMES
      return apszStandardDuplexNames[id];
#else
      return apszDuplexNames[id];
#endif
   }
   else
   {
      static char achUnknown[21];

      sprintf (achUnknown, "Unknown (%d)", id + DUPLEX_SIMPLEX);

      return achUnknown;
   }
}

string DeviceDuplex::
toString (ostringstream& oss)
{
   oss << "{"
       << STRING_OF (id_d)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const DeviceDuplex& const_self)
{
   DeviceDuplex& self = const_cast<DeviceDuplex&>(const_self);
   ostringstream oss;

   os << self.toString (oss);

   return os;
}
