/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DeviceNUp.hpp"
#include "StringResource.hpp"
#include <stdio.h>
#include <string.h>

/* Function prototypes...
*/
static int    omniNameToID          (char             *pszId);
static int    standardNameToID      (char             *pszId);

#define STRING_OF(x) (apszNUpNames[x - DeviceNUp::NUP_UNLISTED])

static char *apszNUpNames[] = {
   "NUP_UNLISTED",
   "NUP_1",
   "NUP_2_V",
   "NUP_2_H",
   "NUP_4",
   "NUP_6_V",
   "NUP_6_H",
   "NUP_9",
   "NUP_16"
};

#ifdef USE_STANDARD_NAMES

static char *apszStandardNUpNames[] = {
   "unlisted",
   "nup-1",
   "nup-2-v",
   "nup-2-h",
   "nup-4",
   "nup-6-v",
   "nup-6-h",
   "nup-9",
   "nup-16"
};

#endif

DeviceNUp::
DeviceNUp (Device     *pDevice,
           int         id,
           BinaryData *pbdData,
           bool        fSimulationRequired)
{
   pDevice_d             = pDevice;
   id_d                  = id;
   iHorizontalPages_d    = 0;
   iVerticalPages_d      = 0;
   pbdData_d             = pbdData;
   fSimulationRequired_d = fSimulationRequired;

   switch (id)
   {
   case NUP_1:   iHorizontalPages_d = 1; iVerticalPages_d = 1; break;
   case NUP_2_V: iHorizontalPages_d = 2; iVerticalPages_d = 1; break;
   case NUP_2_H: iHorizontalPages_d = 1; iVerticalPages_d = 2; break;
   case NUP_4:   iHorizontalPages_d = 2; iVerticalPages_d = 2; break;
   case NUP_6_V: iHorizontalPages_d = 2; iVerticalPages_d = 3; break;
   case NUP_6_H: iHorizontalPages_d = 3; iVerticalPages_d = 2; break;
   case NUP_9:   iHorizontalPages_d = 3; iVerticalPages_d = 3; break;
   case NUP_16:  iHorizontalPages_d = 4; iVerticalPages_d = 4; break;
   }
}

DeviceNUp::
~DeviceNUp ()
{
   delete pbdData_d;

   pDevice_d             = 0;
   id_d                  = 0;
   iHorizontalPages_d    = 0;
   iVerticalPages_d      = 0;
   pbdData_d             = 0;
   fSimulationRequired_d = false;
}

char * DeviceNUp::
getName (Device *pDevice,
         int     id)
{
   return StringResource::getString (pDevice->getLanguageResource (),
                                     StringResource::STRINGGROUP_NUPS,
                                     id);
}

char * DeviceNUp::
getName ()
{
   return StringResource::getString (pDevice_d->getLanguageResource (),
                                     StringResource::STRINGGROUP_NUPS,
                                     id_d);
}

int DeviceNUp::
getHorizontalPages ()
{
   return iHorizontalPages_d;
}

int DeviceNUp::
getVerticalPages ()
{
   return iVerticalPages_d;
}

bool DeviceNUp::
isID (int id)
{
   return id == id_d;
}

int DeviceNUp::
getID ()
{
   return id_d;
}

char * DeviceNUp::
getDeviceID ()
{
   return 0;
}

BinaryData * DeviceNUp::
getData ()
{
   return pbdData_d;
}

bool DeviceNUp::
needsSimulation ()
{
   return fSimulationRequired_d;
}

DeviceNUp * DeviceNUp::
create (char *pszId)
{
   return this->createV (nameToID (pszId));
}

int DeviceNUp::
nameToID (char *pszId)
{
   if (  !pszId
      || !*pszId
      )
   {
      return NUP_UNLISTED;
   }

   int iNUpID = omniNameToID (pszId);

   if (NUP_UNLISTED != iNUpID)
      return iNUpID;

   iNUpID = standardNameToID (pszId);

   return iNUpID;
}

int
omniNameToID (char *pszId)
{
   static short int asiEntries[] = {
      DeviceNUp::NUP_1,
      DeviceNUp::NUP_16,
      DeviceNUp::NUP_2_H,
      DeviceNUp::NUP_2_V,
      DeviceNUp::NUP_4,
      DeviceNUp::NUP_6_H,
      DeviceNUp::NUP_6_V,
      DeviceNUp::NUP_9
   };

   int iLow  = 0;
   int iMid  = (int)dimof (asiEntries) / 2;
   int iHigh = (int)dimof (asiEntries) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, STRING_OF (asiEntries[iMid]));

      if (0 == iResult)
      {
         return asiEntries[iMid];
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceNUp::NUP_UNLISTED;
}

typedef struct _NUpMap {
   int   iOmniId;
   char *pszStandardName;
} NUPMAP, *PNUPMAP;

static NUPMAP vaStandardNUpMapping[] = {
   { DeviceNUp::NUP_1,        "nup-1"    },
   { DeviceNUp::NUP_16,       "nup-16"   },
   { DeviceNUp::NUP_2_H,      "nup-2-h"  },
   { DeviceNUp::NUP_2_V,      "nup-2-v"  },
   { DeviceNUp::NUP_4,        "nup-4"    },
   { DeviceNUp::NUP_6_H,      "nup-6-h"  },
   { DeviceNUp::NUP_6_V,      "nup-6-v"  },
   { DeviceNUp::NUP_9,        "nup-9"    },
   { DeviceNUp::NUP_UNLISTED, "unlisted" }
};

int
standardNameToID (char *pszId)
{
   int   iLow    = 0;
   int   iMid    = (int)dimof (vaStandardNUpMapping) / 2;
   int   iHigh   = (int)dimof (vaStandardNUpMapping) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszId, vaStandardNUpMapping[iMid].pszStandardName);

      if (0 == iResult)
      {
         return vaStandardNUpMapping[iMid].iOmniId;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return DeviceNUp::NUP_UNLISTED;
}

char * DeviceNUp::
IDToName (int id)
{
   id -= NUP_UNLISTED;

   if (  0 <= id
      && id < (int)dimof (apszNUpNames)
      )
   {
#ifdef USE_STANDARD_NAMES
      return apszStandardNUpNames[id];
#else
      return apszNUpNames[id];
#endif
   }
   else
   {
      static char achUnknown[21];

      sprintf (achUnknown, "Unknown (%d)", id + NUP_1);

      return achUnknown;
   }
}

string DeviceNUp::
toString (ostringstream& oss)
{
   oss << "{"
       << STRING_OF (id_d)
       << ", iHorizontalPages_d = " << iHorizontalPages_d
       << ", iVerticalPages_d = " << iVerticalPages_d
       << ", fSimulationRequired_d = " << fSimulationRequired_d
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const DeviceNUp& const_self)
{
   DeviceNUp& self = const_cast<DeviceNUp&>(const_self);
   ostringstream  oss;

   os << self.toString (oss);

   return os;
}
