/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sstream>
#include <fstream>

#include "Device.hpp"
#include "Omni.hpp"
#include "omni2foo.hpp"

#include <sys/stat.h>
#include <sys/types.h>
#include <glib.h>
#include <gmodule.h>

#include <map>

string *
getCleanName (string *rawName, int operation)
{
   // Clean up Foomatic names to get valid file names/Foomatic IDs or
   // the correct make/model pair
   //
   // operation = 0: Return string suitable as Foomatic ID/XML file name
   // Mangle the device name to be Make-Model_Name
   // First space (after manufacturer's name) needs to be a -, and then
   // all spaces after that need to be a _
   // Ex:  Canon-BJC-2000, Epson-Stylus_Color_740
   //
   // operation = 1: Return Make and Model, separated by a "|", names
   // cleaned up from certain weirdnesses
   string device = *rawName;

   // Find the first space, ie, the gap between the make and model
   // and in case of operation 0 replace the " " with a "-", in case
   // of operation 1 with a "|"
   string::size_type pos = device.find_first_of (" ", 0);

   if (operation == 0)
   {
      device.replace (pos, 1, "-");
   }
   else
   {
      device.replace (pos,1,"|");
   }

   // The first word of the device name is not necessarily the
   // manufacturer's name, so determine the manufacturer here and
   // also do a clean-up of the model name
   string::size_type idx;

   if (device.find ("OkiP") == 0)
   {
      string make = "Okidata";

      // Replace the old Make/Model separator by a space and
      // build the output string according to the operation
      if (operation == 0)
      {
         if ((idx = device.find ("-")) != string::npos)
         {
            device.replace (idx, 1, " ");
         }

         device = make + "-" + device;
      }
      else
      {
         if ((idx = device.find ("|")) != string::npos)
         {
            device.replace (idx, 1, " ");
         }

         device = make + "|" + device;
      }
   }
   else if (device.find ("HP") == 0)
   {
      // Replace roman numbers in model names of HP LaserJet printers
      if ((idx = device.find ("III")) != string::npos)
      {
         device.replace (idx, 3, "3");
      }
      else if ((idx = device.find ("II")) != string::npos)
      {
         device.replace (idx, 2, "2");
      }
   }
   else if (device.find ("KS") == 0)
   {
      // "KS" is not really a manufacturer, but the name of a subdriver
      // of Omni for printers of several manufacturers.
      // Remove "KS "
      string s     = device.substr (3);
      string make;
      string model;

      if (  (  (s[0] >= '0')
            && (s[0] <= '9')
            )
         || (s.find ("KS") == 0)
         )
      {
         // Generic printers
         make = "Generic";

         if (s.find ("KS") == 0)
         {
            model = s.substr (3);
         }
         else
         {
            model = s;
         }

         if (model.find ("Generic") == 0)
         {
            model = model.substr (8);
         }
      }
      else if (s.find ("BJ") == 0)
      {
         // This is probably not the Canon BJ-230 because the "Canon"
         // subdriver already supports the Canon BJ-230
         make = "Others";
         model = s;
      }
      else if (s.find ("Gold Star") == 0)
      {
         make = "GoldStar";
         model = s.substr (10);
      }
      else if (  (s.find ("HDMF") == 0)
              || (s.find ("VP-") == 0)
              || (s.find ("LG ") == 0)
              )
      {
         make = "Others";
         model = s;
      }
      else
      {
         string::size_type idx = s.find_first_of (' ',0);

         make = s.substr (0, idx);
         make[0] = toupper (make[0]);

         for (unsigned int i = 1;
              i < make.size ();
              i ++)
         {
            make[i] = tolower (make[i]);
         }

         model = s.substr (idx+1);
      }

      if (operation == 0)
      {
         device = make + "-" + model;
      }
      else
      {
         device = make + "|" + model;
      }
   }

   // Remove trailing spaces
   while (device[device.size () - 1] == ' ')
   {
      device = device.substr (0, device.size () - 1);
   }

   if (operation == 0)
   {
      // Replace all slashes in the model name (breaks file names)
      while ((idx = device.find ("/")) != string::npos)
      {
         device.replace (idx, 1, " ");
      }

      // Replace "Plus", "PLUS", and "+" by "plus"
      while ((idx = device.find ("plus")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find ("Plus")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find ("PLUS")) != string::npos)
      {
         device.replace (idx, 4, "+");
      }

      while ((idx = device.find (" +")) != string::npos)
      {
         device.replace (idx, 2, "+");
      }

      while ((idx = device.find ("+")) != string::npos)
      {
         device.replace (idx, 1, "plus");
      }

      // Remove "(", ")", ",", and " "
      // thanks to jason@alteredminds.com for pointing out a bug
      while ((idx = device.find_first_of ("(), ")) != string::npos)
      {
         device.replace (idx, 1, "_");
      }

      // Remove trailing underscores
      while (device[device.size () - 1] == '_')
      {
         device = device.substr (0, device.size () - 1);
      }
   }

   return new string (device);
}

string *
getDeviceName (Device *pDevice, omni2foo *db, int operation)
{
   // operation = 0: Return Foomatic ID (also used as the XML file name)
   // operation = 1: Return Make and Model (for printer's DB entry, only
   //                if Foomatic ID does not match)
   //
   //                If a Foomatic ID of the printer exists, no printer
   //                XML file will be created because Foomatic provides
   //                already one (empty string returned in operation 1).
   //
   //                If a printer entry does not exist, Omni must provide
   //                one, here for the manufacturer and model entry the
   //                returned string of operation 1 is used. The returned
   //                string of operation 0 provides the Foomatic ID and
   //                the file name.
   string devName = pDevice->getDeviceName ();

   // check the db
   string cleanname = *getCleanName (&devName, 0);
   string id        = db->getFooIDFromString (cleanname);

   // if there is no match, the string that is returned is empty
   if (id.length () > 0)
   {
      if (operation == 0)
      {
         // operation = 0:
         // found a match, return the fooid string
         return new string (id);
      }
      else
      {
         // operation = 1:
         // match, return empty string
         return new string ("");
      }
   }
   else
   {
      // operation = 0:
      // no match, return Foomatic-compliant device name
      // Mangle the device name to be Make-Model_Name
      // First space (after manufacturer's name) needs to be a -, and then
      // all spaces after that need to be a _
      // Ex:  Canon-BJC-2000, Epson-Stylus_Color_740
      //
      // operation = 1:
      // Return cleaned up make and model names

      string device = *getCleanName (&devName, operation);

      return new string (device);
   }
}

// Derive command-line suitable shortnames out of the longnames. Most of
// them are also equal to the shortnames used by the Foomatic options of
// other drivers (as GIMP-Print, ljet4, ...)

// Remove spaces and '#' out of a C-style string, make letters after
// former spaces capital (as long as the character before the space is not a
// number)

const char *
generateShortName (const char *str)
{
   // Allocate memory of same length of original string
   char *result;

   result = (char *)malloc (sizeof (char)*(strlen (str)+1));

   int j          = 0;
   int afterspace = 0;
   int lastnumber = 0;

   for (unsigned int i = 0; i < strlen (str); i++)
   {
      if (  (str[i] != ' ')
         && (str[i] != '#')
	 && (str[i] != '-') // "-" in a short name breaks the Foomatic
	                    // filters
         )
      {
         if (afterspace)
         {
            result[j] = toupper (str[i]);
            afterspace = 0;
         }
         else
         {
            result[j] = str[i];
         }

         if (  (str[i] >= '0')
            && (str[i] <= '9')
            )
         {
            lastnumber = 1;
         }
         else
         {
            lastnumber = 0;
         }

         j++;
      }
      else
      {
         if (  (str[i] == ' ')
            && (!lastnumber)
            )
         {
            afterspace = 1;
         }
      }
   }

   result[j] = 0;

   return result;
}

bool
createDirectory (string   stringDirectoryName,
                 char    *pszProgramName)
{
   struct stat statFile;
   mode_t      mode       = S_IRUSR | S_IWUSR | S_IXUSR
                          | S_IRGRP | S_IWGRP | S_IXGRP
                          | S_IROTH | S_IXOTH;

   if (0 == stat (stringDirectoryName.c_str (), &statFile))
   {
      if (!S_ISDIR (statFile.st_mode))
      {
         cerr << pszProgramName << ": Error: " << stringDirectoryName << " is not a directory!" << endl;

         return false;
      }

      return true;
   }
   else
   {
      cout << pszProgramName << ": Warning: creating directory \"" << stringDirectoryName << "\"" << endl;

      if (0 == mkdir (stringDirectoryName.c_str (), mode))
      {
         return true;
      }
      else
      {
         return false;
      }
   }
}

int
main (int argc, char *argv[])
{
   typedef map <string, Device *> KeyList;

   typedef struct _OmniDeviceHandle {
      GModule *hmodDevice;
      Device  *pDevice;
      string  *pstringDeviceName;
      string  *pstringMakeModel;
   } OMNIHANDLE, *POMNIHANDLE;

   POMNIHANDLE   pohDevices     = 0;
   ofstream     *pofstream      = 0;
   string        basedir        = "foomatic-db/db/source";
   string        foomaticDB     = "/opt/Omni/data/foo2omni";
   bool          fUseLocalBuild = false;
   ostringstream oss;
   int           iArgOrder      = 200;
   KeyList       listOptions;
   int           iRc            = 3;

   Omni::initialize ();

   // Command Line parsing options:
   // -b : base dir, the location where the foomatic-db dir will be made
   // -d : the foomatic id db.  it is a space separated file, specify where that file is, or use the default
   // -l : use the build root
   for (int i = 1; i < argc; i++)
   {
      if (argv[i][0] == '-')
      {
         switch (argv[i][1])
         {
         case 'D':
         case 'd':
         {
            foomaticDB = argv[i + 1];
            break;
         }

         case 'b':
         case 'B':
         {
            basedir = argv[i + 1];
            break;
         }

         case 'l':
         case 'L':
         {
            fUseLocalBuild  = true;
            foomaticDB      = "bin/foo2omni";
            break;
         }
         }
      }
   }

   // Check basedir arg for trailing slash, if not present, add
   if (  basedir.length () > 0
      && basedir[basedir.length ()-1] != '/'
      )
      basedir = basedir + "/";

   // Make sure that the directories exist
   if (basedir.length () > 0)
   {
      string::size_type posCurrent = basedir.find ('/', 1);

      do
      {
         string stringDirectory;

         if (string::npos == posCurrent)
         {
            stringDirectory = basedir;
         }
         else
         {
            stringDirectory = basedir.substr (0, posCurrent);
         }

         if (!createDirectory (stringDirectory, argv[0]))
         {
            break;
         }

         posCurrent = basedir.find ('/', posCurrent + 1);

      } while (posCurrent != string::npos);
   }
   if (  !createDirectory (basedir + "printer", argv[0])
      || !createDirectory (basedir + "driver", argv[0])
      || !createDirectory (basedir + "opt", argv[0])
      )
   {
      return 1;
   }

   if (!g_module_supported ())
   {
      cerr << "This program needs glib's module routines!" << endl;

      return 2;
   }

   cout << argv[0] << ": using " << foomaticDB << endl;

   // Now make the omni2foo DB
   omni2foo DB (foomaticDB);

   cout << argv[0] << ": Counting devices..." << endl;

   Enumeration *pEnum          = Omni::listDevices (fUseLocalBuild);
   int          iNumDevices    = 0;
   int          iCurrentDevice = 0;

   oss.str (""); oss << basedir << "driver/omni.xml" << ends;
   ofstream ofstreamOmni (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-print-model.xml" << ends;
   ofstream ofstreamPrintModel (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-form.xml" << ends;
   ofstream ofstreamForm (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-tray.xml" << ends;
   ofstream ofstreamTray (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-media.xml" << ends;
   ofstream ofstreamInputSlots (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-resolution.xml" << ends;
   ofstream ofstreamResolutions (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-printmode.xml" << ends;
   ofstream ofstreamPrintModes (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-dither.xml" << ends;
   ofstream ofstreamDithers (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-orientation.xml" << ends;
   ofstream ofstreamOrientation (oss.str ().c_str ());
   oss.str (""); oss << basedir << "opt/omni-monodither.xml" << ends;
   ofstream ofstreamMonoDither (oss.str ().c_str ());

   while (pEnum->hasMoreElements ())
   {
      OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

      if (pOD)
      {
         delete pOD;

         iNumDevices++;
      }
   }

   delete pEnum;

   cout << argv[0] << ": There are " << iNumDevices << " devices." << endl;

   if (!iNumDevices)
   {
     cout << argv[0] << ": No devices found, aborting ..." << endl;

     delete[] pohDevices;

     exit (1);
   }

   pohDevices = new OMNIHANDLE [iNumDevices];
   pEnum      = Omni::listDevices (fUseLocalBuild);

   while (pEnum->hasMoreElements ())
   {
      OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

      if (pOD)
      {
         char              *pszLibName        = pOD->getLibraryName ();
         char              *pszJobProperties  = pOD->getJobProperties ();
         PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;

         pohDevices[iCurrentDevice].hmodDevice        = g_module_open (pszLibName, (GModuleFlags)0);
         pohDevices[iCurrentDevice].pDevice           = 0;
         pohDevices[iCurrentDevice].pstringDeviceName = 0;
         pohDevices[iCurrentDevice].pstringMakeModel  = 0;

         if (pohDevices[iCurrentDevice].hmodDevice)
         {
            g_module_symbol (pohDevices[iCurrentDevice].hmodDevice,
                             "newDeviceW_JopProp_Advanced",
                             (gpointer *)&pfnNewDeviceWArgs);

            if (pfnNewDeviceWArgs)
            {
               Device *pDevice = 0;

               pDevice = pfnNewDeviceWArgs (pszJobProperties, true);

               if (pDevice)
               {
                  pohDevices[iCurrentDevice].pDevice           = pDevice;
                  pohDevices[iCurrentDevice].pstringDeviceName = getDeviceName (pDevice, &DB, 0);
                  pohDevices[iCurrentDevice].pstringMakeModel  = getDeviceName (pDevice, &DB, 1);

                  string devName   = pDevice->getDeviceName ();
                  string cleanname = *getCleanName (&devName, 0);

                  cout << argv[0]
                       << ": Initialized "
                       << cleanname
                       << " ("
                       << iCurrentDevice + 1
                       << "/"
                       << iNumDevices
                       << ")"
                       << endl;
               }
               else
               {
                  cerr << "Warning: Could not instantiate " << pszLibName << " with \"" << pszJobProperties << "\"" << endl;
               }
            }
         }

         delete pOD;

         iCurrentDevice++;
      }
   }

   delete pEnum;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device          *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DevicePrintMode *pDPM              = pDevice->getCurrentPrintMode ();
      string          *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;
      string          *pstringMakeModel  = pohDevices[iCurrentDevice].pstringMakeModel;

      // Produce only printer XML files which Foomatic does not already
      // provide.

      if (pstringMakeModel->empty ())
      {
         cerr << argv[0]
              << ": Warning \""
              << pDevice->getDriverName ()
              << "."
              << pDevice->getDeviceName ()
              << "\" MakeModel is empty!"
              << endl;

         continue;
      }

      oss.str ("");
      oss << basedir
          << "printer/"
          << *pstringDeviceName
          << ".xml"
          << ends;

      ofstream  ofstreamPrinter (oss.str ().c_str ());
      pofstream = &ofstreamPrinter;

      cout << argv[0] << ": Generating " << oss.str () << endl;

      *pofstream << "<printer id=\"printer/" << *pstringDeviceName << "\">" << endl;

      if (pofstream->bad ())
      {
         cerr << argv[0] << ": Error: could not write to file." << endl;

         goto done;
      }

      string::size_type idx   = pstringMakeModel->find_first_of ('|',0);
      string            make  = pstringMakeModel->substr (0, idx);
      string            model = pstringMakeModel->substr (idx+1);

      *pofstream << "  <make>" << make << "</make>" << endl;
      *pofstream << "  <model>" << model << "</model>" << endl;
      *pofstream << "  <!-- " << pDevice->getShortName () << " -->" << endl;
      *pofstream << "  <mechanism>" << endl;

      // @TBD
      if (true)
      {
         // Is the default print mode monochrome? Then we don't have a color
         // printer
         char *pModeName = DevicePrintMode::IDToName (pDPM->getID ());

         if (  (!strstr(pModeName, "1_ANY"))
            && (!strstr(pModeName, "_K"))
            )
         {
            // Color printer
            *pofstream << "    <color/>" << endl;
         }
      }
      else
      {
         // Do we have color print modes? If yes, this is a color printer
         for (int iPrintMode = DevicePrintMode::PRINT_MODE_NONE;
              iPrintMode < DevicePrintMode::COUNT;
              iPrintMode++)
         {
            if (pDPM->isPrintModeSupported (iPrintMode))
            {
               char *pModeName = DevicePrintMode::IDToName (iPrintMode);

               // Go on when we have found a monochrome mode
               if (  (strstr(pModeName, "1_ANY"))
                  || (strstr(pModeName, "_K"))
                  )
                  continue;

               // We have found a color mode
               *pofstream << "    <color/>" << endl;
               break;
            }
         }
      }

      *pofstream << "    <resolution>" << endl;
      *pofstream << "      <dpi>" << endl;
      //  Added by raharper@us.ibm.com, get the resolution for each device
      *pofstream << "        <x>" << pDevice->getCurrentResolution ()->getXRes () << "</x>" << endl;
      *pofstream << "        <y>" << pDevice->getCurrentResolution ()->getYRes () << "</y>" << endl;
      *pofstream << "      </dpi>" << endl;
      *pofstream << "    </resolution>" << endl;
      *pofstream << "    <consumables>" << endl;
      //  The <consumables> section can be left blank if the consumables
      //  for this printer are not known.
      *pofstream << "      <!-- no \"comments\" -->" << endl;
      *pofstream << "    <!--one or more \"partno\" elements.--></consumables>" << endl;
      *pofstream << "  </mechanism>" << endl;
      //  Added by raharper@us.ibm.com, changed URL
      *pofstream << "  <url>http://www.sourceforge.net/projects/omniprint</url>" << endl;
      *pofstream << "  <lang>" << endl;
      *pofstream << "    <text>" << endl;
      *pofstream << "      <charset>us-ascii</charset>" << endl;
      *pofstream << "    </text>" << endl;
      *pofstream << "  </lang>" << endl;
      *pofstream << "  <autodetect>" << endl;
      *pofstream << "    <!--no known parport probe information-->" << endl;
      *pofstream << "  </autodetect>" << endl;
      *pofstream << "  <functionality>D</functionality>" << endl;
      *pofstream << "  <driver>omni</driver>" << endl;
      *pofstream << "  <unverified/><!--no \"contrib_url\"-->" << endl;
      *pofstream << "  <comments>" << endl;
      *pofstream << "    <en>" << endl;
      *pofstream << "    Foomatic entry automatically generated by Omni.&lt;p&gt;" << endl;
      *pofstream << "    Printer type and maximum resolution information" << endl;
      *pofstream << "    may be incorrect.&lt;p&gt;" << endl;
      *pofstream << "    The rating &quot;Partially&quot; is given due to the Omni code still" << endl;
      *pofstream << "    being under development and not supporting all features of the" << endl;
      *pofstream << "    supported printer models.&lt;p&gt;" << endl;
      *pofstream << "    </en>" << endl;
      *pofstream << "  </comments>" << endl;
      *pofstream << "</printer>" << endl;
   }

   /**
    * Create Omni
    *
    * list of device names
    */
   pofstream = &ofstreamOmni;

   cout << argv[0] << ": Generating " << basedir << "driver/omni.xml" << endl;

   *pofstream << "<driver id=\"driver/omni\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << " <name>omni</name>" << endl;
   *pofstream << " <url>http://www.ibm.com/linux/ltc/projects/omni/</url>" << endl;
   *pofstream << " <execution>" << endl;
   *pofstream << "  <ghostscript />" << endl;
   // Edited by raharper@us.ibm.com, changed the prototype to what omni wants it to look like
   *pofstream << "  <prototype>gs -q -dPARANOIDSAFER -dBATCH -dNOPAUSE -sOutputFile=- -sDEVICE=omni%B -sproperties=&quot;%A&quot;%Z -</prototype>" << endl; // @TBD
   *pofstream << " </execution>" << endl;
   *pofstream << " <comments>" << endl;
   *pofstream << "  <en>" << endl;
   *pofstream << endl;
   *pofstream << "    &lt;b&gt;";
   *pofstream << "This page provides data for the Omni driver version ";
   if (0 < iNumDevices)
   {
      Device *pDevice = pohDevices[0].pDevice;

      *pofstream << pDevice->getVersion ();
   }
   else
   {
      *pofstream << "Unknown";
   }
   *pofstream << "." << endl;
   *pofstream << "    &lt;/b&gt;&lt;p&gt;" << endl;
   *pofstream << endl;
   *pofstream << "    The Omni printer driver provides support for over 400 printers " << endl;
   *pofstream << "    using the Ghostscript framework. In addition, it provides a " << endl;
   *pofstream << "    model for dynamically loading printer drivers, creating new " << endl;
   *pofstream << "    devices by editing device description files, and simplifies " << endl;
   *pofstream << "    new printer driver development by allowing for the subclassing " << endl;
   *pofstream << "    of previous device features.&lt;p&gt;" << endl;
   *pofstream << endl;
   *pofstream << "  </en>" << endl;
   *pofstream << " </comments>" << endl;
   *pofstream << " <printers>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      if (!pohDevices[iCurrentDevice].pstringDeviceName)
      {
         // Error!
         continue;
      }

      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "  <printer><id>printer/" << *pstringDeviceName << "</id></printer>" << endl;
   }

   *pofstream << " </printers>" << endl;
   *pofstream << "</driver>" << endl;

   /**
    * Create Print Model
    *
    * list of device names
    */
   pofstream = &ofstreamPrintModel;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-print-model.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-print-model\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Printer Model</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Model</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>10</arg_order>" << endl;
   *pofstream << "   <arg_spot>B</arg_spot>" << endl;
   *pofstream << "   <arg_required />" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto> -sDeviceName=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << "" << endl;
   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      if (!pohDevices[iCurrentDevice].pstringDeviceName)
      {
         // Error!
         continue;
      }

      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "      <arg_defval>" << *pstringDeviceName << "</arg_defval>" << endl; // @TBD
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      //raharper@us.ibm.com took out the ev/omni-
      *pofstream << "    <enum_val id='" << *pstringDeviceName << "'>" << endl;
      *pofstream << "      <ev_longname><en>" << *pstringDeviceName << "</en></ev_longname>" << endl;
      *pofstream << "      <ev_shortname><en>" << pDevice->getShortName () << "</en></ev_shortname>" << endl;
      *pofstream << "      <ev_driverval>" << pDevice->getShortName () << "</ev_driverval>" << endl;
      *pofstream << "      <constraints>" << endl;
      *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
      *pofstream << "        <constraint sense='false'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "        </constraint>" << endl;
      *pofstream << "        <!-- ...except to these: -->" << endl;
      *pofstream << "        <constraint sense='true'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "        </constraint>" << endl;
      *pofstream << "      </constraints>" << endl;
      *pofstream << "    </enum_val>" << endl;
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Paper Sizes (Forms)
    *
    * list of forms
    */
   // raharper@us.ibm.com, changed filename to form
   pofstream = &ofstreamForm;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-form.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-form\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Page Size</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>PageSize</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>90</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_required />" << endl;
   // raharper@us.ibm.com, changed the arg_xxxx type to subsitutiuon to get the style=C in the perl dat structure generator, db->getdat ()
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the prototype to what omni uses for commandline options
   *pofstream << "   <arg_proto> form=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DeviceForm *pDF               = pDevice->getCurrentForm ();
      string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing the arg_defval output
      *pofstream << "      <arg_defval>" << DeviceForm::IDToName (pDF->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iForm = DeviceForm::FORM_NONE; iForm < DeviceForm::COUNT; iForm++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DeviceForm *pDF               = pDevice->getCurrentForm ();
         string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDF->isFormSupported (iForm))
         {
            if (fFirstTime)
            {
               char        *pcharLongName;
               const char  *pcharShortName;
               DeviceForm  *pDFDevice       = 0;
               HardCopyCap *pHCCDevice;
               int          iCx;
               int          iCy;

               fFirstTime = false;

               pDFDevice  = pDF->create (DeviceForm::IDToName (iForm));
               pHCCDevice = pDFDevice->getHardCopyCap ();

               // Generate the short name out of the long name
               pcharLongName  = pDFDevice->getName ();
               pcharShortName = generateShortName (pcharLongName);

               // @TBD user defined form is (0, 0) in size
               iCx = (int)(((double)pHCCDevice->getCx ()) / 25400.0 * 72.0 + 0.5);
               iCy = (int)(((double)pHCCDevice->getCy ()) / 25400.0 * 72.0 + 0.5);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceForm::IDToName (iForm) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceForm::IDToName (iForm) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
               delete pDFDevice;
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Input Slots
    *
    * list of trays
    */
   // raharper@us.ibm.com, changed file name to tray
   pofstream = &ofstreamTray;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-tray.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-tray\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Media Source</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>InputSlot</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> tray=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DeviceTray *pDT               = pDevice->getCurrentTray ();
      string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceTray::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iTray = DeviceTray::TRAY_NONE; iTray < DeviceTray::COUNT; iTray++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device     *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DeviceTray *pDT               = pDevice->getCurrentTray ();
         string     *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isTraySupported (iTray))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceTray::getName (pDevice, iTray);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceTray::IDToName (iTray) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceTray::IDToName (iTray) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create MediaTypes
    *
    * list of medias
    */
   // raharper@us.ibm.com, Added Media section
   pofstream = &ofstreamInputSlots;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-media.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-media\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Media Type</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>MediaType</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> media=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device      *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DeviceMedia *pDT               = pDevice->getCurrentMedia ();
      string      *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceMedia::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iMedia = DeviceMedia::MEDIA_NONE; iMedia < DeviceMedia::COUNT; iMedia++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device      *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DeviceMedia *pDT               = pDevice->getCurrentMedia ();
         string      *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isMediaSupported (iMedia))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceMedia::getName (pDevice, iMedia);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceMedia::IDToName (iMedia) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceMedia::IDToName (iMedia) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Resolution
    *
    * list of resolutions
    */
   pofstream = &ofstreamResolutions;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-resolution.xml" << endl;

   // raharper@us.ibm.com, changed GSResolution to Resolution
   *pofstream << "<option type=\"enum\" id=\"opt/omni-resolution\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Resolution</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Resolution</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // edited by raharper@us.ibm.com, changed output to match omni driver options
   *pofstream << "   <arg_proto> resolution=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device           *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DeviceResolution *pDR               = pDevice->getCurrentResolution ();
      string           *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed the defval output
      *pofstream << "      <arg_defval>" << DeviceResolution::IDToName (pDR->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iResolution = DeviceResolution::RESOLUTION_NONE; iResolution < DeviceResolution::COUNT; iResolution++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device           *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DeviceResolution *pDR               = pDevice->getCurrentResolution ();
         string           *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDR->isResolutionSupported (iResolution))
         {
            if (fFirstTime)
            {
               char       *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName = DeviceResolution::getName (pDevice, iResolution);

               if (pcharLongName == NULL)
               {
                  pcharLongName = strdup (DeviceResolution::IDToName (iResolution));
                  pcharLongName += 11;
                  for (unsigned int i = 1;
                       i < strlen (pcharLongName);
                       i ++)
                  {
                     pcharLongName[i] = tolower (pcharLongName[i]);
                  }
               }

               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceResolution::IDToName (iResolution) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceResolution::IDToName (iResolution) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Ink Type
    *
    * list of printmodes
    */
   // raharper@us.ibm.com changed filename to printmode
   pofstream = &ofstreamPrintModes;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-printmode.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-inktype\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // raharper@us.ibm.com, changed longname
   *pofstream << "   <en>Printing Mode</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Mode</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // raharper@us.ibm.com, changed prototype
   *pofstream << "   <arg_proto> printmode=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device          *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DevicePrintMode *pDR               = pDevice->getCurrentPrintMode ();
      string          *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed defval
      *pofstream << "      <arg_defval>" << DevicePrintMode::IDToName (pDR->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iPrintMode = DevicePrintMode::PRINT_MODE_NONE; iPrintMode < DevicePrintMode::COUNT; iPrintMode++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device          *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DevicePrintMode *pDPM              = pDevice->getCurrentPrintMode ();
         string          *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDPM->isPrintModeSupported (iPrintMode))
         {
            if (fFirstTime)
            {
               const char *pcharLongName;
               const char *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DevicePrintMode::getName (pDevice, iPrintMode);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DevicePrintMode::IDToName (iPrintMode) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DevicePrintMode::IDToName (iPrintMode) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Create Image Type
    *
    * list of dithers
    */
   pofstream = &ofstreamDithers;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-dither.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-dither\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Dither Algorithm</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>Dither</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // raharper@us.ibm.com, changed prototype
   *pofstream << "   <arg_proto> dither=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changed defval
      *pofstream << "      <arg_defval>" << pDevice->getCurrentDitherID () << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   pEnum = DeviceDither::getEnumeration ();
   while (pEnum->hasMoreElements ())
   {
      Device     *pDevice        = 0;
      char       *pszID          = (char *)pEnum->nextElement ();
      const char *pcharLongName;
      const char *pcharShortName;
      int         i;

      for (i = 0; !pohDevices[i].pDevice && i < iNumDevices; i++)
         ;
      if (i < iNumDevices)
      {
         pDevice = pohDevices[i].pDevice;
      }
      else
      {
         break;
      }

      // Generate the short name out of the long name
      pcharLongName  = DeviceDither::getName (pDevice, pszID);
      pcharShortName = generateShortName (pcharLongName);

      //raharper@us.ibm.com took out the ev/omni-
      *pofstream << "    <enum_val id='" << pszID << "'>" << endl;
      *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
      *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
      *pofstream << "      <ev_driverval>" << pszID << "</ev_driverval>" << endl;
      *pofstream << "      <constraints>" << endl;
      *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
      *pofstream << "        <constraint sense='false'>" << endl;
      *pofstream << "          <driver>omni</driver>" << endl;
      *pofstream << "        </constraint>" << endl;
      free ((void *)pcharShortName);

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         if (!pohDevices[iCurrentDevice].pstringDeviceName)
         {
            // Error!
            continue;
         }

         string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         *pofstream << "        <constraint sense='true'>" << endl;
         *pofstream << "          <driver>omni</driver>" << endl;
         *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
         *pofstream << "        </constraint>" << endl;
      }

      *pofstream << "      </constraints>" << endl;
      *pofstream << "    </enum_val>" << endl;
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /**
    * Make Orientation section
    *
    * List orientations
    */
   pofstream = &ofstreamOrientation;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-orientation.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-orientation\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Orientation</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <en>Orientation</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   // Edited by raharper@us.ibm.com, changing the output to match omni driver options
   *pofstream << "   <arg_proto> orientation=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      DeviceOrientation *pDT               = pDevice->getCurrentOrientation ();
      string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      // raharper@us.ibm.com, changing what the defval output looks like
      *pofstream << "      <arg_defval>" << DeviceOrientation::IDToName (pDT->getID ()) << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;

   for (int iOrient = DeviceOrientation::ORIENTATION_NONE; iOrient < DeviceOrientation::COUNT; iOrient++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;

         if (!pDevice)
         {
            // Error!
            continue;
         }

         DeviceOrientation *pDT               = pDevice->getCurrentOrientation ();
         string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (pDT->isOrientationSupported (iOrient))
         {
            if (fFirstTime)
            {
               const char     *pcharLongName;
               const char     *pcharShortName;

               fFirstTime = false;

               // Generate the short name out of the long name
               pcharLongName  = DeviceOrientation::getName (pDevice, iOrient);
               pcharShortName = generateShortName (pcharLongName);

               //raharper@us.ibm.com took out the ev/omni-
               *pofstream << "    <enum_val id='" << DeviceOrientation::IDToName (iOrient) << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << pcharLongName << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << pcharShortName << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << DeviceOrientation::IDToName (iOrient) << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;

               free ((void *)pcharShortName);
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;
         }
      }

      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   // Make an option to turn on or off GhostScript dithering. It will apply
   // to all printer models ("-smonodither=GSMONO")

   pofstream = &ofstreamMonoDither;;

   cout << argv[0] << ": Generating " << basedir << "opt/omni-monodither.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-monodither\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>Monochrome Dithering</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>MonoDither</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>90</arg_order>" << endl;
   *pofstream << "   <arg_spot>B</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto>%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   *pofstream << "  <constraints>" << endl;
   *pofstream << "    <constraint sense='true'>" << endl;
   *pofstream << "      <driver>omni</driver>" << endl;
   *pofstream << "      <arg_defval>MONODITHER_GS</arg_defval>" << endl;
   *pofstream << "    </constraint>" << endl;
   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;
   *pofstream << "    <enum_val id='MONODITHER_GS'>" << endl;
   *pofstream << "      <ev_longname><en>GhostScript (Faster)</en></ev_longname>" << endl;
   *pofstream << "      <ev_shortname><en>GS</en></ev_shortname>" << endl;
   *pofstream << "      <ev_driverval> -smonodither=GSMONO</ev_driverval>" << endl;
   *pofstream << "    </enum_val>" << endl;
   *pofstream << "    <enum_val id='MONODITHER_OMNI'>" << endl;
   *pofstream << "      <ev_longname><en>Omni (Better quality)</en></ev_longname>" << endl;
   *pofstream << "      <ev_shortname><en>Omni</en></ev_shortname>" << endl;
   *pofstream << "      <ev_driverval></ev_driverval>" << endl;
   *pofstream << "    </enum_val>" << endl;
   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

   /*
    * printer-model-specific options
    */
   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device *pDevice = pohDevices[iCurrentDevice].pDevice;

      if (!pDevice)
      {
         // Error!
         continue;
      }

      // Only enumerate the device-specific extra options
      Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();

      while (pEnum->hasMoreElements ())
      {
         char *pszKey = (char *)pEnum->nextElement ();

         listOptions[string (pszKey)] = pDevice;
      }

      delete pEnum;
   }

#if 0 // @HACK
     unlink("foomatic-db/db/source/opt/omni-hardwarescaling.xml");
     unlink("foomatic-db/db/source/opt/omni-bidirectional.xml");
     unlink("foomatic-db/db/source/opt/omni-journal.xml");
     unlink("foomatic-db/db/source/opt/omni-cutmode.xml");
#endif

#if 1
     /* This is a generic solution for device specific options
     **
     */
     // Now build the option entries
     for (KeyList::iterator next = listOptions.begin ();
          next != listOptions.end ();
          next++)
     {
        Device *pDevice = 0;
        char   *pszKey  = 0;

        pszKey  = (char *)next->first.c_str ();
        pDevice = next->second;

        cout << argv[0] << ": Extra Options: " << pszKey << endl;

        oss.str ("");
        oss << basedir
            << "opt/omni-"
            << pszKey
            << ".xml"
            << ends;

        ofstream ofstreamOption (oss.str ().c_str ());

        pofstream = &ofstreamOption;

        cout << argv[0] << ": Generating " << oss.str () << endl;

        *pofstream << "<option type=\"enum\" id=\"opt/omni-" << pszKey << "\">" << endl;

        if (pofstream->bad ())
        {
           cerr << argv[0] << ": Error: could not write to file." << endl;

           goto done;
        }

        *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
        *pofstream << "       it should be treated as documentation for the user. -->" << endl;
        *pofstream << "  <arg_longname>" << endl;

////////cout << "Trying " << pDevice->getDriverName () << "." << pDevice->getDeviceName () << endl;

        string *pRet = pDevice->translateKeyValue (pszKey, 0);

        *pofstream << "   <en>" << *pRet << "</en>" << endl;

        delete pRet;

        *pofstream << "  </arg_longname>" << endl;
        *pofstream << "  <arg_shortname>" << endl;
        *pofstream << "   <en>" << pszKey << "</en><!-- backends only know <en> shortnames! -->" << endl;
        *pofstream << "  </arg_shortname>" << endl;
        *pofstream << "  <arg_execution>" << endl;
        *pofstream << "   <arg_order>" << iArgOrder << "</arg_order>" << endl;
        *pofstream << "   <arg_spot>A</arg_spot>" << endl;
        *pofstream << "   <arg_substitution />" << endl;
        *pofstream << "   <arg_proto> " << pszKey << "=%s</arg_proto>" << endl;
        *pofstream << "  </arg_execution>" << endl;
        *pofstream << endl;

        // Put in the printers to which this option applies
        *pofstream << "  <constraints>" << endl;

        for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
        {
           pDevice = pohDevices[iCurrentDevice].pDevice;

           if (!pDevice)
           {
              // Error!
              continue;
           }

           // Check whether this option is under the device-specific extra
           // options
           pRet = pDevice->getJobProperty (pszKey);

           if (pRet)
           {
              string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

              *pofstream << "    <constraint sense='true'>" << endl;
              *pofstream << "      <driver>omni</driver>" << endl;
              *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
              *pofstream << "      <arg_defval>" << *pRet << "</arg_defval>" << endl;
              *pofstream << "    </constraint>" << endl;

              delete pRet;
           }
        }

        *pofstream << "  </constraints>" << endl;
        *pofstream << "  <enum_vals>" << endl;

        pDevice = next->second;

        Enumeration *pEnum = pDevice->listKeyValues (pszKey);

        while (pEnum->hasMoreElements ())
        {
           char *pszValue = (char *)pEnum->nextElement ();

           *pofstream << "    <enum_val id='" << pszValue << "'>" << endl;

           pRet = pDevice->translateKeyValue (pszKey, pszValue);

           if (pRet)
           {
              string::size_type idx;

              idx = pRet->find ("=");

              if (string::npos != idx)
              {
                 *pofstream << "      <ev_longname><en>" << pRet->substr (idx + 1) << "</en></ev_longname>" << endl;
              }
           }

           delete pRet;

           *pofstream << "      <ev_shortname><en>" << pszValue << "</en></ev_shortname>" << endl;
           *pofstream << "      <ev_driverval>" << pszValue << "</ev_driverval>" << endl;
           *pofstream << "    </enum_val>" << endl;
        }

        delete pEnum;

        *pofstream << "  </enum_vals>" << endl;
        *pofstream << "</option>" << endl;

        iArgOrder += 10;
     }

#else

   cout << argv[0] << ": Extra Options: ";

   // Now build the option entries
   for (KeyList::iterator next = listOptions.begin ();
        next != listOptions.end ();
        next++)
   {
      cout << *next << " ";

      if (0 == strcmp (next->c_str (), "HardwareScaling"))
      {
         ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-hardwarescaling.xml");

         pofstream = &ofstreamOption;

         string OptionLongname ("Hardware Scaling");
         string OptionShortname ("HardwareScaling");
         string prototype ("HardwareScaling");

         cout << argv[0] << ": Generating omni-hardwarescaling.xml" << endl;

         *pofstream << "<option type=\"enum\" id=\"opt/omni-hardwarescaling\">" << endl;

         if (pofstream->bad ())
         {
            cerr << argv[0] << ": Error: could not write to file." << endl;

            goto done;
         }

         *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
         *pofstream << "       it should be treated as documentation for the user. -->" << endl;
         *pofstream << "  <arg_longname>" << endl;
         *pofstream << "   <en>" << OptionLongname << "</en>" << endl;
         *pofstream << "  </arg_longname>" << endl;
         *pofstream << "  <arg_shortname>" << endl;
         *pofstream << "   <en>" << OptionShortname << "</en><!-- backends only know <en> shortnames! -->" << endl;
         *pofstream << "  </arg_shortname>" << endl;
         *pofstream << "  <arg_execution>" << endl;
         *pofstream << "   <arg_order>200</arg_order>" << endl;
         *pofstream << "   <arg_spot>A</arg_spot>" << endl;
         *pofstream << "   <arg_substitution />" << endl;
         *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
         *pofstream << "  </arg_execution>" << endl;
         *pofstream << endl;

         // Put in the printers to which this option applies
         *pofstream << "  <constraints>" << endl;

         for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
         {
            Device *pDevice = pohDevices[iCurrentDevice].pDevice;

            // Check whether this option is under the device-specific extra
            // options
            Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (!pszKey)
                  break;

               // Add appropriate constraint
               if (strcmp(pszKey, "HardwareScaling") == 0)
               {
                  string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

                  *pofstream << "    <constraint sense='true'>" << endl;
                  *pofstream << "      <driver>omni</driver>" << endl;
                  *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
                  *pofstream << "      <arg_defval>" << "HardwareScaling_1" << "</arg_defval>" << endl;
                  *pofstream << "    </constraint>" << endl;
                  break;
               }
            }
            delete pEnum;
         }

         *pofstream << "  </constraints>" << endl;
         *pofstream << "  <enum_vals>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_1'>" << endl;
         *pofstream << "      <ev_longname><en>1</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>1</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>1</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_2'>" << endl;
         *pofstream << "      <ev_longname><en>2</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>2</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>2</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_3'>" << endl;
         *pofstream << "      <ev_longname><en>3</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>3</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>3</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_4'>" << endl;
         *pofstream << "      <ev_longname><en>4</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>4</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>4</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_5'>" << endl;
         *pofstream << "      <ev_longname><en>5</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>5</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>5</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_6'>" << endl;
         *pofstream << "      <ev_longname><en>6</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>6</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>6</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_10'>" << endl;
         *pofstream << "      <ev_longname><en>10</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>10</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>10</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_12'>" << endl;
         *pofstream << "      <ev_longname><en>12</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>12</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>12</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_15'>" << endl;
         *pofstream << "      <ev_longname><en>15</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>15</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>15</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_20'>" << endl;
         *pofstream << "      <ev_longname><en>20</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>20</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>20</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_25'>" << endl;
         *pofstream << "      <ev_longname><en>25</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>25</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>25</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_30'>" << endl;
         *pofstream << "      <ev_longname><en>30</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>30</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>30</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_50'>" << endl;
         *pofstream << "      <ev_longname><en>50</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>50</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>50</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_60'>" << endl;
         *pofstream << "      <ev_longname><en>60</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>60</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>60</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='HardwareScaling_100'>" << endl;
         *pofstream << "      <ev_longname><en>100</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>100</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>100</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "  </enum_vals>" << endl;
         *pofstream << "</option>" << endl;
      }
      else if (0 == strcmp (next->c_str (), "bidirectional"))
      {
         ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-bidirectional.xml");
         pofstream = &ofstreamOption;

         string OptionLongname ("Bi-Directional Printing");
         string OptionShortname ("Bidirectional");
         string prototype ("bidirectional");

         cout << argv[0] << ": Generating omni-bidirectional.xml" << endl;

         *pofstream << "<option type=\"enum\" id=\"opt/omni-bidirectional\">" << endl;

         if (pofstream->bad ())
         {
            cerr << argv[0] << ": Error: could not write to file." << endl;

            goto done;
         }

         *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
         *pofstream << "       it should be treated as documentation for the user. -->" << endl;
         *pofstream << "  <arg_longname>" << endl;
         *pofstream << "   <en>" << OptionLongname << "</en>" << endl;
         *pofstream << "  </arg_longname>" << endl;
         *pofstream << "  <arg_shortname>" << endl;
         *pofstream << "   <en>" << OptionShortname << "</en><!-- backends only know <en> shortnames! -->" << endl;
         *pofstream << "  </arg_shortname>" << endl;
         *pofstream << "  <arg_execution>" << endl;
         *pofstream << "   <arg_order>210</arg_order>" << endl;
         *pofstream << "   <arg_spot>A</arg_spot>" << endl;
         *pofstream << "   <arg_substitution />" << endl;
         *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
         *pofstream << "  </arg_execution>" << endl;
         *pofstream << endl;

         // Put in the printers to which this option applies
         *pofstream << "  <constraints>" << endl;

         for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
         {
            Device *pDevice = pohDevices[iCurrentDevice].pDevice;

            // Check whether this option is under the device-specific extra
            // options
            Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (!pszKey)
                  break;

               // Add appropriate constraint
               if (strcmp(pszKey, "bidirectional") == 0)
               {
                  string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

                  *pofstream << "    <constraint sense='true'>" << endl;
                  *pofstream << "      <driver>omni</driver>" << endl;
                  *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
                  *pofstream << "      <arg_defval>" << "bidirectional_on" << "</arg_defval>" << endl;
                  *pofstream << "    </constraint>" << endl;
                  break;
               }
            }

            delete pEnum;
         }

         *pofstream << "  </constraints>" << endl;
         *pofstream << "  <enum_vals>" << endl;
         *pofstream << "    <enum_val id='bidirectional_off'>" << endl;
         *pofstream << "      <ev_longname><en>Off</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>Off</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>false</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='bidirectional_on'>" << endl;
         *pofstream << "      <ev_longname><en>On</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>On</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>true</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "  </enum_vals>" << endl;
         *pofstream << "</option>" << endl;
      }
      else if (0 == strcmp (next->c_str (), "journal"))
      {
         ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-journal.xml");
         pofstream = &ofstreamOption;

         string OptionLongname ("Journal Printing");
         string OptionShortname ("Journal");
         string prototype ("journal");

         cout << argv[0] << ": Generating omni-journal.xml" << endl;

         *pofstream << "<option type=\"enum\" id=\"opt/omni-journal\">" << endl;

         if (pofstream->bad ())
         {
            cerr << argv[0] << ": Error: could not write to file." << endl;

            goto done;
         }

         *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
         *pofstream << "       it should be treated as documentation for the user. -->" << endl;
         *pofstream << "  <arg_longname>" << endl;
         *pofstream << "   <en>" << OptionLongname << "</en>" << endl;
         *pofstream << "  </arg_longname>" << endl;
         *pofstream << "  <arg_shortname>" << endl;
         *pofstream << "   <en>" << OptionShortname << "</en><!-- backends only know <en> shortnames! -->" << endl;
         *pofstream << "  </arg_shortname>" << endl;
         *pofstream << "  <arg_execution>" << endl;
         *pofstream << "   <arg_order>220</arg_order>" << endl;
         *pofstream << "   <arg_spot>A</arg_spot>" << endl;
         *pofstream << "   <arg_substitution />" << endl;
         *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
         *pofstream << "  </arg_execution>" << endl;
         *pofstream << endl;

         // Put in the printers to which this option applies
         *pofstream << "  <constraints>" << endl;

         for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
         {
            Device *pDevice = pohDevices[iCurrentDevice].pDevice;

            // Check whether this option is under the device-specific extra
            // options
            Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (!pszKey)
                  break;

               // Add appropriate constraint
               if (strcmp(pszKey, "journal") == 0)
               {
                  string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

                  *pofstream << "    <constraint sense='true'>" << endl;
                  *pofstream << "      <driver>omni</driver>" << endl;
                  *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
                  *pofstream << "      <arg_defval>" << "journal_off" << "</arg_defval>" << endl;
                  *pofstream << "    </constraint>" << endl;
                  break;
               }
            }

            delete pEnum;
         }

         *pofstream << "  </constraints>" << endl;
         *pofstream << "  <enum_vals>" << endl;
         *pofstream << "    <enum_val id='journal_off'>" << endl;
         *pofstream << "      <ev_longname><en>Off</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>Off</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>false</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='journal_on'>" << endl;
         *pofstream << "      <ev_longname><en>On</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>On</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>true</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "  </enum_vals>" << endl;
         *pofstream << "</option>" << endl;
      }
      else if (0 == strcmp (next->c_str (), "cutmode"))
      {
         ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-cutmode.xml");
         pofstream = &ofstreamOption;

         string OptionLongname ("Paper Cutting");
         string OptionShortname ("CutMode");
         string prototype ("cutmode");

         cout << argv[0] << ": Generating omni-cutmode.xml" << endl;

         *pofstream << "<option type=\"enum\" id=\"opt/omni-cutmode\">" << endl;

         if (pofstream->bad ())
         {
            cerr << argv[0] << ": Error: could not write to file." << endl;

            goto done;
         }

         *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
         *pofstream << "       it should be treated as documentation for the user. -->" << endl;
         *pofstream << "  <arg_longname>" << endl;
         *pofstream << "   <en>" << OptionLongname << "</en>" << endl;
         *pofstream << "  </arg_longname>" << endl;
         *pofstream << "  <arg_shortname>" << endl;
         *pofstream << "   <en>" << OptionShortname << "</en><!-- backends only know <en> shortnames! -->" << endl;
         *pofstream << "  </arg_shortname>" << endl;
         *pofstream << "  <arg_execution>" << endl;
         *pofstream << "   <arg_order>230</arg_order>" << endl;
         *pofstream << "   <arg_spot>A</arg_spot>" << endl;
         *pofstream << "   <arg_substitution />" << endl;
         *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
         *pofstream << "  </arg_execution>" << endl;
         *pofstream << endl;

         // Put in the printers to which this option applies
         *pofstream << "  <constraints>" << endl;

         for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
         {
            Device *pDevice = pohDevices[iCurrentDevice].pDevice;

            // Check whether this option is under the device-specific extra
            // options
            Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
            while (pEnum->hasMoreElements ())
            {
               char *pszKey = (char *)pEnum->nextElement ();

               if (!pszKey)
                  break;

               // Add appropriate constraint
               if (strcmp(pszKey, "cutmode") == 0)
               {
                  string *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

                  *pofstream << "    <constraint sense='true'>" << endl;
                  *pofstream << "      <driver>omni</driver>" << endl;
                  *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
                  *pofstream << "      <arg_defval>" << "cutmode_none" << "</arg_defval>" << endl;
                  *pofstream << "    </constraint>" << endl;
                  break;
               }
            }

            delete pEnum;
         }

         *pofstream << "  </constraints>" << endl;
         *pofstream << "  <enum_vals>" << endl;
         *pofstream << "    <enum_val id='cutmode_none'>" << endl;
         *pofstream << "      <ev_longname><en>Off</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>None</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>none</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='cutmode_partial'>" << endl;
         *pofstream << "      <ev_longname><en>Cut Partially</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>Partial</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>partial</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "    <enum_val id='cutmode_full'>" << endl;
         *pofstream << "      <ev_longname><en>Cut Fully</en></ev_longname>" << endl;
         *pofstream << "      <ev_shortname><en>Full</en></ev_shortname>" << endl;
         *pofstream << "      <ev_driverval>full</ev_driverval>" << endl;
         *pofstream << "    </enum_val>" << endl;
         *pofstream << "  </enum_vals>" << endl;
         *pofstream << "</option>" << endl;
      }
   }

   cout << endl;

#endif

   /*
   ** Make Option section
   **
   ** List Omni driver options
   **
   ** Here is what needs to be done.  There are additional options, that are in
   ** the Omni driver that each printer supports, or rather, Omni does for each
   ** printer, regardless.  For each of these options (to be defined in a file),
   ** this method must generate an omni-<option>.xml describing the properites
   ** of the option in question.  The way this will work is the Foomatic.cpp
   ** code will look for a specific file ( there will be a command line override
   ** added for the user to specify by hand the list of options they want
   ** generated) and Foomatic will read the options from the file and generate
   ** an xml file for each of the options in the file
   **
   ** The reading of the file will be one method, and then there will be a
   ** skeleton function that will take the name of the option and generate the
   ** omni-<option>.xml file with the specifics.  Each option will have a
   ** boolean flag, either on or off.  The only thing left is to determine
   ** whether the default it on or off, that can be specified in the file.
   **
   **
   ** Variables so far:
   ** String OptionLongname
   ** String OptionShortname
   ** String prototype

   ofstream ofstreamOption ("foomatic-db/db/source/opt/omni-myoption.xml");
   pofstream = &ofstreamOption;

   string OptionLongname ("OptionLongname");
   string OptionShortname ("OptionShortname");
   string prototype ("myoption");

   cout << "Generating omni-myoption.xml" << endl;

   *pofstream << "<option type=\"enum\" id=\"opt/omni-myoption\">" << endl;

   if (pofstream->bad ())
   {
      cerr << argv[0] << ": Error: could not write to file." << endl;

      goto done;
   }

   *pofstream << "  <!-- A multilingual <comments> block can appear here, too;" << endl;
   *pofstream << "       it should be treated as documentation for the user. -->" << endl;
   *pofstream << "  <arg_longname>" << endl;
   *pofstream << "   <en>" << OptionLongname << "</en>" << endl;
   *pofstream << "  </arg_longname>" << endl;
   *pofstream << "  <arg_shortname>" << endl;
   *pofstream << "   <en>" << OptionShortname << "</en><!-- backends only know <en> shortnames! -->" << endl;
   *pofstream << "  </arg_shortname>" << endl;
   *pofstream << "  <arg_execution>" << endl;
   *pofstream << "   <arg_order>100</arg_order>" << endl;
   *pofstream << "   <arg_spot>A</arg_spot>" << endl;
   *pofstream << "   <arg_substitution />" << endl;
   *pofstream << "   <arg_proto> " << prototype << "=%s</arg_proto>" << endl;
   *pofstream << "  </arg_execution>" << endl;
   *pofstream << endl;

   // Put in default value for each printer
   // Note to self:  find out whether this option defaults to 1 or 0
   *pofstream << "  <constraints>" << endl;

   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
      string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

      *pofstream << "    <constraint sense='true'>" << endl;
      *pofstream << "      <driver>omni</driver>" << endl;
      *pofstream << "      <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
      *pofstream << "      <arg_defval>" << 1 << "</arg_defval>" << endl;
      *pofstream << "    </constraint>" << endl;
   }

   *pofstream << "  </constraints>" << endl;
   *pofstream << endl;
   *pofstream << "  <enum_vals>" << endl;


   // Here is where we print out what printers support what options.  This can be stripped down to just
   // producing what the option is, and that all devices support the current option.

   // Iterate over all the possible option values (0 or 1)
   for (int iCurrent = 0; iCurrent < 2; iCurrent++)
   {
      bool fFirstTime = true;

      for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
      {
         Device            *pDevice           = pohDevices[iCurrentDevice].pDevice;
         string            *pstringDeviceName = pohDevices[iCurrentDevice].pstringDeviceName;

         if (iCurrent)
         {
            if (fFirstTime)
            {
               fFirstTime = false;
               *pofstream << "    <enum_val id='" << OptionShortname << "'>" << endl;
               *pofstream << "      <ev_longname><en>" << OptionLongname << "</en></ev_longname>" << endl;
               *pofstream << "      <ev_shortname><en>" << OptionShortname << "</en></ev_shortname>" << endl;
               *pofstream << "      <ev_driverval>" << iCurrent << "</ev_driverval>" << endl;
               *pofstream << "      <constraints>" << endl;
               *pofstream << "        <!-- Assume the option doesn't apply... -->" << endl;
               *pofstream << "        <constraint sense='false'>" << endl;
               *pofstream << "          <driver>omni</driver>" << endl;
               *pofstream << "        </constraint>" << endl;
            }

            *pofstream << "        <constraint sense='true'>" << endl;
            *pofstream << "          <driver>omni</driver>" << endl;
            *pofstream << "          <printer>printer/" << *pstringDeviceName << "</printer>" << endl;
            *pofstream << "        </constraint>" << endl;

         }
      }
      if (!fFirstTime)
      {
         *pofstream << "      </constraints>" << endl;
         *pofstream << "    </enum_val>" << endl;
      }
   }

   *pofstream << "  </enum_vals>" << endl;
   *pofstream << "</option>" << endl;

ENDOFOPTIONS
   */

   iRc = 0;

done:
   // Clean up
   for (iCurrentDevice = 0; iCurrentDevice < iNumDevices; iCurrentDevice++)
   {
      delete pohDevices[iCurrentDevice].pstringDeviceName;
      delete pohDevices[iCurrentDevice].pDevice;
      g_module_close (pohDevices[iCurrentDevice].hmodDevice);
   }

   delete[] pohDevices;

   Omni::terminate ();

   return iRc;
}
