/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"
#include "Omni.hpp"
#include "DebugOutput.hpp"

#include <glob.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>

#include <set>
#include <fstream>

static char *vapszLibraryPaths[] = {
   "/usr/lib/Omni/",
///"/opt/Omni/bin/",
   "/opt/Omni/lib/",
///"./",
   ""                 // give a chance for $LD_LIBRARY_PATH to work
};
static char *vapszDataPaths[] = {
   "/opt/Omni/data/",
   ""                 // allow for a fully qualified filename to work
};

void Omni::
initialize ()
{
#ifndef RETAIL
   char *pszDebugOptions = 0;

   pszDebugOptions = getenv ("OMNI_DEBUG_OPTIONS");
   if (   pszDebugOptions
      && *pszDebugOptions
      )
   {
      char *pszString = 0;
      char *pszWord   = 0;
      char *pszSpace  = 0;

      pszString = (char *)malloc (strlen (pszDebugOptions) + 1);
      if (pszString)
      {
         strcpy (pszString, pszDebugOptions);

         pszWord = pszString;
         while (  pszWord
               && *pszWord
               )
         {
            pszSpace = strchr (pszWord, ' ');
            if (pszSpace)
            {
               *pszSpace = '\0';
            }

            DebugOutput::setDebugOutput (pszWord);

            if (pszSpace)
            {
               pszWord = pszSpace + 1;
            }
            else
            {
               pszWord = 0;
            }
         }

         free (pszString);
      }
   }
#endif
}

void Omni::
terminate ()
{
}

void Omni::
setOutputStream (Device *pDevice,
                 FILE   *pfpNew)
{
   if (  pDevice
      && pfpNew
      )
      pDevice->setOutputStream (pfpNew);
}

void Omni::
setErrorStream (Device *pDevice,
                FILE   *pfpNew)
{
   if (!pfpNew)
   {
      return;
   }

   DebugOutput::setErrorStream (pfpNew);

   if (pDevice)
   {
      pDevice->setErrorStream (pfpNew);
   }
}

bool Omni::
libraryValid (GModule *hDevice,
              char    *pszLibraryName,
              char    *pszVersion,
              bool     fVerbose)
{
   bool     rc  = false;
   gpointer ptr = NULL;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": libraryValid (" << hex << hDevice << dec << ", " << pszVersion << ")" << endl;
#endif

   // The library must have the following symbols in it

   if (  g_module_symbol (hDevice, "getVersion", &ptr)
      && g_module_symbol (hDevice, "getDeviceEnumeration", &ptr)
      && g_module_symbol (hDevice, "newDeviceW_Advanced", &ptr)
      && g_module_symbol (hDevice, "newDeviceW_JopProp_Advanced", &ptr)
      && g_module_symbol (hDevice, "deleteDevice", &ptr)
      )
   {
      PFNGETVERSION  pfnGetVersion     = 0;
      char          *pszLibraryVersion = 0;

      g_module_symbol (hDevice, "getVersion", (gpointer *)&pfnGetVersion);

      pszLibraryVersion = pfnGetVersion ();

      if (  pszLibraryVersion
         && 0 == strcmp (VERSION, pszLibraryVersion)
         )
      {
         rc = true;
      }
#ifndef RETAIL
      else if (fVerbose)
      {
         if (!pszLibraryVersion)
            DebugOutput::getErrorStream () << "for " << pszLibraryName << " getVersion returns null!" << endl;
         if (0 != strcmp (VERSION, pszLibraryVersion))
            DebugOutput::getErrorStream () << "for " << pszLibraryName << " " << VERSION << " != " << pszLibraryVersion << endl;
      }
#endif
   }
   else if (fVerbose)
   {
      g_module_symbol (hDevice, "getDeviceEnumeration", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " getDeviceEnumeration is missing." << endl;
      g_module_symbol (hDevice, "newDeviceW_Advanced", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " newDeviceW_Advanced is missing." << endl;
      g_module_symbol (hDevice, "newDeviceW_JopProp_Advanced", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " newDeviceW_JopProp_Advanced is missing." << endl;
      g_module_symbol (hDevice, "deleteDevice", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " deleteDevice is missing." << endl;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": returning " << rc << endl;
#endif

   return rc;
}

static bool
createLibName (char *pszLibName,
               int   cbLibName,
               char *pszDriver,
               char *pszDevice)
{
   int cbDriver = strlen (pszDriver);

   // Is there enough space for everything?
   if ( cbLibName < (int)( 3                   // "lib"
                         + cbDriver            // driver
                         + 1                   // "_"
                         + strlen (pszDevice)  // device
                         + 3                   // ".so"
                         )
      )
      return false;

   strcpy (pszLibName, "lib");
   strcat (pszLibName, pszDriver);
   strcat (pszLibName, "_");

   char *pszStart = pszLibName + strlen (pszLibName);

   // For example, the driver "Epson" with the device "Epson LQ-2550" is
   // called libEpson_LQ_2550.so.  The "Epson" in the device is redundant.
   if (  0 == strncasecmp (pszDevice,
                           pszDriver,
                           cbDriver)
      && ' ' == pszDevice[cbDriver]
      )
   {
      pszDevice += cbDriver + 1;
   }

   strcat (pszLibName, pszDevice);

   while (*pszStart)
   {
      switch (*pszStart)
      {
      case '-':
      case '+':
      case '(':
      case ')':
         *pszStart = '_';
         break;
      }

      pszStart++;
   }

   strcat (pszLibName, ".so");

   return true;
}

bool Omni::
openLibrary (char     *pszLibName,
             GModule **phLibrary)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": g_module_open (" << pszLibName << ")" << endl;
#endif

   if (!phLibrary)
   {
      return false;
   }

   *phLibrary = 0;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "This program needs glib's module routines!" << endl;

      return false;
   }

   char *pszDeviceLib = 0;

   for (int i = 0; i < (int)dimof (vapszLibraryPaths) && !*phLibrary; i++)
   {
      pszDeviceLib = (char *)malloc ( strlen (pszLibName)
                                    + strlen (vapszLibraryPaths[i])
                                    + 1);
      if (pszDeviceLib)
      {
         sprintf (pszDeviceLib, "%s%s", vapszLibraryPaths[i], pszLibName);

         *phLibrary = g_module_open (pszDeviceLib, (GModuleFlags)0);
      }
      else
      {
         return false;
      }

      free (pszDeviceLib);
   }

   if (!*phLibrary)
   {
      DebugOutput::getErrorStream () << endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << endl;
      DebugOutput::getErrorStream () << endl << endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": g_module_error returns " << g_module_error () << endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": pszLibName = " << pszLibName << endl;
      DebugOutput::getErrorStream () << endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": LD_LIBRARY_PATH =  " << getenv ("LD_LIBRARY_PATH") << endl;
      DebugOutput::getErrorStream () << endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Omni device library not found in the following paths:" << endl;
      for (int i = 0; i < (int)dimof (vapszLibraryPaths) - 1 && !*phLibrary; i++)
      {
         DebugOutput::getErrorStream () << "\t" << vapszLibraryPaths[i] << "." << endl;
      }
      DebugOutput::getErrorStream () << "\t$LD_LIBRARY_PATH" << endl;

      return 0;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": phLibrary = " << hex << (int)phLibrary << dec << endl;
#endif

   if (!*phLibrary)
      return false;

   return true;
}

bool Omni::
openDeviceLibrary (char     *pszLibName,
                   GModule **phDevice)
{
   if (!openLibrary (pszLibName, phDevice))
      return false;

   return libraryValid (*phDevice, pszLibName, VERSION);
}

Device * Omni::
createDevice (char     *pszDriver,
              char     *pszDevice,
              char     *pszJobProperties,
              bool      fAdvanced,
              GModule **phDevice)
{
   char               achLibDeviceName[256]; // @TBD - need max path size
   PFNNEWDEVICEWARGS  pfnNewDeviceWArgs     = 0;
   Device            *pDevice               = 0;

   *phDevice = 0;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "This program needs glib's module routines!" << endl;
      return 0;
   }

   if (createLibName (achLibDeviceName,
                      sizeof (achLibDeviceName),
                      pszDriver,
                      pszDevice))
   {
      if (openDeviceLibrary (achLibDeviceName, phDevice))
      {
         g_module_symbol (*phDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

         pDevice = pfnNewDeviceWArgs (pszJobProperties,
                                      fAdvanced);

         if (pDevice)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ())
#endif
            {
               DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDriver << ", " << pDevice->getDriverName () << ") = " << dec << strcasecmp (pszDriver, pDevice->getDriverName ()) << endl;
               DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDevice << ", " << pDevice->getDeviceName () << ") = " << strcasecmp (pszDevice, pDevice->getDeviceName ()) << endl;
            }

            if (  0 != strcasecmp (pszDriver, pDevice->getDriverName ())
               && 0 != strcasecmp (pszDevice, pDevice->getDeviceName ())
               )
            {
               // The driver and device name are not the same!
               delete pDevice;
               pDevice = 0;

#ifndef RETAIL
               int rc =
#endif
                        g_module_close (*phDevice);

#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice @ " << dec << __LINE__ << " : g_module_close (" << hex << *phDevice << ") = " << rc << endl;
#endif

               *phDevice = 0;
            }
         }
      }
   }

   if (!pDevice)
   {
      // Do it the hard way
      Enumeration *pEnum  = Omni::listDevices ();
      bool         fFound = false;

      while (  !fFound
            && pEnum->hasMoreElements ()
            )
      {
         OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

         if (pOD)
         {
            char *pszLibName = pOD->getLibraryName ();

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice g_module_open (" << pszLibName << ")" << endl;
#endif

            *phDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (  *phDevice
               && libraryValid (*phDevice, pszLibName, VERSION)
               )
            {
               g_module_symbol (*phDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

               pDevice = pfnNewDeviceWArgs (pszJobProperties, fAdvanced);

#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ())
               {
                  DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDriver << ", " << pDevice->getDriverName () << ") = " << dec << strcasecmp (pszDriver, pDevice->getDriverName ()) << endl;
                  DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDevice << ", " << pDevice->getDeviceName () << ") = " << strcasecmp (pszDevice, pDevice->getDeviceName ()) << endl;
               }
#endif

               if (  0 == strcasecmp (pszDriver, pDevice->getDriverName ())
                  && 0 == strcasecmp (pszDevice, pDevice->getDeviceName ())
                  )
                  fFound = true;
            }

            delete pOD;
         }

         if (!fFound)
         {
            delete pDevice;
            pDevice = 0;

#ifndef RETAIL
            int rc =
#endif
                     g_module_close (*phDevice);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice @ " << dec <<  __LINE__ << " : g_module_close (" << hex << *phDevice << ") = " << rc << endl;
#endif

            *phDevice = 0;
         }
      }

      delete pEnum;
   }

   return pDevice;
}

Device * Omni::
createDevice (PDL      *pPDL,
              char     *pszJobProperties,
              bool      fAdvanced,
              GModule **phDevice)
{
   // @TBD
   return 0;
}

char * Omni::
openXMLFile (char *pszXMLFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": pszXMLFile = " << pszXMLFile << endl;
#endif

   if (  !pszXMLFile
      || !*pszXMLFile
      )
   {
      return 0;
   }

   char *pszTargetXMLFile = 0;

   for (int i = 0; i < (int)dimof (vapszDataPaths); i++)
   {
      pszTargetXMLFile = (char *)malloc ( strlen (pszXMLFile)
                                        + strlen (vapszDataPaths[i])
                                        + 1);
      if (pszTargetXMLFile)
      {
         struct stat statFile;

         sprintf (pszTargetXMLFile, "%s%s", vapszDataPaths[i], pszXMLFile);

         if (0 == stat (pszTargetXMLFile, &statFile))
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Found it at \"" << pszTargetXMLFile << "\"" << endl;
#endif

            return pszTargetXMLFile;
         }
      }
      else
      {
         DebugOutput::getErrorStream () << endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << endl;
         DebugOutput::getErrorStream () << endl << endl;
         DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Error: Out of memory @ " << __LINE__ << endl;

         return 0;
      }

      free (pszTargetXMLFile);
      pszTargetXMLFile = 0;
   }

   DebugOutput::getErrorStream () << endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << endl;
   DebugOutput::getErrorStream () << endl << endl;
   DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Omni XML file (" << pszXMLFile << ") not found in the following paths:" << endl;
   for (int i = 0; i < (int)dimof (vapszDataPaths); i++)
   {
      DebugOutput::getErrorStream () << "\t" << vapszDataPaths[i] << endl;
   }

   return 0;
}

static int
my_system (const char *pszCommand)
{
   int pid, status;

   if (pszCommand == 0)
      return 1;

   pid = fork();
   if (pid == -1)
      return -1;

   if (pid > 0)
   {
      // parent
      do
      {
         if (waitpid (pid, &status, 0) == -1)
         {
            if (errno != EINTR)
               return -1;
         }
         else
         {
            return status;
         }

      } while (1);
   }
   else
   {
      // child
      char *argv[4];

      argv[0] = "sh";
      argv[1] = "-c";
      argv[2] = (char *)pszCommand;
      argv[3] = 0;

      execvp ("/bin/sh", argv);

      exit (1);
   }
}

class BuildDeviceEnumerator : public Enumeration
{
public:
   BuildDeviceEnumerator (bool fVerbose)
   {
      // Initialize our variables
      fVerbose_d           = fVerbose;
      pszLD_LIBRARY_PATH_d = getenv ("LD_LIBRARY_PATH");
      fGlobActive_d        = false;
      hLibrary_d           = 0;
      pDeviceEnum_d        = 0;
   }

   virtual
   ~BuildDeviceEnumerator ()
   {
      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      delete pDeviceEnum_d;
      pDeviceEnum_d = 0;
      if (hLibrary_d)
      {
         g_module_close (hLibrary_d);
         hLibrary_d = 0;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Do we have an active device enumeration?
      if (pDeviceEnum_d)
      {
         if (pDeviceEnum_d->hasMoreElements ())
         {
            return true;
         }

         delete pDeviceEnum_d;
         pDeviceEnum_d = 0;
         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }
      }

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

      if (  !fGlobActive_d        // Did we not find one?
         && pszLD_LIBRARY_PATH_d  // Still more paths to search
         )
      {
         // Try with another path
         pszLD_LIBRARY_PATH_d = globPath (pszLD_LIBRARY_PATH_d);
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (pDeviceEnum_d)
      {
         return pDeviceEnum_d->nextElement ();
      }
      else if (fGlobActive_d)
      {
         char *pszLibraryName = globbuf_d.gl_pathv[iCurrentFile_d];

         // See if this device has subdevices
         hLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

         if (hLibrary_d)
         {
            PFNENUMERATEDEVICES pfnEnumerateDevices = 0;

            g_module_symbol (hLibrary_d,
                             "getDeviceEnumeration",
                             (gpointer *)&pfnEnumerateDevices);

            if (pfnEnumerateDevices)
            {
               pDeviceEnum_d = pfnEnumerateDevices (pszLibraryName);
            }

            if (pDeviceEnum_d)
            {
               if (pDeviceEnum_d->hasMoreElements ())
               {
                  // Return the first subdevice
                  return pDeviceEnum_d->nextElement ();
               }
            }

            // Clean up
            delete pDeviceEnum_d;
            pDeviceEnum_d = 0;
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }

         // Return the one that we found
         return new OmniDevice (pszLibraryName, 0);
      }
      else
      {
         // Error!
         return 0;
      }
   }

private:
   char *
   globPath (char *pszLD_LIBRARY_PATH)
   {
      char  achPath[256]; // @TBD - need max path size
      char *pszSep       = 0;
      bool  fFound       = false;
      int   rc;

      do
      {
         // See if there are two or more paths
         pszSep = strpbrk (pszLD_LIBRARY_PATH, ":");

         if (pszSep)
         {
            int iChars = pszSep - pszLD_LIBRARY_PATH;

            // Just copy the first one
            strncpy (achPath, pszLD_LIBRARY_PATH, iChars);
            achPath[iChars] = '\0';

            pszLD_LIBRARY_PATH = pszSep + 1;
         }
         else
         {
            // Copy the one one
            strcpy (achPath, pszLD_LIBRARY_PATH);

            pszLD_LIBRARY_PATH = 0;
         }

         // Add our search parameter
         strcat (achPath, "/lib*.so");

         if (fVerbose_d)
         {
            DebugOutput::getErrorStream () << "Searching: glob (" << achPath << ") = ";
         }

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (achPath, 0, NULL, &globbuf_d);

         DebugOutput::getErrorStream () << rc << endl;

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }

      } while (  !fFound            // Haven't found any yet
              && pszLD_LIBRARY_PATH // More to search
              );

      return pszLD_LIBRARY_PATH;
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         char *pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];

         if (fVerbose_d)
         {
            DebugOutput::getErrorStream () << "Checking (" << pszCurrentName << ")" << endl;
         }

         string stringSystem;

/////////stringSystem  = "strace OmniDeviceValid ";
         stringSystem  = "OmniDeviceValid ";
         if (fVerbose_d)
            stringSystem += "-v ";
         stringSystem += "\"";
         stringSystem += pszCurrentName;
         stringSystem += "\" ";
         stringSystem += VERSION;

         if (0 == my_system (stringSystem.c_str ()))
         {
            fFound = true;
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   bool         fVerbose_d;
   char        *pszLD_LIBRARY_PATH_d;
   glob_t       globbuf_d;
   int          iCurrentFile_d;
   bool         fGlobActive_d;

   GModule     *hLibrary_d;
   Enumeration *pDeviceEnum_d;
};

class SystemDeviceEnumerator : public Enumeration
{
public:
   SystemDeviceEnumerator ()
   {
      // Initialize our variables

      // Add the Omni builtin paths
      for (int i = 0; i < (int)dimof (vapszLibraryPaths); i++)
      {
         string stringPath = vapszLibraryPaths[i];

         if (*vapszLibraryPaths[i])
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": inserting (" << stringPath << ")" << endl;
#endif

            setDirectories_d.insert (stringPath);
         }
      }

      // Add /etc/ld.so.conf
      char       achLine[512];             // @TBD
      ifstream   ifIn ("/etc/ld.so.conf");

      while (0 < ifIn.getline (achLine, sizeof (achLine)))
      {
         char *pszEnd = achLine + strlen (achLine) - 1;

         if ('/' != *pszEnd)
         {
            strcat (achLine, "/");
         }

         string stringLine = achLine;

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": inserting (" << achLine << ")" << endl;
#endif

         setDirectories_d.insert (stringLine);
      }

      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
      hLibrary_d    = 0;
      pDeviceEnum_d = 0;
   }

   virtual
   ~SystemDeviceEnumerator ()
   {
      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      delete pDeviceEnum_d;
      pDeviceEnum_d = 0;
      if (hLibrary_d)
      {
         g_module_close (hLibrary_d);
         hLibrary_d = 0;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Do we have an active device enumeration?
      if (pDeviceEnum_d)
      {
         if (pDeviceEnum_d->hasMoreElements ())
         {
            return true;
         }

         delete pDeviceEnum_d;
         pDeviceEnum_d = 0;
         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }
      }

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": fGlobActive_d = " << fGlobActive_d << ", " << (iterCurrent_d != setDirectories_d.end ()) << endl;
#endif

      if (  !fGlobActive_d                           // Did we not find one?
         && iterCurrent_d != setDirectories_d.end () // Still more directories to search
         )
      {
         // Try with another path
         globPath ();
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (pDeviceEnum_d)
      {
         return pDeviceEnum_d->nextElement ();
      }
      else if (fGlobActive_d)
      {
         char *pszLibraryName = globbuf_d.gl_pathv[iCurrentFile_d];

         // See if this device has subdevices
         hLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

         if (hLibrary_d)
         {
            PFNENUMERATEDEVICES pfnEnumerateDevices = 0;

            g_module_symbol (hLibrary_d,
                             "getDeviceEnumeration",
                             (gpointer *)&pfnEnumerateDevices);

            if (pfnEnumerateDevices)
            {
               pDeviceEnum_d = pfnEnumerateDevices (pszLibraryName);
            }

            if (pDeviceEnum_d)
            {
               if (pDeviceEnum_d->hasMoreElements ())
               {
                  // Return the first subdevice
                  return pDeviceEnum_d->nextElement ();
               }
            }

            // Clean up
            delete pDeviceEnum_d;
            pDeviceEnum_d = 0;
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }

         // Return the one that we found
         return new OmniDevice (pszLibraryName, 0);
      }
      else
      {
         // Error!
         return 0;
      }
   }

private:
   void
   globPath ()
   {
      bool fFound = false;
      int  rc     = 0;

      while (  !fFound                                  // Haven't found any yet
            && iterCurrent_d != setDirectories_d.end () // Haven't reached the end
            )
      {
         string stringGlob = *iterCurrent_d++;

         stringGlob += "lib*.so";

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (stringGlob.c_str (), 0, NULL, &globbuf_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << " (" << stringGlob << ") = " << rc << endl;
#endif

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }
      }
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         char *pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": findDeviceInGlob: checking (" << pszCurrentName << ")" << endl;
#endif

         string stringSystem;

         stringSystem  = "OmniDeviceValid ";
/////////stringSystem += "-v ";
         stringSystem += "\"";
         stringSystem += pszCurrentName;
         stringSystem += "\" ";
         stringSystem += VERSION;

         if (0 == my_system (stringSystem.c_str ()))
         {
            fFound = true;
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   set<string>           setDirectories_d;
   set<string>::iterator iterCurrent_d;
   glob_t                globbuf_d;
   int                   iCurrentFile_d;
   bool                  fGlobActive_d;

   GModule              *hLibrary_d;
   Enumeration          *pDeviceEnum_d;
};

Enumeration * Omni::
listDevices (bool fBuildOnly)
{
   if (fBuildOnly)
   {
      return new BuildDeviceEnumerator (true);
   }
   else
   {
      return new SystemDeviceEnumerator ();
   }
}

class MasterXMLEnumerator : public Enumeration
{
public:
   MasterXMLEnumerator ()
   {
      // Initialize our variables

      // Add the Omni builtin paths
      for (int i = 0; i < (int)dimof (vapszDataPaths); i++)
      {
         string stringPath = vapszDataPaths[i];

         if (*vapszDataPaths[i])
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::MasterXMLEnumerator::" << __FUNCTION__ << ": inserting (" << stringPath << ")" << endl;
#endif

            setDirectories_d.insert (stringPath);
         }
      }

      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
   }

   virtual
   ~MasterXMLEnumerator ()
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::MasterXMLEnumerator::~" << __FUNCTION__ << endl;
#endif

      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::MasterXMLEnumerator::" << __FUNCTION__ << ": fGlobActive_d = " << fGlobActive_d << ", " << (iterCurrent_d != setDirectories_d.end ()) << endl;
#endif

      if (  !fGlobActive_d                           // Did we not find one?
         && iterCurrent_d != setDirectories_d.end () // Still more directories to search
         )
      {
         // Try with another path
         globPath ();
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (fGlobActive_d)
         // Return the one that we found
         return globbuf_d.gl_pathv[iCurrentFile_d];
      else
         // Error!
         return 0;
   }

private:
   void
   globPath ()
   {
      bool fFound = false;
      int  rc     = 0;

      while (  !fFound                                  // Haven't found any yet
            && iterCurrent_d != setDirectories_d.end () // Haven't reached the end
            )
      {
         string stringGlob = *iterCurrent_d++;

         stringGlob += "*.xml";

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (stringGlob.c_str (), 0, NULL, &globbuf_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::MasterXMLEnumerator::" << __FUNCTION__ << " (" << stringGlob << ") = " << rc << endl;
#endif

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }
      }
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         char *pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];
         FILE *fpFile         = 0;

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::MasterXMLEnumerator::" << __FUNCTION__ << ": findDeviceInGlob: checking (" << pszCurrentName << ")" << endl;
#endif

         fpFile = fopen (pszCurrentName, "r");

         if (fpFile)                          //           1         2          3         4
         {                                    // 0123456789012345678901234 567890123456789012 34
            static const char *pszTest        = "<!DOCTYPE Device SYSTEM \"../OmniDevice.dtd\">";
            static int         cbTest         = strlen (pszTest);
            char               achBuffer[512];
            char              *pszBuffer      = achBuffer;
            size_t             cbBuffer       = sizeof (achBuffer);
            bool               fContinue      = true;

            do
            {
               // Read a block
               if (fread (pszBuffer, cbBuffer, 1, fpFile))
               {
                  // Start at the end
                  char *pszSearch = achBuffer + sizeof (achBuffer) - cbTest;

                  // While within the buffer
                  while (pszSearch >= achBuffer)
                  {
                     // Is is found
                     if (  *pszSearch == *pszTest
                        && 0 == strncmp (pszSearch, pszTest, cbTest)
                        )
                     {
                        fFound = true;
                        break;
                     }
                     else
                     {
                        // Move backwards based on current character
                        switch (*pszSearch)
                        {
                        case '!': pszSearch -= 1;      break;
                        case 'D': pszSearch -= 2;      break;
                        case 'O': pszSearch -= 3;      break;
                        case 'C': pszSearch -= 4;      break;
                        case 'T': pszSearch -= 5;      break;
                        case 'Y': pszSearch -= 6;      break;
                        case 'P': pszSearch -= 7;      break;
                        case 'E': pszSearch -= 8;      break;
                        case ' ': pszSearch -= 9;      break;
                        case 'e': pszSearch -= 11;     break;
                        case 'v': pszSearch -= 12;     break;
                        case 'i': pszSearch -= 13;     break;
                        case 'c': pszSearch -= 14;     break;
                        case 'S': pszSearch -= 17;     break;
                        case 'M': pszSearch -= 22;     break;
                        case '"': pszSearch -= 24;     break;
                        case '/': pszSearch -= 27;     break;
                        case 'm': pszSearch -= 29;     break;
                        case 'n': pszSearch -= 30;     break;
                        case '.': pszSearch -= 38;     break;
                        case 'd': pszSearch -= 39;     break;
                        case 't': pszSearch -= 40;     break;
                        case '>': pszSearch -= 44;     break;
                        default:  pszSearch -= cbTest; break;
                        }
                     }
                  }

                  if (!fFound)
                  {
                     /* In order to handle the case where the search string might
                     ** be spanning buffers, move the end of the buffer to the
                     ** front and shrink the buffer
                     */
                     memmove (achBuffer,
                              achBuffer + sizeof (achBuffer) - cbTest,
                              cbTest);

                     pszBuffer = achBuffer + cbTest;
                     cbBuffer  = sizeof (achBuffer) - cbTest;
                  }
               }
               else
               {
                  // File read error!
                  fContinue = false;
               }

            } while (  fContinue
                    && !fFound
                    );

            // Clean up
            fclose (fpFile);
         }

         if (fFound)
         {
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   set<string>           setDirectories_d;
   set<string>::iterator iterCurrent_d;
   glob_t                globbuf_d;
   int                   iCurrentFile_d;
   bool                  fGlobActive_d;
};

Enumeration * Omni::
listXMLDevices ()
{
   return new MasterXMLEnumerator ();
}
