/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"
#include "Omni.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <memory>
#include <string>

#include <glib.h>
#include <gmodule.h>

typedef int (*PFNSHOWOPTIONS) (Device *pDevice);

void
printUsage (char *pszProgramName)
{
   cerr << "Usage: "
        << pszProgramName
        << endl
        << " (UsePDC | !UsePDC | driver printer_library_name | jobprop JobPropKey=JobPropValue | Verbose | all | debugoutput xxx)+"
        << endl;
}

int
executeDevice (char           *pszFullDeviceName,
               bool            fUsePDC,
               PFNSHOWOPTIONS  pfnShowOptions,
               char           *pszJobProperties)
{
   Device       *pDevice      = 0;
   GModule      *hmodDevice   = 0;
   int           rc           = 0;

#ifndef RETAIL
   DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": pszFullDeviceName = \"" << pszFullDeviceName << "\" pszJobProperties = \"" << pszJobProperties << "\"" << endl;
#endif

   if (fUsePDC)
   {
      char *pszSlash = strchr (pszFullDeviceName, '/');

      while (pszSlash)
      {
         pszFullDeviceName = pszSlash + 1;

         pszSlash = strchr (pszFullDeviceName, '/');
      }

      pDevice = new OmniPDCProxy (0,                 // client exe to spawn
                                  pszFullDeviceName, // device name
                                  0,                 // job properties
                                  true);             // is renderer
   }
   else
   {
      PFNNEWDEVICE       pfnNewDevice      = 0;
      PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;

      hmodDevice = g_module_open (pszFullDeviceName, (GModuleFlags)0);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": hmodDevice = " << hex << hmodDevice << dec << endl;
#endif

      if (!hmodDevice)
      {
         DebugOutput::getErrorStream ()
            << "Error: trying to load \""
            << pszFullDeviceName
            << "\" g_module_error returns "
            << g_module_error ()
            << endl;

         return 2;
      }

      g_module_symbol (hmodDevice, "newDeviceW_Advanced",         (gpointer *)&pfnNewDevice);
      g_module_symbol (hmodDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << hex << ": pfnNewDevice      = 0x" << (int)pfnNewDevice << endl;
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << hex << ": pfnNewDeviceWArgs = 0x" << (int)pfnNewDeviceWArgs << endl;
#endif

      if (  !pfnNewDevice
         || !pfnNewDeviceWArgs
         )
      {
         DebugOutput::getErrorStream () << "Error: trying to load newDeviceW[_JobProp]_Advanced. g_module_error returns " << g_module_error () << endl;

         return 3;
      }

      if (  !pszJobProperties
         || !*pszJobProperties
         )
      {
         pDevice = pfnNewDevice (true);
      }
      else
      {
         pDevice = pfnNewDeviceWArgs (pszJobProperties, true);
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << ": pDevice = " << *pDevice << endl;
#endif

   if (!pDevice)
   {
      DebugOutput::getErrorStream () << "Error: No Device was created!" << endl;

      return 4;
   }

   // @HACK - cause the device code to be executed.
   Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
   delete pEnum;

   // Check for an error
   if (pDevice->hasError ())
   {
      DebugOutput::getErrorStream () << "Error: The device has an error." << endl;

      return 5;
   }

   rc = pfnShowOptions (pDevice);

   if (pDevice->hasError ())
   {
      DebugOutput::getErrorStream () << "Error: The device has an error." << endl;

      return 6;
   }

   delete pDevice;

   if (hmodDevice)
   {
#ifndef RETAIL
      int rc2 =
#endif

                g_module_close (hmodDevice);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": g_module_close returns " << rc2 << endl;
#endif
   }

   return rc;
}

int
showDeviceGS (Device *pDevice)
{
   cout << "For gs add the following: -sproperties=\"...\"" << endl;
   cout << "\twhere you replace ... with the following space separated key=value (* - default):" << endl;
   cout << endl;

   Enumeration *pEnum = pDevice->listJobPropertyKeys ();

   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
         break;

      string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (pstringValue)
      {
         Enumeration       *pEnumValues        = pDevice->listKeyValues (pszKey);
         string::size_type  idxSpace1          = string::npos;
         string::size_type  idxSpace2          = string::npos;
         string             stringDefaultValue;

         idxSpace1 = pstringValue->find (" ");

         if (string::npos != idxSpace1)
         {
            idxSpace2 = pstringValue->find (" ", idxSpace1 + 1);
         }
         if (string::npos != idxSpace2)
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1,
                                                       idxSpace2 - idxSpace1 - 1);
         }
         else
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1);
         }

         if (pEnumValues->hasMoreElements ())
         {
            cout << pszKey << "=" << endl;
         }

         while (pEnumValues->hasMoreElements ())
         {
            char *pszValue = (char *)pEnumValues->nextElement ();

            cout << "\t" << pszValue;

            if (0 == stringDefaultValue.compare (pszValue))
            {
               cout << " (*)";
            }

            cout << endl;
         }

         delete pEnumValues;
         delete pstringValue;
      }
   }

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
         break;

      cout << pszKey << "=" << endl;

      string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (pstringValue)
      {
         string::size_type idxSpace1          = string::npos;
         string::size_type idxSpace2          = string::npos;
         string::size_type idxSpace3          = string::npos;
         string            stringDefaultValue;

         idxSpace1 = pstringValue->find (" ");

         if (string::npos != idxSpace1)
         {
            idxSpace2 = pstringValue->find (" ", idxSpace1 + 1);
         }
         if (string::npos != idxSpace2)
         {
            idxSpace3 = pstringValue->find (" ", idxSpace2 + 1);
         }

         if (  string::npos != idxSpace2
            && (  ! (  pstringValue->find ("string ") == 0
                    || pstringValue->find ("boolean ") == 0
                    )
               )
            )
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1,
                                                       idxSpace2 - idxSpace1 - 1);
         }
         else
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1);
         }

         if (  pstringValue->find ("string ") == 0
            || pstringValue->find ("boolean ") == 0
            )
         {
            Enumeration *pEnumValues = pDevice->listKeyValues (pszKey);

            while (pEnumValues->hasMoreElements ())
            {
               char *pszValue = (char *)pEnumValues->nextElement ();

               cout << "\t" << pszValue;

               if (0 == stringDefaultValue.compare (pszValue))
               {
                  cout << " (*)";
               }

               cout << endl;
            }

            delete pEnumValues;
         }
         else if (pstringValue->find ("integer ") == 0)
         {
            string stringMin = "1";
            string stringMax = "n";

            if (string::npos != idxSpace2)
            {
               if (string::npos != idxSpace3)
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1,
                                                    idxSpace3 - idxSpace2 - 1);
               }
               else
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1);
               }
            }
            if (string::npos != idxSpace3)
            {
               stringMax = pstringValue->substr (idxSpace3 + 1);
            }

            cout << "\t"
                 << stringMin
                 << " ... "
                 << stringMax
                 << " ("
                 << stringDefaultValue
                 << ")"
                 << endl;
         }
         else if (pstringValue->find ("float ") == 0)
         {
            string stringMin = "1.0";
            string stringMax = "n";

            if (string::npos != idxSpace2)
            {
               if (string::npos != idxSpace3)
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1,
                                                    idxSpace3 - idxSpace2 - 1);
               }
               else
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1);
               }
            }
            if (string::npos != idxSpace3)
            {
               stringMax = pstringValue->substr (idxSpace3 + 1);
            }

            cout << "\t"
                 << stringMin
                 << " ... "
                 << stringMax
                 << " ("
                 << stringDefaultValue
                 << ")"
                 << endl;
         }

         delete pstringValue;
      }
   }

   delete pEnum;

   return 0;
}

int
showDeviceVerbose (Device *pDevice)
{
   cout << "Driver version is : "
        << pDevice->getVersion ()
        << endl
        << "Driver name is : "
        << pDevice->getDriverName ()
        << endl
        << "Device name is : "
        << pDevice->getDeviceName ()
        << endl;

   cout << "The PDL level is "
        << pDevice->getPDLLevel ()
        << endl
        << "The PDL sublevel is "
        << pDevice->getPDLSubLevel ()
        << endl
        << "The PDL major revision level is "
        << pDevice->getPDLMajorRevisionLevel ()
        << endl
        << "The PDL minor revision level is "
        << pDevice->getPDLMinorRevisionLevel ()
        << endl;

   DeviceOrientation *pDO   = pDevice->getCurrentOrientation ();
   Enumeration       *pEnum = pDO->getEnumeration ();

   cout << "Current orientation is:" << endl;
   cout << DeviceOrientation::IDToName (pDO->getID ()) << endl;

   cout << "Enumerating orientations:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDO = (DeviceOrientation *)pEnum->nextElement ();

      cout << DeviceOrientation::IDToName (pDO->getID ()) << endl;

      delete pDO;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DeviceForm  *pDF = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   cout << "Current form is:" << endl;
   cout << DeviceForm::IDToName (pDF->getID ()) << endl;
   HardCopyCap *pHCC = pDF->getHardCopyCap ();
   cout << "Form dimensions are: ("
        << pHCC->getCx ()
        << ","
        << pHCC->getCy ()
        << ") (L:"
        << pHCC->getLeftClip ()
        << ",T:"
        << pHCC->getTopClip ()
        << ",R:"
        << pHCC->getRightClip ()
        << ",B:"
        << pHCC->getBottomClip ()
        << ")"
        << endl;

   cout << "Enumerating forms:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDF = (DeviceForm *)pEnum->nextElement ();

      cout << DeviceForm::IDToName (pDF->getID ()) << endl;

      delete pDF;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DeviceTray *pDT = pDevice->getCurrentTray ();
   pEnum = pDT->getEnumeration ();

   cout << "Current tray is:" << endl;
   cout << DeviceTray::IDToName (pDT->getID ()) << endl;

   cout << "Enumerating trays:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDT = (DeviceTray *)pEnum->nextElement ();

      cout << DeviceTray::IDToName (pDT->getID ()) << endl;

      delete pDT;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DeviceMedia *pDM = pDevice->getCurrentMedia ();
   pEnum = pDM->getEnumeration ();

   cout << "Current media is:" << endl;
   cout << DeviceMedia::IDToName (pDM->getID ()) << endl;

   cout << "Enumerating medias:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDM = (DeviceMedia *)pEnum->nextElement ();

      cout << DeviceMedia::IDToName (pDM->getID ()) << endl;

      delete pDM;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DeviceResolution *pDR = pDevice->getCurrentResolution ();
   pEnum = pDR->getEnumeration ();

   cout << "Current resolution is:" << endl;
   cout << DeviceResolution::IDToName (pDR->getID ()) << endl;

   cout << "Resolution dimensions are: ("
        << pDR->getXRes ()
        << ","
        << pDR->getYRes ()
        << ")"
        << endl;

   cout << "Enumerating resolutions:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDR = (DeviceResolution *)pEnum->nextElement ();

      cout << DeviceResolution::IDToName (pDR->getID ()) << endl;

      delete pDR;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DevicePrintMode *pDPM = pDevice->getCurrentPrintMode ();
   pEnum = pDPM->getEnumeration ();

   cout << "Current print mode is:" << endl;
   cout << DevicePrintMode::IDToName (pDPM->getID ()) << endl;

   cout << "Print Mode dimensions are: (CT:"
        << pDPM->getColorTech ()
        << ",P:"
        << pDPM->getPhysicalCount ()
        << ",L:"
        << pDPM->getLogicalCount ()
        << ",#planes:"
        << pDPM->getNumPlanes ()
        << ")"
        << endl;

   cout << "Enumerating print modes:" << endl;
   while (pEnum->hasMoreElements ())
   {
      pDPM = (DevicePrintMode *)pEnum->nextElement ();

      cout << DevicePrintMode::IDToName (pDPM->getID ()) << endl;

      delete pDPM;
   }
   cout << "Done!" << endl;

   delete pEnum;

   DeviceGamma *pDG = pDevice->getCurrentGamma ();

   cout << "Current gamma is:" << endl;

   if (pDG)
   {
      cout << "{ "
           << pDG->getCGamma ()
           << " "
           << pDG->getMGamma ()
           << " "
           << pDG->getYGamma ()
           << " "
           << pDG->getKGamma ()
           << " "
           << pDG->getCBias ()
           << " "
           << pDG->getMBias ()
           << " "
           << pDG->getYBias ()
           << " "
           << pDG->getKBias ()
           << "}"
           << endl;
   }
   else
   {
      cout << "Non existant!" << endl;
   }

   pEnum = DeviceDither::getEnumeration ();

///cout << "Current dither is:" << endl;
///cout << pDevice->getCurrentDither () << endl;

   cout << "Current dither id is:" << endl;
   cout << pDevice->getCurrentDitherID () << endl;

   cout << "Enumerating dithering ids:" << endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszID = (char *)pEnum->nextElement ();

      cout << pszID << endl;
   }
   cout << "Done!" << endl;

   delete pEnum;

   pEnum = pDevice->listJobPropertyKeys ();

   cout << "Enumerating job property types:" << endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << endl;
         break;
      }

      string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobPropertyType (" << pszKey << ") return null!" << endl;
         break;
      }

      cout << pszKey << " = " << *pstringValue << endl;

      delete pstringValue;
   }
   cout << "Done!" << endl;

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   cout << "Enumerating device job property types:" << endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << endl;
         break;
      }

      string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobPropertyType (" << pszKey << ") return null!" << endl;
         break;
      }

      cout << pszKey << " = " << *pstringValue << endl;

      delete pstringValue;
   }
   cout << "Done!" << endl;

   delete pEnum;

   pEnum = pDevice->listJobPropertyKeys ();

   cout << "Enumerating job properties:" << endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << endl;
         break;
      }

      string *pstringValue = pDevice->getJobProperty (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobProperty (" << pszKey << ") return null!" << endl;
         break;
      }

      cout << pszKey << " = " << *pstringValue << endl;

      delete pstringValue;
   }
   cout << "Done!" << endl;

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   cout << "Enumerating device job properties:" << endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << endl;
         break;
      }

      string *pstringValue = pDevice->getJobProperty (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobProperty (" << pszKey << ") return null!" << endl;
         break;
      }

      cout << pszKey << " = " << *pstringValue << endl;

      delete pstringValue;
   }
   cout << "Done!" << endl;

   delete pEnum;

   PrintDevice *pPrintDevice = dynamic_cast <PrintDevice *>(pDevice);
   if (pPrintDevice)
   {
      string *pstringRet = 0;

      cout << "Job Properties in Device Values:" << endl;

      pstringRet = pPrintDevice->queryDeviceJobProperties ();

      if (pstringRet)
         cout << *pstringRet << endl;
      else
         cout << "NONE!" << endl;
   }

   if (pDevice->hasCapability (Capability::COLOR))
   {
      cout << "This device can print in color." << endl;
   }

   if (pDevice->hasCapability (Capability::MONOCHROME))
   {
      cout << "This device can print in monochrome." << endl;
   }

   if (pDevice->hasCapability (Capability::MIRROR))
   {
      cout << "This device can print in mirror image." << endl;
   }

   cout << "Portrait is ";
   pDO = pDevice->getCurrentOrientation ();
   if (!pDO->isOrientationSupported (DeviceOrientation::ORIENTATION_PORTRAIT))
   {
      cout << "not ";
   }
   cout << "supported" << endl;

   cout << "Letter is ";
   pDF = pDevice->getCurrentForm ();
   if (!pDF->isFormSupported (DeviceForm::FORM_LETTER))
   {
      cout << "not ";
   }
   cout << "supported" << endl;

   cout << "Auto is ";
   pDT = pDevice->getCurrentTray ();
   if (!pDT->isTraySupported (DeviceTray::TRAY_AUTO))
   {
      cout << "not ";
   }
   cout << "supported" << endl;

   cout << "Plain is ";
   pDM = pDevice->getCurrentMedia ();
   if (!pDM->isMediaSupported (DeviceMedia::MEDIA_PLAIN))
   {
      cout << "not ";
   }
   cout << "supported" << endl;

   cout << "360x360 is ";
   pDR = pDevice->getCurrentResolution ();
   if (!pDR->isResolutionSupported (DeviceResolution::RESOLUTION_360_X_360))
   {
      cout << "not ";
   }
   cout << "supported" << endl;

   cout << "CMYK is ";
   pDPM = pDevice->getCurrentPrintMode ();
   if (!pDPM->isPrintModeSupported (DevicePrintMode::PRINT_MODE_24_CMYK))
   {
      cout << "not ";
   }
   cout << "supported" << endl;
   return 0;
}

int
main (int argc, char *argv[])
{
#define JP_BLOCK_SIZE  256
   int            rc                = 0;
   bool           fUsePDC           = false;
   PFNSHOWOPTIONS pfnShowOptions    = showDeviceGS;
   char          *pszJobProperties  = 0;
   int            cbJobPropertyLeft = 0;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "Error: This program needs glib's module routines!" << endl;

      return 1;
   }

   if (2 > argc)
   {
      printUsage (argv[0]);

      return 2;
   }

   Omni::initialize ();

   for (int i = 1; i < argc; i++)
   {
      if (0 == strcasecmp (argv[i], "all"))
      {
         Enumeration *pEnum = Omni::listDevices ();

         while (pEnum->hasMoreElements ())
         {
            OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

            if (pOD)
            {
               char *pszLibName = pOD->getLibraryName ();

               cout << "Checking: " << pszLibName << endl;

               rc |= executeDevice (pszLibName,
                                    fUsePDC,
                                    pfnShowOptions,
                                    pszJobProperties);

               delete pOD;
            }
            else
            {
               cerr << "Error" << endl;

               rc = 1;
            }
         }

         delete pEnum;
      }
      else if (0 == strcasecmp (argv[i], "UsePDC"))
      {
         fUsePDC = true;
      }
      else if (0 == strcasecmp (argv[i], "!UsePDC"))
      {
         fUsePDC = false;
      }
      else if (0 == strcasecmp (argv[i], "Verbose"))
      {
         pfnShowOptions = showDeviceVerbose;
      }
#ifndef RETAIL
      else if (0 == strcmp (argv[i], "debugoutput"))
      {
         DebugOutput::setDebugOutput (argv[i + 1]);
         i++;
      }
#endif
      else if (0 == strcasecmp (argv[i], "driver"))
      {
         char achLibDeviceName[512];  // @TBD

         i++;

         if (0 != strncasecmp (argv[i] + strlen (argv[i]) - 3,
                               ".so",
                               3))
         {
            sprintf (achLibDeviceName, "lib%s.so", argv[i]);
         }
         else
         {
            strcpy (achLibDeviceName, argv[i]);
         }

         rc |= executeDevice (achLibDeviceName,
                              fUsePDC,
                              pfnShowOptions,
                              pszJobProperties);
      }
      else if (0 == strcasecmp (argv[i], "jobprop"))
      {
         int  iSpaceNeeded = strlen (argv[i + 1]) + 1;

         // Out of space?
         if (  !pszJobProperties
            || cbJobPropertyLeft < iSpaceNeeded
            )
         {
            cbJobPropertyLeft = JP_BLOCK_SIZE;
            if (pszJobProperties)
               cbJobPropertyLeft += strlen (pszJobProperties) + 1;

            char *pszTemp = (char *)calloc (1, cbJobPropertyLeft);

            if (!pszTemp)
            {
               DebugOutput::getErrorStream () << "Error: Out of memory!" << endl;

               rc = 3;
               goto done;
            }

            *pszTemp = '\0';

            if (pszJobProperties)
            {
               strcpy (pszTemp, pszJobProperties);
               free (pszJobProperties);
            }
            pszJobProperties = pszTemp;
         }

         strcat (pszJobProperties, argv[i + 1]);

         cbJobPropertyLeft -= strlen (argv[i + 1]) + 1;

         i++;
      }
      else
      {
         bool         fFoundDevice = false;
         Enumeration *pEnum        = Omni::listDevices ();

         cout << "Info: Trying to figure out which device library to use..." << endl;

         while (  !fFoundDevice
               && pEnum->hasMoreElements ()
               )
         {
            OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

            if (pOD)
            {
               char              *pszLibName        = pOD->getLibraryName ();
               char              *pszJobProperties  = pOD->getJobProperties ();
               Device            *pDevice           = 0;
               PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;
               GModule           *hmodDevice        = 0;

               hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

               if (hmodDevice)
               {
                  g_module_symbol (hmodDevice,
                                   "newDeviceW_JopProp_Advanced",
                                   (gpointer *)&pfnNewDeviceWArgs);

                  if (pfnNewDeviceWArgs)
                  {
                     pDevice = pfnNewDeviceWArgs (pszJobProperties, true);

                     if (pDevice)
                     {
                        if (0 == strcmp (argv[i], pDevice->getShortName ()))
                        {
                           cout << "Info: For faster results use"
                                << endl
                                << "\t"
                                << argv[0];

                           if (pszJobProperties)
                           {
                              cout << " jobprop '"
                                   << pszJobProperties
                                   << "'";
                           }

                           cout << " driver "
                                << pszLibName
                                << endl
                                << endl;

                           pfnShowOptions (pDevice);

                           fFoundDevice = true;
                        }

                        delete pDevice;
                     }
                  }

                  g_module_close (hmodDevice);
               }

               delete pOD;
            }
         }

         delete pEnum;

         if (!fFoundDevice)
         {
            DebugOutput::getErrorStream () << "Error: Both key and value is needed!" << endl;

            rc = 4;
            goto done;
         }
      }
   }

done:
   Omni::terminate ();

   return rc;
}
