/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Idea borrowed from http://hpinkjet.sourceforge.net
 */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <errno.h>
#include <sys/wait.h>

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun {
        int val;                    /* value for SETVAL */
        struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
        unsigned short int *array;  /* array for GETALL, SETALL */
        struct seminfo *__buf;      /* buffer for IPC_INFO */
};
#endif


#include "OmniPDCProxy.hpp"

static char *vpszServerToClient     = "PDC_SRV_TO_CLIENT";
static char *vpszClientToServer     = "PDC_CLIENT_TO_SRV";
static char *vpszExecProgram        = "PDC_EXEC_PROGRAM";
static char *vpszExecProgramDefault = "OmniPDCServer";

char *
setupName (char *pszName, char *pszFormat)
{
#define ENVIRONMENT_SIZE 64
   char   *pszRet = 0;
   size_t  rc;

   pszRet = (char *)malloc (ENVIRONMENT_SIZE);

   if (!pszRet)
      return 0;

   rc = sprintf (pszRet, pszFormat, pszName, getpid ());
   if (rc > ENVIRONMENT_SIZE)
   {
      exit (EXIT_FAILURE);
   }

   if (-1 == putenv (pszRet))
      return 0;

   return pszRet;
}

OmniPDCProxy::
OmniPDCProxy (char *pszClientExe,
              char *pszDeviceName,
              char *pszJobProperties,
              bool  fAdvanced,
              int   fdStdOut,
              int   fdStdErr)
{
   char        *pszS2C            = 0;
   char        *pszC2S            = 0;
   bool         fSucceeded        = false;
   int          pid;
   int          idSem             = -1;

   pszClientExe_d           = 0;

   if (  !pszClientExe
      || !*pszClientExe
      )
   {
      pszClientExe = getenv (vpszExecProgram);

      if (!pszClientExe)
      {
         pszClientExe = vpszExecProgramDefault;
      }
   }

   pszClientExe_d = (char *)malloc (strlen (pszClientExe) + 1);
   if (pszClientExe_d)
   {
      strcpy (pszClientExe_d, pszClientExe);
   }

   fHasError_d              = false;

   fAdvanced_d              = fAdvanced;

   // Assign fifo names for environment space
   pszS2C_d                 = setupName (vpszServerToClient, "%s=/tmp/PDC_s2c_%d");
   pszC2S_d                 = setupName (vpszClientToServer, "%s=/tmp/PDC_c2s_%d");
   pszS2C                   = getenv (vpszServerToClient);
   pszC2S                   = getenv (vpszClientToServer);

   fdS2C_d                  = -1;
   fdC2S_d                  = -1;

   idBuffer1_d              = -1;
   cbBuffer1_d              = 0;
   pbBuffer1_d              = 0;
   idBuffer2_d              = -1;
   cbBuffer2_d              = 0;
   pbBuffer2_d              = 0;

   pCmd_d                   = 0;

   pszVersion_d             = 0;
   pszDriverName_d          = 0;
   pszDeviceName_d          = 0;
   pszShortName_d           = 0;

   iLanguageID_d            = StringResource::LANGUAGE_DEFAULT;
   pLanguage_d              = StringResource::create (iLanguageID_d, 0);

   pDO_d                    = 0;
   pszDitherID_d            = 0;
   pDF_d                    = 0;
   pDT_d                    = 0;
   pDM_d                    = 0;
   pDR_d                    = 0;
   pDPM_d                   = 0;
   pDG_d                    = 0;

   fQueriedPDLInfo_d        = false;
   iPDLLevel_d              = 0;
   iPDLSubLevel_d           = 0;
   iPDLMajorRevisionLevel_d = 0;
   iPDLMinorRevisionLevel_d = 0;

   if (  !pszS2C_d
      || !pszC2S_d
      )
   {
      goto BUGOUT;
   }

   DebugOutput::applyAllDebugOutput (pszJobProperties);

   /* Create actual fifos in filesystem. */
   if ((mkfifo (pszS2C, 0666)) < 0)
   {
      goto BUGOUT;
   }

   if ((mkfifo (pszC2S, 0666)) < 0)
   {
      goto BUGOUT;
   }

   idSem = semget (IPC_PRIVATE, 1, SHM_R | SHM_W);

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": idSem = " << idSem << endl;
#endif

   if (-1 != idSem)
   {
      union semun sunion;

      sunion.val = 0;

#ifndef RETAIL
      int rc =
#endif
               semctl (idSem, 0, SETVAL, sunion);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": semctl returns " << rc << endl;
#endif
   }

   /* Spawn the server. */
   if ((pid = fork ()) < 0)
   {
      cerr << "Error: Fork PDC failed" << endl;

      goto BUGOUT;
   }
   else if (pid > 0)
   {
      /* parent */
      int           rc;
      int           status = 0;
      struct sembuf sb;

      sb.sem_num = 0;
      sb.sem_op = -1;
      sb.sem_flg = 0;

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Parent requesting sem." << endl;
#endif

      rc = semop (idSem, &sb, 1);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Parent semop " << rc << endl;
#endif

      rc = waitpid (pid, &status, WNOHANG);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ())
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": waitpid returns " << rc << ", status = " << status << endl;
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": waitpid pid = " << pid << endl;
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": waitpid WIFEXITED (status) = " << WIFEXITED (status) << endl;
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": waitpid WEXITSTATUS (status) = " << WEXITSTATUS (status) << endl;
      }
#endif

      if (  rc == pid
         && 0 != WIFEXITED (status)
         && EXIT_FAILURE == WEXITSTATUS (status) // child failed
         )
      {
         cerr << "Error: Cannot run client" << endl;

         goto BUGOUT;
      }

      /* Open fifos. */
      if ((fdS2C_d = open (pszS2C, O_RDONLY)) < 0)
      {
         goto BUGOUT;
      }
      if ((fdC2S_d = open (pszC2S, O_WRONLY)) < 0)
      {
         goto BUGOUT;
      }
   }
   else
   {
      /* child */
      struct sembuf sb;
      int           rc;

      sb.sem_num = 0;
      sb.sem_op  = 1;
      sb.sem_flg = 0;

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Child setting sem." << endl;
#endif

      rc = semop (idSem, &sb, 1);

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Child semop " << rc << endl;
#endif

      /* Redirect print_stream to stdout. */
      if (fdStdOut != STDOUT_FILENO)
      {
         dup2 (fdStdOut, STDOUT_FILENO);
      }
      if (fdStdErr != STDERR_FILENO)
      {
         dup2 (fdStdErr, STDERR_FILENO);
      }

      if (execlp (pszClientExe_d, pszClientExe_d, (char *)0, (char *)0) < 0)
      {
         cerr << "Error: Cannot run " << pszClientExe_d << endl;

	      goto BUGOUT2;
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": execlp succeded." << endl;
#endif

      exit (0);

BUGOUT2:
      exit (EXIT_FAILURE);
   }

   pCmd_d = new PrinterCommand ("OmniPDCProxy");

   if (!pCmd_d)
   {
      goto BUGOUT;
   }

   // Initialize the session
   if (  !pCmd_d->setCommand (PDCCMD_INITIALIZE_SESSION,
                              PDC_VERSION)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Failed to initialize the session!" << endl;

      goto BUGOUT;
   }

   // Tell the server what the device name will be
   if (  !pCmd_d->setCommand (PDCCMD_SET_DEVICE_NAME, pszDeviceName)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Failed to set OmniPDCProxy device name!" << endl;

      goto BUGOUT;
   }

   if (fAdvanced_d)
   {
      // Tell the server to used advanced bitblt mode.  This is optional
      pCmd_d->setCommand (PDCCMD_MODE_IS_RENDERER, 1);
      pCmd_d->sendCommand (fdC2S_d);
      pCmd_d->readCommand (fdS2C_d);
   }

   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      // If there are job properties, pass them in
      if (  !pCmd_d->setCommand (PDCCMD_SET_JOB_PROPERTIES, pszJobProperties)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || PDCCMD_ACK != pCmd_d->getCommandType ()
         )
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Failed to set OmniPDCProxy job properties!" << endl;

         goto BUGOUT;
      }
   }

   // Tell the server to create a device instance
   if (  !pCmd_d->setCommand (PDCCMD_NEW_DEVICE, (char *)0)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": Failed to create a new device instance!" << endl;

      goto BUGOUT;
   }

   fSucceeded = true;

BUGOUT:
   if (-1 != idSem)
   {
      semctl (idSem, 0, IPC_RMID, 0);
   }

   if (!fSucceeded)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": failed!" << endl;
#endif

      // @TBD throw error?
      fHasError_d = true;

      cleanupInstance ();
      return;
   }
}

void OmniPDCProxy::
cleanupInstance ()
{
   char *pszS2C = 0;
   char *pszC2S = 0;
   int   rc;

   // End the connection
   if (pCmd_d)
   {
      if (pCmd_d->setCommand (PDCCMD_CLOSE_SESSION, (char *)0))
      {
         pCmd_d->sendCommand (fdC2S_d);
      }

      delete pCmd_d;
      pCmd_d = 0;
   }

   if (pszClientExe_d)
   {
      free (pszClientExe_d);
      pszClientExe_d = 0;
   }

   pszS2C = getenv (vpszServerToClient);
   pszC2S = getenv (vpszClientToServer);

   // Clean up
   if (-1 != fdS2C_d)
   {
      rc = close (fdS2C_d);
      fdS2C_d = -1;
   }

   if (-1 != fdC2S_d)
   {
      rc = close (fdC2S_d);
      fdC2S_d = -1;
   }

   if (pszS2C)
   {
      rc = remove (pszS2C);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": remove (" << pszS2C << ") = " << errno << endl;
      }
   }
   if (pszC2S)
   {
      rc = remove (pszC2S);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": remove (" << pszC2S << ") = " << errno << endl;
      }
   }

   if (pszS2C_d)
   {
      free (pszS2C_d);
      pszS2C_d = 0;
   }
   if (pszC2S_d)
   {
      free (pszC2S_d);
      pszC2S_d = 0;
   }

   if (pbBuffer1_d)
   {
      shmdt (pbBuffer1_d);
      shmctl (idBuffer1_d, IPC_RMID, 0);
      pbBuffer1_d = 0;
   }
   if (pbBuffer2_d)
   {
      shmdt (pbBuffer2_d);
      shmctl (idBuffer2_d, IPC_RMID, 0);
      pbBuffer2_d = 0;
   }

   if (pszVersion_d)
   {
      free (pszVersion_d);
      pszVersion_d = 0;
   }
   if (pszDriverName_d)
   {
      free (pszDriverName_d);
      pszDriverName_d = 0;
   }
   if (pszDeviceName_d)
   {
      free (pszDeviceName_d);
      pszDeviceName_d = 0;
   }
   if (pszShortName_d)
   {
      free (pszShortName_d);
      pszShortName_d = 0;
   }

   iLanguageID_d = 0;
   delete pLanguage_d; pLanguage_d = 0;

   delete pDO_d;  pDO_d = 0;
   if (pszDitherID_d)
   {
      free (pszDitherID_d);
      pszDitherID_d = 0;
   }
   delete pDF_d;  pDF_d  = 0;
   delete pDT_d;  pDT_d  = 0;
   delete pDM_d;  pDM_d  = 0;
   delete pDR_d;  pDR_d  = 0;
   delete pDPM_d; pDPM_d = 0;
   delete pDG_d;  pDG_d  = 0;
}

OmniPDCProxy::
~OmniPDCProxy ()
{
   cleanupInstance ();

   fHasError_d = false;
}

bool OmniPDCProxy::
hasError ()
{
   return fHasError_d;
}

void OmniPDCProxy::
initialize ()
{
}

char * OmniPDCProxy::
getVersion ()
{
   if (!pszVersion_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (PDCCMD_GET_VERSION, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszVersion_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszVersion_d)
            {
               strcpy (pszVersion_d, pszResponse);
            }
         }
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_GET_VERSION failed!" << endl;
      }
   }

   return pszVersion_d;
}

char * OmniPDCProxy::
getDriverName ()
{
   if (!pszDriverName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (PDCCMD_GET_DRIVER_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDriverName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDriverName_d)
            {
               strcpy (pszDriverName_d, pszResponse);
            }
         }
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_GET_DRIVER_NAME failed!" << endl;
      }
   }

   return pszDriverName_d;
}

char * OmniPDCProxy::
getDeviceName ()
{
   if (!pszDeviceName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (PDCCMD_GET_DEVICE_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDeviceName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDeviceName_d)
            {
               strcpy (pszDeviceName_d, pszResponse);
            }
         }
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_GET_DEVICE_NAME failed!" << endl;
      }
   }

   return pszDeviceName_d;
}

char * OmniPDCProxy::
getShortName ()
{
   if (!pszShortName_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (PDCCMD_GET_SHORT_NAME, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDeviceName_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDeviceName_d)
            {
               strcpy (pszDeviceName_d, pszResponse);
            }
         }
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_GET_SHORT_NAME failed!" << endl;
      }
   }

   return pszShortName_d;
}

class OmniPDCOrientationEnumeration : public Enumeration
{
public:
            OmniPDCOrientationEnumeration (char             *pszOrientations,
                                           PrinterCommand   *pCmd,
                                           int               fdC2S,
                                           int               fdS2C);
   virtual ~OmniPDCOrientationEnumeration ();
   bool     hasMoreElements               ();
   void    *nextElement                   ();

private:
   Device           *pDevice_d;
   char             *pszOrientations_d;
   char             *pszOrientationCurrent_d;
   PrinterCommand   *pCmd_d;
   int               fdC2S_d;
   int               fdS2C_d;
};

class OmniPDCProxyOrientation : public DeviceOrientation
{
public:
   OmniPDCProxyOrientation (Device           *pDevice,
                            int               id,
                            PrinterCommand   *pCmd,
                            int               fdC2S,
                            int               fdS2C)
      : DeviceOrientation (pDevice,
                           id)
   {
      pDevice_d = pDevice;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isOrientationSupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_ORIENTATION_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the orientations of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_ORIENTATIONS, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCOrientationEnumeration (pCmd_d->getCommandString (),
                                                   pCmd_d,
                                                   fdC2S_d,
                                                   fdS2C_d);

      return new OmniPDCOrientationEnumeration (0,
                                                pCmd_d,
                                                fdC2S_d,
                                                fdS2C_d);
   }

   static DeviceOrientation *
   create (Device           *pDevice,
           char             *pszName,
           PrinterCommand   *pCmd,
           int               fdC2S,
           int               fdS2C)
   {
      // Query the orientation of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_ORIENTATION, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd->getCommandString ();

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyOrientation:" << __FUNCTION__ << ": id            = " << pszResponse << endl;
         }
#endif

         DeviceOrientation *pDO;

         pDO = new OmniPDCProxyOrientation (pDevice,
                                            DeviceOrientation::nameToID (pszResponse),
                                            pCmd,
                                            fdC2S,
                                            fdS2C);

         return pDO;
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_ORIENTATION failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceOrientation *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand   *pCmd_d;
   int               fdC2S_d;
   int               fdS2C_d;
};

OmniPDCOrientationEnumeration::
OmniPDCOrientationEnumeration (char             *pszOrientations,
                               PrinterCommand   *pCmd,
                               int               fdC2S,
                               int               fdS2C)
{
   pszOrientations_d       = 0;
   pszOrientationCurrent_d = 0;
   pCmd_d                  = pCmd;
   fdC2S_d                 = fdC2S;
   fdS2C_d                 = fdS2C;

   if (  pszOrientations
      && *pszOrientations
      )
   {
      pszOrientations_d = (char *)malloc (strlen (pszOrientations) + 1);
      if (pszOrientations_d)
      {
         strcpy (pszOrientations_d, pszOrientations);

         pszOrientationCurrent_d = pszOrientations_d;
      }
   }
}

OmniPDCOrientationEnumeration::
~OmniPDCOrientationEnumeration ()
{
   if (pszOrientations_d)
   {
      free (pszOrientations_d);
   }
}

bool OmniPDCOrientationEnumeration::
hasMoreElements ()
{
   if (  !pszOrientationCurrent_d
      || !*pszOrientationCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCOrientationEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszOrientationCurrent_d
      && *pszOrientationCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszOrientationCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyOrientation::create (pDevice_d,
                                               pszOrientationCurrent_d,
                                               pCmd_d,
                                               fdC2S_d,
                                               fdS2C_d);

      if (pszSpace)
         pszOrientationCurrent_d = pszSpace + 1;
      else
         pszOrientationCurrent_d += strlen (pszOrientationCurrent_d);
   }

   return pvRet;
}

DeviceOrientation * OmniPDCProxy::
getCurrentOrientation ()
{
   if (pDO_d)
      return pDO_d;

   pDO_d = OmniPDCProxyOrientation::create (this,
                                            (char *)0,
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);

   return pDO_d;
}

char * OmniPDCProxy::
getCurrentDitherID ()
{
   if (!pszDitherID_d)
   {
      // Query the orientation of that device
      if (  pCmd_d->setCommand (PDCCMD_QUERY_CURRENT_DITHER_ID, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
      {
         char *pszResponse = 0;

         pszResponse = pCmd_d->getCommandString ();

         if (  pszResponse
            && *pszResponse
            )
         {
            pszDitherID_d = (char *)malloc (strlen (pszResponse) + 1);

            if (pszDitherID_d)
            {
               strcpy (pszDitherID_d, pszResponse);
            }
         }
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_DITHER_ID failed!" << endl;
      }
   }

   return pszDitherID_d;
}

class OmniPDCFormEnumeration : public Enumeration
{
public:
            OmniPDCFormEnumeration (char             *pszForms,
                                    DeviceResolution *pDR,
                                    PrinterCommand   *pCmd,
                                    int               fdC2S,
                                    int               fdS2C);
   virtual ~OmniPDCFormEnumeration ();
   bool     hasMoreElements        ();
   void    *nextElement            ();

private:
   Device           *pDevice_d;
   char             *pszForms_d;
   char             *pszFormCurrent_d;
   DeviceResolution *pDR_d;
   PrinterCommand   *pCmd_d;
   int               fdC2S_d;
   int               fdS2C_d;
};

class OmniPDCProxyForm : public DeviceForm
{
public:
   OmniPDCProxyForm (Device           *pDevice,
                     int               id,
                     int               iCapabilities,
                     BinaryData       *data,
                     HardCopyCap      *hcInfo,
                     DeviceResolution *pDR,
                     PrinterCommand   *pCmd,
                     int               fdC2S,
                     int               fdS2C)
      : DeviceForm (pDevice,
                    id,
                    iCapabilities,
                    data,
                    hcInfo)
   {
      pDevice_d = pDevice;
      pDR_d     = pDR;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isFormSupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_FORM_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the forms of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_FORMS, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCFormEnumeration (pCmd_d->getCommandString (),
                                            pDR_d,
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);

      return new OmniPDCFormEnumeration (0,
                                         pDR_d,
                                         pCmd_d,
                                         fdC2S_d,
                                         fdS2C_d);
   }

   static DeviceForm *
   create (Device           *pDevice,
           char             *pszName,
           DeviceResolution *pDR,
           PrinterCommand   *pCmd,
           int               fdC2S,
           int               fdS2C)
   {
      // Query the form of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_FORM, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse   = 0;
         char *pszID         = 0;
         int   iCapabilities = 0;
         int   iLeftClip     = 0;
         int   iTopClip      = 0;
         int   iRightClip    = 0;
         int   iBottomClip   = 0;

         pszID = pCmd->getCommandString ();

         pszResponse = strchr (pszID, ' ');

         if (!pszResponse)
            return 0;

         *pszResponse++ = '\0';

         sscanf (pszResponse,
                 "%d %d %d %d %d",
                 &iCapabilities,
                 &iLeftClip,
                 &iTopClip,
                 &iRightClip,
                 &iBottomClip);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": id            = " << pszID         << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": iCapabilities = " << iCapabilities << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": iLeftClip     = " << iLeftClip     << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": iTopClip      = " << iTopClip      << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": iRightClip    = " << iRightClip    << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyForm:" << __FUNCTION__ << ": iBottomClip   = " << iBottomClip   << endl;
         }
#endif

         DeviceForm *pDF;
         int         id  = DeviceForm::nameToID (pszID);

         pDF = new OmniPDCProxyForm (pDevice,
                                     id,
                                     iCapabilities,
                                     0,
                                     new HardCopyCap (iLeftClip,
                                                      iTopClip,
                                                      iRightClip,
                                                      iBottomClip),
                                     pDR,
                                     pCmd,
                                     fdC2S,
                                     fdS2C);

         if (pDF)
         {
            pDF->associateWith (pDR);
         }

         return pDF;
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_FORM failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceForm *
   createV (int id)
   {
      return 0;
   }

private:
   DeviceResolution *pDR_d;
   PrinterCommand   *pCmd_d;
   int               fdC2S_d;
   int               fdS2C_d;
};

OmniPDCFormEnumeration::
OmniPDCFormEnumeration (char             *pszForms,
                        DeviceResolution *pDR,
                        PrinterCommand   *pCmd,
                        int               fdC2S,
                        int               fdS2C)
{
   pszForms_d       = 0;
   pszFormCurrent_d = 0;
   pDR_d            = pDR;
   pCmd_d           = pCmd;
   fdC2S_d          = fdC2S;
   fdS2C_d          = fdS2C;

   if (  pszForms
      && *pszForms
      )
   {
      pszForms_d = (char *)malloc (strlen (pszForms) + 1);
      if (pszForms_d)
      {
         strcpy (pszForms_d, pszForms);

         pszFormCurrent_d = pszForms_d;
      }
   }
}

OmniPDCFormEnumeration::
~OmniPDCFormEnumeration ()
{
   if (pszForms_d)
   {
      free (pszForms_d);
   }
}

bool OmniPDCFormEnumeration::
hasMoreElements ()
{
   if (  !pszFormCurrent_d
      || !*pszFormCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCFormEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszFormCurrent_d
      && *pszFormCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszFormCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyForm::create (pDevice_d,
                                        pszFormCurrent_d,
                                        pDR_d,
                                        pCmd_d,
                                        fdC2S_d,
                                        fdS2C_d);

      if (pszSpace)
         pszFormCurrent_d = pszSpace + 1;
      else
         pszFormCurrent_d += strlen (pszFormCurrent_d);
   }

   return pvRet;
}

DeviceForm * OmniPDCProxy::
getCurrentForm ()
{
   if (pDF_d)
      return pDF_d;

   pDF_d = OmniPDCProxyForm::create (this,
                                     (char *)0,
                                     getCurrentResolution (),
                                     pCmd_d,
                                     fdC2S_d,
                                     fdS2C_d);

   return pDF_d;
}

class OmniPDCTrayEnumeration : public Enumeration
{
public:
            OmniPDCTrayEnumeration (char           *pszTrays,
                                    PrinterCommand *pCmd,
                                    int             fdC2S,
                                    int             fdS2C);
   virtual ~OmniPDCTrayEnumeration ();
   bool     hasMoreElements        ();
   void    *nextElement            ();

private:
   Device         *pDevice_d;
   char           *pszTrays_d;
   char           *pszTrayCurrent_d;
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

class OmniPDCProxyTray : public DeviceTray
{
public:
   OmniPDCProxyTray (Device         *pDevice,
                     int             id,
                     int             iType,
                     BinaryData     *data,
                     PrinterCommand *pCmd,
                     int             fdC2S,
                     int             fdS2C)
      : DeviceTray (pDevice,
                    id,
                    iType,
                    data)
   {
      pDevice_d = pDevice;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isTraySupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_TRAY_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the trays of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_TRAYS, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCTrayEnumeration (pCmd_d->getCommandString (),
                                            pCmd_d,
                                            fdC2S_d,
                                            fdS2C_d);

      return new OmniPDCTrayEnumeration (0,
                                         pCmd_d,
                                         fdC2S_d,
                                         fdS2C_d);
   }

   static DeviceTray *
   create (Device         *pDevice,
           char           *pszName,
           PrinterCommand *pCmd,
           int             fdC2S,
           int             fdS2C)
   {
      // Query the tray of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_TRAY, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse = 0;
         char *pszID       = 0;
         int   iType       = 0;

         pszID = pCmd->getCommandString ();

         pszResponse = strchr (pszID, ' ');

         if (!pszResponse)
            return 0;

         *pszResponse++ = '\0';

         sscanf (pszResponse,
                 "%d",
                 &iType);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyTray:" << __FUNCTION__ << ": id            = " << pszID << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyTray:" << __FUNCTION__ << ": iType         = " << iType << endl;
         }
#endif

         return new OmniPDCProxyTray (pDevice,
                                      DeviceTray::nameToID (pszID),
                                      iType,
                                      0,
                                      pCmd,
                                      fdC2S,
                                      fdS2C);
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_TRAY failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceTray *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

OmniPDCTrayEnumeration::
OmniPDCTrayEnumeration (char           *pszTrays,
                        PrinterCommand *pCmd,
                        int             fdC2S,
                        int             fdS2C)
{
   pszTrays_d       = 0;
   pszTrayCurrent_d = 0;
   pCmd_d           = pCmd;
   fdC2S_d          = fdC2S;
   fdS2C_d          = fdS2C;

   if (  pszTrays
      && *pszTrays
      )
   {
      pszTrays_d = (char *)malloc (strlen (pszTrays) + 1);
      if (pszTrays_d)
      {
         strcpy (pszTrays_d, pszTrays);

         pszTrayCurrent_d = pszTrays_d;
      }
   }
}

OmniPDCTrayEnumeration::
~OmniPDCTrayEnumeration ()
{
   if (pszTrays_d)
   {
      free (pszTrays_d);
   }
}

bool OmniPDCTrayEnumeration::
hasMoreElements ()
{
   if (  !pszTrayCurrent_d
      || !*pszTrayCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCTrayEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszTrayCurrent_d
      && *pszTrayCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszTrayCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyTray::create (pDevice_d,
                                        pszTrayCurrent_d,
                                        pCmd_d,
                                        fdC2S_d,
                                        fdS2C_d);

      if (pszSpace)
         pszTrayCurrent_d = pszSpace + 1;
      else
         pszTrayCurrent_d += strlen (pszTrayCurrent_d);
   }

   return pvRet;
}

DeviceTray * OmniPDCProxy::
getCurrentTray ()
{
   if (pDT_d)
      return pDT_d;

   pDT_d = OmniPDCProxyTray::create (this,
                                     (char *)0,
                                     pCmd_d,
                                     fdC2S_d,
                                     fdS2C_d);

   return pDT_d;
}

class OmniPDCMediaEnumeration : public Enumeration
{
public:
            OmniPDCMediaEnumeration (char           *pszMedias,
                                     PrinterCommand *pCmd,
                                     int             fdC2S,
                                     int             fdS2C);
   virtual ~OmniPDCMediaEnumeration ();
   bool     hasMoreElements         ();
   void    *nextElement             ();

private:
   Device         *pDevice_d;
   char           *pszMedias_d;
   char           *pszMediaCurrent_d;
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

class OmniPDCProxyMedia : public DeviceMedia
{
public:
   OmniPDCProxyMedia (Device         *pDevice,
                      int             id,
                      BinaryData     *data,
                      int             iColorAdjustRequired,
                      int             iAbsorption,
                      PrinterCommand *pCmd,
                      int             fdC2S,
                      int             fdS2C)
      : DeviceMedia (pDevice,
                     id,
                     data,
                     iColorAdjustRequired,
                     iAbsorption)
   {
      pDevice_d = pDevice;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isMediaSupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_MEDIA_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the medias of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_MEDIAS, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCMediaEnumeration (pCmd_d->getCommandString (),
                                             pCmd_d,
                                             fdC2S_d,
                                             fdS2C_d);

      return new OmniPDCMediaEnumeration (0,
                                          pCmd_d,
                                          fdC2S_d,
                                          fdS2C_d);
   }

   static DeviceMedia *
   create (Device           *pDevice,
           char           *pszName,
           PrinterCommand *pCmd,
           int             fdC2S,
           int             fdS2C)
   {
      // Query the media of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_MEDIA, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse          = 0;
         char *pszID                = 0;
         int   iColorAdjustRequired = 0;
         int   iAbsorption          = 0;

         pszID = pCmd->getCommandString ();

         pszResponse = strchr (pszID, ' ');

         if (!pszResponse)
            return 0;

         *pszResponse++ = '\0';

         sscanf (pszResponse,
                 "%d %d",
                 &iColorAdjustRequired,
                 &iAbsorption);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyMedia:" << __FUNCTION__ << ": id                   = " << pszID                << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyMedia:" << __FUNCTION__ << ": iColorAdjustRequired = " << iColorAdjustRequired << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyMedia:" << __FUNCTION__ << ": iAbsorption          = " << iAbsorption          << endl;
         }
#endif

         return new OmniPDCProxyMedia (pDevice,
                                       DeviceMedia::nameToID (pszID),
                                       0,
                                       iColorAdjustRequired,
                                       iAbsorption,
                                       pCmd,
                                       fdC2S,
                                       fdS2C);
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_MEDIA failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceMedia *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

OmniPDCMediaEnumeration::
OmniPDCMediaEnumeration (char           *pszMedias,
                         PrinterCommand *pCmd,
                         int             fdC2S,
                         int             fdS2C)
{
   pszMedias_d       = 0;
   pszMediaCurrent_d = 0;
   pCmd_d            = pCmd;
   fdC2S_d           = fdC2S;
   fdS2C_d           = fdS2C;

   if (  pszMedias
      && *pszMedias
      )
   {
      pszMedias_d = (char *)malloc (strlen (pszMedias) + 1);
      if (pszMedias_d)
      {
         strcpy (pszMedias_d, pszMedias);

         pszMediaCurrent_d = pszMedias_d;
      }
   }
}

OmniPDCMediaEnumeration::
~OmniPDCMediaEnumeration ()
{
   if (pszMedias_d)
   {
      free (pszMedias_d);
   }
}

bool OmniPDCMediaEnumeration::
hasMoreElements ()
{
   if (  !pszMediaCurrent_d
      || !*pszMediaCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCMediaEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszMediaCurrent_d
      && *pszMediaCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszMediaCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyMedia::create (pDevice_d,
                                         pszMediaCurrent_d,
                                         pCmd_d,
                                         fdC2S_d,
                                         fdS2C_d);

      if (pszSpace)
         pszMediaCurrent_d = pszSpace + 1;
      else
         pszMediaCurrent_d += strlen (pszMediaCurrent_d);
   }

   return pvRet;
}

DeviceMedia * OmniPDCProxy::
getCurrentMedia ()
{
   if (pDM_d)
      return pDM_d;

   pDM_d = OmniPDCProxyMedia::create (this,
                                      (char *)0,
                                      pCmd_d,
                                      fdC2S_d,
                                      fdS2C_d);

   return pDM_d;
}

class OmniPDCResolutionEnumeration : public Enumeration
{
public:
           OmniPDCResolutionEnumeration (char           *pszResolutions,
                                         PrinterCommand *pCmd,
                                         int             fdC2S,
                                         int             fdS2C);
   virtual ~OmniPDCResolutionEnumeration ();
   bool     hasMoreElements              ();
   void    *nextElement                  ();

private:
   Device         *pDevice_d;
   char           *pszResolutions_d;
   char           *pszResolutionCurrent_d;
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

class OmniPDCProxyResolution : public DeviceResolution
{
public:
   OmniPDCProxyResolution (Device         *pDevice,
                           int             id,
                           int             iXRes,
                           int             iYRes,
                           int             iXInternalRes,
                           int             iYInternalRes,
                           BinaryData     *pbdData,
                           int             iCapabilities,
                           int             iDestinationBitsPerPel,
                           int             iScanlineMultiple,
                           PrinterCommand *pCmd,
                           int             fdC2S,
                           int             fdS2C)
      : DeviceResolution (pDevice,
                          id,
                          iXRes,
                          iYRes,
                          iXInternalRes,
                          iYInternalRes,
                          pbdData,
                          iCapabilities,
                          iDestinationBitsPerPel,
                          iScanlineMultiple)
   {
      pDevice_d = pDevice;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isResolutionSupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_RESOLUTION_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the resolutions of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_RESOLUTIONS, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCResolutionEnumeration (pCmd_d->getCommandString (),
                                                  pCmd_d,
                                                  fdC2S_d,
                                                  fdS2C_d);

      return new OmniPDCResolutionEnumeration (0,
                                               pCmd_d,
                                               fdC2S_d,
                                               fdS2C_d);
   }

   static DeviceResolution *
   create (Device         *pDevice,
           char           *pszName,
           PrinterCommand *pCmd,
           int             fdC2S,
           int             fdS2C)
   {
      // Query the resolution of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_RESOLUTION, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse            = 0;
         char *pszID                  = 0;
         int   iXRes                  = 0;
         int   iYRes                  = 0;
         int   iXInternalRes          = 0;
         int   iYInternalRes          = 0;
         int   iCapabilities          = 0;
         int   iDestinationBitsPerPel = 0;
         int   iScanlineMultiple      = 0;

         pszID = pCmd->getCommandString ();

         pszResponse = strchr (pszID, ' ');

         if (!pszResponse)
            return 0;

         *pszResponse++ = '\0';

         sscanf (pszResponse,
                 "%d %d %d %d %d %d %d",
                 &iXRes,
                 &iYRes,
                 &iXInternalRes,
                 &iYInternalRes,
                 &iCapabilities,
                 &iDestinationBitsPerPel,
                 &iScanlineMultiple);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": id                     = " << pszID                  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iXRes                  = " << iXRes                  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iYRes                  = " << iYRes                  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iXInternalRes          = " << iXInternalRes          << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iYInternalRes          = " << iYInternalRes          << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iCapabilities          = " << iCapabilities          << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iDestinationBitsPerPel = " << iDestinationBitsPerPel << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyResolution:" << __FUNCTION__ << ": iScanlineMultiple      = " << iScanlineMultiple      << endl;
         }
#endif

         return new OmniPDCProxyResolution (pDevice,
                                            DeviceResolution::nameToID (pszID),
                                            iXRes,
                                            iYRes,
                                            iXInternalRes,
                                            iYInternalRes,
                                            0,
                                            iCapabilities,
                                            iDestinationBitsPerPel,
                                            iScanlineMultiple,
                                            pCmd,
                                            fdC2S,
                                            fdS2C);
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_RESOLUTION failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceResolution *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

OmniPDCResolutionEnumeration::
OmniPDCResolutionEnumeration (char           *pszResolutions,
                              PrinterCommand *pCmd,
                              int             fdC2S,
                              int             fdS2C)
{
   pszResolutions_d       = 0;
   pszResolutionCurrent_d = 0;
   pCmd_d                 = pCmd;
   fdC2S_d                = fdC2S;
   fdS2C_d                = fdS2C;

   if (  pszResolutions
      && *pszResolutions
      )
   {
      pszResolutions_d = (char *)malloc (strlen (pszResolutions) + 1);
      if (pszResolutions_d)
      {
         strcpy (pszResolutions_d, pszResolutions);

         pszResolutionCurrent_d = pszResolutions_d;
      }
   }
}

OmniPDCResolutionEnumeration::
~OmniPDCResolutionEnumeration ()
{
   if (pszResolutions_d)
   {
      free (pszResolutions_d);
   }
}

bool OmniPDCResolutionEnumeration::
hasMoreElements ()
{
   if (  !pszResolutionCurrent_d
      || !*pszResolutionCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCResolutionEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszResolutionCurrent_d
      && *pszResolutionCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszResolutionCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyResolution::create (pDevice_d,
                                              pszResolutionCurrent_d,
                                              pCmd_d,
                                              fdC2S_d,
                                              fdS2C_d);

      if (pszSpace)
         pszResolutionCurrent_d = pszSpace + 1;
      else
         pszResolutionCurrent_d += strlen (pszResolutionCurrent_d);
   }

   return pvRet;
}

DeviceResolution * OmniPDCProxy::
getCurrentResolution ()
{
   if (pDR_d)
      return pDR_d;

   pDR_d = OmniPDCProxyResolution::create (this,
                                           (char *)0,
                                           pCmd_d,
                                           fdC2S_d,
                                           fdS2C_d);

   return pDR_d;
}

class OmniPDCPrintModeEnumeration : public Enumeration
{
public:
            OmniPDCPrintModeEnumeration (char           *pszPrintModes,
                                         PrinterCommand *pCmd,
                                         int             fdC2S,
                                         int             fdS2C);
   virtual ~OmniPDCPrintModeEnumeration ();
   bool     hasMoreElements             ();
   void    *nextElement                 ();

private:
   Device         *pDevice_d;
   char           *pszPrintModes_d;
   char           *pszPrintModeCurrent_d;
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

class OmniPDCProxyPrintMode : public DevicePrintMode
{
public:
   OmniPDCProxyPrintMode (Device         *pDevice,
                          int             id,
                          int             iPhysicalCount,
                          int             iLogicalCount,
                          int             iPlanes,
                          PrinterCommand *pCmd,
                          int             fdC2S,
                          int             fdS2C)
      : DevicePrintMode (pDevice,
                         id,
                         iPhysicalCount,
                         iLogicalCount,
                         iPlanes)
   {
      pDevice_d = pDevice;
      pCmd_d    = pCmd;
      fdC2S_d   = fdC2S;
      fdS2C_d   = fdS2C;
   }

   bool
   isPrintModeSupported (int id)
   {
      if (  pCmd_d->setCommand (PDCCMD_IS_PRINT_MODE_SUPPORTED, id)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;

      return false;
   }

   Enumeration *
   getEnumeration ()
   {
      // Enumerate the print modes of that device
      if (  pCmd_d->setCommand (PDCCMD_ENUM_PRINT_MODES, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return new OmniPDCPrintModeEnumeration (pCmd_d->getCommandString (),
                                                 pCmd_d,
                                                 fdC2S_d,
                                                 fdS2C_d);

      return new OmniPDCPrintModeEnumeration (0,
                                              pCmd_d,
                                              fdC2S_d,
                                              fdS2C_d);
   }

   static DevicePrintMode *
   create (Device         *pDevice,
           char           *pszName,
           PrinterCommand *pCmd,
           int             fdC2S,
           int             fdS2C)
   {
      // Query the print mode of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_PRINT_MODE, pszName)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse    = 0;
         char *pszID          = 0;
         int   iPhysicalCount = 0;
         int   iLogicalCount  = 0;
         int   iPlanes        = 0;

         pszID = pCmd->getCommandString ();

         pszResponse = strchr (pszID, ' ');

         if (!pszResponse)
            return 0;

         *pszResponse++ = '\0';

         sscanf (pszResponse,
                 "%d %d %d",
                 &iPhysicalCount,
                 &iLogicalCount,
                 &iPlanes);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyPrintMode:" << __FUNCTION__ << ": id                     = " << pszID          << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyPrintMode:" << __FUNCTION__ << ": iPhysicalCount         = " << iPhysicalCount << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyPrintMode:" << __FUNCTION__ << ": iLogicalCount          = " << iLogicalCount  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyPrintMode:" << __FUNCTION__ << ": iPlanes                = " << iPlanes        << endl;
         }
#endif

         return new OmniPDCProxyPrintMode (pDevice,
                                           DevicePrintMode::nameToID (pszID),
                                           iPhysicalCount,
                                           iLogicalCount,
                                           iPlanes,
                                           pCmd,
                                           fdC2S,
                                           fdS2C);
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_PRINT_MODE failed!" << endl;
      }

      return 0;
   }

protected:
   DevicePrintMode *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

OmniPDCPrintModeEnumeration::
OmniPDCPrintModeEnumeration (char           *pszPrintModes,
                             PrinterCommand *pCmd,
                             int             fdC2S,
                             int             fdS2C)
{
   pszPrintModes_d       = 0;
   pszPrintModeCurrent_d = 0;
   pCmd_d                = pCmd;
   fdC2S_d               = fdC2S;
   fdS2C_d               = fdS2C;

   if (  pszPrintModes
      && *pszPrintModes
      )
   {
      pszPrintModes_d = (char *)malloc (strlen (pszPrintModes) + 1);
      if (pszPrintModes_d)
      {
         strcpy (pszPrintModes_d, pszPrintModes);

         pszPrintModeCurrent_d = pszPrintModes_d;
      }
   }
}

OmniPDCPrintModeEnumeration::
~OmniPDCPrintModeEnumeration ()
{
   if (pszPrintModes_d)
   {
      free (pszPrintModes_d);
   }
}

bool OmniPDCPrintModeEnumeration::
hasMoreElements ()
{
   if (  !pszPrintModeCurrent_d
      || !*pszPrintModeCurrent_d
      )
      return false;

   return true;
}

void * OmniPDCPrintModeEnumeration::
nextElement ()
{
   void *pvRet = 0;

   if (  pszPrintModeCurrent_d
      && *pszPrintModeCurrent_d
      )
   {
      char *pszSpace;

      pszSpace = strchr (pszPrintModeCurrent_d, ' ');

      if (pszSpace)
      {
         *pszSpace = '\0';
      }

      pvRet = OmniPDCProxyPrintMode::create (pDevice_d,
                                             pszPrintModeCurrent_d,
                                             pCmd_d,
                                             fdC2S_d,
                                             fdS2C_d);

      if (pszSpace)
         pszPrintModeCurrent_d = pszSpace + 1;
      else
         pszPrintModeCurrent_d += strlen (pszPrintModeCurrent_d);
   }

   return pvRet;
}

DevicePrintMode * OmniPDCProxy::
getCurrentPrintMode ()
{
   if (pDPM_d)
      return pDPM_d;

   pDPM_d = OmniPDCProxyPrintMode::create (this,
                                           (char *)0,
                                           pCmd_d,
                                           fdC2S_d,
                                           fdS2C_d);

   return pDPM_d;
}

DeviceBooklet * OmniPDCProxy::
getCurrentBooklet ()
{
   // @TBD
   return 0;
}

DeviceCollation * OmniPDCProxy::
getCurrentCollation ()
{
   // @TBD
   return 0;
}

DeviceCopies * OmniPDCProxy::
getCurrentCopies ()
{
   // @TBD
   return 0;
}

DeviceDestination * OmniPDCProxy::
getCurrentDestination ()
{
   // @TBD
   return 0;
}

DeviceDuplex * OmniPDCProxy::
getCurrentDuplex ()
{
   // @TBD
   return 0;
}

DeviceJogging * OmniPDCProxy::
getCurrentJogging ()
{
   // @TBD
   return 0;
}

DeviceNUp * OmniPDCProxy::
getCurrentNUp ()
{
   // @TBD
   return 0;
}

DeviceScaling * OmniPDCProxy::
getCurrentScaling ()
{
   // @TBD
   return 0;
}

DeviceStapling * OmniPDCProxy::
getCurrentStapling ()
{
   // @TBD
   return 0;
}

class OmniPDCProxyGamma : public DeviceGamma
{
public:
   OmniPDCProxyGamma (int             iCGamma,
                      int             iMGamma,
                      int             iYGamma,
                      int             iKGamma,
                      int             iCBias,
                      int             iMBias,
                      int             iYBias,
                      int             iKBias,
                      PrinterCommand *pCmd,
                      int             fdC2S,
                      int             fdS2C)
      : DeviceGamma (iCGamma,
                     iMGamma,
                     iYGamma,
                     iKGamma,
                     iCBias,
                     iMBias,
                     iYBias,
                     iKBias)
   {
      pCmd_d  = pCmd;
      fdC2S_d = fdC2S;
      fdS2C_d = fdS2C;
   }

   static DeviceGamma *
   create (PrinterCommand *pCmd,
           int             fdC2S,
           int             fdS2C)
   {
      // Query the print mode of that device
      if (  pCmd->setCommand (PDCCMD_QUERY_CURRENT_GAMMA, (char *)0)
         && pCmd->sendCommand (fdC2S)
         && pCmd->readCommand (fdS2C)
         && PDCCMD_ACK == pCmd->getCommandType ()
         )
      {
         char *pszResponse = 0;
         int   iCGamma     = 0;
         int   iMGamma     = 0;
         int   iYGamma     = 0;
         int   iKGamma     = 0;
         int   iCBias      = 0;
         int   iMBias      = 0;
         int   iYBias      = 0;
         int   iKBias      = 0;

         pszResponse = pCmd->getCommandString ();

         sscanf (pszResponse,
                 "%d %d %d %d %d %d %d %d",
                 &iCGamma,
                 &iMGamma,
                 &iYGamma,
                 &iKGamma,
                 &iCBias,
                 &iMBias,
                 &iYBias,
                 &iKBias);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmniPDCProxy ())
         {
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iCGamma = " << iCGamma << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iMGamma = " << iMGamma << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iYGamma = " << iYGamma << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iKGamma = " << iKGamma << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iCBias  = " << iCBias  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iMBias  = " << iMBias  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iYBias  = " << iYBias  << endl;
            DebugOutput::getErrorStream () << "OmniPDCProxyGamma:" << __FUNCTION__ << ": iKBias  = " << iKBias  << endl;
         }
#endif

         return new OmniPDCProxyGamma (iCGamma,
                                       iMGamma,
                                       iYGamma,
                                       iKGamma,
                                       iCBias,
                                       iMBias,
                                       iYBias,
                                       iKBias,
                                       pCmd,
                                       fdC2S,
                                       fdS2C);
      }
      else
      {
         DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": PDCCMD_QUERY_CURRENT_GAMMA failed!" << endl;
      }

      return 0;
   }

protected:
   DeviceGamma *
   createV (int id)
   {
      return 0;
   }

private:
   PrinterCommand *pCmd_d;
   int             fdC2S_d;
   int             fdS2C_d;
};

DeviceGamma * OmniPDCProxy::
getCurrentGamma ()
{
   if (pDG_d)
      return pDG_d;

   pDG_d = OmniPDCProxyGamma::create (pCmd_d,
                                      fdC2S_d,
                                      fdS2C_d);

   return pDG_d;
}

bool OmniPDCProxy::
queryPDLInfo ()
{
   if (fQueriedPDLInfo_d)
   {
      return fQueriedPDLInfo_d;
   }

   if (  !pCmd_d->setCommand (PDCCMD_GET_PDL_INFO, (char *)0)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      || !pCmd_d->getCommandString ()
      )
      return false;

   fQueriedPDLInfo_d = 4 == sscanf (pCmd_d->getCommandString (),
                                    "%d %d %d %d",
                                    &iPDLLevel_d,
                                    &iPDLSubLevel_d,
                                    &iPDLMajorRevisionLevel_d,
                                    &iPDLMinorRevisionLevel_d);

   return true;
}

int OmniPDCProxy::
getPDLLevel ()
{
   if (!queryPDLInfo ())
   {
      return 0;
   }

   return iPDLLevel_d;
}

int OmniPDCProxy::
getPDLSubLevel ()
{
   if (!queryPDLInfo ())
   {
      return 0;
   }

   return iPDLSubLevel_d;
}

int OmniPDCProxy::
getPDLMajorRevisionLevel ()
{
   if (!queryPDLInfo ())
   {
      return 0;
   }

   return iPDLMajorRevisionLevel_d;
}

int OmniPDCProxy::
getPDLMinorRevisionLevel ()
{
   if (!queryPDLInfo ())
   {
      return 0;
   }

   return iPDLMinorRevisionLevel_d;
}

bool OmniPDCProxy::
hasCapability (long lMask)
{
   if (  pCmd_d->setCommand (PDCCMD_HAS_CAPABILITY, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

bool OmniPDCProxy::
hasRasterCapability (long lMask)
{
   if (  pCmd_d->setCommand (PDCCMD_HAS_RASTER_CAPABILITY, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

bool OmniPDCProxy::
hasDeviceOption (long lMask)
{
   if (  pCmd_d->setCommand (PDCCMD_HAS_DEVICE_OPTION, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

bool OmniPDCProxy::
hasDeviceOption2 (long lMask)
{
   if (  pCmd_d->setCommand (PDCCMD_HAS_DEVICE_OPTION2, lMask)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return pCmd_d->getCommandLong ();

   return 0;
}

int OmniPDCProxy::
getScanlineMultiple ()
{
   if (!pDR_d)
   {
      pDR_d = getCurrentResolution ();
   }

   if (pDR_d)
      return pDR_d->getScanlineMultiple ();
   else
      return 1;
}

bool OmniPDCProxy::
beginJob ()
{
   // Start the job on the device
   if (  pCmd_d->setCommand (PDCCMD_BEGIN_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // Start the page on the device
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool OmniPDCProxy::
beginJob (char *pszJobProperties)
{
   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      // If there are job properties, pass them in
      if (  !pCmd_d->setCommand (PDCCMD_SET_JOB_PROPERTIES, pszJobProperties)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || PDCCMD_ACK != pCmd_d->getCommandType ()
         )
         return false;
   }

   // Start the job on the device
   if (  pCmd_d->setCommand (PDCCMD_BEGIN_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // Start the page on the device
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool OmniPDCProxy::
newFrame ()
{
   // Send a new page to the device
   if (  pCmd_d->setCommand (PDCCMD_END_PAGE, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // Start the page on the device
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool OmniPDCProxy::
newFrame (char *pszJobProperties)
{
   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      // If there are job properties, pass them in
      if (  !pCmd_d->setCommand (PDCCMD_SET_JOB_PROPERTIES, pszJobProperties)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || PDCCMD_ACK != pCmd_d->getCommandType ()
         )
         return false;
   }

   // Send a new page to the device
   if (  pCmd_d->setCommand (PDCCMD_END_PAGE, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // Start the page on the device
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool OmniPDCProxy::
endJob ()
{
   // End the page on the device
   if (  pCmd_d->setCommand (PDCCMD_END_PAGE, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // End the job on the device
      if (  pCmd_d->setCommand (PDCCMD_END_JOB, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool OmniPDCProxy::
abortJob ()
{
   // Abort the job on the device
   if (  pCmd_d->setCommand (PDCCMD_ABORT_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool OmniPDCProxy::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniPDCProxy ())
      DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__
           <<  ": pbmi->cbFix = " << pbmi->cbFix
           << ", cbBuffer1_d = " << cbBuffer1_d;
#endif

   if (pbmi->cbFix > cbBuffer1_d)
   {
      // Release the previous buffer
      if (pbBuffer1_d)
      {
         if (  !pCmd_d->setCommand (PDCCMD_DETACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": could not PDCCMD_DETACH_BUFFER1" << endl;
#endif

            return false;
         }

         shmdt (pbBuffer1_d);
         shmctl (idBuffer1_d, IPC_RMID, 0);

         idBuffer1_d = -1;
         cbBuffer1_d = 0;
         pbBuffer1_d = 0;
      }

      idBuffer1_d = shmget (IPC_PRIVATE, pbmi->cbFix, 0666);

      if (0 < idBuffer1_d)
      {
         cbBuffer1_d = pbmi->cbFix;
         pbBuffer1_d = (byte *)shmat (idBuffer1_d, 0, 0);

         if (-1 == (int)pbBuffer1_d)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": 1st shmat (" << idBuffer1_d << ") failed." << endl;
#endif

            return false;
         }

         if (  !pCmd_d->setCommand (PDCCMD_ATTACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": could not PDCCMD_ATTACH_BUFFER1" << endl;
#endif

            return false;
         }
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ())
         DebugOutput::getErrorStream () << ", now cbBuffer1_d = " << cbBuffer1_d;
#endif
   }

   int cbBuffer2;

   cbBuffer2 = pbmi->cy * (((pbmi->cx * pbmi->cBitCount + 31) >> 5) << 2);

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniPDCProxy ())
      DebugOutput::getErrorStream () << ", cbBuffer2_d = " << cbBuffer2_d;
#endif

   if (cbBuffer2 > cbBuffer2_d)
   {
      // Release the previous buffer
      if (pbBuffer2_d)
      {
         if (  !pCmd_d->setCommand (PDCCMD_DETACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": could not PDCCMD_DETACH_BUFFER2" << endl;
#endif

            return false;
         }

         shmdt (pbBuffer2_d);
         shmctl (idBuffer2_d, IPC_RMID, 0);

         idBuffer2_d = -1;
         cbBuffer2_d = 0;
         pbBuffer2_d = 0;
      }

      idBuffer2_d = shmget (IPC_PRIVATE, cbBuffer2, 0666);

      if (0 < idBuffer2_d)
      {
         cbBuffer2_d = cbBuffer2;
         pbBuffer2_d = (byte *)shmat (idBuffer2_d, 0, 0);

         if (-1 == (int)pbBuffer2_d)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": 2nd shmat (" << idBuffer2_d << ") failed." << endl;
#endif

            return false;
         }

         if (  !pCmd_d->setCommand (PDCCMD_ATTACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmniPDCProxy ())
               DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": could not PDCCMD_ATTACH_BUFFER2" << endl;
#endif

            return false;
         }
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmniPDCProxy ())
         DebugOutput::getErrorStream () << ", now cbBuffer2_d = " << cbBuffer2_d;
#endif
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniPDCProxy ()) DebugOutput::getErrorStream () << endl;
#endif

   memcpy (pbBuffer1_d, pbmi,   pbmi->cbFix);
   memcpy (pbBuffer2_d, pbBits, cbBuffer2);

   char achBuffer[64]; // @TBD

   sprintf (achBuffer,
            "%d %d %d %d %d",
            eType,
            prectlPageLocation->xLeft,
            prectlPageLocation->yBottom,
            prectlPageLocation->xRight,
            prectlPageLocation->yTop);

   if (  pCmd_d->setCommand (PDCCMD_RASTERIZE, achBuffer)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmniPDCProxy ())
      DebugOutput::getErrorStream () << "OmniPDCProxy::" << __FUNCTION__ << ": could not PDCCMD_RASTERIZE" << endl;
#endif

   return false;
}

void OmniPDCProxy::
setOutputStream (FILE *pFile)
{
   // @TBD
}

void OmniPDCProxy::
setOutputFunction (PFNOUTPUTFUNCTION  pfnOutputFunction,
                   void              *pMagicCookie)
{
   // @TBD
}

void OmniPDCProxy::
setErrorStream (FILE *pFile)
{
   // @TBD
}

bool OmniPDCProxy::
setLanguage (int iLanguageID)
{
   if (iLanguageID != iLanguageID_d)
   {
      StringResource *pNewLanguage = StringResource::create (iLanguageID, 0);

      if (pNewLanguage)
      {
         delete pLanguage_d;

         iLanguageID_d = iLanguageID;
         pLanguage_d   = pNewLanguage;

         return true;
      }
      else
      {
         return false;
      }
   }

   return true;
}

int OmniPDCProxy::
getLanguage ()
{
   return iLanguageID_d;
}

Enumeration * OmniPDCProxy::
getLanguages ()
{
   return 0;
}

StringResource * OmniPDCProxy::
getLanguageResource ()
{
   return pLanguage_d;
}

string * OmniPDCProxy::
queryCurrentJobProperties ()
{
   // Query the current job properties
   if (  pCmd_d->setCommand (PDCCMD_QUERY_CURRENT_JOB_PROPERTIES, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

class OmniPDCJobPropertyKeyEnumeration : public Enumeration
{
public:
   OmniPDCJobPropertyKeyEnumeration (char *pszJobPropertyKeyIDs)
   {
      pszJobPropertyKeyIDs_d       = 0;
      pszJobPropertyKeyIDCurrent_d = 0;

      if (  pszJobPropertyKeyIDs
         && *pszJobPropertyKeyIDs
         )
      {
         pszJobPropertyKeyIDs_d = (char *)malloc (strlen (pszJobPropertyKeyIDs) + 1);
         if (pszJobPropertyKeyIDs_d)
         {
            strcpy (pszJobPropertyKeyIDs_d, pszJobPropertyKeyIDs);

            pszJobPropertyKeyIDCurrent_d = pszJobPropertyKeyIDs_d;
         }
      }
   }

   virtual
   ~OmniPDCJobPropertyKeyEnumeration ()
   {
      if (pszJobPropertyKeyIDs_d)
      {
         free (pszJobPropertyKeyIDs_d);
      }
   }

   bool
   hasMoreElements ()
   {
      if (  !pszJobPropertyKeyIDCurrent_d
         || !*pszJobPropertyKeyIDCurrent_d
         )
         return false;

      return true;
   }

   void *
   nextElement ()
   {
      void *pvRet = 0;

      if (  pszJobPropertyKeyIDCurrent_d
         && *pszJobPropertyKeyIDCurrent_d
         )
      {
         char *pszSpace;

         pszSpace = strchr (pszJobPropertyKeyIDCurrent_d, ' ');

         if (pszSpace)
         {
            *pszSpace = '\0';
         }

         pvRet = pszJobPropertyKeyIDCurrent_d;

         if (pszSpace)
            pszJobPropertyKeyIDCurrent_d = pszSpace + 1;
         else
            pszJobPropertyKeyIDCurrent_d += strlen (pszJobPropertyKeyIDCurrent_d);
      }

      return pvRet;
   }

private:
   char *pszJobPropertyKeyIDs_d;
   char *pszJobPropertyKeyIDCurrent_d;
};

Enumeration * OmniPDCProxy::
listJobPropertyKeys ()
{
   // Enumerate the job property keys of omni
   if (  pCmd_d->setCommand (PDCCMD_LIST_JOB_PROPERTY_KEYS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new OmniPDCJobPropertyKeyEnumeration (pCmd_d->getCommandString ());

   return new OmniPDCJobPropertyKeyEnumeration (0);
}

class OmniPDCDeviceJobPropertyKeyEnumeration : public Enumeration
{
public:
   OmniPDCDeviceJobPropertyKeyEnumeration (char *pszDeviceJobPropertyKeyIDs)
   {
      pszDeviceJobPropertyKeyIDs_d       = 0;
      pszDeviceJobPropertyKeyIDCurrent_d = 0;

      if (  pszDeviceJobPropertyKeyIDs
         && *pszDeviceJobPropertyKeyIDs
         )
      {
         pszDeviceJobPropertyKeyIDs_d = (char *)malloc (strlen (pszDeviceJobPropertyKeyIDs) + 1);
         if (pszDeviceJobPropertyKeyIDs_d)
         {
            strcpy (pszDeviceJobPropertyKeyIDs_d, pszDeviceJobPropertyKeyIDs);

            pszDeviceJobPropertyKeyIDCurrent_d = pszDeviceJobPropertyKeyIDs_d;
         }
      }
   }

   virtual
   ~OmniPDCDeviceJobPropertyKeyEnumeration ()
   {
      if (pszDeviceJobPropertyKeyIDs_d)
      {
         free (pszDeviceJobPropertyKeyIDs_d);
      }
   }

   bool
   hasMoreElements ()
   {
      if (  !pszDeviceJobPropertyKeyIDCurrent_d
         || !*pszDeviceJobPropertyKeyIDCurrent_d
         )
         return false;

      return true;
   }

   void *
   nextElement ()
   {
      void *pvRet = 0;

      if (  pszDeviceJobPropertyKeyIDCurrent_d
         && *pszDeviceJobPropertyKeyIDCurrent_d
         )
      {
         char *pszSpace;

         pszSpace = strchr (pszDeviceJobPropertyKeyIDCurrent_d, ' ');

         if (pszSpace)
         {
            *pszSpace = '\0';
         }

         pvRet = pszDeviceJobPropertyKeyIDCurrent_d;

         if (pszSpace)
            pszDeviceJobPropertyKeyIDCurrent_d = pszSpace + 1;
         else
            pszDeviceJobPropertyKeyIDCurrent_d += strlen (pszDeviceJobPropertyKeyIDCurrent_d);
      }

      return pvRet;
   }

private:
   char *pszDeviceJobPropertyKeyIDs_d;
   char *pszDeviceJobPropertyKeyIDCurrent_d;
};

Enumeration * OmniPDCProxy::
listDeviceJobPropertyKeys ()
{
   // Enumerate the job property keys of that device
   if (  pCmd_d->setCommand (PDCCMD_LIST_DEVICE_JOB_PROPERTY_KEYS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new OmniPDCDeviceJobPropertyKeyEnumeration (pCmd_d->getCommandString ());

   return new OmniPDCDeviceJobPropertyKeyEnumeration (0);
}

class OmniPDCJobPropertyKeyValuesEnumeration : public Enumeration
{
public:
   OmniPDCJobPropertyKeyValuesEnumeration (char *pszJobPropertyKeyValuesIDs)
   {
      pszJobPropertyKeyValuesIDs_d       = 0;
      pszJobPropertyKeyValuesIDCurrent_d = 0;

      if (  pszJobPropertyKeyValuesIDs
         && *pszJobPropertyKeyValuesIDs
         )
      {
         pszJobPropertyKeyValuesIDs_d = (char *)malloc (strlen (pszJobPropertyKeyValuesIDs) + 1);
         if (pszJobPropertyKeyValuesIDs_d)
         {
            strcpy (pszJobPropertyKeyValuesIDs_d, pszJobPropertyKeyValuesIDs);

            pszJobPropertyKeyValuesIDCurrent_d = pszJobPropertyKeyValuesIDs_d;
         }
      }
   }

   virtual
   ~OmniPDCJobPropertyKeyValuesEnumeration ()
   {
      if (pszJobPropertyKeyValuesIDs_d)
      {
         free (pszJobPropertyKeyValuesIDs_d);
      }
   }

   bool
   hasMoreElements ()
   {
      if (  !pszJobPropertyKeyValuesIDCurrent_d
         || !*pszJobPropertyKeyValuesIDCurrent_d
         )
         return false;

      return true;
   }

   void *
   nextElement ()
   {
      void *pvRet = 0;

      if (  pszJobPropertyKeyValuesIDCurrent_d
         && *pszJobPropertyKeyValuesIDCurrent_d
         )
      {
         char *pszNextKey = pszJobPropertyKeyValuesIDCurrent_d;

         // Is there a begin quote?
         if ('"' == *pszNextKey)
         {
            // Find the end quote
            pszNextKey = strchr (pszNextKey + 1, '"');
         }

         // Find the space
         pszNextKey = strchr (pszNextKey, ' ');

         if (pszNextKey)
         {
            // Break it up
            *pszNextKey = '\0';
         }

         pvRet = pszJobPropertyKeyValuesIDCurrent_d;

         if (pszNextKey)
            pszJobPropertyKeyValuesIDCurrent_d = pszNextKey + 1;
         else
            pszJobPropertyKeyValuesIDCurrent_d += strlen (pszJobPropertyKeyValuesIDCurrent_d);
      }

      return pvRet;
   }

private:
   char *pszJobPropertyKeyValuesIDs_d;
   char *pszJobPropertyKeyValuesIDCurrent_d;
};

Enumeration * OmniPDCProxy::
listKeyValues (char *pszKey)
{
   // Enumerate the job property keys of that device
   if (  pCmd_d->setCommand (PDCCMD_LIST_JOB_PROPERTY_KEY_VALUES, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new OmniPDCJobPropertyKeyValuesEnumeration (pCmd_d->getCommandString ());

   return new OmniPDCJobPropertyKeyValuesEnumeration (0);
}

string * OmniPDCProxy::
getJobPropertyType (char *pszKey)
{
   // Query the current job properties
   if (  pCmd_d->setCommand (PDCCMD_GET_JOB_PROPERTY_TYPE, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

string * OmniPDCProxy::
getJobProperty (char *pszKey)
{
   // Query the current job properties
   if (  pCmd_d->setCommand (PDCCMD_GET_JOB_PROPERTY, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

string * OmniPDCProxy::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
   string cmd = pszKey;

   cmd += "=";
   cmd += pszValue;

   if (  pCmd_d->setCommand (PDCMD_XLATE_JOB_PROPERTY_KEY_VALUE, &cmd)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new string (pCmd_d->getCommandString ());

   return 0;
}

class OmniPDCDitherEnumeration : public Enumeration
{
public:
   OmniPDCDitherEnumeration (char *pszDitherIDs)
   {
      pszDitherIDs_d       = 0;
      pszDitherIDCurrent_d = 0;

      if (  pszDitherIDs
         && *pszDitherIDs
         )
      {
         pszDitherIDs_d = (char *)malloc (strlen (pszDitherIDs) + 1);
         if (pszDitherIDs_d)
         {
            strcpy (pszDitherIDs_d, pszDitherIDs);

            pszDitherIDCurrent_d = pszDitherIDs_d;
         }
      }
   }

   virtual
   ~OmniPDCDitherEnumeration ()
   {
      if (pszDitherIDs_d)
      {
         free (pszDitherIDs_d);
      }
   }

   bool
   hasMoreElements ()
   {
      if (  !pszDitherIDCurrent_d
         || !*pszDitherIDCurrent_d
         )
         return false;

      return true;
   }

   void *
   nextElement ()
   {
      void *pvRet = 0;

      if (  pszDitherIDCurrent_d
         && *pszDitherIDCurrent_d
         )
      {
         char *pszSpace;

         pszSpace = strchr (pszDitherIDCurrent_d, ' ');

         if (pszSpace)
         {
            *pszSpace = '\0';
         }

         pvRet = pszDitherIDCurrent_d;

         if (pszSpace)
            pszDitherIDCurrent_d = pszSpace + 1;
         else
            pszDitherIDCurrent_d += strlen (pszDitherIDCurrent_d);
      }

      return pvRet;
   }

private:
   char *pszDitherIDs_d;
   char *pszDitherIDCurrent_d;
};

Enumeration * OmniPDCProxy::
getDitherEnumeration ()
{
   // Enumerate the dither ids of that device
   if (  pCmd_d->setCommand (PDCCMD_ENUM_DITHER_IDS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return new OmniPDCDitherEnumeration (pCmd_d->getCommandString ());

   return new OmniPDCDitherEnumeration (0);
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string OmniPDCProxy::
toString (ostringstream& oss)
{
   oss << "{ ";

   // @TBD

   oss << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const OmniPDCProxy& const_self)
{
   OmniPDCProxy& self = const_cast<OmniPDCProxy&>(const_self);
   ostringstream oss;

   os << self.toString (oss);

   return os;
}
