/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "PluggableInstance.hpp"

#include <JobProperties.hpp>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/shm.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/time.h>

static char *vpszServerToClient = "OMNI_BLITTER_S2C";
static char *vpszClientToServer = "OMNI_BLITTER_C2S";

char *setupBlitterName (char *pszName, char *pszFormat);

static char
encodeCharacter (int iCh)
{
   if (26 > iCh)
      return 'a' + iCh;
   else if (52 > iCh)
      return 'A' + (iCh - 26);
   else if (62 > iCh)
      return '0' + (iCh - 52);
   else if (62 == iCh)
      return '+';
   else if (63 == iCh)
      return '-';

   return '?';
}

static char *
setupBlitterName (char *pszName, char *pszFormat, char *pszShortName)
{
   ostringstream   oss;
   struct timeval  tv;
   long int        lTime;
   string          stringResult;
   char           *pszRet       = 0;
   int             rc;

   oss << pszName
       << pszFormat
       << pszShortName
       << "_";

   rc = gettimeofday (&tv, 0);
   if (!rc)
   {
      lTime = tv.tv_usec;
      for (int i = 0; i < 6; i++)
      {
         oss << encodeCharacter (lTime & 0x3F);

         lTime >>= 6;
      }

      lTime = tv.tv_sec;
      for (int i = 0; i < 6; i++)
      {
         oss << encodeCharacter (lTime & 0x3F);

         lTime >>= 6;
      }

      oss << ends;
   }

   stringResult = oss.str ();

   pszRet = (char *)malloc (strlen (stringResult.c_str ()) + 1);

   if (!pszRet)
      return 0;

   strcpy (pszRet, stringResult.c_str ());

   if (-1 == putenv (pszRet))
      return 0;

   return pszRet;
}

PluggableInstance::
PluggableInstance (PrintDevice *pDevice,
                   char        *pszExeName,
                   char        *pszData)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": exename = " << pszExeName << endl;
#endif

   fHasError_d             = false;
   pszExeName_d            = 0;
   pszData_d               = 0;

   fdS2C_d                 = -1;
   fdC2S_d                 = -1;
   fRemoveS2C_d            = false;
   fRemoveC2S_d            = false;
   pszS2C_d                = 0;
   pszC2S_d                = 0;

   idBuffer1_d             = -1;
   cbBuffer1_d             = 0;
   pbBuffer1_d             = 0;
   idBuffer2_d             = -1;
   cbBuffer2_d             = 0;
   pbBuffer2_d             = 0;

   pCmd_d                  = 0;

   fJobPropertiesChanged_d = true;

   if (pszExeName)
   {
      pszExeName_d = (char *)malloc (strlen (pszExeName) + 1);
      if (pszExeName_d)
      {
         strcpy (pszExeName_d, pszExeName);
      }
   }

   if (pszData)
   {
      JobProperties jobProp (pszData);

      while (jobProp.hasKeyNameLeft ())
      {
         char *pszKey   = jobProp.getCurrentKey ();
         char *pszValue = jobProp.getCurrentValue ();

         mapDJP_d[string (pszKey)] = string (pszValue);

         jobProp.getNextKey ();
      }
   }
}

PluggableInstance::
~PluggableInstance ()
{
   stopPDCSession ();
}

void PluggableInstance::
initializeInstance ()
{
}

bool PluggableInstance::
hasError ()
{
   return fHasError_d;
}

void PluggableInstance::
startPDCSession ()
{
   if (pCmd_d)
   {
      // Already initialized
      return;
   }

   if (  fHasError_d
      || !pszExeName_d
      )
   {
      // Something failed somewhere
      return;
   }

   char        *pszS2C       = 0;
   char        *pszC2S       = 0;
   int          fdStdOut     = STDOUT_FILENO; // @TBD
   int          fdStdErr     = STDERR_FILENO; // @TBD
   bool         fSucceeded   = false;
   int          pid;

   fHasError_d = true;

   // Assign fifo names for environment space
   pszS2C_d = setupBlitterName (vpszServerToClient,
                                "=/tmp/PDC_B_s2c_",
                                pDevice_d->getShortName ());
   pszC2S_d = setupBlitterName (vpszClientToServer,
                                "=/tmp/PDC_B_c2s_",
                                pDevice_d->getShortName ());
   pszS2C   = getenv (vpszServerToClient);
   pszC2S   = getenv (vpszClientToServer);
   pCmd_d   = new PrinterCommand ("PluggableInstance");

   if (  !pszS2C_d
      || !pszC2S_d
      || !pCmd_d
      )
   {
      if (!pszS2C_d)
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": pszS2C_d is NULL!" << endl;
      if (!pszC2S_d)
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": pszC2S_d is NULL!" << endl;
      if (!pCmd_d)
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": pCmd_d is NULL!" << endl;

      goto BUGOUT;
   }

   /* Create actual fifos in filesystem. */
   if ((mkfifo (pszS2C, 0666)) < 0)
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": mkfifo (" << pszS2C << ") failed." << endl;

      goto BUGOUT;
   }
   fRemoveS2C_d = true;

   if ((mkfifo (pszC2S, 0666)) < 0)
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": mkfifo (" << pszC2S << ") failed." << endl;

      goto BUGOUT;
   }
   fRemoveC2S_d = true;

   /* Spawn the server. */
   if ((pid = fork ()) < 0)
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": fork of " << pszExeName_d << " failed." << endl;

      goto BUGOUT;
   }
   else if (pid > 0)
   {
      /* parent */

      /* Open fifos. */
      if ((fdS2C_d = open (pszS2C, O_RDONLY)) < 0)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": open (" << pszS2C << ") failed." << endl;

         goto BUGOUT;
      }
      if ((fdC2S_d = open (pszC2S, O_WRONLY)) < 0)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": open (" << pszC2S << ") failed." << endl;

         goto BUGOUT;
      }
   }
   else
   {
      /* child */
      char         *argv[5];
      int           i       = 0;

      /* Redirect print_stream to stdout. */
      if (fdStdOut != STDOUT_FILENO)
      {
         dup2 (fdStdOut, STDOUT_FILENO);
      }
      if (fdStdErr != STDERR_FILENO)
      {
         dup2 (fdStdErr, STDERR_FILENO);
      }

//////argv[i++] = "sh";
//////argv[i++] = "-c";
      argv[i++] = (char *)pszExeName_d;
      argv[i++] = NULL;

      if (execvp (argv[0], argv) < 0)
      {
         DebugOutput::getErrorStream () << "Error: execlp of " << pszExeName_d << " failed!" << endl;

         // Send a negative-reply to the waiting parent process.
         int            fdS2C                           = open (pszS2C, O_WRONLY);
         int            fdC2S                           = open (pszC2S, O_RDONLY);
         PrinterCommand cmd ("PluggableInstance/error");

         if (0 < fdS2C)
         {
            cmd.setCommand (PDCCMD_NACK, (char *)0);
            cmd.sendCommand (fdS2C);

            close (fdS2C);
         }
         else
         {
            DebugOutput::getErrorStream () << "Error: Could not send NACK to parent." << endl;
         }

         if (0 < fdC2S)
         {
            cmd.readCommand (fdC2S);

            close (fdC2S);
         }
         else
         {
            DebugOutput::getErrorStream () << "Error: Could not read parent's response to NACK." << endl;
         }

         exit (EXIT_FAILURE);
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": execlp succeded." << endl;
#endif

      exit (0);
   }

   // Initialize the session
   if (  !pCmd_d->setCommand (PDCCMD_INITIALIZE_SESSION,
                              PDC_VERSION)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Cannot initialize the session!" << endl;

      goto BUGOUT;
   }

   fHasError_d = false;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Setting the device name to \"" << pDevice_d->getDeviceName () << "\"" << endl;
#endif

   // Ask the server if the device name is valid
   if (  !pCmd_d->setCommand (PDCCMD_IS_VALID_DEVICE_NAME,
                              pDevice_d->getDeviceName ())
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Pluggable device name is not valid!" << endl;

      goto BUGOUT;
   }

   // Tell the server what the device name will be
   if (  !pCmd_d->setCommand (PDCCMD_SET_DEVICE_NAME,
                              pDevice_d->getDeviceName ())
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Failed to set Pluggable device name!" << endl;

      goto BUGOUT;
   }

   fSucceeded = true;

BUGOUT:
   if (!fSucceeded)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": failed!" << endl;
#endif

      // @TBD throw error?

      stopPDCSession ();

      return;
   }
}

void PluggableInstance::
stopPDCSession ()
{
   char *pszS2C = 0;
   char *pszC2S = 0;
   int   rc     = 0;

   // End the connection
   if (pCmd_d)
   {
      if (  !fHasError_d
         && pCmd_d->setCommand (PDCCMD_CLOSE_SESSION, (char *)0)
         )
      {
         pCmd_d->sendCommand (fdC2S_d);
      }

      delete pCmd_d;

      pCmd_d = 0;
   }

   if (pszS2C_d)
   {
      pszS2C = strchr (pszS2C_d, '=');
      if (pszS2C)
         pszS2C++;
   }
   if (pszC2S_d)
   {
      pszC2S = strchr (pszC2S_d, '=');
      if (pszC2S)
         pszC2S++;
   }

   // Clean up
   if (-1 != fdS2C_d)
   {
      rc = close (fdS2C_d);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": remove (" << pszS2C << ") = " << errno << endl;
      }
      fdS2C_d = -1;
   }

   if (-1 != fdC2S_d)
   {
      rc = close (fdC2S_d);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": remove (" << pszC2S << ") = " << errno << endl;
      }
      fdC2S_d = -1;
   }

   if (  fRemoveS2C_d
      && pszS2C
      )
   {
      rc = remove (pszS2C);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": remove (" << pszS2C << ") = " << errno << endl;
      }
      fRemoveS2C_d = false;
   }

   if (  fRemoveC2S_d
      && pszC2S
      )
   {
      rc = remove (pszC2S);
      if (-1 == rc)
      {
         DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": remove (" << pszC2S << ") = " << errno << endl;
      }
      fRemoveC2S_d = false;
   }

   if (pszS2C_d)
   {
      free (pszS2C_d);
      pszS2C_d = 0;
   }

   if (pszC2S_d)
   {
      free (pszC2S_d);
      pszC2S_d = 0;
   }

   if (pbBuffer1_d)
   {
      shmdt (pbBuffer1_d);
      shmctl (idBuffer1_d, IPC_RMID, 0);
      pbBuffer1_d = 0;
   }

   if (pbBuffer2_d)
   {
      shmdt (pbBuffer2_d);
      shmctl (idBuffer2_d, IPC_RMID, 0);
      pbBuffer2_d = 0;
   }

   if (pszExeName_d)
   {
      free (pszExeName_d);
      pszExeName_d = 0;
   }
   if (pszData_d)
   {
      free (pszData_d);
      pszData_d = 0;
   }

   fHasError_d = true;
}

bool PluggableInstance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
   if (0 == strcasecmp (pszKey, "debugoutput"))
   {
      // Keep all debugout values that are passed in
      string stringKey = pszKey;
      string stringOld = mapDJP_d[stringKey];

      if (!stringOld.empty ())
      {
         mapDJP_d[stringKey] = string (pszValue) + " debugoutput=" + stringOld;
      }
      else
      {
         mapDJP_d[stringKey] = string (pszValue);
      }
   }
   else
   {
      mapDJP_d[string (pszKey)] = string (pszValue);
   }

   return true;
}

string * PluggableInstance::
queryDeviceJobProperties ()
{
   ostringstream oss;

   for ( DJPMap::iterator next = mapDJP_d.begin ();
         next != mapDJP_d.end () ;
         next++ )
   {
      if (oss.str ().c_str ()[0])
      {
         oss << " ";
      }

      oss << next->first
          << "="
          << next->second;
   }

   if (oss.str ().c_str ()[0])
      return new string (oss.str ());
   else
      return 0;
}

class KeyEnumerator : public Enumeration
{
public:
   KeyEnumerator (char *pszKeys)
   {
      pszKeys_d    = 0;
      pszCurrent_d = 0;

      if (  pszKeys
         && *pszKeys
         )
      {
         pszKeys_d = (char *)malloc (strlen (pszKeys) + 1);
         if (pszKeys_d)
         {
            strcpy (pszKeys_d, pszKeys);
            pszCurrent_d = pszKeys_d;
         }
      }
   }

   virtual ~KeyEnumerator ()
   {
      if (pszKeys_d)
      {
         free (pszKeys_d);
         pszKeys_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return pszCurrent_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = pszCurrent_d;

      if (  pszCurrent_d
         && *pszCurrent_d
         )
      {
         char *pszSpace = strchr (pszCurrent_d, ' ');

         if (pszSpace)
         {
            *pszSpace = '\0';
            pszCurrent_d = pszSpace + 1;
            while (' ' == *pszCurrent_d)
            {
               pszCurrent_d++;
            }
            if (!*pszCurrent_d)
            {
               pszCurrent_d = 0;
            }
         }
         else
         {
            pszCurrent_d = 0;
         }
      }
      else
      {
         pszCurrent_d = 0;
      }

      return pvRet;
   }

private:
   char *pszKeys_d;
   char *pszCurrent_d;
};

Enumeration * PluggableInstance::
listDeviceJobPropertyKeys ()
{
   char *pszKeys = 0;

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return new KeyEnumerator (0);
      }
   }

   if (  pCmd_d->setCommand (PDCCMD_LIST_DEVICE_JOB_PROPERTY_KEYS, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      pszKeys = pCmd_d->getCommandString ();
   }

   return new KeyEnumerator (pszKeys);
}

Enumeration * PluggableInstance::
listKeyValues (char *pszKey)
{
   char *pszKeys = 0;

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   if (  pCmd_d->setCommand (PDCCMD_LIST_JOB_PROPERTY_KEY_VALUES, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      pszKeys = pCmd_d->getCommandString ();
   }

   return new KeyEnumerator (pszKeys);
}

string * PluggableInstance::
getJobPropertyType (char *pszKey)
{
   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   if (  pCmd_d->setCommand (PDCCMD_GET_JOB_PROPERTY_TYPE, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      return new string (pCmd_d->getCommandString ());
   }

   return 0;
}

string * PluggableInstance::
getJobProperty (char *pszKey)
{
   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   if (  pCmd_d->setCommand (PDCCMD_GET_JOB_PROPERTY, pszKey)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      return new string (pCmd_d->getCommandString ());
   }

   return 0;
}

string * PluggableInstance::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   if (!pCmd_d->setCommand (PDCMD_XLATE_JOB_PROPERTY_KEY_VALUE, pszKey))
   {
      return DeviceInstance::translateKeyValue (pszKey, pszValue);
   }

   if (  pszValue
      && (  !pCmd_d->appendCommand ("=")
         || !pCmd_d->appendCommand (pszValue)
         )
      )
   {
      return DeviceInstance::translateKeyValue (pszKey, pszValue);
   }

   if (  pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      return new string (pCmd_d->getCommandString ());
   }

   return DeviceInstance::translateKeyValue (pszKey, pszValue);
}

bool PluggableInstance::
setOutputStream (FILE *pFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Setting the output stream to " << pFile << endl;
#endif

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   char achTBD[25];
   int  iFileNo    = fileno (pFile);

   sprintf (achTBD, "%d", iFileNo);

   if (  !pCmd_d->setCommand (PDCCMD_SET_OUTPUT_STREAM, achTBD)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      return true;
   }
   else
   {
      return false;
   }
}

bool PluggableInstance::
setErrorStream (FILE *pFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Setting the error stream to " << pFile << endl;
#endif

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   char achTBD[25];
   int  iFileNo    = fileno (pFile);

   sprintf (achTBD, "%d", iFileNo);

   if (  !pCmd_d->setCommand (PDCCMD_SET_ERROR_STREAM, achTBD)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      return true;
   }
   else
   {
      return false;
   }
}

bool PluggableInstance::
setLanguage (int iLanguageID)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Setting the language to " << iLanguageID << endl;
#endif

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   char *pszLanguageID = StringResource::IDToName (iLanguageID);

   if (  !pCmd_d->setCommand (PDCMD_SET_TRANSLATABLE_LANGUAGE, pszLanguageID)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      return true;
   }
   else
   {
      return false;
   }
}

void PluggableInstance::
pushDeviceObjects ()
{
   char               achResult[128];     // @TBD
   DeviceOrientation *pDO            = 0;
   DeviceForm        *pDF            = 0;
   HardCopyCap       *pHCC           = 0;
   DeviceTray        *pDT            = 0;
   DeviceMedia       *pDM            = 0;
   DeviceResolution  *pDR            = 0;
   DevicePrintMode   *pDPM           = 0;
   DeviceGamma       *pDG            = 0;

   pDO = pDevice_d->getCurrentOrientation ();
   sprintf (achResult,
            "%s",
            DeviceOrientation::IDToName (pDO->getID ()));

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_ORIENTATION, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_DITHER_ID, pDevice_d->getCurrentDitherID ())
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   // Push the resolution before pushing the form!
   pDR = pDevice_d->getCurrentResolution ();
   sprintf (achResult,
            "%s %d %d %d %d %d %d %d",
            DeviceResolution::IDToName (pDR->getID ()),
            pDR->getExternalXRes (),
            pDR->getExternalYRes (),
            pDR->getInternalXRes (),
            pDR->getInternalYRes (),
            pDR->getCapabilities (),
            pDR->getDstBitsPerPel (),
            pDR->getScanlineMultiple ());

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_RESOLUTION, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   pDF  = pDevice_d->getCurrentForm ();
   pHCC = pDF->getHardCopyCap ();
   sprintf (achResult,
            "%s %d %d %d %d %d",
            DeviceForm::IDToName (pDF->getID ()),
            pDF->getCapabilities (),
            pHCC->getLeftClip (),
            pHCC->getTopClip (),
            pHCC->getRightClip (),
            pHCC->getBottomClip ());

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_FORM, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   pDT = pDevice_d->getCurrentTray ();
   sprintf (achResult,
            "%s %d",
            DeviceTray::IDToName (pDT->getID ()),
            pDT->getType ());

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_TRAY, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   pDM = pDevice_d->getCurrentMedia ();
   sprintf (achResult,
            "%s %d %d",
            DeviceMedia::IDToName (pDM->getID ()),
            pDM->getColorAdjustRequired (),
            pDM->getAbsorption ());

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_MEDIA, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   pDPM = pDevice_d->getCurrentPrintMode ();
   sprintf (achResult,
            "%s %d %d %d",
            DevicePrintMode::IDToName (pDPM->getID ()),
            pDPM->getPhysicalCount (),
            pDPM->getLogicalCount (),
            pDPM->getNumPlanes ());

   if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_PRINT_MODE, achResult)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
      return;

   pDG = pDevice_d->getCurrentGamma ();
   if (pDG)
   {
      sprintf (achResult,
               "%d %d %d %d %d %d %d %d",
               pDG->getCGamma (),
               pDG->getMGamma (),
               pDG->getYGamma (),
               pDG->getKGamma (),
               pDG->getCBias (),
               pDG->getMBias (),
               pDG->getYBias (),
               pDG->getKBias ());

      if (  !pCmd_d->setCommand (PDCCMD_PUSH_CURRENT_GAMMA, achResult)
         || !pCmd_d->sendCommand (fdC2S_d)
         || !pCmd_d->readCommand (fdS2C_d)
         || PDCCMD_ACK != pCmd_d->getCommandType ()
         )
         return;
   }
}

bool PluggableInstance::
setJobProperties ()
{
   string *pstrJobProps = 0;
   bool    fRC          = false;

   pushDeviceObjects ();

   // Tell the server what the job properties are
   pstrJobProps = pDevice_d->queryDeviceJobProperties ();

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Setting the job properties to \"" << (pstrJobProps ? *pstrJobProps : "(null)") << "\"" << endl;
#endif

   if (  !pCmd_d->setCommand (PDCCMD_SET_JOB_PROPERTIES, pstrJobProps)
      || !pCmd_d->sendCommand (fdC2S_d)
      || !pCmd_d->readCommand (fdS2C_d)
      || PDCCMD_ACK != pCmd_d->getCommandType ()
      )
   {
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": Failed to set Pluggable job properties!" << endl;
   }
   else
   {
      fRC = true;
   }

   delete pstrJobProps;

   return fRC;
}

bool PluggableInstance::
commonBeginJob ()
{
   if (fHasError_d)
   {
      return false;
   }

   if (fJobPropertiesChanged_d)
   {
      if (!setJobProperties ())
         return false;

      fJobPropertiesChanged_d = false;
   }

   // Start the job on the device
   if (  pCmd_d->setCommand (PDCCMD_BEGIN_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool PluggableInstance::
beginJob ()
{
   if (fHasError_d)
   {
      return false;
   }

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   return commonBeginJob ();
}

bool PluggableInstance::
beginJob (bool fJobPropertiesChanged)
{
   bool    fRetVal      = false;

   if (fHasError_d)
   {
      return false;
   }

   if (!pCmd_d)
   {
      startPDCSession ();

      if (!pCmd_d)
      {
         return false;
      }
   }

   fJobPropertiesChanged_d = true;

   fRetVal = commonBeginJob ();

   return fRetVal;
}

bool PluggableInstance::
commonNewFrame ()
{
   if (fJobPropertiesChanged_d)
   {
      if (!setJobProperties ())
         return false;

      fJobPropertiesChanged_d = false;
   }

   // Send a new page to the device
   if (  pCmd_d->setCommand (PDCCMD_END_PAGE, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // Send a new page to the device
      if (  pCmd_d->setCommand (PDCCMD_START_PAGE, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool PluggableInstance::
newFrame ()
{
   if (  fHasError_d
      || !pCmd_d
      )
   {
      return false;
   }

   return commonNewFrame ();
}

bool PluggableInstance::
newFrame (bool fJobPropertiesChanged)
{
   bool fRetVal = false;

   if (  fHasError_d
      || !pCmd_d
      )
   {
      return false;
   }

   fJobPropertiesChanged_d = true;

   fRetVal = commonNewFrame ();

   return fRetVal;
}

bool PluggableInstance::
endJob ()
{
   if (  fHasError_d
      || !pCmd_d
      )
   {
      return false;
   }

   // End the job on the device
   if (  pCmd_d->setCommand (PDCCMD_END_PAGE, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
   {
      // End the job on the device
      if (  pCmd_d->setCommand (PDCCMD_END_JOB, (char *)0)
         && pCmd_d->sendCommand (fdC2S_d)
         && pCmd_d->readCommand (fdS2C_d)
         && PDCCMD_ACK == pCmd_d->getCommandType ()
         )
         return true;
   }

   return false;
}

bool PluggableInstance::
abortJob ()
{
   if (  fHasError_d
      || !pCmd_d
      )
   {
      return false;
   }

   // Abort the job on the device
   if (  pCmd_d->setCommand (PDCCMD_ABORT_JOB, (char *)0)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return true;

   return false;
}

bool PluggableInstance::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
   if (  fHasError_d
      || !pCmd_d
      )
   {
      return false;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ())
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__
           <<  ": pbmi->cbFix = " << pbmi->cbFix
           << ", cbBuffer1_d = " << cbBuffer1_d;
#endif

   if (pbmi->cbFix > cbBuffer1_d)
   {
      // Release the previous buffer
      if (pbBuffer1_d)
      {
         if (  !pCmd_d->setCommand (PDCCMD_DETACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": could not PDCCMD_DETACH_BUFFER1" << endl;
#endif

            return false;
         }

         shmdt (pbBuffer1_d);
         shmctl (idBuffer1_d, IPC_RMID, 0);

         idBuffer1_d = -1;
         cbBuffer1_d = 0;
         pbBuffer1_d = 0;
      }

      idBuffer1_d = shmget (IPC_PRIVATE, pbmi->cbFix, 0666);

      if (0 < idBuffer1_d)
      {
         cbBuffer1_d = pbmi->cbFix;
         pbBuffer1_d = (byte *)shmat (idBuffer1_d, 0, 0);

         if (-1 == (int)pbBuffer1_d)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": 1st shmat (" << idBuffer1_d << ") failed." << endl;
#endif

            return false;
         }

         if (  !pCmd_d->setCommand (PDCCMD_ATTACH_BUFFER1, idBuffer1_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": could not PDCCMD_ATTACH_BUFFER1" << endl;
#endif

            return false;
         }
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputPluggableInstance ())
         DebugOutput::getErrorStream () << ", now cbBuffer1_d = " << cbBuffer1_d;
#endif
   }

   int cbBuffer2;

   cbBuffer2 = pbmi->cy * (((pbmi->cx * pbmi->cBitCount + 31) >> 5) << 2);

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ())
      DebugOutput::getErrorStream () << ", cbBuffer2_d = " << cbBuffer2_d;
#endif

   if (cbBuffer2 > cbBuffer2_d)
   {
      // Release the previous buffer
      if (pbBuffer2_d)
      {
         if (  !pCmd_d->setCommand (PDCCMD_DETACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": could not PDCCMD_DETACH_BUFFER2" << endl;
#endif

            return false;
         }

         shmdt (pbBuffer2_d);
         shmctl (idBuffer2_d, IPC_RMID, 0);

         idBuffer2_d = -1;
         cbBuffer2_d = 0;
         pbBuffer2_d = 0;
      }

      idBuffer2_d = shmget (IPC_PRIVATE, cbBuffer2, 0666);

      if (0 < idBuffer2_d)
      {
         cbBuffer2_d = cbBuffer2;
         pbBuffer2_d = (byte *)shmat (idBuffer2_d, 0, 0);

         if (-1 == (int)pbBuffer2_d)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": 2nd shmat (" << idBuffer2_d << ") failed." << endl;
#endif

            return false;
         }

         if (  !pCmd_d->setCommand (PDCCMD_ATTACH_BUFFER2, idBuffer2_d)
            || !pCmd_d->sendCommand (fdC2S_d)
            || !pCmd_d->readCommand (fdS2C_d)
            || PDCCMD_ACK != pCmd_d->getCommandType ()
            )
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputPluggableInstance ())
               DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": could not PDCCMD_ATTACH_BUFFER2" << endl;
#endif

            return false;
         }
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputPluggableInstance ())
         DebugOutput::getErrorStream () << ", now cbBuffer2_d = " << cbBuffer2_d;
#endif
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ()) DebugOutput::getErrorStream () << endl;
#endif

   memcpy (pbBuffer1_d, pbmi,   pbmi->cbFix);
   memcpy (pbBuffer2_d, pbBits, cbBuffer2);

   char achBuffer[64]; // @TBD

   sprintf (achBuffer,
            "%d %d %d %d %d",
            eType,
            prectlPageLocation->xLeft,
            prectlPageLocation->yBottom,
            prectlPageLocation->xRight,
            prectlPageLocation->yTop);

   if (  pCmd_d->setCommand (PDCCMD_RASTERIZE, achBuffer)
      && pCmd_d->sendCommand (fdC2S_d)
      && pCmd_d->readCommand (fdS2C_d)
      && PDCCMD_ACK == pCmd_d->getCommandType ()
      )
      return true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPluggableInstance ())
      DebugOutput::getErrorStream () << "PluggableInstance::" << __FUNCTION__ << ": could not PDCCMD_RASTERIZE" << endl;
#endif

   return false;
}

string PluggableInstance::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{ "
       << DeviceInstance::toString (oss2) // Add parent's output as well
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const PluggableInstance& const_self)
{
   PluggableInstance& self = const_cast<PluggableInstance&>(const_self);
   ostringstream      oss;

   os << self.toString (oss);

   return os;
}
