/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
//#include <UPDFDeviceBooklet.hpp>
//#include <UPDFDeviceCollation.hpp>
#include <UPDFDeviceCopies.hpp>
//#include <UPDFDeviceDestination.hpp>
#include <UPDFDeviceDuplex.hpp>
#include <UPDFDeviceJogging.hpp>
#include <UPDFDeviceForm.hpp>
#include <UPDFDeviceMedia.hpp>
#include <UPDFDeviceNUp.hpp>
#include <UPDFDeviceOrientation.hpp>
#include <UPDFDevicePrintMode.hpp>
#include <UPDFDeviceResolution.hpp>
//#include <UPDFDeviceScaling.hpp>
//#include <UPDFDeviceStapling.hpp>
#include <UPDFDeviceTray.hpp>
//#include <UPDFDeviceCommands.hpp>
#include <UPDFDeviceInstance.hpp>
#include <UPDFDeviceBlitter.hpp>
#include <UPDFDevice.hpp>
#include <JobProperties.hpp>

#include <glob.h>

static bool      isDeviceConfigurationXMLFile       (xmlDocPtr   doc);
static xmlDocPtr getDeviceConfigurationFromFile     (char       *pszMasterXMLFile);
static xmlDocPtr getDefaultDeviceConfiguration      ();
static xmlDocPtr getUDRFromDeviceConfiguration      (xmlDocPtr   docDC);
static bool      getOmniNames                       (xmlDocPtr   docDC,
                                                     char      **ppszDriverName,
                                                     char      **ppszDeviceName,
                                                     char      **ppszShortName);
static void      MyValidityErrorFunc                (void       *ctx,
                                                     const char *msg,
                                                                 ...);

char *
getVersion ()
{
   return VERSION;
}

Enumeration *
getDeviceEnumeration ()
{
   // @TBD
   return 0;
}

Device *
newDeviceW_Advanced (bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": fAdvanced = " << fAdvanced << endl;
#endif

   xmlDocPtr docDC          = getDefaultDeviceConfiguration ();
   xmlDocPtr docUDR         = 0;
   char      *pszDriverName = 0;
   char      *pszDeviceName = 0;
   char      *pszShortName  = 0;

   if (docDC)
   {
      docUDR = getUDRFromDeviceConfiguration (docDC);
   }

   if (  docDC
      && docUDR
      && getOmniNames (docUDR,
                       &pszDriverName,
                       &pszDeviceName,
                       &pszShortName)
      )
   {
      PrintDevice *pPrintDevice = new UPDFDevice (docDC,
                                                  docUDR,
                                                  pszDriverName,
                                                  pszDeviceName,
                                                  pszShortName,
                                                  0,  // @TBD
                                                  0); // @TBD

      if (pPrintDevice)
      {
         pPrintDevice->initialize ();

         if (fAdvanced)
            return pPrintDevice;
         else
            return new OmniProxy (pPrintDevice);
      }
   }

   return 0;
}

Device *
newDeviceW_JopProp_Advanced (char *pszJobProperties, bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": pszJobProperties = \"" << pszJobProperties << "\", fAdvanced = " << fAdvanced << endl;
#endif

   xmlDocPtr  docDC         = 0;
   xmlDocPtr  docUDR        = 0;
   char      *pszDriverName = 0;
   char      *pszDeviceName = 0;
   char      *pszShortName  = 0;

   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      JobProperties jobProp (pszJobProperties);

      while (jobProp.hasKeyNameLeft ())
      {
         char *pszKey   = jobProp.getCurrentKey ();
         char *pszValue = jobProp.getCurrentValue ();

         if (0 == strcmp (pszKey, "UPDFMasterFile"))
         {
            docDC = getDeviceConfigurationFromFile (pszValue);

            if (!docDC)
            {
               DebugOutput::getErrorStream () << "Warning: Could not load Device Configuration from " << pszValue << endl;
            }
         }
#ifndef RETAIL
         else if (0 == strcmp (pszKey, "debugoutput"))
         {
            DebugOutput::setDebugOutput (pszValue);
         }
#endif

         jobProp.getNextKey ();
      }
   }

   if (!docDC)
   {
      docDC = getDefaultDeviceConfiguration ();
   }

   if (docDC)
   {
      docUDR = getUDRFromDeviceConfiguration (docDC);
   }

   if (  docDC
      && docUDR
      && getOmniNames (docUDR,
                       &pszDriverName,
                       &pszDeviceName,
                       &pszShortName)
      )
   {
      PrintDevice *pPrintDevice = new UPDFDevice (docDC,
                                                  docUDR,
                                                  pszDriverName,
                                                  pszDeviceName,
                                                  pszShortName,
                                                  pszJobProperties,
                                                  0);               // @TBD

      pPrintDevice->initialize ();

      if (fAdvanced)
         return pPrintDevice;
      else
         return new OmniProxy (pPrintDevice);
   }

   if (docDC)
   {
      xmlFreeDoc (docDC);
   }

   return 0;
}

void
deleteDevice (Device *pDevice)
{
   delete pDevice;
}

static bool
isDeviceConfigurationXMLFile (xmlDocPtr doc)
{
   xmlNodePtr nodeRoot = xmlDocGetRootElement (doc);
   bool       fRc      = false;

   if (nodeRoot)
   {
      xmlNodePtr nodeElm = 0;

      nodeElm = firstNode (nodeRoot);

      if (nodeElm)
      {
         if (0 == strcmp ((char *)nodeElm->name, "DeviceConfiguration"))
         {
            fRc = true;
         }
      }
   }

   return fRc;
}

static xmlDocPtr
getDeviceConfigurationFromFile (char *pszMasterXMLFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": pszMasterXMLFile = " << pszMasterXMLFile << endl;
#endif

   xmlDocPtr  docRet   = 0;

   docRet = xmlParseFile (pszMasterXMLFile);
///xmlValidateDocument (&validator, docRet); // @TBD

   if (docRet)
   {
      if (isDeviceConfigurationXMLFile (docRet))
      {
         return docRet;
      }

      xmlFreeDoc (docRet);
   }

   return 0;
}

static xmlDocPtr
getDefaultDeviceConfiguration ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": " << endl;
#endif

   xmlDocPtr docRet  = 0;
   glob_t    globbuf;
   int       rc;

   memset (&globbuf, 0, sizeof (globbuf));
   rc = glob ("*.xml", 0, NULL, &globbuf);

   if (0 == rc)
   {
      for (int i = 0; i < (int)globbuf.gl_pathc; i++)
      {
         docRet = xmlParseFile (globbuf.gl_pathv[i]);
/////////xmlValidateDocument (&validator, docRet); // @TBD

         if (docRet)
         {
            if (isDeviceConfigurationXMLFile (docRet))
            {
               break;
            }

            xmlFreeDoc (docRet);
         }
      }
   }

   globfree (&globbuf);

   return docRet;
}

static xmlDocPtr
getUDRFromDeviceConfiguration (xmlDocPtr docDC)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": " << endl;
#endif

   xmlDocPtr   docRet     = 0;
   xmlNodePtr  nodeRootDC = 0;
   char       *pszUDR     = 0;

   nodeRootDC = xmlDocGetRootElement (docDC);

   pszUDR = (char *)xmlGetProp (nodeRootDC, (xmlChar *)"Unit_Description_Reference");

   if (pszUDR)
   {
      string stringUDR = pszUDR;

      stringUDR += ".xml";

      docRet = xmlParseFile (stringUDR.c_str ());

      if (docRet)
      {
/////////xmlValidateDocument (&validator, docRet); // @TBD
      }

      free (pszUDR);
   }

   return docRet;
}

static bool
getOmniNames (xmlDocPtr   docDC,
              char      **ppszDriverName,
              char      **ppszDeviceName,
              char      **ppszShortName)
{
   bool fRc = false;

   if (  docDC
      && ppszDriverName
      )
   {
      xmlNodePtr nodeDeviceHeader = xmlDocGetRootElement (docDC);

      nodeDeviceHeader = UPDFDevice::findEntry (nodeDeviceHeader,
                                                "Device_Header",
                                                DebugOutput::shouldOutputUPDFDevice ());

      if (nodeDeviceHeader)
      {
         *ppszDriverName = (char *)xmlGetProp (nodeDeviceHeader,
                                               (xmlChar *)"Manufacturer");

         *ppszDeviceName = (char *)xmlGetProp (nodeDeviceHeader,
                                               (xmlChar *)"ProductName");

         if (  *ppszDeviceName
            && 0 < strlen (*ppszDeviceName)
            )
         {
            *ppszShortName = (char *)malloc (strlen (*ppszDeviceName) + 1);

            if (*ppszShortName)
            {
               char *pszFrom = *ppszDeviceName;
               char *pszTo   = *ppszShortName;

               while (*pszFrom)
               {
                  switch (*pszFrom)
                  {
                  case '-':
                  case '+':
                  case '(':
                  case ')':
                  case ' ':
                  case '/':
                  case '\\':
                  {
                     *pszTo++ = '_';
                     pszFrom++;
                     break;
                  }
                  default:
                  {
                     *pszTo++ = *pszFrom++;
                  }
                  }
               }
               *pszTo++ = '\0';
            }
         }

         if (  *ppszDriverName
            && *ppszDeviceName
            && *ppszShortName
            )
         {
            fRc = true;
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ())
      DebugOutput::getErrorStream () << "UPDFDevice: "
                                     << __FUNCTION__
                                     << ": *ppszDriverName = "
                                     << (ppszDriverName ? *ppszDriverName : "(null)")
                                     << ", *ppszDeviceName = "
                                     << (ppszDeviceName ? *ppszDeviceName : "(null)")
                                     << ", *ppszShortName = "
                                     << (ppszShortName ? *ppszShortName : "(null)")
                                     << endl;
#endif

   return fRc;
}

static void
MyValidityErrorFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Error: ");
   vprintf (msg, list);

   va_end (list);

   if (ctx)
   {
      // @TBD
   }
}

static void
MyValidityWarningFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Warning: ");
   vprintf (msg, list);

   va_end (list);
}

void UPDFDevice::
commonInit (xmlDocPtr docDC,
            xmlDocPtr docUDR)
{
   xmlValidCtxt validator = {
      this,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };

   setCapabilities ( Capability::MIRROR
                   | Capability::MONOCHROME
                   );

   setRasterCapabilities ( TOP_TO_BOTTOM
                         );

   docDC_d  = docDC;
   docUDR_d = docUDR;

   if (  docDC_d
      && docUDR_d
      )
   {
      nodeRootDC_d  = xmlDocGetRootElement (docDC_d);
      nodeRootUDR_d = xmlDocGetRootElement (docUDR_d);

      xmlNodePtr nodeLocale = 0;

      if ((nodeLocale = findEntry (nodeRootDC_d, "Locale", DebugOutput::shouldOutputUPDFDevice ())) != 0)
      {
         char *pszLocale = (char *)xmlGetProp (nodeLocale, (xmlChar *)"Locale_Reference");

         if (pszLocale)
         {
            string stringLocale = pszLocale;

            stringLocale += ".xml";

            docLocale_d = xmlParseFile (stringLocale.c_str ());

            if (docLocale_d)
            {
               xmlValidateDocument (&validator, docLocale_d);

               nodeRootLocale_d = xmlDocGetRootElement (docLocale_d);
            }

            free (pszLocale);
         }
      }

      char *pszCS = (char *)xmlGetProp (nodeRootDC_d, (xmlChar *)"CommandSequences_Reference");

      if (pszCS)
      {
         string stringCS = pszCS;

         stringCS += ".xml";

         docCS_d = xmlParseFile (stringCS.c_str ());

         if (docCS_d)
         {
            xmlValidateDocument (&validator, docCS_d);

            nodeRootCS_d = xmlDocGetRootElement (docCS_d);
         }

         free (pszCS);
      }
   }

   pInstance_d = new UPDFDeviceInstance (this);

   setDeviceInstance (pInstance_d);
   setDeviceBlitter (new UPDFDeviceBlitter (this));

   setPDL (new PDL (PDL::PDL_Epson,PDL::LEVEL_ESCP_2,1,0)); // @TBD

   setVirtualUnits ();
}

UPDFDevice::
UPDFDevice (xmlDocPtr  docDC,
            xmlDocPtr  docUDR,
            char      *pszDriverName,
            char      *pszDeviceName,
            char      *pszShortName,
            char      *pszJobProperties,
            char      *pszXMLDeviceName)
   : PrintDevice (pszDriverName,
                  pszDeviceName,
                  pszShortName,
                  pszJobProperties)
{
   pszDriverName_d = pszDriverName;
   pszDeviceName_d = pszDeviceName;
   pszShortName_d  = pszShortName;

   commonInit (docDC, docUDR);
}

UPDFDevice::
~UPDFDevice ()
{
   if (docDC_d)
   {
      xmlFreeDoc (docDC_d);
      docDC_d = 0;
   }
   if (docUDR_d)
   {
      xmlFreeDoc (docUDR_d);
      docUDR_d = 0;
   }
   if (docLocale_d)
   {
      xmlFreeDoc (docLocale_d);
      docLocale_d = 0;
   }
   if (docCS_d)
   {
      xmlFreeDoc (docCS_d);
      docCS_d = 0;
   }
   if (pszDriverName_d)
   {
      free (pszDriverName_d);
      pszDriverName_d = 0;
   }
   if (pszDeviceName_d)
   {
      free (pszDeviceName_d);
      pszDeviceName_d = 0;
   }
   if (pszShortName_d)
   {
      free (pszShortName_d);
      pszShortName_d = 0;
   }
}

DeviceBooklet * UPDFDevice::
getDefaultBooklet ()
{
   // @TBD
   return 0;
}

DeviceCollation * UPDFDevice::
getDefaultCollation ()
{
   // @TBD
   return 0;
}

DeviceCopies * UPDFDevice::
getDefaultCopies ()
{
   DeviceCopies *pDCRet   = 0;
   char         *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaCopy", "ID");

      if (pszValue)
      {
         pDCRet = UPDFDeviceCopies::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDCRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDCRet)
      return pDCRet;
   else
      // @TBD - hard defaults
      return 0;
}

DeviceDestination * UPDFDevice::
getDefaultDestination ()
{
   // @TBD
   return 0;
}

char * UPDFDevice::
getDefaultDitherID ()
{
   return "DITHER_STUCKI_DIFFUSION";
}

DeviceDuplex * UPDFDevice::
getDefaultDuplex ()
{
   DeviceDuplex *pDDRet   = 0;
   char         *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaDuplex", "ID");

      if (pszValue)
      {
         pDDRet = UPDFDeviceDuplex::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDDRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDDRet)
      return pDDRet;
   else
      // @TBD - hard defaults
      return 0;
}

DeviceForm * UPDFDevice::
getDefaultForm ()
{
   DeviceForm *pDFRet   = 0;
   char       *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaSize", "ID");

      if (pszValue)
      {
         pDFRet = UPDFDeviceForm::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDFRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDFRet)
      return pDFRet;
   else
      // @TBD - hard defaults
      return UPDFDeviceForm::create (this, DeviceForm::FORM_LETTER);
}

DeviceJogging * UPDFDevice::
getDefaultJogging ()
{
   DeviceJogging *pDJRet   = 0;
   char          *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaJogging", "ID");

      if (pszValue)
      {
         pDJRet = UPDFDeviceJogging::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDJRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDJRet)
      return pDJRet;
   else
      // @TBD - hard defaults
      return 0;
}

DeviceMedia * UPDFDevice::
getDefaultMedia ()
{
   DeviceMedia *pDMRet   = 0;
   char        *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaType", "ID");

      if (pszValue)
      {
         pDMRet = UPDFDeviceMedia::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDMRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDMRet)
      return pDMRet;
   else
      // @TBD - hard defaults
      return UPDFDeviceMedia::create (this, DeviceMedia::MEDIA_PLAIN);
}

DeviceNUp * UPDFDevice::
getDefaultNUp ()
{
   DeviceNUp *pDNUpRet = 0;
   char      *pszValue = 0;

   if (pInstance_d)
   {
      pszValue = pInstance_d->getXMLObjectValue ("MediaNUp", "ID");

      if (pszValue)
      {
         pDNUpRet = UPDFDeviceNUp::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDNUpRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDNUpRet)
      return pDNUpRet;
   else
      // @TBD - hard defaults
      return 0;
}

DeviceOrientation * UPDFDevice::
getDefaultOrientation ()
{
   DeviceOrientation *pDORet         = 0;
   char              *pszOrientation = 0;
   char              *pszRotation    = 0;

   if (pInstance_d)
   {
      pszOrientation = pInstance_d->getXMLObjectValue ("MediaPageOrientation",
                                                       "ID");
      pszRotation    = pInstance_d->getXMLObjectValue ("MediaPageRotation",
                                                       "ID");

      if (  pszOrientation
         && pszRotation
         )
      {
         pDORet = UPDFDeviceOrientation::create (this,
                                                 pszOrientation,
                                                 pszRotation);
      }
   }

#ifndef RETAIL
   if (!pDORet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszOrientation = " << pszOrientation << ", pszRotation = " << pszRotation << endl;
   }
#endif

   if (pszOrientation)
   {
      free (pszOrientation);
   }

   if (pszRotation)
   {
      free (pszRotation);
   }

   if (pDORet)
      return pDORet;
   else
      // @TBD - hard defaults
      return UPDFDeviceOrientation::create (this, DeviceOrientation::ORIENTATION_PORTRAIT);
}

DevicePrintMode * UPDFDevice::
getDefaultPrintMode ()
{
   // @TBD
   return UPDFDevicePrintMode::create (this, DevicePrintMode::PRINT_MODE_1_ANY);
}

DeviceResolution * UPDFDevice::
getDefaultResolution ()
{
   DeviceResolution *pDRRet   = 0;
   char             *pszValue = 0;

   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("DeviceResolution", "ID");

      cout << "pszValue = " << pszValue << endl;

      if (pszValue)
      {
         pDRRet = UPDFDeviceResolution::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDRRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDRRet)
      return pDRRet;
   else
      // @TBD - hard defaults
      return UPDFDeviceResolution::create (this, DeviceResolution::RESOLUTION_300_X_300);
}

DeviceScaling * UPDFDevice::
getDefaultScaling ()
{
   // @TBD
   return 0;
}

DeviceStapling * UPDFDevice::
getDefaultStapling ()
{
   // @TBD
   return 0;
}

DeviceTray * UPDFDevice::
getDefaultTray ()
{
   DeviceTray *pDTRet   = 0;
   char       *pszValue = 0;

   if (pInstance_d)
   {
      char *pszValue = pInstance_d->getXMLObjectValue ("MediaSource", "ID");

      if (pszValue)
      {
         pDTRet = UPDFDeviceTray::create (this, pszValue);
      }
   }

#ifndef RETAIL
   if (!pDTRet)
   {
      if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice: " << __FUNCTION__ << ": Error: pszValue = " << pszValue << endl;
   }
#endif

   if (pszValue)
   {
      free (pszValue);
   }

   if (pDTRet)
      return pDTRet;
   else
      // @TBD - hard defaults
      return UPDFDeviceTray::create (this, DeviceTray::TRAY_AUTO);
}

DeviceCommand * UPDFDevice::
getDefaultCommands ()
{
   // @TBD
   return 0;
///return new UPDFDeviceCommands (this);
}

DeviceData * UPDFDevice::
getDefaultData ()
{
   // @TBD
   return 0;
}

DeviceString * UPDFDevice::
getDefaultString ()
{
   // @TBD
   return 0;
}

DeviceGamma * UPDFDevice::
getCurrentGamma ()
{
   // @TBD
   return 0;
}

UPDFDevice * UPDFDevice::
isAUPDFDevice (Device *pDevice)
{
   return dynamic_cast<UPDFDevice *>(pDevice);
}

xmlNodePtr UPDFDevice::
findDCEntry (xmlNodePtr  root,
             char       *pszName,
             bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootDC_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findUDREntry (xmlNodePtr  root,
              char       *pszName,
              bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootUDR_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findLocaleEntry (xmlNodePtr  root,
                 char       *pszName,
                 bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootLocale_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findCSEntry (xmlNodePtr  root,
             char       *pszName,
             bool        fShouldDebugOutput)
{
   if (root == 0)
   {
      root = nodeRootCS_d;
   }

   return findEntry (root, pszName, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findDCEntryKeyValue (char *pszKey,
                     char *pszValue,
                     bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootDC_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findUDREntryKeyValue (char *pszKey,
                      char *pszValue,
                      bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootUDR_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findLocaleEntryKeyValue (char *pszKey,
                         char *pszValue,
                         bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootLocale_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findCSEntryKeyValue (char *pszKey,
                     char *pszValue,
                     bool  fShouldDebugOutput)
{
   return findEntryKeyValue (nodeRootCS_d, pszKey, pszValue, fShouldDebugOutput);
}

xmlNodePtr UPDFDevice::
findEntry (xmlNodePtr  root,
           char       *pszName,
           bool        fShouldDebugOutput)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm = firstNode (root->xmlChildrenNode);

   while (elm != 0)
   {
#ifndef RETAIL
//////if (DebugOutput::shouldOutputUPDFDevice ()) DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": " << elm->name << " == " << pszName << endl;
#endif

      if (0 == strcmp ((char *)elm->name, pszName))
      {
         break;
      }

      elm = nextNode (elm);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevice ())
   {
      if (!elm)
      {
         DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": Did not find " << pszName << endl;
      }
   }
#endif

   return elm;
}

xmlNodePtr UPDFDevice::
findEntryKeyValue (xmlNodePtr  root,
                   char       *pszKey,
                   char       *pszValue,
                   bool        fShouldDebugOutput)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm      = root;
   xmlNodePtr elmFound = 0;

   while (elm != 0)
   {
      char *pszElmValue = (char *)xmlGetProp (elm, (xmlChar *)pszKey);

      if (  pszElmValue
         && 0 == strcmp (pszElmValue, pszValue)
         )
      {
         elmFound = elm;
      }

      if (pszElmValue)
      {
         free (pszElmValue);
      }

      if (  !elmFound
         && elm->xmlChildrenNode
         )
      {
         elmFound = findEntryKeyValue (elm->xmlChildrenNode, pszKey, pszValue, false);
      }

      if (elmFound)
      {
         elm = elmFound;
         break;
      }

      elm = nextNode (elm);
   }

#ifndef RETAIL
   if (fShouldDebugOutput)
   {
      if (elm)
      {
         DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": Found ";
      }
      else
      {
         DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": Did not find ";
      }
      DebugOutput::getErrorStream () << pszValue << endl;
   }
#endif

   return elm;
}

bool UPDFDevice::
setVirtualUnits ()
{
   xmlNodePtr nodeHeader = 0;
   int        iRet       = 0;

   if (  ((nodeHeader = findUDREntry (nodeHeader, "Device_Features",   DebugOutput::shouldOutputUPDFDevice ())) != 0)
      && ((nodeHeader = findUDREntry (nodeHeader, "PrintCapabilities", DebugOutput::shouldOutputUPDFDevice ())) != 0)
      && ((nodeHeader = findUDREntry (nodeHeader, "Header",            DebugOutput::shouldOutputUPDFDevice ())) != 0)
      && ((nodeHeader = findUDREntry (nodeHeader, "VirtualUnits",      DebugOutput::shouldOutputUPDFDevice ())) != 0)
      )
   {
      char *pszVirtualUnits = (char *)xmlGetProp (nodeHeader,
                                                  (xmlChar *)"Predefined_ID");
      if (pszVirtualUnits)
      {
         int iX   = 0,
             iY   = 0;

         iRet = sscanf (pszVirtualUnits, "virtualUnits_%dx%d", &iX, &iY);

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDevice ())
            DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": iRet = " << iRet << ", iX = " << iX << ", iY = " << iY << endl;
#endif

         if (2 == iRet)
         {
            iVirtualUnitsX_d = iX;
            iVirtualUnitsY_d = iY;
         }

         free (pszVirtualUnits);
      }
   }

#ifndef RETAIL
   if (  !iVirtualUnitsX_d
      || !iVirtualUnitsY_d
      )
   {
      DebugOutput::getErrorStream () << "UPDFDevice::" << __FUNCTION__ << ": Error: Could not find virtual units. nodeHeader = " << hex << (int)nodeHeader << dec << endl;
   }
#endif

   return iRet == true;
}

int UPDFDevice::
getXVirtualUnits ()
{
   return iVirtualUnitsX_d;
}

int UPDFDevice::
getYVirtualUnits ()
{
   return iVirtualUnitsY_d;
}

xmlNodePtr UPDFDevice::
getRootDC ()
{
   return nodeRootDC_d;
}

xmlNodePtr UPDFDevice::
getRootUDR ()
{
   return nodeRootUDR_d;
}

xmlNodePtr UPDFDevice::
getRootLocale ()
{
   return nodeRootLocale_d;
}

xmlNodePtr UPDFDevice::
getRootCS ()
{
   return nodeRootCS_d;
}
