/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceDuplex.hpp>

/*
*/
static int     getOmniDuplexID             (xmlNodePtr      node);

UPDFDeviceDuplex::
UPDFDeviceDuplex (Device      *pDevice,
                  int          id,
                  BinaryData  *data,
                  xmlNodePtr   node,
                  bool         fDeviceFeature)
   : DeviceDuplex (pDevice,
                  id,
                  data,
                  fDeviceFeature)
{
   node_d = node;
}

UPDFDeviceDuplex::
~UPDFDeviceDuplex ()
{
}

DeviceDuplex * UPDFDeviceDuplex::
createV (int id)
{
   return UPDFDeviceDuplex::create (pDevice_d, id);
}

DeviceDuplex * UPDFDeviceDuplex::
create (Device *pDevice, char *pszId)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);
   xmlNodePtr  nodeDuplexs = 0;

   nodeDuplexs = findDuplexs (pUPDFDevice);

   if (nodeDuplexs)
   {
      xmlNodePtr nodeItem  = firstNode (nodeDuplexs->xmlChildrenNode);
      xmlNodePtr nodeFound = 0;

      while (nodeItem != 0)
      {
         char *pszNodeId = (char *)xmlGetProp (nodeItem, (xmlChar *)"ID");

         if (pszNodeId)
         {
            if (0 == strcmp (pszNodeId, pszId))
            {
               nodeFound = nodeItem;
            }

            free (pszNodeId);
         }

         if (nodeFound)
         {
            int id = getOmniDuplexID (nodeFound);

            if (DeviceDuplex::DUPLEX_UNLISTED != id)
            {
               return create (pDevice, id);
            }

            nodeFound = 0;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return 0;
}

static int
getOmniDuplexID (xmlNodePtr node)
{
   int id = DeviceDuplex::DUPLEX_UNLISTED;

   if (!node)
   {
      return id;
   }

   char *pszID = (char *)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      id = DeviceDuplex::nameToID (pszID);

      free (pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidDuplexs (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniDuplexID (node);

   while (  node
         && (id == DeviceDuplex::DUPLEX_UNLISTED)
         )
   {
      node = nextNode (node);
      id   = getOmniDuplexID (node);
   }

   return node;
}

static DeviceDuplex *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   bool  fDeviceFeature   = true;
   char *pszDeviceFeature = 0;

   pszDeviceFeature = (char *)xmlGetProp (node, (xmlChar *)"DeviceFeature");

   if (pszDeviceFeature)
   {
      if (0 == strcasecmp (pszDeviceFeature, "false"))
      {
         fDeviceFeature = false;
      }

      free (pszDeviceFeature);
   }

   BinaryData  *data = 0; // @TBD

   return new UPDFDeviceDuplex (pUPDFDevice,
                                id,
                                data,
                                node,
                                !fDeviceFeature);
}

DeviceDuplex * UPDFDeviceDuplex::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceDuplex ()) DebugOutput::getErrorStream () << "UPDFDeviceDuplex::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeDuplexs = 0;

   nodeDuplexs = findDuplexs (pUPDFDevice);

   if (nodeDuplexs)
   {
      xmlNodePtr nodeItem = firstNode (nodeDuplexs->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iDuplexID = getOmniDuplexID (nodeItem);

         if (  id == iDuplexID
            && !(id == DeviceDuplex::DUPLEX_UNLISTED)
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

char * UPDFDeviceDuplex::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceDuplex::
isDuplexSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeDuplexs = 0;

   nodeDuplexs = findDuplexs (pUPDFDevice);

   if (nodeDuplexs)
   {
      xmlNodePtr nodeItem = firstNode (nodeDuplexs->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iDuplexID = getOmniDuplexID (nodeItem);

         if (  id == iDuplexID
            && !(id == DeviceDuplex::DUPLEX_UNLISTED)
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFDuplex_Enumerator : public Enumeration
{
public:
   UPDFDuplex_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d  = pDevice;
      nodeItem_d = skipInvalidDuplexs (nodeItem);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int id = getOmniDuplexID (nodeItem_d);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    id);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidDuplexs (nodeItem_d);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * UPDFDeviceDuplex::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFDuplex_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeDuplexs = 0;

   nodeDuplexs = findDuplexs (pUPDFDevice);

   if (nodeDuplexs)
   {
      xmlNodePtr nodeItem = firstNode (nodeDuplexs->xmlChildrenNode);

      return new UPDFDuplex_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFDuplex_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceDuplex::
findDuplexs (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeDuplexs = 0;

   if (!pUPDFDevice)
   {
      return nodeDuplexs;
   }

   if (  ((nodeDuplexs = FINDUDRENTRY (pUPDFDevice, nodeDuplexs, "Device_Features",    UPDFDeviceDuplex)) != 0)
      && ((nodeDuplexs = FINDUDRENTRY (pUPDFDevice, nodeDuplexs, "PrintCapabilities",  UPDFDeviceDuplex)) != 0)
      && ((nodeDuplexs = FINDUDRENTRY (pUPDFDevice, nodeDuplexs, "Features",           UPDFDeviceDuplex)) != 0)
      && ((nodeDuplexs = FINDUDRENTRY (pUPDFDevice, nodeDuplexs, "PrintMediaHandling", UPDFDeviceDuplex)) != 0)
      && ((nodeDuplexs = FINDUDRENTRY (pUPDFDevice, nodeDuplexs, "MediaDuplexList",    UPDFDeviceDuplex)) != 0)
      )
      return nodeDuplexs;

   return nodeDuplexs;
}

string UPDFDeviceDuplex::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceDuplex::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const UPDFDeviceDuplex& const_self)
{
   UPDFDeviceDuplex& self = const_cast<UPDFDeviceDuplex&>(const_self);
   ostringstream     oss;

   os << self.toString (oss);

   return os;
}
