/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <fstream>
#include <string>

#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <DebugOutput.hpp>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "XMLDevice.hpp"

static int
my_system (const char *pszCommand)
{
   int pid, status;

   if (pszCommand == 0)
      return 1;

   pid = fork();
   if (pid == -1)
      return -1;

   if (pid > 0)
   {
      // parent
      do
      {
         if (waitpid (pid, &status, 0) == -1)
         {
            if (errno != EINTR)
               return -1;
         }
         else
         {
            return status;
         }

      } while (1);
   }
   else
   {
      // child
      char *argv[4];

      argv[0] = "sh";
      argv[1] = "-c";
      argv[2] = (char *)pszCommand;
      argv[3] = 0;

      execvp ("/bin/sh", argv);

      exit (1);
   }
}

void
convertFilename (char *pszFileName)
{
   while (*pszFileName)
   {
      switch (*pszFileName)
      {
      case '-':
      case '+':
      case '(':
      case ')':
      case ' ':
      case '\\':
         *pszFileName = '_';
      }

      pszFileName++;
   }
}

bool
copyFile (string& stringFromFileNameXML,
          string& stringToFileNameXML)
{
   ostringstream oss;
   struct stat   statFromFile;
   struct stat   statToFile;
   char         *pszFileName  = 0;
   bool          fSuccess     = false;

   pszFileName = (char *)stringToFileNameXML.c_str ();
   while (strchr (pszFileName, '/'))
   {
      pszFileName = strchr (pszFileName, '/') + 1;
   }
   convertFilename (pszFileName);

   if (  0 == stat (stringFromFileNameXML.c_str (), &statFromFile)
      && 0 == stat (stringToFileNameXML.c_str (),   &statToFile)
      )
   {
      if (statToFile.st_ctime >= statFromFile.st_ctime)
      {
         return false;
      }
   }

   oss << "cp \""
       << stringFromFileNameXML
       << "\" "
       << stringToFileNameXML
       << ends;

   fSuccess = 0 == my_system ((const char *)oss.str ().c_str ());

   if (fSuccess)
   {
      cout << oss.str () << endl;

      return true;
   }
   else
   {
      cout << "Error: Could not copy " << stringFromFileNameXML << " to here." << endl;

      return false;
   }
}

int
main (int argc, char *argv[])
{
   char  achLine[512];   // @TBD
   char *pszTargetRoot = "./";

   if (1 == argc)
   {
      cout << "Usage: " << argv[0] << "[-r root] directory [directory]*" << endl;
   }

   ofstream ofLibraries1MAK ("libraries1.mak", ios::out | ios::app);

   for (int i = 1; i < argc; i++)
   {
      if ('-' == argv[i][0])
      {
         switch (argv[i][1])
         {
         case 'r':
         case 'R':
            pszTargetRoot = argv[++i];
            break;
         }
         continue;
      }

      string stringFileNameDeviceList ("../");

      stringFileNameDeviceList += argv[i];
      stringFileNameDeviceList += "/Device List";

      ifstream ifIn (stringFileNameDeviceList.c_str ());

      while (0 < ifIn.getline (achLine, sizeof (achLine)))
      {
         string      stringFileNameMasterXML;
         string      stringFromFileNameXML;
         string      stringToFileNameXML;
         xmlDocPtr   docDevice               = 0;
         xmlNodePtr  nodeRoot                = 0;
         xmlNodePtr  nodeElm                 = 0;

         stringFileNameMasterXML  = "../";
         stringFileNameMasterXML += argv[i];
         stringFileNameMasterXML += "/";
         stringFileNameMasterXML += achLine;
         stringFileNameMasterXML += ".xml";

         docDevice = xmlParseFile (stringFileNameMasterXML.c_str ());

         if (!docDevice)
            goto done;

         nodeRoot = xmlDocGetRootElement (docDevice);

         if (!nodeRoot)
            goto done;

         nodeElm = firstNode (nodeRoot);

         if (!nodeElm)
            goto done;

         if (0 != strcmp ((char *)nodeElm->name, "Device"))
            goto done;

         if (nodeElm)
            nodeElm = firstNode (nodeElm->children);

         while (nodeElm)
         {
            if (  0 == strcmp ((char *)nodeElm->name, "Instance")
               || 0 == strcmp ((char *)nodeElm->name, "Blitter")
               )
            {
               char *pszTargetXMLFile = 0;

               pszTargetXMLFile = (char *)xmlNodeListGetString (docDevice,
                                                                nodeElm->xmlChildrenNode,
                                                                1);

               if (pszTargetXMLFile)
               {
                  stringFromFileNameXML  = "../";
                  stringFromFileNameXML += argv[i];
                  stringFromFileNameXML += "/";
                  stringFromFileNameXML += pszTargetXMLFile;

                  stringToFileNameXML    = pszTargetRoot;
                  stringToFileNameXML   += pszTargetXMLFile;

                  if (copyFile (stringFromFileNameXML, stringToFileNameXML))
                  {
                     convertFilename (pszTargetXMLFile);

/////////////////////ofLibraries1MAK << "CLEANFILES += " << pszTargetXMLFile << endl;

                     char *pszDot = strstr (pszTargetXMLFile, ".cpp");

                     if (pszDot)
                     {
                        *pszDot = '\0';

                        ofLibraries1MAK << "lib_LTLIBRARIES += lib" << pszTargetXMLFile << ".la" << endl;
                        ofLibraries1MAK << "lib" << pszTargetXMLFile << "_la_SOURCES = " << pszTargetXMLFile << ".cpp" << endl;
                        ofLibraries1MAK << "lib" << pszTargetXMLFile << "_la_LDFLAGS = -version-info @LT_CURRENT@:@LT_REVISION@:@LT_AGE@ @XML_LIBS@" << endl;
                     }
                  }

                  free (pszTargetXMLFile);
               }
            }

            nodeElm = nextNode (nodeElm);
         }

done:
         if (docDevice)
         {
            xmlFreeDoc (docDevice);
         }
      }
   }

   return 0;
}
