/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <XMLDeviceForm.hpp>
#include <XMLDeviceMedia.hpp>
#include <XMLDeviceOrientation.hpp>
#include <XMLDevicePrintMode.hpp>
#include <XMLDeviceResolution.hpp>
#include <XMLDeviceTray.hpp>
#include <XMLDeviceInstance.hpp>
#include <XMLDeviceBlitter.hpp>
#include <XMLDevice.hpp>

#include <Omni.hpp>
#include <JobProperties.hpp>

#include <typeinfo>

static bool      parseHexGroup                  (const char   *pszData,
                                                 unsigned int *puiData);
static xmlDocPtr getDeviceConfigurationFromFile (char         *pszMasterXMLFile);
static bool      isDeviceConfigurationXMLFile   (xmlDocPtr     doc);
static void      convertFilename                (char         *pszFileName);
static char     *convertToLibrary               (char         *pszFileName);

char *
getVersion ()
{
   return VERSION;
}

class XMLDeviceEnumeration : public Enumeration
{
public:
   XMLDeviceEnumeration (char *pszLibraryName)
   {
      pszLibraryName_d = 0;
      pEnum_d          = Omni::listXMLDevices ();

      if (  pszLibraryName
         && *pszLibraryName
         )
      {
         pszLibraryName_d = (char *)malloc (strlen (pszLibraryName) + 1);
         if (pszLibraryName_d)
         {
            strcpy (pszLibraryName_d, pszLibraryName);
         }
      }
   }

   virtual
   ~XMLDeviceEnumeration ()
   {
      if (pszLibraryName_d)
      {
         free (pszLibraryName_d);
         pszLibraryName_d = 0;
      }

      delete pEnum_d;
      pEnum_d = 0;
   }

   virtual bool
   hasMoreElements ()
   {
      if (!pEnum_d)
      {
         return false;
      }

      return pEnum_d->hasMoreElements ();
   }

   virtual void *
   nextElement ()
   {
      char *pszMasterXML = 0;
      void *pvRet        = 0;

      if (pEnum_d)
      {
         pszMasterXML = (char *)pEnum_d->nextElement ();
      }

      if (pszMasterXML)
      {
         ostringstream  oss;
         char          *pszJobProperties = 0;

         oss << "XMLMasterFile=\""
             << pszMasterXML
             << "\""
             << ends;

         pszJobProperties = (char *)oss.str ().c_str ();

         pvRet = new OmniDevice (pszLibraryName_d,
                                 pszJobProperties);
      }

      return pvRet;
   }

private:
   char        *pszLibraryName_d;
   Enumeration *pEnum_d;
};

Enumeration *
getDeviceEnumeration (char *pszLibraryName)
{
   return new XMLDeviceEnumeration (pszLibraryName);
}

Device *
newDeviceW_Advanced (bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": fAdvanced = " << fAdvanced << endl;
#endif

   return 0;
}

Device *
newDeviceW_JopProp_Advanced (char *pszJobProperties, bool fAdvanced)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszJobProperties = \"" << pszJobProperties << "\", fAdvanced = " << fAdvanced << endl;
#endif

   if (  !pszJobProperties
      || !*pszJobProperties
      )
   {
      return 0;
   }

   JobProperties  jobProp (pszJobProperties);
   char          *pszXMLDeviceName           = 0;

   while (jobProp.hasKeyNameLeft ())
   {
      char *pszKey   = jobProp.getCurrentKey ();
      char *pszValue = jobProp.getCurrentValue ();

      if (0 == strcmp (pszKey, "XMLMasterFile"))
      {
         pszXMLDeviceName = pszValue;
         break;
      }
#ifndef RETAIL
      else if (0 == strcmp (pszKey, "debugoutput"))
      {
         DebugOutput::setDebugOutput (pszValue);
      }
#endif

      jobProp.getNextKey ();
   }

   if (!pszXMLDeviceName)
   {
      cerr << "Error: XMLMasterFile=\"...\" must be specified in the job properties!" << endl;
   }

   char      *pszMasterXML = Omni::openXMLFile (pszXMLDeviceName);
   xmlDocPtr  docDevice    = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszMasterXML = " << pszMasterXML << endl;
#endif

   if (pszMasterXML)
   {
      docDevice = getDeviceConfigurationFromFile (pszMasterXML);

      free (pszMasterXML);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDevice = " << docDevice << endl;
#endif

   if (docDevice)
   {
      xmlNodePtr  rootDevice    = xmlDocGetRootElement (docDevice);
      xmlNodePtr  elm           = 0;
      char       *pszDriverName = 0;
      char       *pszDeviceName = 0;
      char       *pszShortName  = 0;

      elm = firstNode (rootDevice);

      if (elm)
      {
         pszDeviceName = (char *)xmlGetProp (elm, (xmlChar *)"name");
         pszShortName  = (char *)xmlGetProp (elm, (xmlChar *)"name");
      }

      pszDriverName = getXMLContentString (rootDevice, docDevice, "DriverName");

      if (pszShortName)
      {
         convertFilename (pszShortName);
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ())
      {
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDriverName = " << pszDriverName << endl;
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDeviceName = " << pszDeviceName << endl;
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszShortName  = " << pszShortName << endl;
      }
#endif

      if (   pszDriverName
         && *pszDriverName
         &&  pszDeviceName
         && *pszDeviceName
         &&  pszShortName
         && *pszShortName
         )
      {
         PrintDevice *pPrintDevice = 0;

         pPrintDevice = new XMLDevice (docDevice,
                                       pszDriverName,
                                       pszDeviceName,
                                       pszShortName,
                                       pszJobProperties,
                                       pszXMLDeviceName);

         if (pPrintDevice)
         {
            pPrintDevice->initialize ();

            if (fAdvanced)
               return pPrintDevice;
            else
               return new OmniProxy (pPrintDevice);
         }
      }

      if (pszDriverName)
      {
         free (pszDriverName);
      }
      if (pszDeviceName)
      {
         free (pszDeviceName);
      }
      if (pszShortName)
      {
         free (pszShortName);
      }
      if (docDevice)
      {
         xmlFreeDoc (docDevice);
      }
   }

   return 0;
}

void
deleteDevice (Device *pDevice)
{
   delete pDevice;
}

XMLDevice::
XMLDevice (xmlDocPtr    docDevice,
           char        *pszDriverName,
           char        *pszDeviceName,
           char        *pszShortName,
           char        *pszJobProperties,
           char        *pszXMLDeviceName)
   : PrintDevice (pszDriverName,
                  pszDeviceName,
                  pszShortName,
                  pszJobProperties)
{
   pszXMLDeviceName_d      = 0;

   docDevice_d             = docDevice;
   rootDevice_d            = 0;

   pszDriverName_d         = pszDriverName;
   pszDeviceName_d         = pszDeviceName;
   pszShortName_d          = pszShortName;

   docDeviceBooklets_d     = 0;
   docDeviceForms_d        = 0;
   docDeviceMedias_d       = 0;
   docDeviceOrientations_d = 0;
   docDevicePrintModes_d   = 0;
   docDeviceResolutions_d  = 0;
   docDeviceTrays_d        = 0;
   docDeviceGammas_d       = 0;
   docDeviceCommands_d     = 0;
   docDeviceDatas_d        = 0;
   docDeviceStrings_d      = 0;

   pszDefaultBooklet_d     = 0;
   pszDefaultForm_d        = 0;
   pszDefaultMedia_d       = 0;
   pszDefaultOrientation_d = 0;
   pszDefaultPrintMode_d   = 0;
   pszDefaultResolution_d  = 0;
   pszDefaultTray_d        = 0;
   pszDefaultDitherID_d    = 0;

   pszInstance_d           = 0;
   pszBlitter_d            = 0;

   pDefaultCommands_d      = 0;
   pDefaultDatas_d         = 0;
   pDefaultStrings_d       = 0;

   pInstance_d             = 0;
   pBlitter_d              = 0;

   if (  pszXMLDeviceName
      && *pszXMLDeviceName
      )
   {
      pszXMLDeviceName_d = (char *)malloc (strlen (pszXMLDeviceName) + 1);
      if (pszXMLDeviceName_d)
      {
         strcpy (pszXMLDeviceName_d, pszXMLDeviceName);
      }
   }

   if (docDevice_d)
   {
      rootDevice_d = xmlDocGetRootElement (docDevice);
   }

   initializeDevice ();
}

XMLDevice::
~XMLDevice ()
{
   if (pszXMLDeviceName_d)
   {
      free (pszXMLDeviceName_d);
      pszXMLDeviceName_d = 0;
   }

   if (docDevice_d)
   {
      xmlFreeDoc (docDevice_d);
      docDevice_d = 0;
   }
   if (pszDriverName_d)
   {
      free (pszDriverName_d);
      pszDriverName_d = 0;
   }
   if (pszDeviceName_d)
   {
      free (pszDeviceName_d);
      pszDeviceName_d = 0;
   }
   if (pszShortName_d)
   {
      free (pszShortName_d);
      pszShortName_d = 0;
   }

   if (pszDefaultBooklet_d)
   {
      free (pszDefaultBooklet_d);
      pszDefaultBooklet_d = 0;
   }
   if (pszDefaultForm_d)
   {
      free (pszDefaultForm_d);
      pszDefaultForm_d = 0;
   }
   if (pszDefaultMedia_d)
   {
      free (pszDefaultMedia_d);
      pszDefaultMedia_d = 0;
   }
   if (pszDefaultOrientation_d)
   {
      free (pszDefaultOrientation_d);
      pszDefaultOrientation_d = 0;
   }
   if (pszDefaultPrintMode_d)
   {
      free (pszDefaultPrintMode_d);
      pszDefaultPrintMode_d = 0;
   }
   if (pszDefaultResolution_d)
   {
      free (pszDefaultResolution_d);
      pszDefaultResolution_d = 0;
   }
   if (pszDefaultTray_d)
   {
      free (pszDefaultTray_d);
      pszDefaultTray_d = 0;
   }
   if (pszDefaultDitherID_d)
   {
      free (pszDefaultDitherID_d);
      pszDefaultDitherID_d = 0;
   }

   if (pszInstance_d)
   {
      free (pszInstance_d);
      pszInstance_d = 0;
   }
   if (pszBlitter_d)
   {
      free (pszBlitter_d);
      pszBlitter_d = 0;
   }

   if (docDeviceBooklets_d)
   {
      xmlFreeDoc (docDeviceBooklets_d);
      docDeviceBooklets_d = 0;
   }
   if (docDeviceForms_d)
   {
      xmlFreeDoc (docDeviceForms_d);
      docDeviceForms_d = 0;
   }
   if (docDeviceMedias_d)
   {
      xmlFreeDoc (docDeviceMedias_d);
      docDeviceMedias_d = 0;
   }
   if (docDeviceOrientations_d)
   {
      xmlFreeDoc (docDeviceOrientations_d);
      docDeviceOrientations_d = 0;
   }
   if (docDevicePrintModes_d)
   {
      xmlFreeDoc (docDevicePrintModes_d);
      docDevicePrintModes_d = 0;
   }
   if (docDeviceResolutions_d)
   {
      xmlFreeDoc (docDeviceResolutions_d);
      docDeviceResolutions_d = 0;
   }
   if (docDeviceTrays_d)
   {
      xmlFreeDoc (docDeviceTrays_d);
      docDeviceTrays_d = 0;
   }
   if (docDeviceGammas_d)
   {
      xmlFreeDoc (docDeviceGammas_d);
      docDeviceGammas_d = 0;
   }
   if (docDeviceCommands_d)
   {
      xmlFreeDoc (docDeviceCommands_d);
      docDeviceCommands_d = 0;
   }
   if (docDeviceDatas_d)
   {
      xmlFreeDoc (docDeviceDatas_d);
      docDeviceDatas_d = 0;
   }
   if (docDeviceStrings_d)
   {
      xmlFreeDoc (docDeviceStrings_d);
      docDeviceStrings_d = 0;
   }

   // These will be cleaned up by the base class
   pDefaultCommands_d = 0;
   pDefaultDatas_d    = 0;
   pDefaultStrings_d  = 0;
}

void XMLDevice::
initializeDevice ()
{
   if (  !docDevice_d
      || !rootDevice_d
      )
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: docDevice_d or rootDevice_d is null!" << endl;
#endif

      return;
   }

   // Read and set the capabilities
   xmlNodePtr elm           = 0;
   int        iCapabilities = 0;

   elm = firstNode (rootDevice_d);
   if (elm)
      elm = firstNode (elm->children);

   while (elm)
   {
      if (0 == strcmp ((char *)elm->name, "Capability"))
      {
         char *pszCapability = 0;

         pszCapability = (char *)xmlGetProp (elm, (xmlChar *)"type");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszCapability = " << pszCapability << endl;
#endif

         if (pszCapability)
         {
            iCapabilities |= Capability::getReservedValue (pszCapability);

            free (pszCapability);
         }
      }

      elm = nextNode (elm);
   }

   // Read and set the raster capabilities
   int   iRasterCapabilities = 0;

   elm = firstNode (rootDevice_d);
   if (elm)
      elm = firstNode (elm->children);

   while (elm)
   {
      if (0 == strcmp ((char *)elm->name, "RasterCapabilities"))
      {
         char *pszRasterCapability = 0;

         pszRasterCapability = (char *)xmlGetProp (elm, (xmlChar *)"type");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszRasterCapability = " << pszRasterCapability << endl;
#endif

         if (pszRasterCapability)
         {
            iRasterCapabilities |= RasterCapabilities::getReservedValue (pszRasterCapability);

            free (pszRasterCapability);
         }
      }

      elm = nextNode (elm);
   }

   setRasterCapabilities (iRasterCapabilities);

   // @TBD setDeviceOptions

   elm = findEntry (rootDevice_d, "PDL", false);
   if (elm)
   {
      char *pszLevel              = 0;
      char *pszSubLevel           = 0;
      char *pszMajorRevisionLevel = 0;
      char *pszMinorRevisionLevel = 0;

      pszLevel = (char *)xmlGetProp (elm, (xmlChar *)"level");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszLevel = " << pszLevel << endl;
#endif

      pszSubLevel = (char *)xmlGetProp (elm, (xmlChar *)"sublevel");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszSubLevel = " << pszSubLevel << endl;
#endif

      pszMajorRevisionLevel = (char *)xmlGetProp (elm, (xmlChar *)"major");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszMajorRevisionLevel = " << pszMajorRevisionLevel << endl;
#endif

      pszMinorRevisionLevel = (char *)xmlGetProp (elm, (xmlChar *)"minor");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszMinorRevisionLevel = " << pszMinorRevisionLevel << endl;
#endif

      if (  pszLevel
         && pszSubLevel
         && pszMajorRevisionLevel
         && pszMinorRevisionLevel
         )
      {
         int  iLevel              = 0;
         int  iSubLevel           = 0;
         int  iMajorRevisionLevel = 0;
         int  iMinorRevisionLevel = 0;
         PDL *pPDL                = 0;

         if (1 != sscanf (pszLevel, "%d", &iLevel))
         {
            iLevel = PDL::getReservedValue (pszLevel);
         }
         if (1 != sscanf (pszSubLevel, "%d", &iSubLevel))
         {
            iSubLevel = PDL::getReservedValue (pszSubLevel);
         }
         if (1 != sscanf (pszMajorRevisionLevel, "%d", &iMajorRevisionLevel))
         {
            iMajorRevisionLevel = PDL::getReservedValue (pszMajorRevisionLevel);
         }
         if (1 != sscanf (pszMinorRevisionLevel, "%d", &iMinorRevisionLevel))
         {
            iMinorRevisionLevel = PDL::getReservedValue (pszMinorRevisionLevel);
         }

         pPDL = new PDL (iLevel,
                         iSubLevel,
                         iMajorRevisionLevel,
                         iMinorRevisionLevel);

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ())
         {
            if (pPDL)
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": setting PDL " << *pPDL << endl;
            }
            else
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": setting NULL PDL! " << endl;
            }
         }
#endif

         setPDL (pPDL);
      }

      if (pszLevel)
      {
         free (pszLevel);
      }
      if (pszSubLevel)
      {
         free (pszSubLevel);
      }
      if (pszMajorRevisionLevel)
      {
         free (pszMajorRevisionLevel);
      }
      if (pszMinorRevisionLevel)
      {
         free (pszMinorRevisionLevel);
      }
   }

   pszInstance_d = getXMLContentString (rootDevice_d, docDevice_d, "Instance");
   pszBlitter_d  = getXMLContentString (rootDevice_d, docDevice_d, "Blitter");

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ())
   {
      DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszInstance_d = " << pszInstance_d << endl;
      DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszBlitter_d = " << pszBlitter_d << endl;
   }
#endif

   if (pszInstance_d)
   {
      pszInstance_d = convertToLibrary (pszInstance_d);
   }
   if (pszBlitter_d)
   {
      pszBlitter_d = convertToLibrary (pszBlitter_d);
   }

   if (pszInstance_d)
   {
      GModule *hmodInstanceLibrary = 0;
      bool     fSucccess           = false;

      fSucccess = Omni::openLibrary (pszInstance_d, &hmodInstanceLibrary);

      if (fSucccess)
      {
         pInstance_d = new XMLDeviceInstance (hmodInstanceLibrary, this);

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pInstance_d = " << pInstance_d << endl;
#endif

         fSucccess = pInstance_d ? true : false;
      }

      if (!fSucccess)
      {
         if (pInstance_d)
         {
            delete pInstance_d;
            pInstance_d = 0;
         }
         if (hmodInstanceLibrary)
         {
            g_module_close (hmodInstanceLibrary);
            hmodInstanceLibrary = 0;
         }
      }
      else
      {
         setDeviceInstance (pInstance_d);
      }
   }

   if (pszBlitter_d)
   {
      GModule *hmodBlitterLibrary = 0;
      bool     fSucccess          = false;

      fSucccess = Omni::openLibrary (pszBlitter_d, &hmodBlitterLibrary);

      if (fSucccess)
      {
         pBlitter_d  = new XMLDeviceBlitter (hmodBlitterLibrary, this);

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pBlitter_d = " << pBlitter_d << endl;
#endif

         fSucccess = pBlitter_d ? true : false;
      }

      if (!fSucccess)
      {
         if (pBlitter_d)
         {
            delete pBlitter_d;
            pBlitter_d = 0;
         }
         if (hmodBlitterLibrary)
         {
            g_module_close (hmodBlitterLibrary);
            hmodBlitterLibrary = 0;
         }
      }
      else
      {
         setDeviceBlitter (pBlitter_d);
      }
   }

///DeviceCommand *pDC     = getDefaultCommands ();
///BinaryData    *pbdData = pDC->getCommandData ("cmdBeginRasterGraphics");
///cout << *pbdData << endl;

///DeviceData *pDD      = getDefaultData ();
///int         iValue   = 0;
///bool        fSuccess = false;
///fSuccess = pDD->getIntData ("Nozzle_Spacing", &iValue);
///cout << "Nozzle_Spacing = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;
///fSuccess = pDD->getIntData ("Nozzle_Number", &iValue);
///cout << "Nozzle_Number = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;
///fSuccess = pDD->getIntData ("Positioning_x", &iValue);
///cout << "Positioning_x = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;
///fSuccess = pDD->getIntData ("DotSize360", &iValue);
///cout << "DotSize360 = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;
///fSuccess = pDD->getIntData ("DotSize720", &iValue);
///cout << "DotSize720 = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;
///fSuccess = pDD->getIntData ("DotSize1440", &iValue);
///cout << "DotSize1440 = " << iValue << endl;
///cout << "fSuccess = " << fSuccess << endl;

///DeviceString *pDS           = getDefaultString ();
///char         *apszStrings[] = {
///   "color-mode",
///   "draft",
///   "fine",
///   "graph",
///   "high-speed",
///   "micro-weave",
///   "none",
///   "off",
///   "on",
///   "photo",
///   "quality",
///   "standard",
///   "super-fine",
///   "super-photo",
///};
///char         *pszString     = 0;
///cout << *pDS << endl;
///pDS->setLanguage (StringResource::LANGUAGE_EN);
///for (int i = 0; i < (int)dimof (apszStrings); i++)
///{
///   pszString = pDS->getStringV (StringResource::STRINGGROUP_UNKNOWN,
///                                apszStrings[i]);
///   cout << "for \""
///        << apszStrings[i]
///        << "\" returns \""
///        << pszString
///        << "\"."
///        << endl;
///}
}

DeviceBooklet * XMLDevice::
getDefaultBooklet ()
{
   DeviceBooklet *pDBRet = 0;

   if (!docDeviceBooklets_d)
   {
      docDeviceBooklets_d = getDeviceXML ("deviceBooklets");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceBooklets_d = " << docDeviceBooklets_d << endl;
#endif
   }

   if (  !pszDefaultBooklet_d
      && docDeviceBooklets_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultBooklet_d = getXMLContentString (elm, docDevice_d, "booklet");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultBooklet_d = " << pszDefaultBooklet_d << endl;
#endif
      }
   }

   // @TBD
   return 0;
}

DeviceCollation * XMLDevice::
getDefaultCollation ()
{
   // @TBD
   return 0;
}

DeviceCopies * XMLDevice::
getDefaultCopies ()
{
   // @TBD
   return 0;
}

DeviceDestination * XMLDevice::
getDefaultDestination ()
{
   // @TBD
   return 0;
}

char * XMLDevice::
getDefaultDitherID ()
{
   if (!pszDefaultDitherID_d)
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultDitherID_d = getXMLContentString (elm, docDevice_d, "dither");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultDitherID_d = " << pszDefaultDitherID_d << endl;
#endif
      }
   }

   return pszDefaultDitherID_d;
}

DeviceDuplex * XMLDevice::
getDefaultDuplex ()
{
   // @TBD
   return 0;
}

DeviceForm * XMLDevice::
getDefaultForm ()
{
   DeviceForm *pDFRet = 0;

   if (!docDeviceForms_d)
   {
      docDeviceForms_d = getDeviceXML ("deviceForms");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceForms_d = " << docDeviceForms_d << endl;
#endif
   }

   if (  !pszDefaultForm_d
      && docDeviceForms_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultForm_d = getXMLContentString (elm, docDevice_d, "form");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultForm_d = " << pszDefaultForm_d << endl;
#endif
      }
   }

   if (pszDefaultForm_d)
   {
      pDFRet = XMLDeviceForm::create (this, pszDefaultForm_d);

//////// Using member function pointers in C++
//////// http://www.vmlinux.org/jakov/community.borland.com/15837.html
//////Enumeration * (DeviceForm::*PM) (void);
//////PM = &DeviceForm::getEnumeration;
//////cout << "pDFRet_d->getEnumeration = " << hex << &(pDFRet_d->*PM) << dec << endl;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDFRet << endl;
#endif

   return pDFRet;
}

DeviceJogging * XMLDevice::
getDefaultJogging ()
{
   // @TBD
   return 0;
}

DeviceMedia * XMLDevice::
getDefaultMedia ()
{
   DeviceMedia *pDMRet = 0;

   if (!docDeviceMedias_d)
   {
      docDeviceMedias_d = getDeviceXML ("deviceMedias");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceMedias_d = " << docDeviceMedias_d << endl;
#endif
   }

   if (  !pszDefaultMedia_d
      && docDeviceMedias_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultMedia_d = getXMLContentString (elm, docDevice_d, "media");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultMedia_d = " << pszDefaultMedia_d << endl;
#endif
      }
   }

   if (pszDefaultMedia_d)
   {
      pDMRet = XMLDeviceMedia::create (this, pszDefaultMedia_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDMRet << endl;
#endif

   return pDMRet;
}

DeviceNUp * XMLDevice::
getDefaultNUp ()
{
   // @TBD
   return 0;
}

DeviceOrientation * XMLDevice::
getDefaultOrientation ()
{
   DeviceOrientation *pDORet = 0;

   if (!docDeviceOrientations_d)
   {
      docDeviceOrientations_d = getDeviceXML ("deviceOrientations");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceOrientations_d = " << docDeviceOrientations_d << endl;
#endif
   }

   if (  !pszDefaultOrientation_d
      && docDeviceOrientations_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultOrientation_d = getXMLContentString (elm, docDevice_d, "orientation");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultOrientation_d = " << pszDefaultOrientation_d << endl;
#endif
      }
   }

   if (pszDefaultOrientation_d)
   {
      pDORet = XMLDeviceOrientation::create (this, pszDefaultOrientation_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDORet << endl;
#endif

   return pDORet;
}

DevicePrintMode * XMLDevice::
getDefaultPrintMode ()
{
   DevicePrintMode *pDPMRet = 0;

   if (!docDevicePrintModes_d)
   {
      docDevicePrintModes_d = getDeviceXML ("devicePrintModes");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDevicePrintModes_d = " << docDevicePrintModes_d << endl;
#endif
   }

   if (  !pszDefaultPrintMode_d
      && docDevicePrintModes_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultPrintMode_d = getXMLContentString (elm, docDevice_d, "printmode");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultPrintMode_d = " << pszDefaultPrintMode_d << endl;
#endif
      }
   }

   if (pszDefaultPrintMode_d)
   {
      pDPMRet = XMLDevicePrintMode::create (this, pszDefaultPrintMode_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDPMRet << endl;
#endif

   return pDPMRet;
}

DeviceResolution * XMLDevice::
getDefaultResolution ()
{
   DeviceResolution *pDRRet = 0;

   if (!docDeviceResolutions_d)
   {
      docDeviceResolutions_d = getDeviceXML ("deviceResolutions");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceResolutions_d = " << docDeviceResolutions_d << endl;
#endif
   }

   if (  !pszDefaultResolution_d
      && docDeviceResolutions_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultResolution_d = getXMLContentString (elm, docDevice_d, "resolution");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultResolution_d = " << pszDefaultResolution_d << endl;
#endif
      }
   }

   if (pszDefaultResolution_d)
   {
      pDRRet = XMLDeviceResolution::create (this, pszDefaultResolution_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDRRet << endl;
#endif

   return pDRRet;
}

DeviceScaling * XMLDevice::
getDefaultScaling ()
{
   // @TBD
   return 0;
}

DeviceStapling * XMLDevice::
getDefaultStapling ()
{
   // @TBD
   return 0;
}

DeviceTray * XMLDevice::
getDefaultTray ()
{
   DeviceTray *pDTRet = 0;

   if (!docDeviceTrays_d)
   {
      docDeviceTrays_d = getDeviceXML ("deviceTrays");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceTrays_d = " << docDeviceTrays_d << endl;
#endif
   }

   if (  !pszDefaultTray_d
      && docDeviceTrays_d
      )
   {
      xmlNodePtr elm = 0;

      elm = findEntry (rootDevice_d, "DefaultJobProperties", false);

      if (elm)
      {
         pszDefaultTray_d = getXMLContentString (elm, docDevice_d, "tray");

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDefaultTray_d = " << pszDefaultTray_d << endl;
#endif
      }
   }

   if (pszDefaultTray_d)
   {
      pDTRet = XMLDeviceTray::create (this, pszDefaultTray_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDTRet << endl;
#endif

   return pDTRet;
}

DeviceCommand * XMLDevice::
getDefaultCommands ()
{
   if (!docDeviceCommands_d)
   {
      docDeviceCommands_d = getDeviceXML ("deviceCommands");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceCommands_d = " << docDeviceCommands_d << endl;
#endif
   }

   if (  !pDefaultCommands_d
      && docDeviceCommands_d
      )
   {
      xmlNodePtr  rootDeviceCommands = xmlDocGetRootElement (docDeviceCommands_d);
      xmlNodePtr  elmDeviceCommand   = 0;

      elmDeviceCommand = firstNode (rootDeviceCommands);
      if (elmDeviceCommand)
         elmDeviceCommand = firstNode (elmDeviceCommand->children);

      pDefaultCommands_d = new DeviceCommand ();

      while (elmDeviceCommand)
      {
         char       *pszName    = 0;
         char       *pszCommand = 0;
         BinaryData *pbdData    = 0;

         try
         {
            // Read in the name
            pszName = (char *)xmlGetProp (elmDeviceCommand, (xmlChar *)"name");

            // Read in the command
            pszCommand = getXMLContentString (elmDeviceCommand, docDeviceCommands_d, 0);

            if (pszCommand)
            {
               byte *pbData = 0;
               int   cbData = 0;

               if (XMLDevice::parseBinaryData (pszCommand,
                                               &pbData,
                                               &cbData))
               {
                  pbdData = new BinaryData (pbData, cbData);
               }
            }

#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ())
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszName = " << pszName << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszCommand = " << pszCommand << endl;
            }
#endif

            if (  pszName
               && pbdData
               )
            {
               pDefaultCommands_d->add (pszName, pbdData);
            }
         }
         catch (string *pstringError)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

            delete pstringError;
         }

         if (pszName)
         {
            free (pszName);
         }
         if (pszCommand)
         {
            free (pszCommand);
         }

         elmDeviceCommand = nextNode (elmDeviceCommand);
      }
   }

   return pDefaultCommands_d;
}

DeviceData * XMLDevice::
getDefaultData ()
{
   if (!docDeviceDatas_d)
   {
      docDeviceDatas_d = getDeviceXML ("deviceDatas");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceDatas_d = " << docDeviceDatas_d << endl;
#endif
   }

   if (  !pDefaultDatas_d
      && docDeviceDatas_d
      )
   {
      xmlNodePtr rootDeviceDatas = xmlDocGetRootElement (docDeviceDatas_d);
      xmlNodePtr elmDeviceData   = 0;

      elmDeviceData = firstNode (rootDeviceDatas);
      if (elmDeviceData)
         elmDeviceData = firstNode (elmDeviceData->children);

      pDefaultDatas_d = new DeviceData ();

      while (elmDeviceData)
      {
         char       *pszName = 0;
         char       *pszType = 0;
         char       *pszData = 0;
         BinaryData *pbdData = 0;
         byte       *pbData  = 0;
         int         cbData  = 0;

         try
         {
            // Read in the name
            pszName = (char *)xmlGetProp (elmDeviceData, (xmlChar *)"name");

            // Read in the type
            pszType = (char *)xmlGetProp (elmDeviceData, (xmlChar *)"type");

            // Read in the command
            pszData = getXMLContentString (elmDeviceData, docDeviceDatas_d, 0);

#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ())
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszName = " << pszName << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszType = " << pszType << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszData = " << pszData << endl;
            }
#endif

            if (pszData)
            {
               if (0 == strcmp (pszType, "string"))
               {
                  cbData = strlen (pszData) + 1;
                  pbData = new byte [cbData];

                  if (pbData)
                  {
                     strcpy ((char *)pbData, pszData);
                     pbdData = new BinaryData (pbData, cbData);
                     if (!pbdData)
                     {
                        delete[] pbData;
                        pbData = 0;
                        // @TBD
                     }
                  }
                  else
                  {
                     // @TBD
                  }
               }
               else if (0 == strcmp (pszType, "boolean"))
               {
                  cbData = 4;
                  pbData = new byte [cbData];

                  if (pbData)
                  {
                     memset (pbData, 0, cbData);

                     if (0 == strcasecmp (pszData, "true"))
                     {
                        pbData[0] = 1;
                     }
                     else if (0 == strcasecmp (pszData, "false"))
                     {
                        pbData[0] = 0;
                     }
                     else
                     {
                        delete[] pbData;
                        pbData = 0;
                        // @TBD
                     }

                     if (pbData)
                     {
                        pbdData = new BinaryData (pbData, cbData);
                        if (!pbdData)
                        {
                           delete[] pbData;
                           pbData = 0;
                           // @TBD
                        }
                     }
                  }
                  else
                  {
                     // @TBD
                  }
               }
               else if (0 == strcmp (pszType, "integer"))
               {
                  cbData = 4;
                  pbData = new byte [cbData];

                  if (pbData)
                  {
                     if (0 == sscanf (pszData, "%d", (int *)pbData))
                     {
                        delete[] pbData;
                        pbData = 0;
                        // @TBD
                     }

                     if (pbData)
                     {
                        pbdData = new BinaryData (pbData, cbData);
                        if (!pbdData)
                        {
                           delete[] pbData;
                           pbData = 0;
                           // @TBD
                        }
                     }
                  }
                  else
                  {
                     // @TBD
                  }
               }
               else if (0 == strcmp (pszType, "byte"))
               {
                  cbData = 1;
                  pbData = new byte [cbData];

                  if (pbData)
                  {
                     if (0 == sscanf (pszData, "%c", (char *)pbData))
                     {
                        delete[] pbData;
                        pbData = 0;
                        // @TBD
                     }

                     if (pbData)
                     {
                        pbdData = new BinaryData (pbData, cbData);
                        if (!pbdData)
                        {
                           delete[] pbData;
                           pbData = 0;
                           // @TBD
                        }
                     }
                  }
                  else
                  {
                     // @TBD
                  }
               }
               else if (0 == strcmp (pszType, "binary"))
               {
                  if (XMLDevice::parseBinaryData (pszData,
                                                  &pbData,
                                                  &cbData))
                  {
                     pbdData = new BinaryData (pbData, cbData);
                  }
               }
               else if (0 == strcmp (pszType, "bytearray"))
               {
                  int iCnt,
                      iTempCnt = 0;

                  for (iCnt = 0; *(pszData + iCnt); iCnt++)
                  {
                     if (!isxdigit (*(pszData + iCnt)))
                     {
                        if (isspace (*(pszData + iCnt)) == 0)
                        {
                           cerr << "Error: Data in bytearray is not understood \""
                                << *(pszData + iCnt)
                                << "\" for "
                                << pszData
                                << "."
                                << endl;
                           // @TBD
                        }
                     }
                     else
                     {
                        iTempCnt++;
                     }
                  }

                  cbData = iTempCnt / 2;
                  pbData = new byte [cbData];
                  if (!pbData)
                  {
                     // @TBD
                  }

                  int          iEnd = strlen (pszData);
                  int          iLoc = 0;
                  bool         bOK;
                  char         cTemp[2];
                  unsigned int iValue;

                  iTempCnt = 0;

                  for (iCnt = 0; iCnt < iEnd; iCnt++)
                  {
                     if (isxdigit (*(pszData + iCnt)))
                     {
                        cTemp[iLoc++] = *(pszData + iCnt);
                        if (iLoc == 2)
                        {
                           bOK = parseHexGroup ((const char *)cTemp, &iValue);
                           if (!bOK)
                           {
                              cerr << "Error: Data in bytearray is not understood \""
                                   << *(pszData + iCnt)
                                   << "\" for "
                                   << pszData
                                   << "."
                                   << endl;

                              delete[] pbData;
                              pbData = 0;
                              // @TBD
                           }

                           *(pbData + iTempCnt++) = (unsigned char)iValue;
                           iLoc = 0;
                        }
                     }
                  }

                  if (iLoc)
                  {
                     cerr << "Error: Data in bytearray is missing bytes\""
                          << *(pszData+iCnt)
                          << "\" for "
                          << pszData
                          << "."
                          << endl;

                     delete[] pbData;
                     pbData = 0;
                     // @TBD
                  }

                  if (pbData)
                  {
                     pbdData = new BinaryData (pbData, cbData);
                     if (!pbdData)
                     {
                        delete[] pbData;
                        pbData = 0;
                        // @TBD
                     }
                  }
               }
               else
               {
                  // @TBD
               }

               if (pbdData)
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Adding data (\"" << pszName << "\", " << *pbdData << ")" << endl;
#endif

                  pDefaultDatas_d->add (pszName, pbdData);
               }
            }

#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ())
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszName = " << pszName << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszData = " << pszData << endl;
            }
#endif
         }
         catch (string *pstringError)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

            delete pstringError;
         }

         if (pszName)
         {
            free (pszName);
         }
         if (pszType)
         {
            free (pszType);
         }
         if (pszData)
         {
            free (pszData);
         }

         elmDeviceData = nextNode (elmDeviceData);
      }
   }

   return pDefaultDatas_d;
}

DeviceString * XMLDevice::
getDefaultString ()
{
   if (!docDeviceStrings_d)
   {
      docDeviceStrings_d = getDeviceXML ("deviceStrings");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceStrings_d = " << docDeviceStrings_d << endl;
#endif
   }

   if (  !pDefaultStrings_d
      && docDeviceStrings_d
      )
   {
      xmlNodePtr  rootDeviceStrings  = xmlDocGetRootElement (docDeviceStrings_d);
      xmlNodePtr  elmDeviceString    = 0;
      xmlNodePtr  elmDeviceLanguages = 0;
      xmlNodePtr  elmDeviceLanguage  = 0;

      elmDeviceString = firstNode (rootDeviceStrings);
      if (elmDeviceString)
         elmDeviceString = firstNode (elmDeviceString->children);

      pDefaultStrings_d = new DeviceString ();

      while (elmDeviceString)
      {
         char *pszLanguage = 0;
         char *pszInternal = 0;
         char *pszExternal = 0;

         // Read in the internal name
         elmDeviceLanguages = findEntry (elmDeviceString, "name", false);
         if (elmDeviceLanguages)
         {
            pszInternal = getXMLContentString (elmDeviceLanguages,
                                               docDeviceStrings_d,
                                               0);
         }

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszInternal = " << pszInternal << endl;
#endif

         if (pszInternal)
         {
            elmDeviceLanguage  = 0;
            elmDeviceLanguages = findEntry (elmDeviceString,
                                            "languages",
                                            false);
            if (elmDeviceLanguages)
               elmDeviceLanguage = elmDeviceLanguages->children;
            if (elmDeviceLanguage)
               elmDeviceLanguage = firstNode (elmDeviceLanguage);

            while (elmDeviceLanguage)
            {
               if (elmDeviceLanguage)
               {
                  // Read in the external name
                  pszExternal = getXMLContentString (elmDeviceLanguage,
                                                     docDeviceStrings_d,
                                                     0);

#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszExternal = " << pszExternal << endl;
#endif

                  pszLanguage = (char *)elmDeviceLanguage->name;

#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszLanguage = " << pszLanguage << endl;
#endif

                  if (  pszLanguage
                     && pszInternal
                     && pszExternal
                     )
                  {
                     pDefaultStrings_d->add (pszLanguage,
                                             pszInternal,
                                             pszExternal);
                  }

                  if (pszExternal)
                  {
                     free (pszExternal);
                  }
               }

               elmDeviceLanguage = nextNode (elmDeviceLanguage);
            }
         }

         if (pszInternal)
         {
            free (pszInternal);
         }

         elmDeviceString = nextNode (elmDeviceString);
      }
   }

   return pDefaultStrings_d;
}

DeviceGamma * XMLDevice::
getCurrentGamma ()
{
   if (!docDeviceGammas_d)
   {
      docDeviceGammas_d = getDeviceXML ("deviceGammaTables");

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": docDeviceGammas_d = " << docDeviceGammas_d << endl;
#endif
   }

   DeviceResolution *pDR         = 0;
   DeviceMedia      *pDM         = 0;
   DevicePrintMode  *pDPM        = 0;
   char             *pszDitherID = 0;
   DeviceGamma      *pDGRet      = 0;

                                          // NOTE: for testing at init time
   pDR         = getCurrentResolution (); // getDefaultResolution ();
   pDM         = getCurrentMedia ();      // getDefaultMedia ();
   pDPM        = getCurrentPrintMode ();  // getDefaultPrintMode ();
   pszDitherID = getCurrentDitherID ();   // getDefaultDitherID ();

   if (  docDeviceGammas_d
      && pDR
      && pDM
      && pDPM
      && pszDitherID
      )
   {
      char      *pszDitherCatagoryID = 0;
      int        iResolutionID       = 0;
      int        iMediaID            = 0;
      int        iPrintModeID        = 0;
      xmlNodePtr rootDeviceGammas    = xmlDocGetRootElement (docDeviceGammas_d);
      xmlNodePtr elmDeviceGamma      = 0;

      iResolutionID       = pDR->getID ();
      iMediaID            = pDM->getID ();
      iPrintModeID        = pDPM->getID ();
      pszDitherCatagoryID = DeviceDither::getDitherCatagory (pszDitherID);

      elmDeviceGamma = firstNode (rootDeviceGammas);
      if (elmDeviceGamma)
         elmDeviceGamma = firstNode (elmDeviceGamma->children);

      while (  elmDeviceGamma
            && !pDGRet
            )
      {
         char *pszResolution     = 0;
         char *pszMedia          = 0;
         char *pszPrintMode      = 0;
         char *pszDitherCatagory = 0;

         try
         {
            // Read in the resolution
            pszResolution = getXMLContentString (elmDeviceGamma, docDeviceGammas_d, "gammaTableResolution");
            // Read in the media
            pszMedia = getXMLContentString (elmDeviceGamma, docDeviceGammas_d, "gammaTableMedia");
            // Read in the print mode
            pszPrintMode = getXMLContentString (elmDeviceGamma, docDeviceGammas_d, "gammaTablePrintMode");
            // Read in the dither catagory
            pszDitherCatagory = getXMLContentString (elmDeviceGamma, docDeviceGammas_d, "gammaTableDitherCatagory");

#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ())
            {
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszResolution = " << pszResolution << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszMedia = " << pszMedia << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszPrintMode = " << pszPrintMode << endl;
               DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pszDitherCatagory = " << pszDitherCatagory << endl;
            }
#endif

            if (  iResolutionID == DeviceResolution::nameToID (pszResolution)
               && iMediaID      == DeviceMedia::nameToID (pszMedia)
               && iPrintModeID  == DevicePrintMode::nameToID (pszPrintMode)
               && 0 == strcmp (pszDitherCatagory, pszDitherCatagoryID)
               )
            {
               try
               {
                  int iCGamma = 0;
                  int iMGamma = 0;
                  int iYGamma = 0;
                  int iKGamma = 0;
                  int iCBias  = 0;
                  int iMBias  = 0;
                  int iYBias  = 0;
                  int iKBias  = 0;

                  iCGamma = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableCGamma");
                  iMGamma = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableMGamma");
                  iYGamma = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableYGamma");
                  iKGamma = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableKGamma");
                  iCBias  = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableCBias");
                  iMBias  = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableMBias");
                  iYBias  = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableYBias");
                  iKBias  = getXMLContentInt (elmDeviceGamma, docDeviceGammas_d, "gammaTableKBias");

#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ())
                  {
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iCGamma = " << iCGamma << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iMGamma = " << iMGamma << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iYGamma = " << iYGamma << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iKGamma = " << iKGamma << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iCBias  = " << iCBias  << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iMBias  = " << iMBias  << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iYBias  = " << iYBias  << endl;
                     DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": iKBias  = " << iKBias  << endl;
                  }
#endif

                  pDGRet = new DeviceGamma (iCGamma,
                                            iMGamma,
                                            iYGamma,
                                            iKGamma,
                                            iCBias,
                                            iMBias,
                                            iYBias,
                                            iKBias);
               }
               catch (string *pstringError)
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

                  delete pstringError;
               }
            }
         }
         catch (string *pstringError)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

            delete pstringError;
         }

         if (pszResolution)
         {
            free (pszResolution);
         }
         if (pszMedia)
         {
            free (pszMedia);
         }
         if (pszPrintMode)
         {
            free (pszPrintMode);
         }
         if (pszDitherCatagory)
         {
            free (pszDitherCatagory);
         }

         elmDeviceGamma = nextNode (elmDeviceGamma);
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ())
   {
      if (!pDGRet)
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Error: returning NULL gamma!" << endl;
      else
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": returns " << pDGRet << endl;
   }
#endif

   return pDGRet;
}

XMLDevice * XMLDevice::
isAXMLDevice (Device *pDevice)
{
   return dynamic_cast<XMLDevice *>(pDevice);
}

xmlNodePtr XMLDevice::
findEntryKeyValue (xmlNodePtr  root,
                   char       *pszKey,
                   char       *pszValue,
                   bool        fShouldDebugOutput)
{
   if (root == 0)
      return 0;

   xmlNodePtr elm      = root;
   xmlNodePtr elmFound = 0;

   while (elm != 0)
   {
      char *pszElmValue = (char *)xmlGetProp (elm, (xmlChar *)pszKey);

      if (  pszElmValue
         && 0 == strcmp (pszElmValue, pszValue)
         )
      {
         elmFound = elm;
      }

      if (pszElmValue)
      {
         free (pszElmValue);
      }

      if (  !elmFound
         && elm->xmlChildrenNode
         )
      {
         elmFound = findEntryKeyValue (elm->xmlChildrenNode, pszKey, pszValue, false);
      }

      if (elmFound)
      {
         elm = elmFound;
         break;
      }

      elm = nextNode (elm);
   }

#ifndef RETAIL
   if (fShouldDebugOutput)
   {
      if (elm)
      {
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Found ";
      }
      else
      {
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Did not find ";
      }
      DebugOutput::getErrorStream () << pszValue << endl;
   }
#endif

   return elm;
}

xmlDocPtr XMLDevice::
getDeviceXML (char *pszDeviceType)
{
   xmlNodePtr elm = 0;

   elm = firstNode (rootDevice_d);
   if (elm)
      elm = firstNode (elm->children);

   while (elm)
   {
      if (  0 == strcmp ((char *)elm->name, "Has")
         || 0 == strcmp ((char *)elm->name, "Uses")
         )
      {
         char *pszTargetXMLFile = 0;

         pszTargetXMLFile = (char *)xmlNodeListGetString (docDevice_d,
                                                          elm->xmlChildrenNode,
                                                          1);

         if (pszTargetXMLFile)
         {
            char      *pszFullTargetXMLFile = 0;
            xmlDocPtr  docTargetXML         = 0;

            pszFullTargetXMLFile = Omni::openXMLFile (pszTargetXMLFile);

            if (pszFullTargetXMLFile)
            {
               docTargetXML = xmlParseFile (pszFullTargetXMLFile);
            }

            if (docTargetXML)
            {
               xmlNodePtr  elmTargetElm = 0;

               elmTargetElm = xmlDocGetRootElement (docTargetXML);
               elmTargetElm = firstNode (elmTargetElm);

               if (0 == strcmp ((const char *)elmTargetElm->name, pszDeviceType))
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDevice ()) DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": Loading \"" << pszFullTargetXMLFile << "\"" << endl;
#endif

                  free (pszTargetXMLFile);

                  free (pszFullTargetXMLFile);

                  return docTargetXML;
               }

               xmlFreeDoc (docTargetXML);
            }

            free (pszTargetXMLFile);

            free (pszFullTargetXMLFile);
         }
      }

      elm = nextNode (elm);
   }

   return 0;
}

xmlDocPtr XMLDevice::
getDocBooklets ()
{
   return docDeviceBooklets_d;
}

xmlDocPtr XMLDevice::
getDocForms ()
{
   return docDeviceForms_d;
}

xmlDocPtr XMLDevice::
getDocMedias ()
{
   return docDeviceMedias_d;
}

xmlDocPtr XMLDevice::
getDocOrientations ()
{
   return docDeviceOrientations_d;
}

xmlDocPtr XMLDevice::
getDocPrintModes ()
{
   return docDevicePrintModes_d;
}

xmlDocPtr XMLDevice::
getDocResolutions ()
{
   return docDeviceResolutions_d;
}

xmlDocPtr XMLDevice::
getDocTrays ()
{
   return docDeviceTrays_d;
}

xmlDocPtr XMLDevice::
getDocGammas ()
{
   return docDeviceGammas_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

char * XMLDevice::
getXMLDeviceName ()
{
   return pszXMLDeviceName_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

DeviceInstance * XMLDevice::
getDeviceInstance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ())
   {
      if (pInstance_d)
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": " << typeid (*pInstance_d).name () << endl;
      else
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pInstance_d is NULL!" << endl;
   }
#endif

   // Ask the instance for the child instance
   return pInstance_d->getDeviceInstance ();
}

DeviceBlitter * XMLDevice::
getDeviceBlitter ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevice ())
   {
      if (pBlitter_d)
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": " << typeid (*pBlitter_d).name () << endl;
      else
         DebugOutput::getErrorStream () << "XMLDevice::" << __FUNCTION__ << ": pBlitter_d is NULL!" << endl;
   }
#endif

   // Ask the blitter for the child blitter
   return pBlitter_d->getDeviceBlitter ();
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

static bool
parseHexGroup (const char   *pszData,
               unsigned int *puiData)
{
   if (  isxdigit (pszData[0])
      && isxdigit (pszData[1])
      )
   {
      return 1 == sscanf (pszData, "%x", puiData);
   }
   else
   {
      return false;
   }
}

bool XMLDevice::
parseBinaryData (const char  *pszData,
                 byte       **ppbData,
                 int         *pcbData)
   throw (string *)
{
   int    cbDataAlloc    = 64;
   int    cbDataLeft     = 64;
   byte  *pbdData        = new byte [cbDataAlloc];
   byte  *pbdDataCurrent = pbdData;
   bool   fContinue      = true;
   bool   fInString      = false;
   byte   bChar          = 0;
   byte   bNextChar      = 0;
   bool   fCharFound     = false;
   bool   fNextCharFound = false;
   string stringError;

   *ppbData = pbdData;
   *pcbData = cbDataAlloc - cbDataLeft;

   while (  *pszData
         && fContinue
         )
   {
      if (fNextCharFound)
      {
         bChar          = bNextChar;
         fNextCharFound = false;
         fCharFound     = true;
      }
      else
      {
         if (fInString)
         {
            switch (*pszData)
            {
            case '\\':
            {
               break;
            }

            case '"':
            {
               pszData++;

               fInString = false;
               continue;
            }

            default:
            {
               bChar      = *pszData++;
               fCharFound = true;
               break;
            }
            }
         }
         else
         {
            while (  '\x0a' == *pszData
                  || '\x0d' == *pszData
                  || ' '    == *pszData
                  )
               pszData++;

            switch (*pszData)
            {
            case '_':
            {
               pszData++;

               if (0 == strncmp (pszData, "ESC_", 4))
               {
                  bChar      = 0x1b;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "NUL_", 4))
               {
                  bChar      = 0;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "LF_", 3))
               {
                  bChar      = 0x0a;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "CR_", 3))
               {
                  bChar      = 0x0d;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "FF_", 3))
               {
                  bChar      = 0x0c;
                  fCharFound = true;
                  pszData += 3;
               }
               else
               {
                  stringError  = "Error: Unknown define ";
                  stringError += pszData;
                  fContinue    = false;
               }
               break;
            }

            case 'H':
            {
               pszData++;

               if (0 == strncmp (pszData, "EX2S", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData  = 0;
                     unsigned int uiData2 = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (',' == *pszData)
                        {
                           pszData++;

                           while (' ' == *pszData)
                              pszData++;

                           if (parseHexGroup (pszData, &uiData2))
                           {
                              pszData += 2;

                              while (' ' == *pszData)
                                 pszData++;

                              if (')' == *pszData)
                              {
                                 pszData++;

                                 bChar          = (byte)uiData2;
                                 bNextChar      = (byte)uiData;
                                 fCharFound     = true;
                                 fNextCharFound = true;
                              }
                              else
                              {
                                 stringError  = "Error: Expecting ')', found '";
                                 stringError += *pszData;
                                 stringError += "' @ ";
                                 stringError += __LINE__;
                                 fContinue    = false;
                              }
                           }
                           else
                           {
                              stringError  = "Error: Expecting ',', found '";
                              stringError += *pszData;
                              stringError += "' @ ";
                              stringError += __LINE__;
                              fContinue    = false;
                           }
                        }
                        else
                        {
                           stringError  = "Error: Expecting ',', found '";
                           stringError += *pszData;
                           stringError += "' @ ";
                           stringError += __LINE__;
                           fContinue    = false;
                        }
                     }
                     else
                     {
                        stringError  = "Error: Expecting 2 hex digits, found '";
                        stringError += pszData[0];
                        stringError += pszData[1];
                        stringError += "' @ ";
                        stringError += __LINE__;
                        fContinue    = false;
                     }
                  }
                  else
                  {
                     stringError  = "Error: Expecting '(', found '";
                     stringError += *pszData;
                     stringError += "' @ ";
                     stringError += __LINE__;
                     fContinue    = false;
                  }
               }
               else if (0 == strncmp (pszData, "EX", 2))
               {
                  pszData += 2;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (')' == *pszData)
                        {
                           pszData++;

                           bChar      = (byte)uiData;
                           fCharFound = true;
                        }
                        else
                        {
                           stringError  = "Error: Expecting ')', found '";
                           stringError += *pszData;
                           stringError += "' @ ";
                           stringError += __LINE__;
                           fContinue    = false;
                        }
                     }
                     else
                     {
                        stringError  = "Error: Expecting 2 hex digits, found '";
                        stringError += pszData[0];
                        stringError += pszData[1];
                        stringError += "' @ ";
                        stringError += __LINE__;
                        fContinue    = false;
                     }
                  }
                  else
                  {
                     stringError  = "Error: Expecting '(', found '";
                     stringError += *pszData;
                     stringError += "' @ ";
                     stringError += __LINE__;
                     fContinue    = false;
                  }
               }
               break;
            }

            case 'A':
            {
               pszData++;

               if (0 == strncmp (pszData, "SCII", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     bChar      = (byte)*pszData++;
                     fCharFound = true;

                     if (')' == *pszData)
                     {
                        pszData++;
                     }
                     else
                     {
                        stringError  = "Error: Expecting ')', found '";
                        stringError += *pszData;
                        stringError += "' @ ";
                        stringError += __LINE__;
                        fContinue    = false;
                     }
                  }
                  else
                  {
                     stringError  = "Error: Expecting '(', found '";
                     stringError += *pszData;
                     stringError += "' @ ";
                     stringError += __LINE__;
                     fContinue    = false;
                  }
               }
               break;
            }

            case '"':
            {
               pszData++;
               fInString = true;
               break;
            }
            }
         }
      }

      if (fCharFound)
      {
/////////if (isprint (bChar))
/////////   cerr << "bChar = '" << bChar << "'" << endl;
/////////else
/////////   cerr << "bChar = 0x" << hex << (int)bChar << endl;

         *pbdDataCurrent++ = bChar;
         cbDataLeft--;

         bChar      = 0;
         fCharFound = false;

         if (0 == cbDataLeft)
         {
            byte *pbTemp = new byte [cbDataAlloc + 64];

            if (pbTemp)
            {
               memcpy (pbTemp, pbdData, cbDataAlloc);

               delete[] pbdData;

               pbdData        = pbTemp;
               pbdDataCurrent = pbdData + cbDataAlloc;
               cbDataAlloc   += 64;
               cbDataLeft     = 64;
            }
            else
            {
               stringError  = "Error: Allocation of ";
               stringError += (cbDataAlloc + 64);
               stringError += " bytes failed @ ";
               stringError += __LINE__;
               fContinue    = false;
            }
         }
      }
      else if (  !fInString
              && *pszData
              )
      {
         delete[] pbdData;

         return false;
      }
   }

   if (fContinue)
   {
      *ppbData = pbdData;
      *pcbData = cbDataAlloc - cbDataLeft;
   }
   else
   {
      delete[] pbdData;

      throw new string (stringError);
   }

   return fContinue;
}

static xmlDocPtr
getDeviceConfigurationFromFile (char *pszMasterXMLFile)
{
   xmlDocPtr  docRet   = 0;

   docRet = xmlParseFile (pszMasterXMLFile);
///xmlValidateDocument (&validator, docRet); // @TBD

   if (docRet)
   {
      if (isDeviceConfigurationXMLFile (docRet))
      {
         return docRet;
      }

      xmlFreeDoc (docRet);
   }

   return 0;
}

static bool
isDeviceConfigurationXMLFile (xmlDocPtr doc)
{
   xmlNodePtr nodeRoot = xmlDocGetRootElement (doc);
   bool       fRc      = false;

   if (nodeRoot)
   {
      xmlNodePtr nodeElm = 0;

      nodeElm = firstNode (nodeRoot);

      if (nodeElm)
      {
         if (0 == strcmp ((char *)nodeElm->name, "Device"))
         {
            fRc = true;
         }
      }
   }

   return fRc;
}

static void
convertFilename (char *pszFileName)
{
   while (*pszFileName)
   {
      switch (*pszFileName)
      {
      case '-':
      case '+':
      case '(':
      case ')':
      case ' ':
      case '/':
      case '\\':
         *pszFileName = '_';
      }

      pszFileName++;
   }
}

char *
convertToLibrary (char *pszFileName)
{
   if (  !pszFileName
      || !*pszFileName
      )
   {
      return pszFileName;
   }

   convertFilename (pszFileName);

   char *pszDot = strchr (pszFileName, '.');

   if (pszDot)
   {
      *pszDot = '\0';
   }

   char *pszRet     = 0;
   int   cbFileName = 0;

   cbFileName = 3                    // 'lib'
              + strlen (pszFileName)
              + 3                    // '.so'
              + 1;                   // '\0'

   pszRet = (char *)malloc (cbFileName);
   if (pszRet)
   {
      strcpy (pszRet, "lib");
      strcat (pszRet, pszFileName);
      strcat (pszRet, ".so");

      free (pszFileName);

      return pszRet;
   }
   else
   {
      return pszFileName;
   }
}

string XMLDevice::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << PrintDevice::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const XMLDevice& const_self)
{
   XMLDevice&    self = const_cast<XMLDevice&>(const_self);
   ostringstream oss;

   os << self.toString (oss);

   return os;
}
