/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <XMLDeviceForm.hpp>

XMLDeviceForm::
XMLDeviceForm (Device      *pDevice,
               int          id,
               int          iCapabilities,
               BinaryData  *pbdData,
               HardCopyCap *hcInfo,
               xmlNodePtr   node)
   : DeviceForm (pDevice,
                 id,
                 iCapabilities,
                 pbdData,
                 hcInfo)
{
   node_d = node;
}

XMLDeviceForm::
~XMLDeviceForm ()
{
}

DeviceForm * XMLDeviceForm::
createV (int id)
{
   return XMLDeviceForm::create (pDevice_d, id);
}

DeviceForm * XMLDeviceForm::
create (Device *pDevice, char *pszId)
{
   int id = DeviceForm::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": id = " << id << endl;
#endif

   if (DeviceForm::FORM_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DeviceForm * XMLDeviceForm::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": !pXMLDevice " << endl;
#endif

      return 0;
   }

   xmlDocPtr   docDeviceForms  = pXMLDevice->getDocForms ();
   xmlNodePtr  rootDeviceForms = xmlDocGetRootElement (docDeviceForms);
   xmlNodePtr  elmDeviceForm   = 0;
   DeviceForm *pDFRet          = 0;

   if (!rootDeviceForms)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": !rootDeviceForms " << endl;
#endif

      return 0;
   }

   elmDeviceForm = firstNode (rootDeviceForms);
   if (!elmDeviceForm)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": !elmDeviceForm " << endl;
#endif

      return 0;
   }

   elmDeviceForm = firstNode (elmDeviceForm->children);

   while (  elmDeviceForm
         && !pDFRet
         )
   {
      char *pszForm = 0;

      pszForm = getXMLContentString (elmDeviceForm, docDeviceForms, "name");

      if (pszForm)
      {
         int idForm = DeviceForm::nameToID (pszForm);

         if (idForm == id)
         {
            char        *pszComma        = 0;
            char        *pszCapabilities = 0;
            int          iCapabilities   = 0;
            char        *pszCommand      = 0;
            BinaryData  *pbdData         = 0;
            int          iLeft           = 0;
            int          iTop            = 0;
            int          iRight          = 0;
            int          iBottom         = 0;
            HardCopyCap *hcInfo          = 0;
            xmlNodePtr   elmHCC          = 0;

            pszComma = strchr (pszForm, ',');

            try
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": Creating \"" << pszForm << "\"" << endl;
#endif

               // Read in the capabilities
               pszCapabilities = getXMLContentString (elmDeviceForm, docDeviceForms, "formCapabilities");

               if (pszCapabilities)
               {
                  iCapabilities = DeviceForm::getReservedValue (pszCapabilities);

                  free (pszCapabilities);
               }

               // Read in the command
               pszCommand = getXMLContentString (elmDeviceForm, docDeviceForms, "command");

               if (pszCommand)
               {
                  byte       *pbData          = 0;
                  int         cbData          = 0;

                  if (XMLDevice::parseBinaryData (pszCommand,
                                                  &pbData,
                                                  &cbData))
                  {
                     pbdData = new BinaryData (pbData, cbData);
                  }

                  free (pszCommand);
               }

               // Read in the HardCopyCap
               elmHCC  = findEntry (elmDeviceForm, "hardCopyCap", false);
               iLeft   = getXMLContentInt (elmHCC, docDeviceForms, "hardCopyCapLeft");
               iTop    = getXMLContentInt (elmHCC, docDeviceForms, "hardCopyCapTop");
               iRight  = getXMLContentInt (elmHCC, docDeviceForms, "hardCopyCapRight");
               iBottom = getXMLContentInt (elmHCC, docDeviceForms, "hardCopyCapBottom");
               hcInfo  = new HardCopyCap (iLeft, iTop, iRight, iBottom);

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceForm ())
               {
                  DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": iCapabilities = 0x" << hex << iCapabilities << dec << endl;
                  if (pbdData)
                     DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": pbdData = " << *pbdData << endl;
                  else
                     DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": pbdData is null!" << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": iLeft   = " << iLeft << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": iTop    = " << iTop << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": iRight  = " << iRight << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": iBottom = " << iBottom << endl;
               }
#endif

               // Create the object
               pDFRet = new XMLDeviceForm (pDevice,
                                           id,
                                           iCapabilities,
                                           pbdData,
                                           hcInfo,
                                           elmDeviceForm);

               if (  pszComma
                  && pDFRet
                  )
               {
                  int iCy = 0;

                  // @TBD
                  if (1 == sscanf (pszComma + 1, "%d", &iCy))
                  {
                     pDFRet->setCy (iCy);
                  }
               }

               if (hcInfo)
               {
                  hcInfo->setOwner (pDFRet);
               }
            }
            catch (string *pstringError)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

               delete pstringError;
            }
         }

         free (pszForm);
      }

      elmDeviceForm = nextNode (elmDeviceForm);
   }

#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceForm::" << __FUNCTION__ << ": returning " << pDFRet << endl;
#endif

   return pDFRet;
}

char * XMLDeviceForm::
getDeviceID ()
{
   return 0;
}

bool XMLDeviceForm::
isFormSupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDeviceForms  = pXMLDevice->getDocForms ();
   xmlNodePtr  rootDeviceForms = xmlDocGetRootElement (docDeviceForms);
   xmlNodePtr  elmDeviceForm   = 0;
   bool        fFound          = false;

   if (!rootDeviceForms)
      return false;

   elmDeviceForm = firstNode (rootDeviceForms);
   if (!elmDeviceForm)
      return false;

   elmDeviceForm = firstNode (elmDeviceForm->children);

   while (  elmDeviceForm
         && !fFound
         )
   {
      char *pszForm = 0;

      pszForm = getXMLContentString (elmDeviceForm, docDeviceForms, "name");

      if (pszForm)
      {
         int idForm = DeviceForm::nameToID (pszForm);

         if (idForm == id)
         {
            fFound = true;
         }

         free (pszForm);
      }

      elmDeviceForm = nextNode (elmDeviceForm);
   }

   return fFound;
}

class XMLForm_Enumerator : public Enumeration
{
public:
   XMLForm_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d     = pXMLDevice;
      docDeviceForms_d = 0;
      nodeItem_d       = nodeItem;

      if (pXMLDevice)
      {
         docDeviceForms_d = pXMLDevice->getDocForms ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DeviceForm *pDFRet = 0;

      if (nodeItem_d)
      {
         char *pszFormName = 0;

         pszFormName = getXMLContentString (nodeItem_d, docDeviceForms_d, "name");

         if (pszFormName)
         {
            pDFRet = XMLDeviceForm::create (pXMLDevice_d, pszFormName);

            free (pszFormName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDFRet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDeviceForms_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDeviceForm::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLForm_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDeviceForms  = pXMLDevice->getDocForms ();
   xmlNodePtr  rootDeviceForms = xmlDocGetRootElement (docDeviceForms);
   xmlNodePtr  elmDeviceForm   = 0;

   if (!rootDeviceForms)
      return new XMLForm_Enumerator (pDevice_d, 0);

   elmDeviceForm = firstNode (rootDeviceForms);
   if (!elmDeviceForm)
      return new XMLForm_Enumerator (pDevice_d, 0);

   elmDeviceForm = firstNode (elmDeviceForm->children);

   return new XMLForm_Enumerator (pDevice_d, elmDeviceForm);
}

string XMLDeviceForm::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceForm::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const XMLDeviceForm& const_self)
{
   XMLDeviceForm& self = const_cast<XMLDeviceForm&>(const_self);
   ostringstream  oss;

   os << self.toString (oss);

   return os;
}
