/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "XMLDeviceInstance.hpp"
#include "XMLDevice.hpp"

#include <typeinfo>

XMLDeviceInstance::
XMLDeviceInstance (GModule     *hmodLibrary,
                   PrintDevice *pDevice)
   : DeviceInstance (pDevice)
{
   hmodLibrary_d       = hmodLibrary;
   pfnCreateInstance_d = 0;
   pfnDeleteInstance_d = 0;
   pInstance_d         = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": hmodLibrary_d = " << hmodLibrary_d << endl;
#endif

   if (hmodLibrary_d)
   {
      int rc = 0;

      rc = ::g_module_symbol (hmodLibrary_d,
                              "createInstance",
                              (void **)&pfnCreateInstance_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pfnCreateInstance_d = " << (void *)pfnCreateInstance_d << endl;
#endif

      if (!rc)
      {
         cerr << "g_module_error returns " << g_module_error () << endl;
      }

      rc = ::g_module_symbol (hmodLibrary_d,
                              "deleteInstance",
                              (void **)&pfnDeleteInstance_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pfnDeleteInstance_d = " << (void *)pfnDeleteInstance_d << endl;
#endif

      if (!rc)
      {
         cerr << "g_module_error returns " << g_module_error () << endl;
      }

      if (  pfnCreateInstance_d
         || pfnDeleteInstance_d
         )
      {
         pInstance_d = pfnCreateInstance_d (pDevice);

#ifndef RETAIL
         if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pInstance_d = " << pInstance_d << endl;
#endif
      }
   }
}

XMLDeviceInstance::
~XMLDeviceInstance ()
{
   if (pInstance_d)
   {
      pfnDeleteInstance_d (pInstance_d);
   }

   if (hmodLibrary_d)
   {
      g_module_close (hmodLibrary_d);
      hmodLibrary_d = 0;
   }
}

void XMLDeviceInstance::
initializeInstance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
   {
      pInstance_d->initializeInstance ();
   }
}

bool XMLDeviceInstance::
hasError ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->hasError ();
   else
      return false;
}

bool XMLDeviceInstance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pszKey = " << pszKey << ", pszValue = " << pszValue << endl;
#endif

   if (pInstance_d)
      return pInstance_d->handleDeviceJobProperty (pszKey, pszValue);
   else
      return false;
}

string * XMLDeviceInstance::
queryDeviceJobProperties ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->queryDeviceJobProperties ();
   else
      return 0;
}

static char *vapszExtraKeys[] = {
   "XMLMasterFile"
};

#define KEY_DEVICENAME vapszExtraKeys[0]

class ChildWithExtraEnumerator : public Enumeration
{
public:
   ChildWithExtraEnumerator (Enumeration *pChild,
                             char       **apszExtras,
                             int          cExtras)
   {
      pChild_d     = pChild;
      apszExtras_d = apszExtras;
      cExtras_d    = cExtras;
   }

   virtual bool hasMoreElements ()
   {
      bool fReturn = false;

#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pChild_d = " << pChild_d << ", apszExtras_d = " << apszExtras_d << ", cExtras_d = " << cExtras_d << endl;
#endif

      if (pChild_d)
      {
         fReturn = pChild_d->hasMoreElements ();
      }

      if (  !fReturn
         && apszExtras_d
         )
      {
         fReturn = cExtras_d > 0;
      }

      return fReturn;
   }

   virtual void *nextElement ()
   {
      void *pvReturn = 0;

      if (pChild_d)
      {
         pvReturn = pChild_d->nextElement ();

         if (!pvReturn)
         {
            pChild_d = 0;
         }
      }

      if (  !pvReturn
         && apszExtras_d
         && 0 < cExtras_d
         )
      {
         pvReturn = apszExtras_d[--cExtras_d];
      }

      return pvReturn;
   }

private:
   Enumeration *pChild_d;
   char       **apszExtras_d;
   int          cExtras_d;
};

Enumeration * XMLDeviceInstance::
listDeviceJobPropertyKeys ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   return new ChildWithExtraEnumerator (pInstance_d
                                        ? pInstance_d->listDeviceJobPropertyKeys ()
                                        : 0,
                                        vapszExtraKeys,
                                        dimof (vapszExtraKeys));
}

class NullValueEnumerator : public Enumeration
{
public:
   NullValueEnumerator (Enumeration *pChild)
   {
      pChild_d = pChild;
   }

   virtual bool hasMoreElements ()
   {
      return false;
   }

   virtual void *nextElement ()
   {
      return 0;
   }

private:
   Enumeration *pChild_d;
};

class DeviceNameValueEnumerator : public Enumeration
{
public:
   DeviceNameValueEnumerator (string *pstringXMLDeviceName)
   {
      pstringXMLDeviceName_d = pstringXMLDeviceName;

      if (pstringXMLDeviceName)
         fSentString_d = false;
      else
         fSentString_d = true;
   }

   virtual ~DeviceNameValueEnumerator ()
   {
      delete pstringXMLDeviceName_d;
      pstringXMLDeviceName_d = 0;
   }

   virtual bool hasMoreElements ()
   {
      if (!fSentString_d)
      {
         return true;
      }
      else
      {
         return false;
      }
   }

   virtual void *nextElement ()
   {
      void *pvReturn = 0;

      if (!fSentString_d)
      {
         if (pstringXMLDeviceName_d)
         {
            pvReturn = (void *)pstringXMLDeviceName_d->c_str ();
         }

         fSentString_d = true;
      }

      return pvReturn;
   }

private:
   string *pstringXMLDeviceName_d;
   bool    fSentString_d;
};

Enumeration * XMLDeviceInstance::
listKeyValues (char *pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pszKey = " << pszKey << endl;
#endif

   if (0 == strcmp (pszKey, KEY_DEVICENAME))
   {
      string *pstringXMLDeviceName = 0;

      if (pDevice_d)
      {
         XMLDevice *pDevice = dynamic_cast<XMLDevice *>(pDevice_d);

         if (pDevice)
         {
            pstringXMLDeviceName = new string ("\"");
            *pstringXMLDeviceName += pDevice->getXMLDeviceName ();
            *pstringXMLDeviceName += "\"";
         }
      }

      return new DeviceNameValueEnumerator (pstringXMLDeviceName);
   }

   if (pInstance_d)
      return pInstance_d->listKeyValues (pszKey);
   else
      return new NullValueEnumerator (0);
}

string * XMLDeviceInstance::
getJobPropertyType (char *pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pszKey = " << pszKey << endl;
#endif

   if (0 == strcmp (pszKey, KEY_DEVICENAME))
   {
      string *pRet = 0;

      pRet = new string ("string ");

      if (pDevice_d)
      {
         XMLDevice *pDevice = dynamic_cast<XMLDevice *>(pDevice_d);

         if (pDevice)
         {
            *pRet += "\"";
            *pRet += pDevice->getXMLDeviceName ();
            *pRet += "\"";
         }
         else
         {
            delete pRet;
            pRet = 0;
         }
      }
      else
      {
         delete pRet;
         pRet = 0;
      }

      return pRet;
   }

   if (pInstance_d)
      return pInstance_d->getJobPropertyType (pszKey);
   else
      return 0;
}

string * XMLDeviceInstance::
getJobProperty (char *pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pszKey = " << pszKey << endl;
#endif

   if (0 == strcmp (pszKey, KEY_DEVICENAME))
   {
      string *pRet = 0;

      if (pDevice_d)
      {
         XMLDevice *pDevice = dynamic_cast<XMLDevice *>(pDevice_d);

         if (pDevice)
         {
            pRet = new string ("\"");
            *pRet += pDevice->getXMLDeviceName ();
            *pRet += "\"";
         }
      }

      return pRet;
   }

   if (pInstance_d)
      return pInstance_d->getJobProperty (pszKey);
   else
      return 0;
}

string * XMLDeviceInstance::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": pszKey = " << pszKey << ", pszValue = " << pszValue << endl;
#endif

   if (0 == strcmp (pszKey, KEY_DEVICENAME))
   {
      string *pRet = new string (pszKey);

      if (pszValue)
      {
         *pRet += "=";
         *pRet += pszValue;
      }

      return pRet;
   }

   if (pInstance_d)
      return pInstance_d->translateKeyValue (pszKey,
                                             pszValue);
   else
      return 0;
}

bool XMLDeviceInstance::
setOutputStream (FILE *pFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->setOutputStream (pFile);
   else
      return false;
}

bool XMLDeviceInstance::
setErrorStream (FILE *pFile)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->setErrorStream (pFile);
   else
      return false;
}

bool XMLDeviceInstance::
setLanguage (int iLanguageID)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->setLanguage (iLanguageID);
   else
      return false;
}

bool XMLDeviceInstance::
beginJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->beginJob ();
   else
      return false;
}

bool XMLDeviceInstance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->beginJob (fJobPropertiesChanged);
   else
      return false;
}

bool XMLDeviceInstance::
newFrame ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->newFrame ();
   else
      return false;
}

bool XMLDeviceInstance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->newFrame (fJobPropertiesChanged);
   else
      return false;
}

bool XMLDeviceInstance::
endJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->endJob ();
   else
      return false;
}

bool XMLDeviceInstance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << endl;
#endif

   if (pInstance_d)
      return pInstance_d->abortJob ();
   else
      return false;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

DeviceInstance * XMLDeviceInstance::
getDeviceInstance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceInstance ()) DebugOutput::getErrorStream () << "XMLDeviceInstance::" << __FUNCTION__ << ": " << typeid (*pInstance_d).name () << endl;
#endif

   return pInstance_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string XMLDeviceInstance::
toString (ostringstream& oss)
{
   // @TBD
   oss << "{ "
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream&                 os,
            const XMLDeviceInstance& const_self)
{
   XMLDeviceInstance& self = const_cast<XMLDeviceInstance&>(const_self);
   ostringstream      oss;

   os << self.toString (oss);

   return os;
}
