/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <XMLDeviceMedia.hpp>

XMLDeviceMedia::
XMLDeviceMedia (Device      *pDevice,
                int          id,
                BinaryData  *pbdData,
                int          iColorAdjustRequired,
                int          iAbsorption,
                xmlNodePtr   node)
   : DeviceMedia (pDevice,
                  id,
                  pbdData,
                  iColorAdjustRequired,
                  iAbsorption)
{
   node_d = node;
}

XMLDeviceMedia::
~XMLDeviceMedia ()
{
}

DeviceMedia * XMLDeviceMedia::
createV (int id)
{
   return XMLDeviceMedia::create (pDevice_d, id);
}

DeviceMedia * XMLDeviceMedia::
create (Device *pDevice, char *pszId)
{
   int id = DeviceMedia::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": id = " << id << endl;
#endif

   if (DeviceMedia::MEDIA_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DeviceMedia * XMLDeviceMedia::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": !pXMLDevice " << endl;
#endif

      return 0;
   }

   xmlDocPtr    docDeviceMedias  = pXMLDevice->getDocMedias ();
   xmlNodePtr   rootDeviceMedias = xmlDocGetRootElement (docDeviceMedias);
   xmlNodePtr   elmDeviceMedia   = 0;
   DeviceMedia *pDMRet           = 0;

   if (!rootDeviceMedias)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": !rootDeviceMedias " << endl;
#endif

      return 0;
   }

   elmDeviceMedia = firstNode (rootDeviceMedias);
   if (!elmDeviceMedia)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": !elmDeviceMedia " << endl;
#endif

      return 0;
   }

   elmDeviceMedia = firstNode (elmDeviceMedia->children);

   while (  elmDeviceMedia
         && !pDMRet
         )
   {
      char *pszMedia = 0;

      pszMedia = getXMLContentString (elmDeviceMedia, docDeviceMedias, "name");

      if (pszMedia)
      {
         int idMedia = DeviceMedia::nameToID (pszMedia);

         if (idMedia == id)
         {
            char        *pszCommand           = 0;
            BinaryData  *pbdData              = 0;
            int          iColorAdjustRequired = 0;
            char        *pszAbsorption        = 0;
            int          iAbsorption          = 0;

            try
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": Creating \"" << pszMedia << "\"" << endl;
#endif

               // Read in the command
               pszCommand = getXMLContentString (elmDeviceMedia, docDeviceMedias, "command");

               if (pszCommand)
               {
                  byte       *pbData          = 0;
                  int         cbData          = 0;

                  if (XMLDevice::parseBinaryData (pszCommand,
                                                  &pbData,
                                                  &cbData))
                  {
                     pbdData = new BinaryData (pbData, cbData);
                  }

                  free (pszCommand);
               }

               // Read in the color adjust required
               iColorAdjustRequired = getXMLContentInt (elmDeviceMedia, docDeviceMedias, "mediaColorAdjustRequired");

               // Read in the absorption
               pszAbsorption = getXMLContentString (elmDeviceMedia, docDeviceMedias, "mediaAbsorption");

               if (pszAbsorption)
               {
                  iAbsorption = DeviceMedia::getReservedValue (pszAbsorption);

                  free (pszAbsorption);
               }

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceMedia ())
               {
                  if (pbdData)
                     DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": pbdData = " << *pbdData << endl;
                  else
                     DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": pbdData is null!" << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": iColorAdjustRequired = " << iColorAdjustRequired << endl;
                  DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": iAbsorption = 0x" << hex << iAbsorption << dec << endl;
               }
#endif

               // Create the object
               pDMRet = new XMLDeviceMedia (pDevice,
                                            id,
                                            pbdData,
                                            iColorAdjustRequired,
                                            iAbsorption,
                                            elmDeviceMedia);
            }
            catch (string *pstringError)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

               delete pstringError;
            }
         }

         free (pszMedia);
      }

      elmDeviceMedia = nextNode (elmDeviceMedia);
   }


#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceMedia ()) DebugOutput::getErrorStream () << "XMLDeviceMedia::" << __FUNCTION__ << ": returning " << pDMRet << endl;
#endif

   return pDMRet;
}

char * XMLDeviceMedia::
getDeviceID ()
{
   return 0;
}

bool XMLDeviceMedia::
isMediaSupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDeviceMedias  = pXMLDevice->getDocMedias ();
   xmlNodePtr  rootDeviceMedias = xmlDocGetRootElement (docDeviceMedias);
   xmlNodePtr  elmDeviceMedia   = 0;
   bool        fFound           = false;

   if (!rootDeviceMedias)
      return false;

   elmDeviceMedia = firstNode (rootDeviceMedias);
   if (!elmDeviceMedia)
      return false;

   elmDeviceMedia = firstNode (elmDeviceMedia->children);

   while (  elmDeviceMedia
         && !fFound
         )
   {
      char *pszMedia = 0;

      pszMedia = getXMLContentString (elmDeviceMedia, docDeviceMedias, "name");

      if (pszMedia)
      {
         int idMedia = DeviceMedia::nameToID (pszMedia);

         if (idMedia == id)
         {
            fFound = true;
         }

         free (pszMedia);
      }

      elmDeviceMedia = nextNode (elmDeviceMedia);
   }

   return fFound;
}

class XMLMedia_Enumerator : public Enumeration
{
public:
   XMLMedia_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d      = pXMLDevice;
      docDeviceMedias_d = 0;
      nodeItem_d        = nodeItem;

      if (pXMLDevice)
      {
         docDeviceMedias_d = pXMLDevice->getDocMedias ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DeviceMedia *pDMRet = 0;

      if (nodeItem_d)
      {
         char *pszMediaName = 0;

         pszMediaName = getXMLContentString (nodeItem_d, docDeviceMedias_d, "name");

         if (pszMediaName)
         {
            pDMRet = XMLDeviceMedia::create (pXMLDevice_d, pszMediaName);

            free (pszMediaName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDMRet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDeviceMedias_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDeviceMedia::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLMedia_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDeviceMedias  = pXMLDevice->getDocMedias ();
   xmlNodePtr  rootDeviceMedias = xmlDocGetRootElement (docDeviceMedias);
   xmlNodePtr  elmDeviceMedia   = 0;

   if (!rootDeviceMedias)
      return new XMLMedia_Enumerator (pDevice_d, 0);

   elmDeviceMedia = firstNode (rootDeviceMedias);
   if (!elmDeviceMedia)
      return new XMLMedia_Enumerator (pDevice_d, 0);

   elmDeviceMedia = firstNode (elmDeviceMedia->children);

   return new XMLMedia_Enumerator (pDevice_d, elmDeviceMedia);
}

string XMLDeviceMedia::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceMedia::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const XMLDeviceMedia& const_self)
{
   XMLDeviceMedia& self = const_cast<XMLDeviceMedia&>(const_self);
   ostringstream   oss;

   os << self.toString (oss);

   return os;
}
