/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <XMLDevicePrintMode.hpp>

XMLDevicePrintMode::
XMLDevicePrintMode (Device      *pDevice,
                    int          id,
                    int          iPhysicalCount,
                    int          iLogicalCount,
                    int          iPlanes,
                    xmlNodePtr   node)
   : DevicePrintMode (pDevice,
                      id,
                      iPhysicalCount,
                      iLogicalCount,
                      iPlanes)
{
   node_d = node;
}

XMLDevicePrintMode::
~XMLDevicePrintMode ()
{
}

DevicePrintMode * XMLDevicePrintMode::
createV (int id)
{
   return XMLDevicePrintMode::create (pDevice_d, id);
}

DevicePrintMode * XMLDevicePrintMode::
create (Device *pDevice, char *pszId)
{
   int id = DevicePrintMode::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": id = " << id << endl;
#endif

   if (DevicePrintMode::PRINT_MODE_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DevicePrintMode * XMLDevicePrintMode::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": !pXMLDevice " << endl;
#endif

      return 0;
   }

   xmlDocPtr        docDevicePrintModes  = pXMLDevice->getDocPrintModes ();
   xmlNodePtr       rootDevicePrintModes = xmlDocGetRootElement (docDevicePrintModes);
   xmlNodePtr       elmDevicePrintMode   = 0;
   DevicePrintMode *pDPMRet              = 0;

   if (!rootDevicePrintModes)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": !rootDevicePrintModes " << endl;
#endif

      return 0;
   }

   elmDevicePrintMode = firstNode (rootDevicePrintModes);
   if (!elmDevicePrintMode)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": !elmDevicePrintMode " << endl;
#endif

      return 0;
   }

   elmDevicePrintMode = firstNode (elmDevicePrintMode->children);

   while (  elmDevicePrintMode
         && !pDPMRet
         )
   {
      char *pszPrintMode = 0;

      pszPrintMode = getXMLContentString (elmDevicePrintMode, docDevicePrintModes, "name");

      if (pszPrintMode)
      {
         int idPrintMode = DevicePrintMode::nameToID (pszPrintMode);

         if (idPrintMode == id)
         {
            int iPhysicalCount = 0;
            int iLogicalCount  = 0;
            int iPlanes        = 0;

            try
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": Creating \"" << pszPrintMode << "\"" << endl;
#endif

               // Read in the physical count
               iPhysicalCount = getXMLContentInt (elmDevicePrintMode, docDevicePrintModes, "printModePhysicalCount");
               // Read in the logical count
               iLogicalCount = getXMLContentInt (elmDevicePrintMode, docDevicePrintModes, "printModeLogicalCount");
               // Read in the planes
               iPlanes = getXMLContentInt (elmDevicePrintMode, docDevicePrintModes, "printModePlanes");

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDevicePrintMode ())
               {
                  DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": iPhysicalCount = " << iPhysicalCount << endl;
                  DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": iLogicalCount = " << iLogicalCount << endl;
                  DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": iPlanes = " << iPlanes << endl;
               }
#endif

               // Create the object
               pDPMRet = new XMLDevicePrintMode (pDevice,
                                                 id,
                                                 iPhysicalCount,
                                                 iLogicalCount,
                                                 iPlanes,
                                                 elmDevicePrintMode);
            }
            catch (string *pstringError)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

               delete pstringError;
            }
         }

         free (pszPrintMode);
      }

      elmDevicePrintMode = nextNode (elmDevicePrintMode);
   }


#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDevicePrintMode ()) DebugOutput::getErrorStream () << "XMLDevicePrintMode::" << __FUNCTION__ << ": returning " << pDPMRet << endl;
#endif

   return pDPMRet;
}

char * XMLDevicePrintMode::
getDeviceID ()
{
   return 0;
}

bool XMLDevicePrintMode::
isPrintModeSupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDevicePrintModes  = pXMLDevice->getDocPrintModes ();
   xmlNodePtr  rootDevicePrintModes = xmlDocGetRootElement (docDevicePrintModes);
   xmlNodePtr  elmDevicePrintMode   = 0;
   bool        fFound               = false;

   if (!rootDevicePrintModes)
      return false;

   elmDevicePrintMode = firstNode (rootDevicePrintModes);
   if (!elmDevicePrintMode)
      return false;

   elmDevicePrintMode = firstNode (elmDevicePrintMode->children);

   while (  elmDevicePrintMode
         && !fFound
         )
   {
      char *pszPrintMode = 0;

      pszPrintMode = getXMLContentString (elmDevicePrintMode, docDevicePrintModes, "name");

      if (pszPrintMode)
      {
         int idPrintMode = DevicePrintMode::nameToID (pszPrintMode);

         if (idPrintMode == id)
         {
            fFound = true;
         }

         free (pszPrintMode);
      }

      elmDevicePrintMode = nextNode (elmDevicePrintMode);
   }

   return fFound;
}

class XMLPrintMode_Enumerator : public Enumeration
{
public:
   XMLPrintMode_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d          = pXMLDevice;
      docDevicePrintModes_d = 0;
      nodeItem_d            = nodeItem;

      if (pXMLDevice)
      {
         docDevicePrintModes_d = pXMLDevice->getDocPrintModes ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DevicePrintMode *pDPMRet = 0;

      if (nodeItem_d)
      {
         char *pszPrintModeName = 0;

         pszPrintModeName = getXMLContentString (nodeItem_d, docDevicePrintModes_d, "name");

         if (pszPrintModeName)
         {
            pDPMRet = XMLDevicePrintMode::create (pXMLDevice_d, pszPrintModeName);

            free (pszPrintModeName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDPMRet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDevicePrintModes_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDevicePrintMode::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLPrintMode_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDevicePrintModes  = pXMLDevice->getDocPrintModes ();
   xmlNodePtr  rootDevicePrintModes = xmlDocGetRootElement (docDevicePrintModes);
   xmlNodePtr  elmDevicePrintMode   = 0;

   if (!rootDevicePrintModes)
      return new XMLPrintMode_Enumerator (pDevice_d, 0);

   elmDevicePrintMode = firstNode (rootDevicePrintModes);
   if (!elmDevicePrintMode)
      return new XMLPrintMode_Enumerator (pDevice_d, 0);

   elmDevicePrintMode = firstNode (elmDevicePrintMode->children);

   return new XMLPrintMode_Enumerator (pDevice_d, elmDevicePrintMode);
}

string XMLDevicePrintMode::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DevicePrintMode::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const XMLDevicePrintMode& const_self)
{
   XMLDevicePrintMode& self = const_cast<XMLDevicePrintMode&>(const_self);
   ostringstream       oss;

   os << self.toString (oss);

   return os;
}
