/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <XMLDeviceTray.hpp>

XMLDeviceTray::
XMLDeviceTray (Device      *pDevice,
               int          id,
               int          iType,
               BinaryData  *pbdData,
               xmlNodePtr   node)
   : DeviceTray (pDevice,
                 id,
                 iType,
                 pbdData)
{
   node_d = node;
}

XMLDeviceTray::
~XMLDeviceTray ()
{
}

DeviceTray * XMLDeviceTray::
createV (int id)
{
   return XMLDeviceTray::create (pDevice_d, id);
}

DeviceTray * XMLDeviceTray::
create (Device *pDevice, char *pszId)
{
   int id = DeviceTray::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": id = " << id << endl;
#endif

   if (DeviceTray::TRAY_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DeviceTray * XMLDeviceTray::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": !pXMLDevice " << endl;
#endif

      return 0;
   }

   xmlDocPtr   docDeviceTrays  = pXMLDevice->getDocTrays ();
   xmlNodePtr  rootDeviceTrays = xmlDocGetRootElement (docDeviceTrays);
   xmlNodePtr  elmDeviceTray   = 0;
   DeviceTray *pDTRet          = 0;

   if (!rootDeviceTrays)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": !rootDeviceTrays " << endl;
#endif

      return 0;
   }

   elmDeviceTray = firstNode (rootDeviceTrays);
   if (!elmDeviceTray)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": !elmDeviceTray " << endl;
#endif

      return 0;
   }

   elmDeviceTray = firstNode (elmDeviceTray->children);

   while (  elmDeviceTray
         && !pDTRet
         )
   {
      char *pszTray = 0;

      pszTray = getXMLContentString (elmDeviceTray, docDeviceTrays, "name");

      if (pszTray)
      {
         int idTray = DeviceTray::nameToID (pszTray);

         if (idTray == id)
         {
            char        *pszType    = 0;
            int          iType      = 0;
            char        *pszCommand = 0;
            BinaryData  *pbdData    = 0;

            try
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": Creating \"" << pszTray << "\"" << endl;
#endif

               // Read in the Type
               pszType = getXMLContentString (elmDeviceTray, docDeviceTrays, "trayType");

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": pszType = " << pszType << endl;
#endif

               if (pszType)
               {
                  iType = DeviceTray::getReservedValue (pszType);

#ifndef RETAIL
                  if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": iType = 0x" << hex << iType << dec << endl;
#endif

                  free (pszType);
               }

               // Read in the command
               pszCommand = getXMLContentString (elmDeviceTray, docDeviceTrays, "command");

               if (pszCommand)
               {
                  byte       *pbData          = 0;
                  int         cbData          = 0;

                  if (XMLDevice::parseBinaryData (pszCommand,
                                                  &pbData,
                                                  &cbData))
                  {
                     pbdData = new BinaryData (pbData, cbData);
                  }

                  free (pszCommand);
               }

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceTray ())
               {
                  if (pbdData)
                     DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": pbdData = " << *pbdData << endl;
                  else
                     DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": pbdData is null!" << endl;
               }
#endif

               // Create the object
               pDTRet = new XMLDeviceTray (pDevice,
                                           id,
                                           iType,
                                           pbdData,
                                           elmDeviceTray);
            }
            catch (string *pstringError)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": Error: " << *pstringError << endl;
#endif

               delete pstringError;
            }
         }

         free (pszTray);
      }

      elmDeviceTray = nextNode (elmDeviceTray);
   }


#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceTray ()) DebugOutput::getErrorStream () << "XMLDeviceTray::" << __FUNCTION__ << ": returning " << pDTRet << endl;
#endif

   return pDTRet;
}

char * XMLDeviceTray::
getDeviceID ()
{
   return 0;
}

bool XMLDeviceTray::
isTraySupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDeviceTrays  = pXMLDevice->getDocTrays ();
   xmlNodePtr  rootDeviceTrays = xmlDocGetRootElement (docDeviceTrays);
   xmlNodePtr  elmDeviceTray   = 0;
   bool        fFound          = false;

   if (!rootDeviceTrays)
      return false;

   elmDeviceTray = firstNode (rootDeviceTrays);
   if (!elmDeviceTray)
      return false;

   elmDeviceTray = firstNode (elmDeviceTray->children);

   while (  elmDeviceTray
         && !fFound
         )
   {
      char *pszTray = 0;

      pszTray = getXMLContentString (elmDeviceTray, docDeviceTrays, "name");

      if (pszTray)
      {
         int idTray = DeviceTray::nameToID (pszTray);

         if (idTray == id)
         {
            fFound = true;
         }

         free (pszTray);
      }

      elmDeviceTray = nextNode (elmDeviceTray);
   }

   return fFound;
}

class XMLTray_Enumerator : public Enumeration
{
public:
   XMLTray_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d     = pXMLDevice;
      docDeviceTrays_d = 0;
      nodeItem_d       = nodeItem;

      if (pXMLDevice)
      {
         docDeviceTrays_d = pXMLDevice->getDocTrays ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DeviceTray *pDTRet = 0;

      if (nodeItem_d)
      {
         char *pszTrayName = 0;

         pszTrayName = getXMLContentString (nodeItem_d, docDeviceTrays_d, "name");

         if (pszTrayName)
         {
            pDTRet = XMLDeviceTray::create (pXMLDevice_d, pszTrayName);

            free (pszTrayName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDTRet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDeviceTrays_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDeviceTray::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLTray_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDeviceTrays  = pXMLDevice->getDocTrays ();
   xmlNodePtr  rootDeviceTrays = xmlDocGetRootElement (docDeviceTrays);
   xmlNodePtr  elmDeviceTray   = 0;

   if (!rootDeviceTrays)
      return new XMLTray_Enumerator (pDevice_d, 0);

   elmDeviceTray = firstNode (rootDeviceTrays);
   if (!elmDeviceTray)
      return new XMLTray_Enumerator (pDevice_d, 0);

   elmDeviceTray = firstNode (elmDeviceTray->children);

   return new XMLTray_Enumerator (pDevice_d, elmDeviceTray);
}

string XMLDeviceTray::
toString (ostringstream& oss)
{
   ostringstream oss2;

   oss << "{"
       << DeviceTray::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const XMLDeviceTray& const_self)
{
   XMLDeviceTray& self = const_cast<XMLDeviceTray&>(const_self);
   ostringstream  oss;

   os << self.toString (oss);

   return os;
}
