/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <sstream>
#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>

#include <util/PlatformUtils.hpp>
#include <util/XMLString.hpp>
#include <sax/HandlerBase.hpp>
#include <dom/DOM_NamedNodeMap.hpp>

#include "OmniDomParser.hpp"

#ifdef DEBUG
const static bool fDebugOutput = false;
#endif

void convertFilename (char *pszFileName);

extern char *vpszExeName;

OmniDomParser::
OmniDomParser (const char *pszXMLFile,
               DeviceInfo *pdi,
               bool        fAutoconf,
               bool        fAutoconfNoInst)
{
   pdi_d    = pdi;

   parser_d = new DOMParser ();

   parser_d->setDoValidation (true);
   parser_d->setDoNamespaces (true);
   parser_d->setErrorHandler (new MyErrorHandler (this));

   pstrXMLFile_d     = new string (pszXMLFile);
   fErrorCondition_d = false;
   fAutoconf_d       = fAutoconf;
   fAutoconfNoInst_d = fAutoconfNoInst;
   fFileModified_d   = false;
}

OmniDomParser::
~OmniDomParser ()
{
   delete parser_d;
   delete pstrXMLFile_d;
}

void OmniDomParser::
setErrorCondition ()
{
   fErrorCondition_d = true;
}

bool OmniDomParser::
fileModified ()
{
   return fFileModified_d;
}

void
convertFilename (char *pszFileName)
{
   while (*pszFileName)
   {
      switch (*pszFileName)
      {
      case '-':
      case '+':
      case '(':
      case ')':
      case ' ':
      case '/':
      case '\\':
         *pszFileName = '_';
      }

      pszFileName++;
   }
}

string * OmniDomParser::
getNameOut ()
{
   int     iDotPos    = pstrXMLFile_d->find_last_of ('.');
   string *pstringRet = new string (pstrXMLFile_d->substr (0, iDotPos));

   convertFilename (const_cast<char*>(pstringRet->c_str ())); // @HACK?

   return pstringRet;
}

string *
converToClassName (const char *pszFileName)
{
   string  stringFileName (pszFileName);

   int     iDotPos    = stringFileName.find_last_of ('.');
   string *pstringRet = new string (stringFileName.substr (0, iDotPos));

   convertFilename (const_cast<char*>(pstringRet->c_str ())); // @HACK?

   return pstringRet;
}

bool OmniDomParser::
shouldCreateFile (string strFileName)
{
   string stringFileNameHPP (strFileName.c_str ());
   string stringFileNameCPP (strFileName.c_str ());

   convertFilename (const_cast<char*>(stringFileNameHPP.c_str ())); // @HACK?
   convertFilename (const_cast<char*>(stringFileNameCPP.c_str ())); // @HACK?

   int iDotPos = stringFileNameHPP.find_last_of ('.');

   stringFileNameHPP.replace (iDotPos, 4, ".hpp");

   iDotPos = stringFileNameCPP.find_last_of ('.');

   stringFileNameCPP.replace (iDotPos, 4, ".cpp");

   struct stat statExe;
   struct stat statIn;
   struct stat statOutHPP;
   struct stat statOutCPP;

   if (-1 == stat (vpszExeName, &statExe))
   {
      cerr << "Error: File not found \"" << vpszExeName << "\" (could not stat it)!" << endl;
      setErrorCondition ();
      abort ();
   }

   if (-1 == stat (strFileName.c_str (), &statIn))
   {
      cerr << "Error: File not found \"" << strFileName << "\" (could not stat it)!" << endl;
      setErrorCondition ();
      abort ();
   }

#ifdef DEBUG
   if (fDebugOutput)
   {
      cerr << "stats for \"" << strFileName << "\"" << endl;
      cerr << dec << "   atime = " << statIn.st_atime << endl;
      cerr        << "   mtime = " << statIn.st_mtime << endl;
      cerr        << "   ctime = " << statIn.st_ctime << endl;
   }
#endif

   if (-1 == stat (stringFileNameHPP.c_str (), &statOutHPP))
      return true;

#ifdef DEBUG
   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameHPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutHPP.st_atime << endl;
      cerr        << "   mtime = " << statOutHPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutHPP.st_ctime << endl;
   }
#endif

   if (-1 == stat (stringFileNameCPP.c_str (), &statOutCPP))
      return true;

#ifdef DEBUG
   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameCPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutCPP.st_atime << endl;
      cerr        << "   mtime = " << statOutCPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutCPP.st_ctime << endl;
   }
#endif

   if (  statExe.st_ctime > statOutHPP.st_ctime
      || statExe.st_ctime > statOutCPP.st_ctime
      || statIn.st_ctime > statOutHPP.st_ctime
      || statIn.st_ctime > statOutCPP.st_ctime
      || 0 == statOutHPP.st_size
      || 0 == statOutCPP.st_size
      )
      return true;
   else
      return false;
}

void OmniDomParser::
copyFile (const char *pszFileNameIn)
{
   string stringFileNameOut (pszFileNameIn);

   convertFilename (const_cast<char*>(stringFileNameOut.c_str ())); // @HACK?

   struct stat statIn;
   struct stat statOut;
   bool        fShouldCopy = false;

   if (-1 == stat (pszFileNameIn, &statIn))
   {
      cerr << "Error: File not found \"" << pszFileNameIn << "\" (could not stat it)!" << endl;
      setErrorCondition ();

      return;
   }

   if (-1 == stat (stringFileNameOut.c_str (), &statOut))
      fShouldCopy = true;

   if (  statIn.st_ctime > statOut.st_ctime
       || 0 == statOut.st_size
      )
      fShouldCopy = true;

   if (fShouldCopy)
   {
      cerr << "copying \"" << pszFileNameIn << "\" to \"" << stringFileNameOut << "\"" << endl;

      ifstream  ifIn  (pszFileNameIn, ios::in | ios::binary);
      ofstream  ofOut (stringFileNameOut.c_str (), ios::out | ios::binary);

      // copy the file in one line o' code!
      ofOut << ifIn.rdbuf ();

      if (fAutoconf_d)
      {
         ofstream ofstreamAM3 ("libraries3.mak", ios::app);

         ofstreamAM3 << "CLEANFILES += " << stringFileNameOut << endl;

         ofstreamAM3.close ();
      }
   }
}

DOM_Node
firstNode (DOM_Node elm)
{
   while (  elm != 0
         && (  DOM_Node::TEXT_NODE    == elm.getNodeType ()
            || DOM_Node::COMMENT_NODE == elm.getNodeType ()
            )
         )
   {
      // Move to the next node
      elm = elm.getNextSibling ();
   }

   return elm;
}

DOM_Node
nextNode (DOM_Node elm)
{
   do
   {
      // Move to the next node
      elm = elm.getNextSibling ();

   } while (  elm != 0
           && (  DOM_Node::TEXT_NODE    == elm.getNodeType ()
              || DOM_Node::COMMENT_NODE == elm.getNodeType ()
              )
           );

   return elm;
}

void
printNodeType (DOM_Node& node)
{
#ifdef DEBUG
   if (fDebugOutput)
   {
      switch (node.getNodeType ())
      {
      case DOM_Node::ELEMENT_NODE:                cerr << "<ELEMENT_NODE>"; break;
      case DOM_Node::ATTRIBUTE_NODE:              cerr << "<ATTRIBUTE_NODE>"; break;
      case DOM_Node::TEXT_NODE:                   cerr << "<TEXT_NODE>"; break;
      case DOM_Node::CDATA_SECTION_NODE:          cerr << "<CDATA_SECTION_NODE>"; break;
      case DOM_Node::ENTITY_REFERENCE_NODE:       cerr << "<ENTITY_REFERENCE_NODE>"; break;
      case DOM_Node::ENTITY_NODE:                 cerr << "<ENTITY_NODE>"; break;
      case DOM_Node::PROCESSING_INSTRUCTION_NODE: cerr << "<PROCESSING_INSTRUCTION_NODE>"; break;
      case DOM_Node::COMMENT_NODE:                cerr << "<COMMENT_NODE>"; break;
      case DOM_Node::DOCUMENT_NODE:               cerr << "<DOCUMENT_NODE>"; break;
      case DOM_Node::DOCUMENT_TYPE_NODE:          cerr << "<DOCUMENT_TYPE_NODE>"; break;
      case DOM_Node::DOCUMENT_FRAGMENT_NODE:      cerr << "<DOCUMENT_FRAGMENT_NODE>"; break;
      case DOM_Node::NOTATION_NODE:               cerr << "<NOTATION_NODE>"; break;
      default:                                    cerr << "<\?\?\?>"; break;
      }
   }
#endif
}

void
printDOMNode (DOM_Node& node, int iLevel)
{
#ifdef DEBUG
   if (fDebugOutput)
   {
      // Get the name and value out for convenience
      DOMString nodeName  = node.getNodeName ();
      DOMString nodeValue = node.getNodeValue ();

      if (0 <= iLevel)
         for (int i = 0; i < iLevel; i++)
            cerr << "   ";

      printNodeType (node);

      switch (node.getNodeType ())
      {
      case DOM_Node::TEXT_NODE:
      {
         const char *pszText = nodeValue.transcode ();

         while (*pszText)
         {
            switch (*pszText)
            {
            case 0x09: cerr << "\\t"; break;
            case 0x0a: cerr << "\\r"; break;
            case 0x0d: cerr << "\\n"; break;
            default:   cerr << *pszText; break;
            }
            pszText++;
         }
         break;
      }

      case DOM_Node::PROCESSING_INSTRUCTION_NODE:
      {
         cerr << "<?" << nodeName.transcode () << " " << nodeValue.transcode () << "?>" << endl;
         break;
      }

      case DOM_Node::DOCUMENT_NODE:
      {
         cerr << "<?xml version='1.0' encoding='ISO-8859-1' ?>" << endl;

         DOM_Node child = node.getFirstChild ();

         while (child != 0)
         {
            cerr << endl;
            printDOMNode (child, iLevel + 1);

            child = child.getNextSibling ();
         }
         break;
      }

      case DOM_Node::ELEMENT_NODE:
      {
         // Output the element start tag.
         cerr << '<' << nodeName.transcode ();

         // Output any attributes on this element
         DOM_NamedNodeMap attributes = node.getAttributes ();
         int attrCount = attributes.getLength ();
         for (int i = 0; i < attrCount; i++)
         {
            DOM_Node attribute = attributes.item (i);

            cerr << " " << attribute.getNodeName ().transcode ()
                 << " = \"";

            //  Note that "<" must be escaped in attribute values.
            cerr << attribute.getNodeValue ().transcode ();

            cerr << '"';
         }

         //
         //  Test for the presence of children, which includes both
         //  text content and nested elements.
         //
         DOM_Node child = node.getFirstChild ();
         if (child != 0)
         {
            // There are children. Close start-tag, and output children.
            cerr << ">";
            while (child != 0)
            {
               cerr << endl;
               printDOMNode (child, iLevel + 1);

               child = child.getNextSibling ();
            }

            // Done with children.  Output the end tag.
            cerr << "</" << nodeName.transcode () << ">";
         }
         else
         {
            //
            //  There were no children.  Output the short form close of the
            //  element start tag, making it an empty-element tag.
            //
            cerr << "/>";
         }
         break;
      }

      case DOM_Node::ENTITY_REFERENCE_NODE:
      {
         DOM_Node child;
         for (child = node.getFirstChild (); child != 0; child = child.getNextSibling ())
         {
            cerr << endl;
            printDOMNode (child, iLevel + 1);
         }
         break;
      }

      case DOM_Node::CDATA_SECTION_NODE:
      {
         cerr << "<![CDATA[" << nodeValue.transcode () << "]]>";
         break;
      }

      case DOM_Node::COMMENT_NODE:
      {
         cerr << "<!--" << nodeValue.transcode () << "-->";
         break;
      }

      default:
      {
         break;
      }
      }
   }
#endif
}

void
skipTextOrCommentNodes (DOM_Node& node)
{
   while (  node != 0
         && (  DOM_Node::TEXT_NODE    == node.getNodeType ()
            || DOM_Node::COMMENT_NODE == node.getNodeType ()
            )
         )
   {
      node = node.getNextSibling ();
   }
}

bool OmniDomParser::
openOutputFiles (string    *pstringName,
                 ofstream **ppofstreamHPP,
                 ofstream **ppofstreamCPP)
{
   char     achFilename[512]; //@TBD

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".hpp");
   convertFilename (achFilename);

   *ppofstreamHPP = new ofstream (achFilename);

   if (fAutoconf_d)
   {
      ofstream ofstreamAM3 ("libraries3.mak", ios::app);

      ofstreamAM3 << "CLEANFILES += " << achFilename << endl;

      ofstreamAM3.close ();
   }

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".cpp");
   convertFilename (achFilename);

   *ppofstreamCPP = new ofstream (achFilename);

   if (fAutoconf_d)
   {
      ofstream ofstreamAM3 ("libraries3.mak", ios::app);

      ofstreamAM3 << "CLEANFILES += " << achFilename << endl;

      ofstreamAM3.close ();
   }

   return true;
}

bool
openMakeFile (string    *pstringName,
              ofstream **ppofstreamMAK)
{
   char achFilename[512]; //@TBD

   strcpy (achFilename, pstringName->c_str ());
   strcat (achFilename, ".mak");
   convertFilename (achFilename);

   *ppofstreamMAK = new ofstream (achFilename);

   return true;
}

void
outputHeader (ofstream *pofstream)
{
   *pofstream << "/*" << endl;
   *pofstream << " *   IBM Omni driver" << endl;
   *pofstream << " *   Copyright (c) International Business Machines Corp., 2000" << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   This library is free software; you can redistribute it and/or modify" << endl;
   *pofstream << " *   it under the terms of the GNU Lesser General Public License as published" << endl;
   *pofstream << " *   by the Free Software Foundation; either version 2.1 of the License, or" << endl;
   *pofstream << " *   (at your option) any later version." << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   This library is distributed in the hope that it will be useful," << endl;
   *pofstream << " *   but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl;
   *pofstream << " *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See" << endl;
   *pofstream << " *   the GNU Lesser General Public License for more details." << endl;
   *pofstream << " *" << endl;
   *pofstream << " *   You should have received a copy of the GNU Lesser General Public License" << endl;
   *pofstream << " *   along with this library; if not, write to the Free Software" << endl;
   *pofstream << " *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA" << endl;
   *pofstream << " */" << endl;
}

bool
getNameValue (DOM_Node    &node,
              const char **ppszName,
              const char **ppszValue)
{
   *ppszName = node.getNodeName ().transcode ();

   DOM_Node child = node.getFirstChild ();

   if (child == 0)
      return false;

   *ppszValue = child.getNodeValue ().transcode ();

   return true;
}

bool
getName (DOM_Node    &node,
         const char **ppszName)
{
   *ppszName = node.getNodeName ().transcode ();

   return true;
}

bool
getValue (DOM_Node    &node,
          const char **ppszValue)
{
   DOM_Node child = node.getFirstChild ();

   if (child == 0)
      return false;

   *ppszValue = child.getNodeValue ().transcode ();

   return true;
}

bool
getValue (DOM_Node    &node,
          int         *piValue)
{
   DOM_Node    child    = node.getFirstChild ();
   const char *pszValue;

   if (child == 0)
      return false;

   pszValue = child.getNodeValue ().transcode ();

   return 1 == sscanf (pszValue, "%d", piValue);
}

bool
parseHexGroup (const char   *pszData,
               unsigned int *puiData)
{
   if (  isxdigit (pszData[0])
      && isxdigit (pszData[1])
      )
   {
      return 1 == sscanf (pszData, "%x", puiData);
   }
   else
   {
      return false;
   }
}

bool OmniDomParser::
parseBinaryData (const char  *pszData,
                 byte       **ppbData,
                 int         *pcbData)
{
   int   cbDataAlloc    = 64;
   int   cbDataLeft     = 64;
   byte *pbData         = new byte [cbDataAlloc];
   byte *pbDataCurrent  = pbData;
   bool  fContinue      = true;
   bool  fInString      = false;
   byte  bChar          = 0;
   byte  bNextChar      = 0;
   bool  fCharFound     = false;
   bool  fNextCharFound = false;

   *ppbData = pbData;
   *pcbData = cbDataAlloc - cbDataLeft;

   while (  *pszData
         && fContinue
         )
   {
      if (fNextCharFound)
      {
         bChar          = bNextChar;
         fNextCharFound = false;
         fCharFound     = true;
      }
      else
      {
         if (fInString)
         {
            switch (*pszData)
            {
            case '\\':
            {
               break;
            }

            case '"':
            {
               pszData++;

               fInString = false;
               continue;
            }

            default:
            {
               bChar      = *pszData++;
               fCharFound = true;
               break;
            }
            }
         }
         else
         {
            while (  '\x0a' == *pszData
                  || '\x0d' == *pszData
                  || ' '    == *pszData
                  )
               pszData++;

            switch (*pszData)
            {
            case '_':
            {
               pszData++;

               if (0 == strncmp (pszData, "ESC_", 4))
               {
                  bChar      = 0x1b;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "NUL_", 4))
               {
                  bChar      = 0;
                  fCharFound = true;
                  pszData += 4;
               }
               else if (0 == strncmp (pszData, "LF_", 3))
               {
                  bChar      = 0x0a;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "CR_", 3))
               {
                  bChar      = 0x0d;
                  fCharFound = true;
                  pszData += 3;
               }
               else if (0 == strncmp (pszData, "FF_", 3))
               {
                  bChar      = 0x0c;
                  fCharFound = true;
                  pszData += 3;
               }
               else
               {
                  cerr << "Error: Unknown define " << pszData << endl;
                  setErrorCondition ();
                  fContinue = false;
               }
               break;
            }

            case 'H':
            {
               pszData++;

               if (0 == strncmp (pszData, "EX2S", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData  = 0;
                     unsigned int uiData2 = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (',' == *pszData)
                        {
                           pszData++;

                           while (' ' == *pszData)
                              pszData++;

                           if (parseHexGroup (pszData, &uiData2))
                           {
                              pszData += 2;

                              while (' ' == *pszData)
                                 pszData++;

                              if (')' == *pszData)
                              {
                                 pszData++;

                                 bChar          = (byte)uiData2;
                                 bNextChar      = (byte)uiData;
                                 fCharFound     = true;
                                 fNextCharFound = true;
                              }
                              else
                              {
                                 cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                                 setErrorCondition ();
                                 fContinue = false;
                              }
                           }
                           else
                           {
                              cerr << "Error: Expecting ',', found '" << *pszData << "' @ " << __LINE__ << endl;
                              setErrorCondition ();
                              fContinue = false;
                           }
                        }
                        else
                        {
                           cerr << "Error: Expecting ',', found '" << *pszData << "' @ " << __LINE__ << endl;
                           setErrorCondition ();
                           fContinue = false;
                        }
                     }
                     else
                     {
                        cerr << "Error: Expecting 2 hex digits, found '" << pszData[0] << pszData[1] << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               else if (0 == strncmp (pszData, "EX", 2))
               {
                  pszData += 2;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     unsigned int uiData = 0;

                     if (parseHexGroup (pszData, &uiData))
                     {
                        pszData += 2;

                        while (' ' == *pszData)
                           pszData++;

                        if (')' == *pszData)
                        {
                           pszData++;

                           bChar      = (byte)uiData;
                           fCharFound = true;
                        }
                        else
                        {
                           cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                           setErrorCondition ();
                           fContinue = false;
                        }
                     }
                     else
                     {
                        cerr << "Error: Expecting 2 hex digits, found '" << pszData[0] << pszData[1] << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               break;
            }

            case 'A':
            {
               pszData++;

               if (0 == strncmp (pszData, "SCII", 4))
               {
                  pszData += 4;

                  while (' ' == *pszData)
                     pszData++;

                  if ('(' == *pszData)
                  {
                     pszData++;
                     while (' ' == *pszData)
                        pszData++;

                     bChar      = (byte)*pszData++;
                     fCharFound = true;

                     if (')' == *pszData)
                     {
                        pszData++;
                     }
                     else
                     {
                        cerr << "Error: Expecting ')', found '" << *pszData << "' @ " << __LINE__ << endl;
                        setErrorCondition ();
                        fContinue = false;
                     }
                  }
                  else
                  {
                     cerr << "Error: Expecting '(', found '" << *pszData << "' @ " << __LINE__ << endl;
                     setErrorCondition ();
                     fContinue = false;
                  }
               }
               break;
            }

            case '"':
            {
               pszData++;
               fInString = true;
               break;
            }
            }
         }
      }

      if (fCharFound)
      {
/////////if (isprint (bChar))
/////////   cerr << "bChar = '" << bChar << "'" << endl;
/////////else
/////////   cerr << "bChar = 0x" << hex << (int)bChar << endl;

         *pbDataCurrent++ = bChar;
         cbDataLeft--;

         bChar      = 0;
         fCharFound = false;

         if (0 == cbDataLeft)
         {
            byte *pbTemp = new byte [cbDataAlloc + 64];

            if (pbTemp)
            {
               memcpy (pbTemp, pbData, cbDataAlloc);

               delete[] pbData;

               pbData        = pbTemp;
               pbDataCurrent = pbData + cbDataAlloc;
               cbDataAlloc  += 64;
               cbDataLeft    = 64;
            }
            else
            {
               cerr << "Error: Allocation of " << (cbDataAlloc + 64) << " bytes failed @ " << __LINE__ << endl;
               setErrorCondition ();
               fContinue = false;
            }
         }
      }
      else if (  !fInString
              && *pszData
              )
      {
         delete[] pbData;
         return false;
      }
   }

   if (fContinue)
   {
      *ppbData = pbData;
      *pcbData = cbDataAlloc - cbDataLeft;
   }
   else
   {
      delete[] pbData;
   }

   return fContinue;
}

bool OmniDomParser::
processDeviceOrientations (DOM_Node deviceOrientationsNode)
{
   if (deviceOrientationsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringOrientationClassName = pstringNameOut;

   typedef map <string, string *> OrientationMapElement;
   typedef map <string, string>   DeviceIDList;

   DOM_Node               deviceOrientationNode = firstNode (deviceOrientationsNode.getFirstChild ());
   ofstream              *pofstreamHPP          = 0;
   ofstream              *pofstreamCPP          = 0;
   OrientationMapElement  mapElement;
   DeviceIDList           mapDeviceID;
   int                    rc                    = false;

   if (deviceOrientationNode == 0)
      goto done;

   while (deviceOrientationNode != 0)
   {
      DOM_Node    nodeItem = firstNode (deviceOrientationNode.getFirstChild ());
      const char *pszId    = 0;
      int         id       = -1;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DeviceOrientation::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceOrientation \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listOrientations.push_back (string (pszId));

      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << "); }";

      mapElement[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput) cerr << dec << "id              = " << id << endl;
#endif

      // Move to the next deviceOrientation node
      deviceOrientationNode = nextNode (deviceOrientationNode);
   }

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << "#include <DeviceOrientation.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceOrientation" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int          id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceOrientation  *create                 (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DeviceOrientation  *create                 (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char               *getDeviceID            ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool                isOrientationSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration        *getEnumeration         ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceOrientation  *createV                (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                     int id)" << endl;
      *pofstreamCPP << "      : DeviceOrientation (pDevice," << endl;
      *pofstreamCPP << "                           id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DeviceOrientation::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( OrientationMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedOrientations[] = {" << endl;

      int iCount = mapElement.size ();
      for ( OrientationMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceOrientation::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isOrientationSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedOrientations); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedOrientations[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class OrientationEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      OrientationEnumerator (Device *pDevice, int cOrientations, int aOrientations[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d       = pDevice;" << endl;
      *pofstreamCPP << "         iOrientation_d  = 0;" << endl;
      *pofstreamCPP << "         cOrientations_d = cOrientations;" << endl;
      *pofstreamCPP << "         aOrientations_d = aOrientations;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iOrientation_d < cOrientations_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iOrientation_d > cOrientations_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aOrientations_d[iOrientation_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iOrientation_d;" << endl;
      *pofstreamCPP << "      int     cOrientations_d;" << endl;
      *pofstreamCPP << "      int    *aOrientations_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new OrientationEnumerator (pDevice_d, dimof (aSupportedOrientations), aSupportedOrientations);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceOrientationsNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringOrientationClassName = 0;
   }

   for ( OrientationMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceCommands (DOM_Node deviceCommandsNode)
{
   if (deviceCommandsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringCommandClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, BinaryData *> CommandMapElement;
   typedef map <string, byte *>       CommandMapData;

   DOM_Node          deviceCommandNode = firstNode (deviceCommandsNode.getFirstChild ());
   ofstream         *pofstreamHPP      = 0;
   ofstream         *pofstreamCPP      = 0;
   CommandMapElement mapElement;
   CommandMapData    mapData;
   int               rc                = false;

   if (deviceCommandNode == 0)
      goto done;

   while (deviceCommandNode != 0)
   {
      DOM_NamedNodeMap commandAttributes = deviceCommandNode.getAttributes ();

      if (commandAttributes == 0)
         goto done;

      DOMString nameString ("name");
      DOM_Node  name       = commandAttributes.getNamedItem (nameString);

      if (name == 0)
         goto done;

      const char *pszCmdName  = 0;
      const char *pszCmdValue = 0;

      getValue (deviceCommandNode, &pszCmdValue);
      getValue (name,        &pszCmdName);

#ifdef DEBUG
      if (fDebugOutput) cerr << pszCmdName << " = " << pszCmdValue << endl;
#endif

      byte *pbData = 0;
      int   cbData = 0;

      if (parseBinaryData (pszCmdValue, &pbData, &cbData))
      {
         mapData[string (pszCmdName)]    = pbData;
         mapElement[string (pszCmdName)] = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszCmdValue << "\" for " << pszCmdName << "." << endl;
         goto done;
      }

      // Move to the next one
      deviceCommandNode = nextNode (deviceCommandNode);
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   fFileModified_d = true;

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <DeviceCommand.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceCommand" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << "()" << endl;
   *pofstreamCPP << "{" << endl;

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      BinaryData *bdValue = (BinaryData *)next->second;
      PBYTE       pbValue = bdValue->getData ();
      int         cbValue = bdValue->getLength ();

      *pofstreamCPP << "   static byte " << next->first << "[] = {" << dec;

      for (int i = 0; i < cbValue; i++)
      {
         *pofstreamCPP << (int)pbValue[i];
         if (i < cbValue - 1)
            *pofstreamCPP << ",";
      }

      *pofstreamCPP << "};" << endl;

#ifdef DEBUG
//////if (fDebugOutput) cerr << next->first << " = " << *(next->second) << endl;
#endif
   }

   *pofstreamCPP << endl;

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      *pofstreamCPP << "   add (\""
                    << next->first
                    << "\", new BinaryData ("
                    << next->first
                    << ", dimof ("
                    << next->first << ")));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceCommandsNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringCommandClassName = 0;
   }

   for ( CommandMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( CommandMapData::iterator next = mapData.begin () ;
         next != mapData.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceResolutions (DOM_Node deviceResolutionsNode)
{
   if (deviceResolutionsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringResolutionClassName = pstringNameOut;

   typedef map <string, string *>     ResolutionMapElement;
   typedef map <string, BinaryData *> ResolutionMapData1;
   typedef map <string, byte *>       ResolutionMapData2;
   typedef map <string, string>       DeviceIDList;

   DOM_Node             deviceResolutionNode = firstNode (deviceResolutionsNode.getFirstChild ());
   ofstream            *pofstreamHPP         = 0;
   ofstream            *pofstreamCPP         = 0;
   ResolutionMapElement mapResolution;
   ResolutionMapData1   mapData1;
   ResolutionMapData2   mapData2;
   DeviceIDList         mapDeviceID;
   int                  iCount               = 0;
   int                  rc                   = false;

   if (deviceResolutionNode == 0)
      goto done;

   while (deviceResolutionNode != 0)
   {
      DOM_Node    nodeItem               = firstNode (deviceResolutionNode.getFirstChild ());
      const char *pszId                  = 0;
      int         id                     = -1;
      int         iXRes;
      int         iYRes;
      int         iXInternalRes          = 0;
      int         iYInternalRes          = 0;
      const char *pszBinaryData          = 0;
      byte       *pbData                 = 0;
      int         cbData                 = 0;
      BinaryData *pdbData;
      int         iCapabilities;
      int         iDestinationBitsPerPel;
      int         iScanlineMultiple;
      const char *pszNodeName            = 0;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DeviceResolution::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceResolution \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listResolutions.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iXRes);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iYRes);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);
      if (0 == strcmp (pszNodeName, "xInternalRes"))
      {
         getValue (nodeItem, &iXInternalRes);

         // Move to the next one
         nodeItem = nextNode (nodeItem);
         if (nodeItem == 0)
            goto done;

         getValue (nodeItem, &iYInternalRes);

         // Move to the next one
         nodeItem = nextNode (nodeItem);
         if (nodeItem == 0)
            goto done;
      }

      getValue (nodeItem, &pszBinaryData);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iCapabilities);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iDestinationBitsPerPel);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iScanlineMultiple);

      // Move to the next one
      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << ","
          << iXRes
          << ","
          << iYRes
          << ","
          << iXInternalRes
          << ","
          << iYInternalRes
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),"
          << iCapabilities
          << ","
          << iDestinationBitsPerPel
          << ","
          << iScanlineMultiple
          << "); }";

      mapData2[string (pszId)]      = pbData;
      mapData1[string (pszId)]      = new BinaryData (pbData, cbData);
      mapResolution[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "id                     = " << id                     << endl;
         cerr        << "iXRes                  = " << iXRes                  << endl;
         cerr        << "iYRes                  = " << iYRes                  << endl;
         cerr        << "iXInternalRes          = " << iXInternalRes          << endl;
         cerr        << "iYInternalRes          = " << iYInternalRes          << endl;
         cerr        << "pszBinaryData          = " << pszBinaryData          << endl;
         cerr        << "pdbData                = " << *pdbData               << endl;
         cerr << dec << "iCapabilities          = " << iCapabilities          << endl;
         cerr        << "iDestinationBitsPerPel = " << iDestinationBitsPerPel << endl;
         cerr        << "iScanlineMultiple      = " << iScanlineMultiple      << endl;
      }
#endif

      // Move to the next deviceResolution node
      deviceResolutionNode = nextNode (deviceResolutionNode);
   }

   pdi_d->listResolutions.sort ();
   pdi_d->listResolutions.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceResolution.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceResolution" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int          id," << endl;
      *pofstreamHPP << "                          int         iXRes," << endl;
      *pofstreamHPP << "                          int         iYRes," << endl;
      *pofstreamHPP << "                          int         iXInternalRes," << endl;
      *pofstreamHPP << "                          int         iYInternalRes," << endl;
      *pofstreamHPP << "                          BinaryData *data," << endl;
      *pofstreamHPP << "                          int         iCapabilities," << endl;
      *pofstreamHPP << "                          int         iDestinationBitsPerPel," << endl;
      *pofstreamHPP << "                          int         iScanlineMultiple);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceResolution *create                (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DeviceResolution *create                (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char             *getDeviceID           ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool              isResolutionSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration      *getEnumeration        ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceResolution *createV               (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                          int         id," << endl;
      *pofstreamCPP << "                          int         iXRes," << endl;
      *pofstreamCPP << "                          int         iYRes," << endl;
      *pofstreamCPP << "                          int         iXInternalRes," << endl;
      *pofstreamCPP << "                          int         iYInternalRes," << endl;
      *pofstreamCPP << "                          BinaryData *data," << endl;
      *pofstreamCPP << "                          int         iCapabilities," << endl;
      *pofstreamCPP << "                          int         iDestinationBitsPerPel," << endl;
      *pofstreamCPP << "                          int         iScanlineMultiple)" << endl;
      *pofstreamCPP << "      : DeviceResolution (pDevice," << endl;
      *pofstreamCPP << "                          id," << endl;
      *pofstreamCPP << "                          iXRes," << endl;
      *pofstreamCPP << "                          iYRes," << endl;
      *pofstreamCPP << "                          iXInternalRes," << endl;
      *pofstreamCPP << "                          iYInternalRes," << endl;
      *pofstreamCPP << "                          data," << endl;
      *pofstreamCPP << "                          iCapabilities," << endl;
      *pofstreamCPP << "                          iDestinationBitsPerPel," << endl;
      *pofstreamCPP << "                          iScanlineMultiple)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DeviceResolution::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( ResolutionMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
            next != mapResolution.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedResolutions[] = {" << endl;

      iCount = mapResolution.size ();
      for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
            next != mapResolution.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceResolution::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isResolutionSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedResolutions); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedResolutions[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class ResolutionEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      ResolutionEnumerator (Device *pDevice, int cResolutions, int aResolutions[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d      = pDevice;" << endl;
      *pofstreamCPP << "         iResolution_d  = 0;" << endl;
      *pofstreamCPP << "         cResolutions_d = cResolutions;" << endl;
      *pofstreamCPP << "         aResolutions_d = aResolutions;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iResolution_d < cResolutions_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iResolution_d > cResolutions_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aResolutions_d[iResolution_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iResolution_d;" << endl;
      *pofstreamCPP << "      int     cResolutions_d;" << endl;
      *pofstreamCPP << "      int    *aResolutions_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new ResolutionEnumerator (pDevice_d, dimof (aSupportedResolutions), aSupportedResolutions);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceResolutionsNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringResolutionClassName = 0;
   }

   for ( ResolutionMapElement::iterator next = mapResolution.begin () ;
         next != mapResolution.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( ResolutionMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( ResolutionMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDevicePrintModes (DOM_Node devicePrintModesNode)
{
   if (devicePrintModesNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringPrintModeClassName = pstringNameOut;

   typedef map <string, string *> PrintModeMapElement;
   typedef map <string, string>   DeviceIDList;

   DOM_Node             devicePrintModeNode = firstNode (devicePrintModesNode.getFirstChild ());
   ofstream            *pofstreamHPP        = 0;
   ofstream            *pofstreamCPP        = 0;
   PrintModeMapElement  mapElement;
   DeviceIDList         mapDeviceID;
   int                  iCount              = 0;
   int                  rc                  = false;

   if (devicePrintModeNode == 0)
      goto done;

   while (devicePrintModeNode != 0)
   {
      DOM_Node    nodeItem       = firstNode (devicePrintModeNode.getFirstChild ());
      const char *pszId          = 0;
      int         id             = -1;
      int         iPhysicalCount;
      int         iLogicalCount;
      int         iPlanes;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DevicePrintMode::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DevicePrintMode \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listPrintModes.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iPhysicalCount);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iLogicalCount);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iPlanes);

      // Move to the next one
      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << ","
          << iPhysicalCount
          << ","
          << iLogicalCount
          << ","
          << iPlanes
          << "); }";

      mapElement[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "id             = " << id             << endl;
         cerr        << "iPhysicalCount = " << iPhysicalCount << endl;
         cerr        << "iLogicalCount  = " << iLogicalCount  << endl;
         cerr        << "iPlanes        = " << iPlanes        << endl;
      }
#endif

      // Get the next devicePrintMode node
      devicePrintModeNode = nextNode (devicePrintModeNode);
   }

   pdi_d->listPrintModes.sort ();
   pdi_d->listPrintModes.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DevicePrintMode.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DevicePrintMode" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;
      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int         id," << endl;
      *pofstreamHPP << "                          int         iPhysicalCount," << endl;
      *pofstreamHPP << "                          int         iLogicalCount," << endl;
      *pofstreamHPP << "                          int         iPlanes);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DevicePrintMode *create               (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DevicePrintMode *create               (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char            *getDeviceID          ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool             isPrintModeSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration     *getEnumeration       ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DevicePrintMode *createV              (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                          int         id," << endl;
      *pofstreamCPP << "                          int         iPhysicalCount," << endl;
      *pofstreamCPP << "                          int         iLogicalCount," << endl;
      *pofstreamCPP << "                          int         iPlanes)" << endl;
      *pofstreamCPP << "      : DevicePrintMode (pDevice," << endl;
      *pofstreamCPP << "                         id," << endl;
      *pofstreamCPP << "                         iPhysicalCount," << endl;
      *pofstreamCPP << "                         iLogicalCount," << endl;
      *pofstreamCPP << "                         iPlanes)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DevicePrintMode::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( PrintModeMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "static int aSupportedPrintModes[] = {" << endl;

      iCount = mapElement.size ();
      for ( PrintModeMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DevicePrintMode::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isPrintModeSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedPrintModes); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedPrintModes[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class PrintModeEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      PrintModeEnumerator (Device *pDevice, int cPrintModes, int aPrintModes[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d     = pDevice;" << endl;
      *pofstreamCPP << "         iPrintMode_d  = 0;" << endl;
      *pofstreamCPP << "         cPrintModes_d = cPrintModes;" << endl;
      *pofstreamCPP << "         aPrintModes_d = aPrintModes;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iPrintMode_d < cPrintModes_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iPrintMode_d > cPrintModes_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aPrintModes_d[iPrintMode_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iPrintMode_d;" << endl;
      *pofstreamCPP << "      int     cPrintModes_d;" << endl;
      *pofstreamCPP << "      int    *aPrintModes_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new PrintModeEnumerator (pDevice_d, dimof (aSupportedPrintModes), aSupportedPrintModes);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (devicePrintModesNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringPrintModeClassName = 0;
   }

   for ( PrintModeMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceTrays (DOM_Node deviceTraysNode)
{
   if (deviceTraysNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringTrayClassName = pstringNameOut;

   typedef map <string, string *>     TrayMapElement;
   typedef map <string, BinaryData *> TrayMapData1;
   typedef map <string, byte *>       TrayMapData2;
   typedef map <string, string>       DeviceIDList;

   DOM_Node        deviceTrayNode = firstNode (deviceTraysNode.getFirstChild ());
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   TrayMapElement  mapElement;
   TrayMapData1    mapData1;
   TrayMapData2    mapData2;
   TrayList        listType;
   DeviceIDList    mapDeviceID;
   int             iCount         = 0;
   int             rc             = false;

   if (deviceTrayNode == 0)
      goto done;

   while (deviceTrayNode != 0)
   {
      DOM_Node    nodeItem      = firstNode (deviceTrayNode.getFirstChild ());
      const char *pszId         = 0;
      int         id            = -1;
      const char *pszType;
      const char *pszBinaryData = 0;
      byte       *pbData        = 0;
      int         cbData        = 0;
      BinaryData *pdbData;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DeviceTray::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceTray \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listTrays.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszType);

      listType.push_back (string (pszType));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszBinaryData);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << ","
          << pszType
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << "))); }";

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "id            = " << id            << endl;
         cerr        << "pszType       = " << pszType       << endl;
         cerr        << "pszBinaryData = " << pszBinaryData << endl;
         cerr        << "pdbData       = " << *pdbData      << endl;
      }
#endif

      // Get the next deviceTray node
      deviceTrayNode = nextNode (deviceTrayNode);
   }

   pdi_d->listTrays.sort ();
   pdi_d->listTrays.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceTray.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceTray" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << "   enum {" << endl;
      listType.sort ();
      listType.unique ();
      iCount = listType.size ();
      for ( TrayList::iterator next = listType.begin () ;
            next != listType.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   };" << endl;

      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int         id," << endl;
      *pofstreamHPP << "                          int         iType," << endl;
      *pofstreamHPP << "                          BinaryData *data);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceTray  *create          (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DeviceTray  *create          (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char        *getDeviceID     ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool         isTraySupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration  ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceTray  *createV         (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                          int         id," << endl;
      *pofstreamCPP << "                          int         iType," << endl;
      *pofstreamCPP << "                          BinaryData *data)" << endl;
      *pofstreamCPP << "      : DeviceTray (pDevice," << endl;
      *pofstreamCPP << "                    id," << endl;
      *pofstreamCPP << "                    iType," << endl;
      *pofstreamCPP << "                    data)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DeviceTray::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( TrayMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( TrayMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << "" << endl;
      *pofstreamCPP << "static int aSupportedTrays[] = {" << endl;

      iCount = mapElement.size ();
      for ( TrayMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceTray::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "   };" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isTraySupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedTrays); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedTrays[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class TrayEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      TrayEnumerator (Device *pDevice, int cTrays, int aTrays[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d = pDevice;" << endl;
      *pofstreamCPP << "         iTray_d   = 0;" << endl;
      *pofstreamCPP << "         cTrays_d  = cTrays;" << endl;
      *pofstreamCPP << "         aTrays_d  = aTrays;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iTray_d < cTrays_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iTray_d > cTrays_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aTrays_d[iTray_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iTray_d;" << endl;
      *pofstreamCPP << "      int     cTrays_d;" << endl;
      *pofstreamCPP << "      int    *aTrays_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new TrayEnumerator (pDevice_d, dimof (aSupportedTrays), aSupportedTrays);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceTraysNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringTrayClassName = 0;
   }

   for ( TrayMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( TrayMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( TrayMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceForms (DOM_Node deviceFormsNode)
{
   if (deviceFormsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringFormClassName = pstringNameOut;

   typedef map <string, string *>     FormMapElement;
   typedef map <string, BinaryData *> FormMapData1;
   typedef map <string, byte *>       FormMapData2;
   typedef list <string>              FormList;
   typedef map <string, string>       DeviceIDList;

   DOM_Node        deviceFormNode = firstNode (deviceFormsNode.getFirstChild ());
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   FormMapElement  mapElement;
   FormMapData1    mapData1;
   FormMapData2    mapData2;
   FormList        listType;
   DeviceIDList    mapDeviceID;
   int             iCount         = 0;
   int             rc             = false;

   if (deviceFormNode == 0)
      goto done;

   while (deviceFormNode != 0)
   {
      DOM_Node    nodeItem        = firstNode (deviceFormNode.getFirstChild ());
      DOM_Node    nodeHCC;
      const char *pszId           = 0;
      int         id              = -1;
      const char *pszCapabilities = 0;
      const char *pszBinaryData   = 0;
      byte       *pbData          = 0;
      int         cbData          = 0;
      BinaryData *pdbData;
      int         iLeft           = 0;
      int         iTop            = 0;
      int         iRight          = 0;
      int         iBottom         = 0;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DeviceForm::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceForm \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listForms.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszCapabilities);

      if (!DeviceForm::isReservedKeyword (pszCapabilities))
      {
         listType.push_back (string (pszCapabilities));
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszBinaryData);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // Move down into the HardCopyCap tree
      nodeHCC = firstNode (nodeItem.getFirstChild ());
      if (nodeHCC == 0)
         goto done;

      getValue (nodeHCC, &iLeft);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      getValue (nodeHCC, &iTop);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      getValue (nodeHCC, &iRight);

      // Move to the next one
      nodeHCC = nextNode (nodeHCC);
      if (nodeHCC == 0)
         goto done;

      getValue (nodeHCC, &iBottom);

      // Move to the next one
      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << ","
          << pszCapabilities
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),new HardCopyCap ("
          << iLeft
          << ","
          << iTop
          << ","
          << iRight
          << ","
          << iBottom
          << ")); }";

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "id              = " << id              << endl;
         cerr        << "pszCapabilities = " << pszCapabilities << endl;
         cerr        << "pszBinaryData   = " << pszBinaryData   << endl;
         cerr        << "pdbData         = " << *pdbData        << endl;
         cerr << dec << "   iLeft        = " << iLeft           << endl;
         cerr        << "   iTop         = " << iTop            << endl;
         cerr        << "   iRight       = " << iRight          << endl;
         cerr        << "   iBottom      = " << iBottom         << endl;
      }
#endif

      // Get the next deviceForm node
      deviceFormNode = nextNode (deviceFormNode);
   }

   pdi_d->listForms.sort ();
   pdi_d->listForms.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceForm.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceForm" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << "   enum {" << endl;
      listType.sort ();
      listType.unique ();
      iCount = listType.size ();
      for ( FormList::iterator next = listType.begin () ;
            next != listType.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   };" << endl;
      *pofstreamHPP << endl;

      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int          id," << endl;
      *pofstreamHPP << "                          int          iCapabilities," << endl;
      *pofstreamHPP << "                          BinaryData  *data," << endl;
      *pofstreamHPP << "                          HardCopyCap *hcInfo);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceForm  *create          (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DeviceForm  *create          (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char        *getDeviceID     ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool         isFormSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration  ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceForm  *createV         (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                          int          id," << endl;
      *pofstreamCPP << "                          int          iCapabilities," << endl;
      *pofstreamCPP << "                          BinaryData  *data," << endl;
      *pofstreamCPP << "                          HardCopyCap *hcInfo)" << endl;
      *pofstreamCPP << "      : DeviceForm (pDevice," << endl;
      *pofstreamCPP << "                    id," << endl;
      *pofstreamCPP << "                    iCapabilities," << endl;
      *pofstreamCPP << "                    data," << endl;
      *pofstreamCPP << "                    hcInfo)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DeviceForm::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( FormMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( FormMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "static int aSupportedForms[] = {" << endl;

      iCount = mapElement.size ();
      for ( FormMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceForm::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isFormSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedForms); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedForms[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class FormEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      FormEnumerator (Device *pDevice, int cForms, int aForms[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d = pDevice;" << endl;
      *pofstreamCPP << "         iForm_d   = 0;" << endl;
      *pofstreamCPP << "         cForms_d  = cForms;" << endl;
      *pofstreamCPP << "         aForms_d  = aForms;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iForm_d < cForms_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iForm_d > cForms_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aForms_d[iForm_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iForm_d;" << endl;
      *pofstreamCPP << "      int     cForms_d;" << endl;
      *pofstreamCPP << "      int    *aForms_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new FormEnumerator (pDevice_d, dimof (aSupportedForms), aSupportedForms);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceFormsNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringFormClassName = 0;
   }

   for ( FormMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( FormMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( FormMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceMedias (DOM_Node deviceMediasNode)
{
   if (deviceMediasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringMediaClassName = pstringNameOut;

   typedef map <string, string *>     MediaMapElement;
   typedef map <string, BinaryData *> MediaMapData1;
   typedef map <string, byte *>       MediaMapData2;
   typedef map <string, string>       DeviceIDList;

   DOM_Node        deviceMediaNode = firstNode (deviceMediasNode.getFirstChild ());
   ofstream       *pofstreamHPP    = 0;
   ofstream       *pofstreamCPP    = 0;
   MediaMapElement mapElement;
   MediaMapData1   mapData1;
   MediaMapData2   mapData2;
   DeviceIDList    mapDeviceID;
   int             iCount          = 0;
   int             rc              = false;

   if (deviceMediaNode == 0)
      goto done;

   while (deviceMediaNode != 0)
   {
      DOM_Node    nodeItem             = firstNode (deviceMediaNode.getFirstChild ());
      const char *pszId                = 0;
      int         id                   = -1;
      const char *pszBinaryData        = 0;
      byte       *pbData               = 0;
      int         cbData               = 0;
      BinaryData *pdbData;
      int         iColorAdjustRequired;
      const char *pszAbsorption        = 0;

      if (nodeItem == 0)
         goto done;

      if (getValue (nodeItem, &pszId))
      {
         id = DeviceMedia::nameToID (const_cast<char*>(pszId));
         if (-1 == id)
         {
            cerr << "Error: Unknown DeviceMedia \"" << pszId << "\"" << endl;
            setErrorCondition ();
            goto done;
         }
      }

      pdi_d->listMedias.push_back (string (pszId));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszBinaryData);

      if (parseBinaryData (pszBinaryData, &pbData, &cbData))
      {
         pdbData = new BinaryData (pbData, cbData);
      }
      else
      {
         cerr << "Error: binary data not understood \"" << pszBinaryData << "\" for " << pszId << "." << endl;
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iColorAdjustRequired);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszAbsorption);

      if (  0 != strcmp (pszAbsorption, "MEDIA_NO_ABSORPTION")
         && 0 != strcmp (pszAbsorption, "MEDIA_LIGHT_ABSORPTION")
         && 0 != strcmp (pszAbsorption, "MEDIA_HEAVY_ABSORPTION")
         )
         goto done;

      // Move to the next one
      nodeItem = nextNode (nodeItem);

      if (nodeItem != 0)
      {
         const char *pszDeviceID = 0;

         if (getValue (nodeItem, &pszDeviceID))
         {
            mapDeviceID[string (pszId)] = string (pszDeviceID);
         }
      }

      ostringstream oss;

      oss << "   case "
          << pszId
          << ": { return new "
          << *pstringNameOut
          << " (pDevice, "
          << pszId
          << ",new BinaryData (abData"
          << pszId
          << ",dimof (abData"
          << pszId
          << ")),"
          << iColorAdjustRequired
          << ","
          << pszAbsorption
          << "); }";

      mapData2[string (pszId)]   = pbData;
      mapData1[string (pszId)]   = new BinaryData (pbData, cbData);
      mapElement[string (pszId)] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "id                   = " << id                   << endl;
         cerr        << "pszBinaryData        = " << pszBinaryData        << endl;
         cerr        << "pdbData              = " << *pdbData             << endl;
         cerr        << "iColorAdjustRequired = " << iColorAdjustRequired << endl;
         cerr        << "pszAbsorption        = " << pszAbsorption        << endl;
      }
#endif

      // Get the next deviceMedia node
      deviceMediaNode = nextNode (deviceMediaNode);
   }

   pdi_d->listMedias.sort ();
   pdi_d->listMedias.unique ();

   rc = true;

   if (shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

      openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

      fFileModified_d = true;

      outputHeader (pofstreamHPP);

      *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
      *pofstreamHPP << "#define _" << *pstringNameOut << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#include <BinaryData.hpp>" << endl;
      *pofstreamHPP << "#include <DeviceMedia.hpp>" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceMedia" << endl;
      *pofstreamHPP << "{" << endl;
      *pofstreamHPP << "public:" << endl;

      *pofstreamHPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamHPP << "                          int          id," << endl;
      *pofstreamHPP << "                           BinaryData *data," << endl;
      *pofstreamHPP << "                           int         iColorAdjustRequired," << endl;
      *pofstreamHPP << "                           int         iAbsorption);" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "   static  DeviceMedia *create           (Device *pDevice, int   id);" << endl;
      *pofstreamHPP << "   static  DeviceMedia *create           (Device *pDevice, char *pszId);" << endl;
      *pofstreamHPP << endl;
      if (0 < mapDeviceID.size ())
      {
         *pofstreamHPP << "   virtual char        *getDeviceID      ();" << endl;
         *pofstreamHPP << endl;
      }
      *pofstreamHPP << "   virtual bool         isMediaSupported (int   id);" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "   virtual Enumeration *getEnumeration   ();" << endl;

      *pofstreamHPP << endl;
      *pofstreamHPP << "protected:" << endl;
      *pofstreamHPP << "   virtual DeviceMedia *createV          (int   id);" << endl;

      *pofstreamHPP << "};" << endl;
      *pofstreamHPP << endl;
      *pofstreamHPP << "#endif" << endl;

      outputHeader (pofstreamCPP);

      *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << *pstringNameOut << "::" << endl;
      *pofstreamCPP << *pstringNameOut << " (Device *pDevice," << endl;
      *pofstreamCPP << "                           int         id," << endl;
      *pofstreamCPP << "                           BinaryData *data," << endl;
      *pofstreamCPP << "                           int         iColorAdjustRequired," << endl;
      *pofstreamCPP << "                           int         iAbsorption)" << endl;
      *pofstreamCPP << "      : DeviceMedia (pDevice," << endl;
      *pofstreamCPP << "                     id," << endl;
      *pofstreamCPP << "                     data," << endl;
      *pofstreamCPP << "                     iColorAdjustRequired," << endl;
      *pofstreamCPP << "                     iAbsorption)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "}" << endl;

      if (0 < mapDeviceID.size ())
      {
         *pofstreamCPP << endl;
         *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
         *pofstreamCPP << "getDeviceID ()" << endl;
         *pofstreamCPP << "{" << endl;
         *pofstreamCPP << "   switch (id_d)" << endl;
         *pofstreamCPP << "   {" << endl;

         for ( DeviceIDList::iterator nextDeviceID = mapDeviceID.begin () ;
               nextDeviceID != mapDeviceID.end () ;
               nextDeviceID++ )
         {
            *pofstreamCPP << "   case "
                          << nextDeviceID->first
                          << ": return \""
                          << nextDeviceID->second
                          << "\";"
                          << endl;
         }

         *pofstreamCPP << "   default: return 0;" << endl;
         *pofstreamCPP << "   }" << endl;
         *pofstreamCPP << "}" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "createV (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return " << *pstringNameOut << "::create (pDevice_d, id);" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, char *pszId)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   return create (pDevice, DeviceMedia::nameToID (pszId));" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "create (Device *pDevice, int id)" << endl;
      *pofstreamCPP << "{" << endl;

      for ( MediaMapData1::iterator next = mapData1.begin () ;
            next != mapData1.end () ;
            next++ )
      {
         BinaryData *bdValue = (BinaryData *)next->second;
         PBYTE       pbValue = bdValue->getData ();
         int         cbValue = bdValue->getLength ();

         *pofstreamCPP << "   static byte abData" << next->first << "[] = {" << dec;

         for (int i = 0; i < cbValue; i++)
         {
            *pofstreamCPP << (int)pbValue[i];
            if (i < cbValue - 1)
               *pofstreamCPP << ",";
         }

         *pofstreamCPP << "};" << endl;
      }

      *pofstreamCPP << endl;
      *pofstreamCPP << "   switch (id)" << endl;
      *pofstreamCPP << "   {" << endl;

      for ( MediaMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++ )
      {
         string *pElement = (string *)next->second;

         *pofstreamCPP << *pElement << endl;
      }

      *pofstreamCPP << "   default: return 0;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "static int aSupportedMedias[] = {" << endl;

      iCount = mapElement.size ();
      for ( MediaMapElement::iterator next = mapElement.begin () ;
            next != mapElement.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "   DeviceMedia::" << next->first;
         if (1 < iCount)
            *pofstreamCPP << ",";
         *pofstreamCPP << endl;
      }

      *pofstreamCPP << "};" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "bool " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "isMediaSupported (int id)" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   for (int i = 0; i < (int)dimof (aSupportedMedias); i++)" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "      if (aSupportedMedias[i] == id)" << endl;
      *pofstreamCPP << "         return true;" << endl;
      *pofstreamCPP << "   }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return false;" << endl;
      *pofstreamCPP << "}" << endl;

      *pofstreamCPP << endl;
      *pofstreamCPP << "Enumeration * " << *pstringNameOut << "::" << endl;
      *pofstreamCPP << "getEnumeration ()" << endl;
      *pofstreamCPP << "{" << endl;
      *pofstreamCPP << "   class MediaEnumerator : public Enumeration" << endl;
      *pofstreamCPP << "   {" << endl;
      *pofstreamCPP << "   public:" << endl;
      *pofstreamCPP << "      MediaEnumerator (Device *pDevice, int cMedias, int aMedias[])" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         pDevice_d = pDevice;" << endl;
      *pofstreamCPP << "         iMedia_d  = 0;" << endl;
      *pofstreamCPP << "         cMedias_d = cMedias;" << endl;
      *pofstreamCPP << "         aMedias_d = aMedias;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual bool hasMoreElements ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iMedia_d < cMedias_d)" << endl;
      *pofstreamCPP << "            return true;" << endl;
      *pofstreamCPP << "         else" << endl;
      *pofstreamCPP << "            return false;" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "      virtual void *nextElement ()" << endl;
      *pofstreamCPP << "      {" << endl;
      *pofstreamCPP << "         if (iMedia_d > cMedias_d - 1)" << endl;
      *pofstreamCPP << "            return 0;" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "         return create (pDevice_d, aMedias_d[iMedia_d++]);" << endl;
      *pofstreamCPP << "      }" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   private:" << endl;
      *pofstreamCPP << "      Device *pDevice_d;" << endl;
      *pofstreamCPP << "      int     iMedia_d;" << endl;
      *pofstreamCPP << "      int     cMedias_d;" << endl;
      *pofstreamCPP << "      int    *aMedias_d;" << endl;
      *pofstreamCPP << "   };" << endl;
      *pofstreamCPP << endl;
      *pofstreamCPP << "   return new MediaEnumerator (pDevice_d, dimof (aSupportedMedias), aSupportedMedias);" << endl;
      *pofstreamCPP << "}" << endl;
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceMediasNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringMediaClassName = 0;
   }

   for ( MediaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( MediaMapData1::iterator next = mapData1.begin () ;
         next != mapData1.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( MediaMapData2::iterator next = mapData2.begin () ;
         next != mapData2.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceConnections (DOM_Node deviceConnectionsNode)
{
   if (deviceConnectionsNode == 0)
      return false;

   DOM_Node deviceConnectionNode = firstNode (deviceConnectionsNode.getFirstChild ());
   int      rc                   = false;

   if (deviceConnectionNode == 0)
      goto done;

   while (deviceConnectionNode != 0)
   {
      DOM_Node nodeItem = firstNode (deviceConnectionNode.getFirstChild ());

      if (nodeItem == 0)
         goto done;

      // name

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // form

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // tray

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      // media

      // Get the next deviceConnection node
      deviceConnectionNode = nextNode (deviceConnectionNode);
   }

   rc = true;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceConnectionsNode, 0);
   }

   // Clean up!

   return rc;
}

bool OmniDomParser::
processDeviceGammas (DOM_Node deviceGammasNode)
{
   if (deviceGammasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringGammaClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   if (!pdi_d->pstringResolutionClassName)
   {
      cerr << "Error: DeviceResolution not filled in." << endl;
      return false;
   }

   if (!pdi_d->pstringMediaClassName)
   {
      cerr << "Error: DeviceMedia not filled in." << endl;
      return false;
   }

   if (!pdi_d->pstringPrintModeClassName)
   {
      cerr << "Error: DevicePrintMode not filled in." << endl;
      return false;
   }

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, string *> GammaMapElement;

   DOM_Node        deviceGammaNode = firstNode (deviceGammasNode.getFirstChild ());
   ofstream       *pofstreamHPP    = 0;
   ofstream       *pofstreamCPP    = 0;
   GammaMapElement mapElement;
   int             iCount          = 0;
   int             rc              = false;

   if (deviceGammaNode == 0)
      goto done;

   while (deviceGammaNode != 0)
   {
      DOM_Node    nodeItem            = firstNode (deviceGammaNode.getFirstChild ());
      const char *pszResolutionID     = 0;
      const char *pszMediaID          = 0;
      const char *pszPrintModeID      = 0;
      const char *pszDitherCatagoryID = 0;
      int         iCGamma;
      int         iMGamma;
      int         iYGamma;
      int         iKGamma;
      int         iCBias;
      int         iMBias;
      int         iYBias;
      int         iKBias;

      if (nodeItem == 0)
         goto done;

      if (  !getValue (nodeItem, &pszResolutionID)
         || -1 == DeviceResolution::nameToID (const_cast<char*>(pszResolutionID))
         )
      {
         cerr << "Error: Unknown DeviceResolution \"" << pszResolutionID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !getValue (nodeItem, &pszMediaID)
         || -1 == DeviceMedia::nameToID (const_cast<char*>(pszMediaID))
         )
      {
         cerr << "Error: Unknown DeviceMedia \"" << pszMediaID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !getValue (nodeItem, &pszPrintModeID)
         || -1 == DevicePrintMode::nameToID (const_cast<char*>(pszPrintModeID))
         )
      {
         cerr << "Error: Unknown DevicePrintMode \"" << pszPrintModeID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      if (  !getValue (nodeItem, &pszDitherCatagoryID)
         || !DeviceDither::ditherCatagoryValid (const_cast<char*>(pszDitherCatagoryID))
         )
      {
         cerr << "Error: Unknown Dither type \"" << pszDitherCatagoryID << "\"" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iCGamma);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iMGamma);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iYGamma);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iKGamma);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iCBias);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iMBias);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iYBias);

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &iKBias);

      ostringstream oss;

      oss << "   "
          << "{"
          << *pdi_d->pstringResolutionClassName
          << "::"
          << pszResolutionID
          << ","
          << *pdi_d->pstringMediaClassName
          << "::"
          << pszMediaID
          << ","
          << *pdi_d->pstringPrintModeClassName
          << "::"
          << pszPrintModeID
          << ","
          << "\""
          << pszDitherCatagoryID
          << "\","
          << iCGamma
          << ","
          << iMGamma
          << ","
          << iYGamma
          << ","
          << iKGamma
          << ","
          << iCBias
          << ","
          << iMBias
          << ","
          << iYBias
          << ","
          << iKBias
          << "}";

      ostringstream ossName;

      // Create a unique name
      ossName << pszResolutionID
              << "/"
              << pszMediaID
              << "/"
              << pszPrintModeID
              << "/"
              << pszDitherCatagoryID;

      mapElement[string (ossName.str ())] = new string (oss.str ());

#ifdef DEBUG
      if (fDebugOutput)
      {
         cerr << dec << "pszResolutionID     = " << pszResolutionID     << endl;
         cerr <<        "pszMediaID          = " << pszMediaID          << endl;
         cerr <<        "pszPrintModeID      = " << pszPrintModeID      << endl;
         cerr <<        "pszDitherCatagoryID = " << pszDitherCatagoryID << endl;
         cerr <<        "iCGamma =             " << iCGamma             << endl;
         cerr <<        "iMGamma =             " << iMGamma             << endl;
         cerr <<        "iYGamma =             " << iYGamma             << endl;
         cerr <<        "iKGamma =             " << iKGamma             << endl;
         cerr <<        "iCBias =              " << iCBias              << endl;
         cerr <<        "iMBias =              " << iMBias              << endl;
         cerr <<        "iYBias =              " << iYBias              << endl;
         cerr <<        "iKBias =              " << iKBias              << endl;
      }
#endif

      // Get the next deviceGamma node
      deviceGammaNode = nextNode (deviceGammaNode);
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   fFileModified_d = true;

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <DeviceGammaTable.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceGammaTable" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;

   *pofstreamHPP << "static DeviceGamma *getDeviceGamma (int   iResolutionID," << endl;
   *pofstreamHPP << "                                    int   iMediaID," << endl;
   *pofstreamHPP << "                                    int   iPrintModeID," << endl;
   *pofstreamHPP << "                                    char *pszDitherID);" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <defines.hpp>" << endl;
   *pofstreamCPP << "#include <Device.hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringResolutionClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringMediaClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pdi_d->pstringPrintModeClassName << ".hpp>" << endl;
   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "typedef struct _GAMMATABLE {" << endl;
   *pofstreamCPP << "   int                iResolutionID;" << endl;
   *pofstreamCPP << "   int                iMediaID;" << endl;
   *pofstreamCPP << "   int                iPrintModeID;" << endl;
   *pofstreamCPP << "   char              *pszDitherCatagoryID;" << endl;
   *pofstreamCPP << "   unsigned short int usCGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usMGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usYGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usKGamma;" << endl;
   *pofstreamCPP << "   unsigned short int usCBias;" << endl;
   *pofstreamCPP << "   unsigned short int usMBias;" << endl;
   *pofstreamCPP << "   unsigned short int usYBias;" << endl;
   *pofstreamCPP << "   unsigned short int usKBias;" << endl;
   *pofstreamCPP << "} GAMMATABLE, *PGAMMATABLE;" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "GAMMATABLE gt" << *pstringNameOut << "[] = {" << endl;

   iCount = mapElement.size ();
   for ( GammaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      string *pElement = (string *)next->second;

      *pofstreamCPP << *pElement;
      if (1 < iCount)
         *pofstreamCPP << ",";
      *pofstreamCPP << endl;
   }

   *pofstreamCPP << "};" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceGamma * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDeviceGamma (int   iResolutionID," << endl;
   *pofstreamCPP << "                int   iMediaID," << endl;
   *pofstreamCPP << "                int   iPrintModeID," << endl;
   *pofstreamCPP << "                char *pszDitherID)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   char *pszDitherCatagoryID = DeviceDither::getDitherCatagory (pszDitherID);" << endl;
   *pofstreamCPP << "" << endl;
   *pofstreamCPP << "   for (int i = 0; i < (int)dimof (gt" << *pstringNameOut << "); i++)" << endl;
   *pofstreamCPP << "   {" << endl;
   *pofstreamCPP << "      if (  iResolutionID     == gt" << *pstringNameOut << "[i].iResolutionID" << endl;
   *pofstreamCPP << "         && iMediaID          == gt" << *pstringNameOut << "[i].iMediaID" << endl;
   *pofstreamCPP << "         && iPrintModeID      == gt" << *pstringNameOut << "[i].iPrintModeID" << endl;
   *pofstreamCPP << "         && 0 == strcmp (pszDitherCatagoryID, gt" << *pstringNameOut << "[i].pszDitherCatagoryID)" << endl;
   *pofstreamCPP << "         )" << endl;
   *pofstreamCPP << "      {" << endl;
   *pofstreamCPP << "         return new DeviceGamma ((int)gt" << *pstringNameOut << "[i].usCGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usMGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usYGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usKGamma," << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usCBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usMBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usYBias,"  << endl;
   *pofstreamCPP << "                                 (int)gt" << *pstringNameOut << "[i].usKBias);" << endl;
   *pofstreamCPP << "      }" << endl;
   *pofstreamCPP << "   }" << endl;
   *pofstreamCPP << "" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceGammasNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringGammaClassName = 0;
   }

   for ( GammaMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceDatas (DOM_Node deviceDatasNode)
{
   if (deviceDatasNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringDataClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, BinaryData *> DataMapElement;
   typedef map <string, byte *>       DataMapData;

   DOM_Node        deviceDataNode = firstNode (deviceDatasNode.getFirstChild ());
   ofstream       *pofstreamHPP   = 0;
   ofstream       *pofstreamCPP   = 0;
   DataMapElement  mapElement;
   DataMapData     mapData;
   int             rc             = false;

   if (deviceDataNode == 0)
      goto done;

   while (deviceDataNode != 0)
   {
      DOM_NamedNodeMap dataAttributes = deviceDataNode.getAttributes ();

      if (dataAttributes == 0)
         goto done;

      DOMString   nameString ("name");
      DOMString   typeString ("type");
      DOM_Node    name       = dataAttributes.getNamedItem (nameString);
      DOM_Node    type       = dataAttributes.getNamedItem (typeString);
      byte       *pbData     = 0;
      int         cbData;

      if (name == 0)
         goto done;
      if (type == 0)
         goto done;

      const char *pszDataName  = 0;
      const char *pszDataValue = 0;
      const char *pszTypeName  = 0;

      getValue (name,            &pszDataName);
      getValue (deviceDataNode, &pszDataValue);
      getValue (type,            &pszTypeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "name  = " << pszDataName  << ", type = " << pszTypeName << endl;
      if (fDebugOutput) cerr << "value = " << pszDataValue << endl;
#endif

      pbData = 0;

      if (0 == strcmp (pszTypeName, "string"))
      {
         cbData = strlen (pszDataValue) + 1;
         pbData = new byte [cbData];

         if (pbData)
         {
            strcpy ((char *)pbData, pszDataValue);
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "boolean"))
      {
         cbData = 4;
         pbData = new byte [cbData];

         if (pbData)
         {
            memset (pbData, 0, cbData);

            if (0 == strcasecmp (pszDataValue, "true"))
            {
               pbData[0] = 1;
            }
            else if (0 == strcasecmp (pszDataValue, "false"))
            {
               pbData[0] = 0;
            }
            else
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "integer"))
      {
         cbData = 4;
         pbData = new byte [cbData];

         if (pbData)
         {
            if (0 == sscanf (pszDataValue, "%d", (int *)pbData))
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "byte"))
      {
         cbData = 1;
         pbData = new byte [cbData];

         if (pbData)
         {
            if (0 == sscanf (pszDataValue, "%c", (char *)pbData))
            {
               delete[] pbData;
               goto done;
            }
         }
         else
         {
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "binary"))
      {
         if (!parseBinaryData (pszDataValue, &pbData, &cbData))
         {
            cerr << "Error: binary data not understood \"" << pszDataValue << "\" for " << pszDataName << "." << endl;
            delete[] pbData;
            goto done;
         }
      }
      else if (0 == strcmp (pszTypeName, "bytearray"))
      {
         int iCnt,
             iTempCnt = 0;

         for (iCnt = 0; *(pszDataValue+iCnt); iCnt++)
         {
            if (!isxdigit (*(pszDataValue+iCnt)))
            {
               if (isspace (*(pszDataValue+iCnt)) == 0)
               {
                  cerr << "Error: Data in bytearray is not understood \"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
                  goto done;
               }
            }
            else
            {
               iTempCnt++;
            }
         }

         cbData = iTempCnt/2;
         pbData = new byte [cbData];
         if (!pbData)
         {
            goto done;
         }

         int          iEnd = strlen (pszDataValue);
         int          iLoc = 0;
         bool         bOK;
         char         cTemp[2];
         unsigned int iValue;

         iTempCnt = 0;

         for (iCnt = 0; iCnt < iEnd; iCnt++)
         {
            if (isxdigit (*(pszDataValue+iCnt)))
            {
               cTemp[iLoc++] = *(pszDataValue+iCnt);
               if (iLoc == 2)
               {
                  bOK = parseHexGroup ((const char *)cTemp, &iValue);
                  if (!bOK)
                  {
                     cerr << "Error: Data in bytearray is not understood \"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
                     delete[] pbData;
                     goto done;
                  }

                  *(pbData+iTempCnt++) = (unsigned char)iValue;
                  iLoc = 0;
               }
            }
         }

         if (iLoc)
         {
            cerr << "Error: Data in bytearray is missing bytes\"" << *(pszDataValue+iCnt) << "\" for " << pszDataName << "." << endl;
            delete[] pbData;
            goto done;
         }
      }
      else
      {
         cerr << "Error: Unknown type " << pszTypeName << endl;
         goto done;
      }

      if (!pbData)
      {
         cerr << "Error: pbData is not allocated." << endl;
         goto done;
      }

      mapData[string (pszDataName)]    = pbData;
      mapElement[string (pszDataName)] = new BinaryData (pbData, cbData);

      // Move to the next one
      deviceDataNode = nextNode (deviceDataNode);
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   fFileModified_d = true;

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << "#include <DeviceData.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceData" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " ()" << endl;
   *pofstreamCPP << "{" << endl;

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      BinaryData *bdValue = (BinaryData *)next->second;
      PBYTE       pbValue = bdValue->getData ();
      int         cbValue = bdValue->getLength ();

      *pofstreamCPP << "   static byte " << next->first << "[] = {" << dec;

      for (int i = 0; i < cbValue; i++)
      {
         *pofstreamCPP << (int)pbValue[i];
         if (i < cbValue - 1)
            *pofstreamCPP << ",";
      }

      *pofstreamCPP << "};" << endl;
   }

   *pofstreamCPP << endl;

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      *pofstreamCPP << "   add (\""
                    << next->first
                    << "\", new BinaryData ("
                    << next->first
                    << ", dimof ("
                    << next->first << ")));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceDatasNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringDataClassName = 0;
   }

   for ( DataMapElement::iterator next = mapElement.begin () ;
         next != mapElement.end () ;
         next++ )
   {
      delete (*next).second;
   }
   for ( DataMapData::iterator next = mapData.begin () ;
         next != mapData.end () ;
         next++ )
   {
      delete[] (*next).second;
   }

   return rc;
}

bool OmniDomParser::
processDeviceStrings (DOM_Node deviceStringsNode)
{
   if (deviceStringsNode == 0)
      return false;

   string *pstringNameOut = getNameOut ();

   pdi_d->pstringStringClassName = pstringNameOut;

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;

   typedef map <string, string>             StringMapElement;
   typedef map <string, StringMapElement *> StringMapLanguage;

   DOM_Node           deviceStringNode = firstNode (deviceStringsNode.getFirstChild ());
   ofstream          *pofstreamHPP     = 0;
   ofstream          *pofstreamCPP     = 0;
   StringMapLanguage  mapLanguage;
   int                rc               = false;

   if (deviceStringNode == 0)
      goto done;

   while (deviceStringNode != 0)
   {
      DOM_Node    nodeItem      = firstNode (deviceStringNode.getFirstChild ());
      DOM_Node    nodeLanguages;
      const char *pszName       = 0;
      string      stringElement;

      if (nodeItem == 0)
         goto done;

      getValue (nodeItem, &pszName);

      stringElement = pszName;

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      nodeLanguages = firstNode (nodeItem.getFirstChild ());

      while (nodeLanguages != 0)
      {
         const char       *pszTranslation      = 0;
         DOMString         nodeName            = nodeLanguages.getNodeName ();
         string            stringLanguageName;
         string            stringTranslation;
         StringMapElement *psme                = 0;

         getValue (nodeLanguages, &pszTranslation);

         stringLanguageName = (char *)nodeName.transcode ();
         stringTranslation  = pszTranslation;

         psme = mapLanguage[stringLanguageName];

         if (!psme)
         {
            psme = new StringMapElement ();

            mapLanguage[stringLanguageName] = psme;
         }

         (*psme)[stringElement] = stringTranslation;

         // Move to the next one
         nodeLanguages = nextNode (nodeLanguages);
      }

      // Move to the next one
      deviceStringNode = nextNode (deviceStringNode);
   }

   rc = true;

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   fFileModified_d = true;

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << "#include <DeviceString.hpp>" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "class " << *pstringNameOut << " : public DeviceString" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " ()" << endl;
   *pofstreamCPP << "{" << endl;

   for ( StringMapLanguage::iterator nextLanguage = mapLanguage.begin () ;
         nextLanguage != mapLanguage.end () ;
         nextLanguage++ )
   {
      StringMapElement *pEntries = (StringMapElement *)nextLanguage->second;

      for ( StringMapElement::iterator nextEntry = pEntries->begin () ;
            nextEntry != pEntries->end () ;
            nextEntry++ )
      {
         *pofstreamCPP << "   add (\""
                       << nextLanguage->first
                       << "\", \""
                       << nextEntry->first
                       << "\", \""
                       << nextEntry->second
                       << "\");"
                       << endl;
      }
   }

   *pofstreamCPP << "};" << endl;

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (deviceStringsNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   if (!rc)
   {
      delete pstringNameOut;
      pdi_d->pstringStringClassName = 0;
   }

   for ( StringMapLanguage::iterator nextLanguage = mapLanguage.begin () ;
         nextLanguage != mapLanguage.end () ;
         nextLanguage++ )
   {
      StringMapElement *pEntries = (StringMapElement *)nextLanguage->second;

      delete pEntries;
   }

   return rc;
}

bool OmniDomParser::
preprocessDriver (DOM_Node driverNode)
{
   // This routine does two things.
   //    It fill is pstringInstanceClassName and pstringBlitterClassName
   //       in the DeviceInfo structure.
   //    It sees if it needs to process (copy) the instance and blitter files.

   DOM_Node          nodeItem               = firstNode (driverNode.getFirstChild ());
   const char       *pszNodeName               = 0;
   DOM_NamedNodeMap  nodeAttributes;
   DOMString         exenameString ("exename");
   DOM_Node          exename;
   DOMString         dataString ("data");
   DOM_Node          data;
   const char       *pszExeName                = 0;
   const char       *pszData                   = 0;
   bool              fRetVal                   = false;

   getName (nodeItem, &pszNodeName);

   while (  nodeItem != 0
         && !(  0 == strcmp (pszNodeName, "Instance")
             || 0 == strcmp (pszNodeName, "PluggableInstance")
             )
         )
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);

      getName (nodeItem, &pszNodeName);
   }

   if (nodeItem == 0)
      goto done;

   if (0 == strcmp (pszNodeName, "Instance"))
   {
      getValue (nodeItem, &pszNodeName);

      pdi_d->pstringInstanceClassName = converToClassName (pszNodeName);

      copyFile (pszNodeName);

      // Move to the next node
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Instance"))
         goto done;

      getValue (nodeItem, &pszNodeName);
      copyFile (pszNodeName);

      // Move to the next node
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Blitter"))
         goto done;

      getValue (nodeItem, &pszNodeName);

      pdi_d->pstringBlitterClassName = converToClassName (pszNodeName);

      copyFile (pszNodeName);

      // Move to the next node
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Blitter"))
         goto done;

      getValue (nodeItem, &pszNodeName);
      copyFile (pszNodeName);

      fRetVal = true;
   }
   else if (0 == strcmp (pszNodeName, "PluggableInstance"))
   {
      nodeAttributes = nodeItem.getAttributes ();
      exename        = nodeAttributes.getNamedItem (exenameString);
      data           = nodeAttributes.getNamedItem (dataString);

      getValue (exename, &pszExeName);

      pdi_d->pstringExeName = new string (pszExeName);
      if (data != 0)
      {
         getValue (data, &pszData);
         pdi_d->pstringData = new string (pszData);
      }

#ifdef DEBUG
      if (fDebugOutput) cerr << "exename = " << pszExeName << endl;
      if (fDebugOutput) cerr << "data = " << pszData << endl;
#endif

      // Move to the next node
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "PluggableBlitter"))
         goto done;

      fRetVal = true;
   }
   else
   {
      cerr << "Error: Expecting either an Instance tag or a PluggableInstance tag." << endl;
      goto done;
   }

done:
   if (!fRetVal)
   {
      cerr << "Error: preprocessDriver failed." << endl;
      setErrorCondition ();
   }

   return fRetVal;
}

static int
parseSize (char **ppszComma)
{
   int iReturn = 0;

   while (  '0' <= **ppszComma
         && **ppszComma <= '9'
         )
   {
      iReturn = iReturn * 10 + (**ppszComma - '0');

      (*ppszComma)++;
   }

   return iReturn;
}

bool OmniDomParser::
processDriver (DOM_Node driverNode)
{
   bool  fFoundItem = false;
   char *pszComma   = 0;

   if (driverNode == 0)
      return false;

   if (!preprocessDriver (driverNode))
      return false;

   if (!shouldCreateFile (*pstrXMLFile_d))
   {
      cout << "Parsing:    \"" << *pstrXMLFile_d << "\"" << endl;
   }
   else
   {
      cout << "Generating: \"" << *pstrXMLFile_d << "\"" << endl;
   }

   // Make sure required fields in the Device Info struture are filled in
   if (!pdi_d->pstringCommandClassName)
   {
      cerr << "Error: DeviceCommands not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringResolutionClassName)
   {
      cerr << "Error: DeviceResolution not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringPrintModeClassName)
   {
      cerr << "Error: DevicePrintMode not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringTrayClassName)
   {
      cerr << "Error: DeviceTray not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringFormClassName)
   {
      cerr << "Error: DeviceForm not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringMediaClassName)
   {
      cerr << "Error: DeviceMedia not filled in." << endl;
      setErrorCondition ();
      return false;
   }
   if (!pdi_d->pstringOrientationClassName)
   {
      cerr << "Error: DeviceOrientation not filled in." << endl;
      setErrorCondition ();
      return false;
   }

   if (fAutoconf_d)
   {
      ofstream *pofstreamAM1   = 0;
      ofstream *pofstreamAM2   = 0;
      string   *pstringNameOut = getNameOut ();

      pofstreamAM1 = new ofstream ("libraries1.mak", ios::app);
      pofstreamAM2 = new ofstream ("libraries2.mak", ios::app);

      if (fAutoconfNoInst_d)
         *pofstreamAM1 << "noinst_LTLIBRARIES";
      else
         *pofstreamAM1 << "lib_LTLIBRARIES";

      *pofstreamAM1 << " += lib" << *pstringNameOut << ".la" << endl;

      *pofstreamAM2 << "lib" << *pstringNameOut << "_la_SOURCES = " << *pdi_d->pstringPrintModeClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringResolutionClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringOrientationClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringCommandClassName << ".cpp \\" << endl;
      if (pdi_d->pstringGammaClassName)
         *pofstreamAM2 << "                        " << *pdi_d->pstringGammaClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringTrayClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringFormClassName << ".cpp \\" << endl;
      if (pdi_d->pstringInstanceClassName)
         *pofstreamAM2 << "                        " << *pdi_d->pstringInstanceClassName << ".cpp \\" << endl;
      if (pdi_d->pstringBlitterClassName)
         *pofstreamAM2 << "                        " << *pdi_d->pstringBlitterClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pdi_d->pstringMediaClassName << ".cpp \\" << endl;
      if (pdi_d->pstringDataClassName)
         *pofstreamAM2 << "                        " << *pdi_d->pstringDataClassName << ".cpp \\" << endl;
      if (pdi_d->pstringStringClassName)
         *pofstreamAM2 << "                        " << *pdi_d->pstringStringClassName << ".cpp \\" << endl;
      *pofstreamAM2 << "                        " << *pstringNameOut << ".cpp" << endl;
      *pofstreamAM2 << "lib" << *pstringNameOut << "_la_LDFLAGS = -version-info @LT_CURRENT@:@LT_REVISION@:@LT_AGE@" << endl;

      pofstreamAM1->close ();
      pofstreamAM2->close ();

      delete pofstreamAM1;
      delete pofstreamAM2;
      delete pstringNameOut;
   }

   if (!shouldCreateFile (*pstrXMLFile_d))
      return true;

   typedef list <string> DriverList;

   DOM_NamedNodeMap driverNodeAttributes   = driverNode.getAttributes ();
   DOM_NamedNodeMap nodeSiblingAttributes;
   DOMString        nameString ("name");
   DOMString        typeString ("type");
   DOMString        levelString ("level");
   DOMString        majorString ("major");
   DOMString        minorString ("minor");
   DOMString        sublevelString ("sublevel");
   DOM_Node         value;
   const char      *pszDeviceName          = 0;
   const char      *pszDriverName          = 0;
   const char      *pszNodeName            = 0;
   const char      *pszTypeName            = 0;
   const char      *pszMajorPDLLevel       = 0;
   const char      *pszMinorPDLLevel       = 0;
   const char      *pszMajorRevisionLevel  = 0;
   const char      *pszMinorRevisionLevel  = 0;
   const char      *pszOrientation         = 0;
   const char      *pszForm                = 0;
   const char      *pszTray                = 0;
   const char      *pszMedia               = 0;
   const char      *pszResolution          = 0;
   const char      *pszDither              = 0;
   const char      *pszPrintMode           = 0;
   DOM_Node         nodeItem               = firstNode (driverNode.getFirstChild ());
   string          *pstringNameOut         = getNameOut ();
   ofstream        *pofstreamHPP           = 0;
   ofstream        *pofstreamCPP           = 0;
   ofstream        *pofstreamMAK           = 0;
   DriverList       listCapabilities;
   DriverList       listRasterCapabilities;
   DriverList       listDeviceOptions;
   DriverList       listDeviceOptions2;
   int              iCount                 = 0;
   DOM_Node         nodeOther;
   int              rc                     = false;

   if (driverNodeAttributes == 0)
      goto done;

   value = driverNodeAttributes.getNamedItem (nameString);

   if (value == 0)
      goto done;

   getValue (value, &pszDeviceName);

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszDeviceName = " << pszDeviceName << endl;
#endif

   getValue (nodeItem, &pszDriverName);

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszDriverName = " << pszDriverName << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getName (nodeItem, &pszNodeName);

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif

   while (0 == strcmp (pszNodeName, "Capability"))
   {
      nodeSiblingAttributes = nodeItem.getAttributes ();
      value                 = nodeSiblingAttributes.getNamedItem (typeString);

      if (value == 0)
         goto done;

      getValue (value, &pszTypeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;
#endif

      listCapabilities.push_back (string (pszTypeName));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif
   }

   while (0 == strcmp (pszNodeName, "RasterCapabilities"))
   {
      nodeSiblingAttributes = nodeItem.getAttributes ();
      value                 = nodeSiblingAttributes.getNamedItem (typeString);

      if (value == 0)
         goto done;

      getValue (value, &pszTypeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;
#endif

      listRasterCapabilities.push_back (string (pszTypeName));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif
   }

   while (0 == strcmp (pszNodeName, "DeviceOptions"))
   {
      nodeSiblingAttributes = nodeItem.getAttributes ();
      value                 = nodeSiblingAttributes.getNamedItem (typeString);

      if (value == 0)
         goto done;

      getValue (value, &pszTypeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;
#endif

      listDeviceOptions.push_back (string (pszTypeName));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif
   }

   while (0 == strcmp (pszNodeName, "DeviceOptions2"))
   {
      nodeSiblingAttributes = nodeItem.getAttributes ();
      value                 = nodeSiblingAttributes.getNamedItem (typeString);

      if (value == 0)
         goto done;

      getValue (value, &pszTypeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszTypeName = " << pszTypeName << endl;
#endif

      listDeviceOptions2.push_back (string (pszTypeName));

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif
   }

   if (0 != strcmp (pszNodeName, "PDL"))
   {
      cerr << "Error: Could not find PDL tag" << endl;
      setErrorCondition ();
      goto done;
   }

   nodeSiblingAttributes = nodeItem.getAttributes ();

   value = nodeSiblingAttributes.getNamedItem (levelString);
   if (value == 0)
      goto done;
   getValue (value, &pszMajorPDLLevel);
#ifdef DEBUG
   if (fDebugOutput) cerr << "pszMajorPDLLevel = " << pszMajorPDLLevel << endl;
#endif

   value = nodeSiblingAttributes.getNamedItem (sublevelString);
   if (value == 0)
      goto done;
   getValue (value, &pszMinorPDLLevel);
#ifdef DEBUG
   if (fDebugOutput) cerr << "pszMinorPDLLevel = " << pszMinorPDLLevel << endl;
#endif

   value = nodeSiblingAttributes.getNamedItem (majorString);
   if (value == 0)
      goto done;
   getValue (value, &pszMajorRevisionLevel);
#ifdef DEBUG
   if (fDebugOutput) cerr << "pszMajorRevisionLevel = " << pszMajorRevisionLevel << endl;
#endif

   value = nodeSiblingAttributes.getNamedItem (minorString);
   if (value == 0)
      goto done;
   getValue (value, &pszMinorRevisionLevel);
#ifdef DEBUG
   if (fDebugOutput) cerr << "pszMinorRevisionLevel = " << pszMinorRevisionLevel << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getName (nodeItem, &pszNodeName);

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif

   while (  0 == strcmp (pszNodeName, "Uses")
         || 0 == strcmp (pszNodeName, "Has")
         )
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "pszNodeName = " << pszNodeName << endl;
#endif
   }

   if (0 == strcmp (pszNodeName, "Instance"))
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Instance"))
      {
         cerr << "Error: Could not find Instance tag" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Blitter"))
      {
         cerr << "Error: Could not find Blitter tag" << endl;
         setErrorCondition ();
         goto done;
      }

      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "Blitter"))
      {
         cerr << "Error: Could not find Blitter tag" << endl;
         setErrorCondition ();
         goto done;
      }
   }
   else if (0 == strcmp (pszNodeName, "PluggableInstance"))
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;

      getName (nodeItem, &pszNodeName);

      if (0 != strcmp (pszNodeName, "PluggableBlitter"))
      {
         cerr << "Error: Could not find PluggableBlitter tag" << endl;
         setErrorCondition ();
         goto done;
      }

      if (  0 != listDeviceOptions.size ()
         || 0 != listDeviceOptions2.size ()
         )
      {
         cerr << "Error: Cannot have <DeviceOptions> or <DeviceOptions2> with <PluggableInstance> tag" << endl;
         setErrorCondition ();
         goto done;
      }
   }
   else
   {
      cerr << "Error: Could not find Instance or PluggableInstance tag" << endl;
      setErrorCondition ();
      goto done;
   }

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getName (nodeItem, &pszNodeName);

   if (0 == strcmp (pszNodeName, "DeviceData"))
   {
      // Move to the next one
      nodeItem = nextNode (nodeItem);
      if (nodeItem == 0)
         goto done;
   }

   // Move down into the <DefaultJobProperties> tree
   nodeItem  = firstNode (nodeItem.getFirstChild ());

   getValue (nodeItem, &pszOrientation);

   if (  0 == pszOrientation
      || -1 == DeviceOrientation::nameToID (const_cast<char*>(pszOrientation))
      )
   {
      cerr << "Error: Unknown DeviceOrientation \"" << pszOrientation << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( OrientationList::iterator next = pdi_d->listOrientations.begin () ;
         next != pdi_d->listOrientations.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszOrientation))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszOrientation << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszOrientation = " << pszOrientation << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszForm);

   if (pszForm)
   {
      pszComma = strchr (pszForm, ',');
      if (pszComma)
      {
         *pszComma = '\0';
      }
      if (-1 == DeviceForm::nameToID (const_cast<char*>(pszForm)))
      {
         cerr << "Error: Unknown DeviceForm \"" << pszForm << "\"" << endl;
         setErrorCondition ();
         goto done;
      }
   }
   else
   {
      cerr << "Error: 0 DeviceForm \"" << pszForm << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( FormList::iterator next = pdi_d->listForms.begin () ;
         next != pdi_d->listForms.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszForm))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszForm << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

   if (pszComma)
   {
      *pszComma = ',';
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszForm = " << pszForm << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszTray);

   if (  0 == pszTray
      || -1 == DeviceTray::nameToID (const_cast<char*>(pszTray))
      )
   {
      cerr << "Error: Unknown DeviceTray \"" << pszTray << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( TrayList::iterator next = pdi_d->listTrays.begin () ;
         next != pdi_d->listTrays.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszTray))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszTray << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszTray = " << pszTray << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszMedia);

   if (  0 == pszMedia
      || -1 == DeviceMedia::nameToID (const_cast<char*>(pszMedia))
      )
   {
      cerr << "Error: Unknown DeviceMedia \"" << pszMedia << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( MediaList::iterator next = pdi_d->listMedias.begin () ;
         next != pdi_d->listMedias.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszMedia))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszMedia << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszMedia = " << pszMedia << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszResolution);

   if (  0 == pszResolution
      || -1 == DeviceResolution::nameToID (const_cast<char*>(pszResolution))
      )
   {
      cerr << "Error: Unknown DeviceResolution \"" << pszResolution << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( ResolutionList::iterator next = pdi_d->listResolutions.begin () ;
         next != pdi_d->listResolutions.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszResolution))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszResolution << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszResolution = " << pszResolution << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszDither);

   if (  0 == pszDither
      || !DeviceDither::ditherNameValid (const_cast<char*>(pszDither))
      )
   {
      cerr << "Error: Unknown DeviceDither \"" << pszDither << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszDither = " << pszDither << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);
   if (nodeItem == 0)
      goto done;

   getValue (nodeItem, &pszPrintMode);

   if (  0 == pszPrintMode
      || -1 == DevicePrintMode::nameToID (const_cast<char*>(pszPrintMode))
      )
   {
      cerr << "Error: Unknown DevicePrintMode \"" << pszPrintMode << "\"" << endl;
      setErrorCondition ();
      goto done;
   }

   fFoundItem = false;
   for ( PrintModeList::iterator next = pdi_d->listPrintModes.begin () ;
         next != pdi_d->listPrintModes.end () ;
         next++ )
   {
      if (0 == strcmp (next->c_str (), pszPrintMode))
      {
         fFoundItem = true;
      }
   }

   if (!fFoundItem)
   {
      cerr << "Error: \"" << pszPrintMode << "\" is not defined for this device." << endl;
      setErrorCondition ();
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) cerr << "pszPrintMode = " << pszPrintMode << endl;
#endif

   // Move to the next one
   nodeItem = nextNode (nodeItem);

   nodeOther = nodeItem;

   // process optional other(s)
   while (nodeItem != 0)
   {
      getName (nodeItem, &pszNodeName);

#ifdef DEBUG
      if (fDebugOutput) cerr << "other = " << pszNodeName << endl;
#endif

      if (0 != (strcmp (pszNodeName, "other")))
          goto done;

      // Move to the next one
      nodeItem = nextNode (nodeItem);
   }

   rc = true;

   listCapabilities.sort ();
   listRasterCapabilities.sort ();
   listDeviceOptions.sort ();
   listDeviceOptions2.sort ();
   listCapabilities.unique ();
   listRasterCapabilities.unique ();
   listDeviceOptions.unique ();
   listDeviceOptions2.unique ();

   openOutputFiles (pstringNameOut, &pofstreamHPP, &pofstreamCPP);

   fFileModified_d = true;

   outputHeader (pofstreamHPP);

   *pofstreamHPP << "#ifndef _" << *pstringNameOut << endl;
   *pofstreamHPP << "#define _" << *pstringNameOut << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringFormClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringTrayClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringMediaClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringResolutionClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringOrientationClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringCommandClassName << ".hpp>" << endl;
   if (pdi_d->pstringDataClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringDataClassName << ".hpp>" << endl;
   if (pdi_d->pstringStringClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringStringClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <" << *pdi_d->pstringPrintModeClassName << ".hpp>" << endl;
   if (pdi_d->pstringInstanceClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringInstanceClassName << ".hpp>" << endl;
   if (pdi_d->pstringBlitterClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringBlitterClassName << ".hpp>" << endl;
   if (pdi_d->pstringGammaClassName)
      *pofstreamHPP << "#include <" << *pdi_d->pstringGammaClassName << ".hpp>" << endl;
   *pofstreamHPP << "#include <Device.hpp>" << endl;
   *pofstreamHPP << "#include <PrintDevice.hpp>" << endl;
   *pofstreamHPP << "#include <JobProperties.hpp>" << endl;
   *pofstreamHPP << "#include <OmniProxy.hpp>" << endl;
   *pofstreamHPP << endl;

   *pofstreamHPP << "class " << *pstringNameOut << " : public PrintDevice" << endl;
   *pofstreamHPP << "{" << endl;
   *pofstreamHPP << "public:" << endl;

   iCount = listRasterCapabilities.size ();
   if (0 < iCount)
   {
      int iValue = 1;

      *pofstreamHPP << "   enum {" << endl;

      for ( DriverList::iterator next = listRasterCapabilities.begin () ;
            next != listRasterCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamHPP << "      " << *next << " = " << iValue;
         if (1 < iCount)
            *pofstreamHPP << ",";
         *pofstreamHPP << endl;

         iValue *= 2;
      }

      *pofstreamHPP << "   };" << endl;
      *pofstreamHPP << endl;
   }

   *pofstreamHPP << "   " << *pstringNameOut << " ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   " << *pstringNameOut << " (char *pszJobProperties);" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   virtual DeviceBooklet     *getDefaultBooklet        ();" << endl;
   *pofstreamHPP << "   virtual DeviceCollation   *getDefaultCollation      ();" << endl;
   *pofstreamHPP << "   virtual DeviceCopies      *getDefaultCopies         ();" << endl;
   *pofstreamHPP << "   virtual DeviceDestination *getDefaultDestination    ();" << endl;
   *pofstreamHPP << "   virtual char              *getDefaultDitherID       ();" << endl;
   *pofstreamHPP << "   virtual DeviceDuplex      *getDefaultDuplex         ();" << endl;
   *pofstreamHPP << "   virtual DeviceForm        *getDefaultForm           ();" << endl;
   *pofstreamHPP << "   virtual DeviceJogging     *getDefaultJogging        ();" << endl;
   *pofstreamHPP << "   virtual DeviceMedia       *getDefaultMedia          ();" << endl;
   *pofstreamHPP << "   virtual DeviceNUp         *getDefaultNUp            ();" << endl;
   *pofstreamHPP << "   virtual DeviceOrientation *getDefaultOrientation    ();" << endl;
   *pofstreamHPP << "   virtual DevicePrintMode   *getDefaultPrintMode      ();" << endl;
   *pofstreamHPP << "   virtual DeviceResolution  *getDefaultResolution     ();" << endl;
   *pofstreamHPP << "   virtual DeviceScaling     *getDefaultScaling        ();" << endl;
   *pofstreamHPP << "   virtual DeviceStapling    *getDefaultStapling       ();" << endl;
   *pofstreamHPP << "   virtual DeviceTray        *getDefaultTray           ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   virtual DeviceGamma       *getCurrentGamma          ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "   virtual DeviceCommand     *getDefaultCommands       ();" << endl;
   *pofstreamHPP << "   virtual DeviceData        *getDefaultData           ();" << endl;
   *pofstreamHPP << "   virtual DeviceString      *getDefaultString         ();" << endl;
   *pofstreamHPP << endl;
   *pofstreamHPP << "private:" << endl;
   *pofstreamHPP << "   void commonInit ();" << endl;
   *pofstreamHPP << "};" << endl;
   *pofstreamHPP << endl;

   *pofstreamHPP << "extern \"C\" {" << endl;
   *pofstreamHPP << "   char        *getVersion                  ();" << endl;
   *pofstreamHPP << "   Enumeration *getDeviceEnumeration        (char   *pszLibraryName);" << endl;
   *pofstreamHPP << "   Device      *newDeviceW_Advanced         (bool    fAdvanced);" << endl;
   *pofstreamHPP << "   Device      *newDeviceW_JopProp_Advanced (char   *pszJobProperties, bool fAdvanced);" << endl;
   *pofstreamHPP << "   void         deleteDevice                (Device *pDevice);" << endl;
   *pofstreamHPP << "   void         deleteDevice__FP6Device     (Device *pDevice);" << endl;
   *pofstreamHPP << "};" << endl;

   *pofstreamHPP << endl;
   *pofstreamHPP << "#endif" << endl;

   outputHeader (pofstreamCPP);

   *pofstreamCPP << "#include <" << *pstringNameOut << ".hpp>" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "char *" << endl;
   *pofstreamCPP << "getVersion ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return VERSION;" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "Enumeration *" << endl;
   *pofstreamCPP << "getDeviceEnumeration (char *pszLibraryName)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "Device *" << endl;
   *pofstreamCPP << "newDeviceW_Advanced (bool fAdvanced)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   PrintDevice *pPrintDevice = new " << *pstringNameOut << " ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   pPrintDevice->initialize ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   if (fAdvanced)" << endl;
   *pofstreamCPP << "      return pPrintDevice;" << endl;
   *pofstreamCPP << "   else" << endl;
   *pofstreamCPP << "      return new OmniProxy (pPrintDevice);" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "Device *" << endl;
   *pofstreamCPP << "newDeviceW_JopProp_Advanced (char *pszJobProperties, bool fAdvanced)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   PrintDevice *pPrintDevice = new " << *pstringNameOut << " (pszJobProperties);" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   pPrintDevice->initialize ();" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << "   if (fAdvanced)" << endl;
   *pofstreamCPP << "      return pPrintDevice;" << endl;
   *pofstreamCPP << "   else" << endl;
   *pofstreamCPP << "      return new OmniProxy (pPrintDevice);" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "void" << endl;
   *pofstreamCPP << "deleteDevice__FP6Device (Device *pDevice)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   deleteDevice (pDevice);" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "void" << endl;
   *pofstreamCPP << "deleteDevice (Device *pDevice)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   delete pDevice;" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << "void " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "commonInit ()" << endl;
   *pofstreamCPP << "{" << endl;

   iCount = listCapabilities.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << "   setCapabilities ( ";

      for ( DriverList::iterator next = listCapabilities.begin () ;
            next != listCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << "Capability::" << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                   | ";
      }

      *pofstreamCPP << "                   );" << endl;
   }

   iCount = listRasterCapabilities.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << endl;
      *pofstreamCPP << "   setRasterCapabilities ( ";

      for ( DriverList::iterator next = listRasterCapabilities.begin () ;
            next != listRasterCapabilities.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                   | ";
      }

      *pofstreamCPP << "                         );" << endl;
   }

   iCount = listDeviceOptions.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << endl;
      *pofstreamCPP << "   setDeviceOptions ( ";

      for ( DriverList::iterator next = listDeviceOptions.begin () ;
            next != listDeviceOptions.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << *pdi_d->pstringInstanceClassName << "::" << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                    | ";
      }

      *pofstreamCPP << "                    );" << endl;
   }

   iCount = listDeviceOptions2.size ();
   if (0 < iCount)
   {
      *pofstreamCPP << endl;
      *pofstreamCPP << "   setDeviceOptions2 ( ";

      for ( DriverList::iterator next = listDeviceOptions2.begin () ;
            next != listDeviceOptions2.end () ;
            next++, iCount-- )
      {
         *pofstreamCPP << *pdi_d->pstringInstanceClassName << "::" << *next << endl;
         if (1 < iCount)
            *pofstreamCPP << "                     | ";
      }

      *pofstreamCPP << "                     );" << endl;
   }

   *pofstreamCPP << endl;
   if (pdi_d->pstringExeName)
   {
      const char *pszData = "";

      if (pdi_d->pstringData)
         pszData = pdi_d->pstringData->c_str ();
      *pofstreamCPP << "   setDeviceInstance (new PluggableInstance (this," << endl;
      *pofstreamCPP << "                                             \"" << *pdi_d->pstringExeName << "\"," << endl;
      *pofstreamCPP << "                                             \"" << pszData << "\"));" << endl;
      *pofstreamCPP << "   setDeviceBlitter (new PluggableBlitter (this));" << endl;
   }
   else if (  pdi_d->pstringInstanceClassName
           && pdi_d->pstringBlitterClassName
           )
   {
      *pofstreamCPP << "   setDeviceInstance (new " << *pdi_d->pstringInstanceClassName << " (this));" << endl;
      *pofstreamCPP << "   setDeviceBlitter (new " << *pdi_d->pstringBlitterClassName << " (this));" << endl;
   }

   if (  pszMajorPDLLevel
      && pszMinorPDLLevel
      && pszMajorRevisionLevel
      && pszMinorRevisionLevel
      )
   {
      *pofstreamCPP << "   setPDL (new PDL (PDL::"
                    << pszMajorPDLLevel
                    << ",PDL::"
                    << pszMinorPDLLevel
                    << ","
                    << pszMajorRevisionLevel
                    << ","
                    << pszMinorRevisionLevel
                    << "));"
                    << endl;
   }

   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;

   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " ()" << endl;
   *pofstreamCPP << "   : PrintDevice (\"" << pszDriverName << "\"," << endl;
   *pofstreamCPP << "                  \"" << pszDeviceName << "\"," << endl;
   *pofstreamCPP << "                  \"" << *pstringNameOut << "\"," << endl;
   *pofstreamCPP << "                  \"orientation=" << pszOrientation;
   *pofstreamCPP << " dither=" << pszDither;
   *pofstreamCPP << " form=" << pszForm;
   *pofstreamCPP << " tray=" << pszTray;
   *pofstreamCPP << " media=" << pszMedia;
   *pofstreamCPP << " resolution=" << pszResolution;
   *pofstreamCPP << " printmode=" << pszPrintMode;

   // process optional other(s)
   while (nodeOther != 0)
   {
      const char *pszId = 0;

      getValue (nodeOther, &pszId);

      *pofstreamCPP << " " << pszId;

      // Move to the next one
      nodeOther = nextNode (nodeOther);
   }

   *pofstreamCPP << "\")" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   commonInit ();" << endl;
   *pofstreamCPP << "}" << endl;
   *pofstreamCPP << endl;
   *pofstreamCPP << *pstringNameOut << "::" << endl;
   *pofstreamCPP << *pstringNameOut << " (char *pszJobProperties)" << endl;
   *pofstreamCPP << "   : PrintDevice (\"" << pszDriverName << "\"," << endl;
   *pofstreamCPP << "                  \"" << pszDeviceName << "\"," << endl;
   *pofstreamCPP << "                  \"" << *pstringNameOut << "\"," << endl;
   *pofstreamCPP << "                  pszJobProperties)" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   commonInit ();" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceBooklet * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultBooklet ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceCollation * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultCollation ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceCopies * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultCopies ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceDestination * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultDestination ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "char * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultDitherID ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return \"" << pszDither << "\";" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceDuplex * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultDuplex ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceForm * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultForm ()" << endl;
   *pofstreamCPP << "{" << endl;

   if (pszComma)
   {
      int iX = 0,
          iY = 0;

      *pszComma = '\0';

      *pofstreamCPP << "   DeviceForm *pDF = " << *pdi_d->pstringFormClassName << "::create (this, " << *pdi_d->pstringFormClassName << "::" << pszForm << ");"  << endl;

      pszComma++;
      iX = parseSize (&pszComma);

      if (',' == *pszComma)
      {
         pszComma++;
         iY = parseSize (&pszComma);
      }

      if (  iX != 0
         && iY != 0
         )
      {
         *pofstreamCPP << "   pDF->setCx (" << iX << ");" << endl;
         *pofstreamCPP << "   pDF->setCy (" << iY << ");" << endl;
      }
      else if (iX != 0)
      {
         *pofstreamCPP << "   pDF->setCy (" << iX << ");" << endl;
      }

      *pofstreamCPP << "   return pDF;" << endl;
   }
   else
   {
      *pofstreamCPP << "   return " << *pdi_d->pstringFormClassName << "::create (this, " << *pdi_d->pstringFormClassName << "::" << pszForm << ");"  << endl;
   }

   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceJogging * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultJogging ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceMedia * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultMedia ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringMediaClassName << "::create (this, " << *pdi_d->pstringMediaClassName << "::" << pszMedia << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceOrientation * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultOrientation ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringOrientationClassName << "::create (this, " << *pdi_d->pstringOrientationClassName << "::" << pszOrientation << ");" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DevicePrintMode * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultPrintMode ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringPrintModeClassName << "::create (this, " << *pdi_d->pstringPrintModeClassName << "::" << pszPrintMode << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceResolution * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultResolution ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringResolutionClassName << "::create (this, " << *pdi_d->pstringResolutionClassName << "::" << pszResolution << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceNUp * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultNUp ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceScaling * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultScaling ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   // @TBD
   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceStapling * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultStapling ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceTray * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultTray ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return " << *pdi_d->pstringTrayClassName << "::create (this, " << *pdi_d->pstringTrayClassName << "::" << pszTray << ");"  << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceGamma * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getCurrentGamma ()" << endl;
   *pofstreamCPP << "{" << endl;
   if (pdi_d->pstringGammaClassName)
   {
      *pofstreamCPP << "   return " << *pdi_d->pstringGammaClassName << "::getDeviceGamma (getCurrentResolution ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentMedia ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentPrintMode ()->getID ()," << endl;
      *pofstreamCPP << "                            getCurrentDitherID ());" << endl;
   }
   else
   {
      *pofstreamCPP << "   return 0;" << endl;
   }
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceCommand * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultCommands ()" << endl;
   *pofstreamCPP << "{" << endl;
   *pofstreamCPP << "   return new " << *pdi_d->pstringCommandClassName << " ();" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceData * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultData ()" << endl;
   *pofstreamCPP << "{" << endl;
   if (pdi_d->pstringDataClassName)
      *pofstreamCPP << "   return new " << *pdi_d->pstringDataClassName << " ();" << endl;
   else
      *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   *pofstreamCPP << endl;
   *pofstreamCPP << "DeviceString * " << *pstringNameOut << "::" << endl;
   *pofstreamCPP << "getDefaultString ()" << endl;
   *pofstreamCPP << "{" << endl;
   if (pdi_d->pstringStringClassName)
      *pofstreamCPP << "   return new " << *pdi_d->pstringStringClassName << " ();" << endl;
   else
      *pofstreamCPP << "   return 0;" << endl;
   *pofstreamCPP << "}" << endl;

   if (!fAutoconf_d)
   {
      openMakeFile (pstringNameOut, &pofstreamMAK);

      *pofstreamMAK << "DEVICENAME            = " << *pstringNameOut << endl;
      *pofstreamMAK << "DEVCFILES             = " << *pdi_d->pstringPrintModeClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringResolutionClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringOrientationClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringCommandClassName << "~ \\" << endl;
      if (pdi_d->pstringGammaClassName)
         *pofstreamMAK << "                        " << *pdi_d->pstringGammaClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringTrayClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringFormClassName << "~ \\" << endl;
      if (pdi_d->pstringInstanceClassName)
         *pofstreamMAK << "                        " << *pdi_d->pstringInstanceClassName << "~ \\" << endl;
      if (pdi_d->pstringBlitterClassName)
         *pofstreamMAK << "                        " << *pdi_d->pstringBlitterClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pdi_d->pstringMediaClassName << "~ \\" << endl;
      if (pdi_d->pstringDataClassName)
         *pofstreamMAK << "                        " << *pdi_d->pstringDataClassName << "~ \\" << endl;
      if (pdi_d->pstringStringClassName)
         *pofstreamMAK << "                        " << *pdi_d->pstringStringClassName << "~ \\" << endl;
      *pofstreamMAK << "                        " << *pstringNameOut << "~" << endl;
      *pofstreamMAK << "OMNI_PATH             = .." << endl;
      *pofstreamMAK << endl;
      *pofstreamMAK << "all:    lib$(DEVICENAME).so" << endl;
      *pofstreamMAK << endl;
      *pofstreamMAK << "include ../common.mak" << endl;
      *pofstreamMAK << endl;
      *pofstreamMAK << *pstringNameOut
                    << ".o:\t"
                    << *pstringNameOut
                    << ".cpp ";
      if (pdi_d->pstringBlitterClassName)
         *pofstreamMAK << *pdi_d->pstringBlitterClassName << ".hpp ";
      if (pdi_d->pstringInstanceClassName)
         *pofstreamMAK << *pdi_d->pstringInstanceClassName << ".hpp ";
      *pofstreamMAK << *pdi_d->pstringFormClassName
                    << ".o "
                    << *pdi_d->pstringResolutionClassName
                    << ".o "
                    << *pdi_d->pstringMediaClassName
                    << ".o "
                    << *pdi_d->pstringTrayClassName
                    << ".o "
                    << *pdi_d->pstringOrientationClassName
                    << ".o "
                    << *pdi_d->pstringCommandClassName
                    << ".o "
                    << "../Device.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringFormClassName
                    << ".o:\t"
                    << *pdi_d->pstringFormClassName
                    << ".cpp "
                    << *pdi_d->pstringFormClassName
                    << ".hpp "
                    << "../DeviceForm.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringResolutionClassName
                    << ".o:\t"
                    << *pdi_d->pstringResolutionClassName
                    << ".cpp "
                    << *pdi_d->pstringResolutionClassName
                    << ".hpp "
                    << "../DeviceResolution.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringMediaClassName
                    << ".o:\t"
                    << *pdi_d->pstringMediaClassName
                    << ".cpp "
                    << *pdi_d->pstringMediaClassName
                    << ".hpp "
                    << "../DeviceMedia.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringTrayClassName
                    << ".o:\t"
                    << *pdi_d->pstringTrayClassName
                    << ".cpp "
                    << *pdi_d->pstringTrayClassName
                    << ".hpp "
                    << "../DeviceTray.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringOrientationClassName
                    << ".o:\t"
                    << *pdi_d->pstringOrientationClassName
                    << ".cpp "
                    << *pdi_d->pstringOrientationClassName
                    << ".hpp "
                    << "../DeviceOrientation.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringCommandClassName
                    << ".o:\t"
                    << *pdi_d->pstringCommandClassName
                    << ".cpp "
                    << *pdi_d->pstringCommandClassName
                    << ".hpp "
                    << "../DeviceCommand.o"
                    << endl;
      *pofstreamMAK << *pdi_d->pstringPrintModeClassName
                    << ".o:\t"
                    << *pdi_d->pstringPrintModeClassName
                    << ".cpp "
                    << *pdi_d->pstringPrintModeClassName
                    << ".hpp "
                    << "../DevicePrintMode.o"
                    << endl;
//////*pofstreamMAK << *pdi_d->pstringConnectionClassName
//////              << ".o:\t"
//////              << *pdi_d->pstringConnectionClassName
//////              << ".cpp "
//////              << *pdi_d->pstringConnectionClassName
//////              << ".hpp "
//////              << "../DeviceConnection.o"
//////              << endl;
      if (pdi_d->pstringGammaClassName != 0)
      {
         *pofstreamMAK << *pdi_d->pstringGammaClassName
                       << ".o:\t"
                       << *pdi_d->pstringGammaClassName
                       << ".cpp "
                       << *pdi_d->pstringGammaClassName
                       << ".hpp "
                       << *pdi_d->pstringResolutionClassName
                       << ".hpp "
                       << *pdi_d->pstringMediaClassName
                       << ".hpp "
                       << *pdi_d->pstringPrintModeClassName
                       << ".hpp "
                       << "../DeviceGammaTable.hpp "
                       << "../DeviceGamma.hpp "
                       << "../DeviceResolution.hpp "
                       << "../DeviceMedia.hpp "
                       << "../DevicePrintMode.hpp "
                       << "../Enumeration.hpp "
                       << "../BinaryData.hpp "
                       << "../DeviceDither.hpp "
                       << "../defines.hpp"
                       << endl;
      }
      if (pdi_d->pstringBlitterClassName)
      {
         *pofstreamMAK << *pdi_d->pstringBlitterClassName
                       << ".o:\t"
                       << *pdi_d->pstringBlitterClassName
                       << ".cpp "
                       << *pdi_d->pstringBlitterClassName
                       << ".hpp "
                       << "../DeviceBlitter.o ";
         if (pdi_d->pstringInstanceClassName)
            *pofstreamMAK << *pdi_d->pstringInstanceClassName
                          << ".hpp ";
         *pofstreamMAK << *pdi_d->pstringPrintModeClassName
                       << ".hpp "
                       << *pdi_d->pstringResolutionClassName
                       << ".hpp "
                       << "../JobProperties.hpp"
                       << endl;
      }
      if (pdi_d->pstringInstanceClassName)
      {
         *pofstreamMAK << *pdi_d->pstringInstanceClassName
                       << ".o:\t"
                       << *pdi_d->pstringInstanceClassName
                       << ".cpp "
                       << *pdi_d->pstringInstanceClassName
                       << ".hpp "
                       << "../DeviceInstance.o"
                       << endl;
      }
      if (pdi_d->pstringDataClassName)
      {
         *pofstreamMAK << *pdi_d->pstringDataClassName
                       << ".o:\t"
                       << *pdi_d->pstringDataClassName
                       << ".cpp "
                       << *pdi_d->pstringDataClassName
                       << ".hpp "
                       << "../DeviceData.o"
                       << endl;
      }
      if (pdi_d->pstringStringClassName)
      {
         *pofstreamMAK << *pdi_d->pstringStringClassName
                       << ".o:\t"
                       << *pdi_d->pstringStringClassName
                       << ".cpp "
                       << *pdi_d->pstringStringClassName
                       << ".hpp "
                       << "../DeviceString.o"
                       << endl;
      }
   }

done:
   if (!rc)
   {
      cerr << "Error: Could not traverse the tree for \"" << *pstrXMLFile_d << "\"" << endl;
      setErrorCondition ();

      printDOMNode (driverNode, 0);
   }

   // Clean up!
   delete pofstreamHPP;
   delete pofstreamCPP;
   delete pofstreamMAK;
   delete pstringNameOut;

   return rc;
}

void OmniDomParser::
processDOMNode (DOM_Node& node)
{
   switch (node.getNodeType ())
   {
   case DOM_Node::DOCUMENT_NODE:
   {
      DOM_Node child = node.getFirstChild ();

      while (child != 0)
      {
         processDOMNode (child);

         child = child.getNextSibling ();
      }
      break;
   }

   case DOM_Node::ELEMENT_NODE:
   {
      DOMString   nodeName       = node.getNodeName ();
      DOMString   nodeValue      = node.getNodeValue ();
      const char *pszElementName = nodeName.transcode ();

      if (  0 == strcmp ("Uses", pszElementName)
         || 0 == strcmp ("Has", pszElementName)
         )
      {
         DOM_Node    nodeSibling = node.getFirstChild ();
         const char *pszFilename = nodeSibling.getNodeValue ().transcode ();

         OmniDomParser parser (pszFilename,
                               pdi_d,
                               fAutoconf_d,
                               fAutoconfNoInst_d);

         try
         {
            parser.parse ();

            fFileModified_d = fFileModified_d || parser.fileModified ();
         }
         catch (const XMLException& e)
         {
            cerr << "An error occured during Parsing:  "
                 << DOMString (e.getMessage ()).transcode ()
                 << endl;
            setErrorCondition ();
         }
      }
      else if (0 == strcmp ("deviceOrientations", pszElementName))
         processDeviceOrientations (node);
      else if (0 == strcmp ("deviceCommands", pszElementName))
         processDeviceCommands (node);
      else if (0 == strcmp ("deviceResolutions", pszElementName))
         processDeviceResolutions (node);
      else if (0 == strcmp ("devicePrintModes", pszElementName))
         processDevicePrintModes (node);
      else if (0 == strcmp ("deviceTrays", pszElementName))
         processDeviceTrays (node);
      else if (0 == strcmp ("deviceForms", pszElementName))
         processDeviceForms (node);
      else if (0 == strcmp ("deviceMedias", pszElementName))
         processDeviceMedias (node);
      else if (0 == strcmp ("deviceConnections", pszElementName))
         processDeviceConnections (node);
      else if (0 == strcmp ("deviceGammaTables", pszElementName))
         processDeviceGammas (node);
      else if (0 == strcmp ("deviceStrings", pszElementName))
         processDeviceStrings (node);
      else if (0 == strcmp ("deviceDatas", pszElementName))
         processDeviceDatas (node);

      //
      //  Test for the presence of children, which includes both
      //  text content and nested elements.
      //
      DOM_Node child = node.getFirstChild ();

      if (child != 0)
      {
         while (child != 0)
         {
            processDOMNode (child);

            child = child.getNextSibling ();
         }
      }

      if (0 == strcmp ("Device", pszElementName))
         processDriver (node);
      break;
   }
   }
}

bool OmniDomParser::
parse ()
{
   const char *pszXMLFile = pstrXMLFile_d->c_str ();

   try
   {
#ifdef DEBUG
      if (fDebugOutput) cerr << "DOM Parsing: " << pszXMLFile << endl;
#endif

      parser_d->parse (pszXMLFile);

      DOM_Node doc = parser_d->getDocument ();

//////printDOMNode (doc, 0);
      processDOMNode (doc);

      return !fErrorCondition_d;
   }
   catch (const XMLException& e)
   {
      cerr << "Error: File not found: '" << pszXMLFile << "'" << endl;
      cerr << "Exception message is: " << XMLString::transcode (e.getMessage ()) << endl;

      return false;
   }
   catch (...)
   {
      cerr << "Error: Unknown catch occured @ " << __FILE__ << ":" << __LINE__ << endl;

      return false;
   }
}
